unit freereverb;
{
  Unit: Reverb for either KOL or VCL Delphi and Freepascal 1.9.x,
  probably Kylix.
  purpose: Based on the Freeverb (C++) design by Jezar[at]dreampoint.co.uk
  Author: Thaddy de Koning, thaddy[at]thaddy.com
  Copyright: Original in C++ 2000, Jezar
  Delphi and Freepascal version 2003,
  Basm  added 2004, Thaddy de Koning
  Use as you like, copyrighted freeware.

  Remarks: Removed original skip for interleave.
  Comb and Allpass filter processing rewritten in BASM.
  It is as faithfull a translation as I could manage.
  Sound is exactly the same as FreeVerb 3.
  I have writen it to be as portable as my Pascals go
  Although not tested, it should work in BCB6+ and Kylix too.
  It works with Delphi 6/7 and Freepascal 1.9.X under both linux
  and windows.
  There are no dependencies on the windows unit.

  2014-05-16 : Changed to be used in WREN by Blue Hell to work on a single
  sample pair instead of a buffer. This was downloaded from :
  http://members.chello.nl/t.koning8/kolindex.htm .
  It will not work on all platforms anymore.
}

interface


uses

  Classes, KnobsUtils;


//
// Reverb model tuning values, taken from original algoritm by Jezar
//

const

  Kdenorm : single = 1.0e-23;

  numcombs       = 8;
  numallpasses   = 4;
  muted          = 0;
  fixedgain      = 0.015;
  scalewet       = 3;
  scaledry       = 2;
  scaledamp      = 0.9;        // was 0.4 :: 0.9 provides more damping
  scaleroom      = 0.28;       // was 0.28
  offsetroom     = 0.7;
  initialroom    = 0.5;
  initialdamp    = 0.5;
  initialwet     = 1 / scalewet;
  initialdry     = 1 / scaledry;
  initialwidth   = 1;
  initialmode    = 0;
  freezemode     = 0.5;
  stereospread_l = 0;          // Original value
  stereospread_r = +23;        // Original value

//  stereospread_l = -21;        // Original was 23, and it was added only to the right channel
//  stereospread_r = +21;        // Original was 23, and it was added only to the right channel

type

  TAllpass = class( TObject)
  private
    feedback : single;
    buffer   : pointer;
    bufsize  : integer;
    bufidx   : integer;
  public
    constructor Create( Buffersize : integer);                                                                  virtual;
    destructor  Destroy;                                                                                       override;
    function    process( const input : single) : single;                                                       register;
    procedure   mute;
    procedure   setfeedback( Value : single);
    function    getfeedback: single;
  end;

  Tcomb = class( TObject)
  private
    feedback    : single;
    filterstore : single;
    damp1       : single;
    damp2       : single;
    buffer      : pointer;
    bufsize     : integer;
    bufidx      : integer;
  public
    constructor Create( Buffersize : integer);                                                                  virtual;
    destructor  Destroy;                                                                                       override;
    function    process( const input: single): single;                                                         register;
    procedure   mute;
    procedure   setdamp    ( Value : single);
    function    getdamp    : single;
    procedure   setfeedback( Value : single);
    function    getfeedback: single;
  end;


  TReverb = class( TObject)
  private
    combtuningL1    : Integer;
    combtuningR1    : Integer;
    combtuningL2    : Integer;
    combtuningR2    : Integer;
    combtuningL3    : Integer;
    combtuningR3    : Integer;
    combtuningL4    : Integer;
    combtuningR4    : Integer;
    combtuningL5    : Integer;
    combtuningR5    : Integer;
    combtuningL6    : Integer;
    combtuningR6    : Integer;
    combtuningL7    : Integer;
    combtuningR7    : Integer;
    combtuningL8    : Integer;
    combtuningR8    : Integer;
    allpasstuningL1 : Integer;
    allpasstuningR1 : Integer;
    allpasstuningL2 : Integer;
    allpasstuningR2 : Integer;
    allpasstuningL3 : Integer;
    allpasstuningR3 : Integer;
    allpasstuningL4 : Integer;
    allpasstuningR4 : Integer;
  private
    gain            : single;
    roomsize        : single;
    roomsize1       : single;
    damp            : single;
    damp1           : single;
    wet             : single;
    wet1            : single;
    wet2            : single;
    dry             : single;
    Width           : single;
    mode            : single;
    combL           : array[ 0 .. numcombs     - 1] of Tcomb;
    combR           : array[ 0 .. numcombs     - 1] of Tcomb;
    allpassL        : array[ 0 .. numallpasses - 1] of TAllpass;
    allpassR        : array[ 0 .. numallpasses - 1] of TAllpass;
  protected
    procedure   Update;
  public
    constructor Create( aSampleRate: Integer);
    destructor  Destroy;                                                                                       override;
    procedure   mute;
    procedure   processmix( const inputL : single; const inputR : single; var outputL : single; var outputR : single);
    procedure   processreplace( const inputL : single; const inputR : single; var outputL : single; var outputR : single);
    procedure   setroomsize( Value : single);
    function    getroomsize: single;
    procedure   setdamp    ( Value : single);
    function    getdamp    : single;
    procedure   setwet     ( Value : single);
    function    getwet     : single;
    procedure   setdry     ( Value : single);
    function    getdry     : single;
    procedure   setwidth   ( Value : single);
    function    getwidth   : single;
    procedure   setmode    ( Value : single);
    function    getmode    : single;
  end;


implementation


{ ========
  TAllpass = class( TObject)
  private
    feedback : single;
    buffer   : pointer;
    bufsize  : integer;
    bufidx   : integer;
  public
 }

    constructor TAllpass.Create( Buffersize: integer);
    begin
      inherited Create;
      bufsize := Buffersize * 4;
      buffer  := AllocMem(bufsize);
      bufidx  := 0;
    end;

    destructor TAllpass.Destroy;
    begin
      freemem( buffer);
      inherited;
    end;

    function TAllpass.getfeedback: single;
    begin
      Result := feedback;
    end;

    procedure TAllpass.mute;
    begin
      Fillchar( buffer^, bufsize, 0);
    end;

    { I really don't know if this is all as fast as can be,
      but it beats Delphi's compiler generated code hands down,
      Thaddy }
    function TAllpass.process( const input: single): single;
    {$ifdef win32}
    asm
      mov  ecx, [ eax].buffer                 // buffer start in ecx
      mov  edx, [ eax].Bufidx                 // buffer index in edx
      fld  input

      // This checks for very small values that can cause a processor
      // to switch in extra precision mode, which is expensive.
      // Since such small values are irrelevant to audio, avoid this.
      // The code is equivalent to the C inline macro by Jezar
      // This is the same spot where the original C macro appears
      test dword ptr [ ecx + edx], $7F800000  // test if denormal
      jnz @Normal
      mov dword ptr [ ecx + edx], 0           // if so, zero out
    @normal:

      fld  [ ecx + edx].Single                // load current sample from buffer
      fsub st( 0), st( 1)                     // subtract input sample
      // NOT fsub, because delphi 7 translates that into fsubp!
      fxch                                    // this is a zero cycle operant,
      // just renames the stack internally
      fmul [ eax].feedback                    // multiply stored sample with feedback
      fadd input                              // and add the input
      fstp [ ecx + edx].Single;               // store at the current sample pos
      add  edx, 4                             // increment sample position
      cmp  edx, [ eax].BufSize;               // are we at end of buffer?
      jb   @OK
      xor  edx, edx                           // if so, reset buffer index
    @OK:
      mov  [ eax].bufidx, edx                 // and store new index,
      // result already in st(0),
      // hence the fxch
    end;
    {$else}
    begin
      Result := 0;
    end;
    {$endif}

    procedure TAllpass.setfeedback( Value : single);
    begin
      feedback := Value;
    end;

{ ========
  Tcomb = class( TObject)
  private
  public
}

    constructor Tcomb.Create( Buffersize : integer);
    begin
      inherited Create;
      bufsize     := Buffersize * 4;
      buffer      := AllocMem( bufsize);
      filterstore := 0;
      bufidx      := 0;
    end;

    destructor Tcomb.Destroy;
    begin
      freemem( buffer);
      inherited;
    end;

    function Tcomb.getdamp : single;
    begin
      Result := damp1;
    end;

    function Tcomb.getfeedback : single;
    begin
      Result := feedback;
    end;

    procedure Tcomb.mute;
    begin
      Fillchar( buffer^, bufsize, 0);
    end;

    { I really don't know if this is all as fast as can be,
      but it beats Delphi's compiler generated code hands down,
      Thaddy }

    function Tcomb.process(const input : single) : single;
    {$ifdef win32}
    asm
      mov   ecx, [ eax].Buffer                       // buffer start in ecx
      mov   edx, [ eax].Bufidx                       // buffer index in edx

      // This checks for very small values that can cause a processor
      // to switch in extra precision mode, which is expensive.
      // Since such small values are irrelevant to audio, avoid this.
      // This is the same spot where the original C macro appears
      test  dword ptr [ ecx + edx], $7F800000        // test if denormal
      jnz   @Normal
      mov   dword ptr [ ecx + edx], 0                // if so, zero out
    @normal:

      fld   [ ecx + edx].Single;                     // load sample from buffer
      fld   st( 0)                                   // duplicate on the stack
      fmul  [ eax].damp2                             // multiply with damp2
      fld   [ eax].filterstore;                      // load stored filtered sample
      fmul  [ eax].damp1                             // multiply with damp1
      faddp
      // fadd  Kdenorm
      fst   [ eax].filterstore                       // store it back

      // This checks for very small values that can cause a processor
      // to switch in extra precision mode, which is expensive.
      // Since such small values are irrelevant to audio, avoid this.
      // This is the same spot where the original C macro appears
      test  dword ptr [ eax].filterstore, $7F800000  // test if denormal
      jnz   @Normal2
      mov   dword ptr [ eax].filterstore, 0          // if so, zero out
    @normal2:

      fmul  [ eax].feedback                          // multiply with feedback
      fadd  input                                    // and add to input sample

      // fadd  Kdenorm
      fstp  [ ecx + edx].Single                      // store at current buffer pos
      add   edx, 4                                   // update buffer index
      cmp   edx, [ eax].BufSize;                     // end of buffer reached?
      jb    @OK

      xor   edx, edx                                 // if so, reset buffer index
    @OK:
      mov  [ eax].bufidx, edx                        // and store new index.
      // result already in st(0),
      // hence duplicate
    end;
    {$else}
    begin
      Result := 0;
    end;
    {$endif}

    procedure Tcomb.setdamp(Value : single);
    begin
      damp1 := Value;
      damp2 := 1 - Value;
    end;

    procedure Tcomb.setfeedback(Value : single);
    begin
      feedback := Value;
    end;

{ ========
  TReverb = class( TObject)
  private
    combtuningL1    : Integer;
    combtuningR1    : Integer;
    combtuningL2    : Integer;
    combtuningR2    : Integer;
    combtuningL3    : Integer;
    combtuningR3    : Integer;
    combtuningL4    : Integer;
    combtuningR4    : Integer;
    combtuningL5    : Integer;
    combtuningR5    : Integer;
    combtuningL6    : Integer;
    combtuningR6    : Integer;
    combtuningL7    : Integer;
    combtuningR7    : Integer;
    combtuningL8    : Integer;
    combtuningR8    : Integer;
    allpasstuningL1 : Integer;
    allpasstuningR1 : Integer;
    allpasstuningL2 : Integer;
    allpasstuningR2 : Integer;
    allpasstuningL3 : Integer;
    allpasstuningR3 : Integer;
    allpasstuningL4 : Integer;
    allpasstuningR4 : Integer;
  private
    gain            : single;
    roomsize        : single;
    roomsize1       : single;
    damp            : single;
    damp1           : single;
    wet             : single;
    wet1            : single;
    wet2            : single;
    dry             : single;
    Width           : single;
    mode            : single;
    combL           : array[ 0 .. numcombs     - 1] of Tcomb;
    combR           : array[ 0 .. numcombs     - 1] of Tcomb;
    allpassL        : array[ 0 .. numallpasses - 1] of TAllpass;
    allpassR        : array[ 0 .. numallpasses - 1] of TAllpass;
  protected
}

    procedure TReverb.Update;
    var
      i : integer;
    begin
      // Recalculate internal values after parameter change
      wet1 := wet * ( Width / 2 + 0.5);
      wet2 := wet * (( 1 - Width) / 2);

      if mode >= freezemode
      then begin
        roomsize1 := 1;
        damp1     := 0;
        gain      := muted;
      end
      else begin
        roomsize1 := roomsize;
        damp1     := damp;
        gain      := fixedgain;
      end;

      for i := 0 to numcombs - 1
      do begin
        combL[ i].setfeedback( roomsize1);
        combR[ i].setfeedback( roomsize1);
        combL[ i].setdamp( damp1);
        combR[ i].setdamp( damp1);
      end;
    end;

//  public

    constructor TReverb.Create( aSampleRate: Integer);
    begin
      inherited Create;

      // These values assume 44.1KHz sample rate
      // they will probably be OK for 48KHz sample rate
      // but would need scaling for 96KHz (or other) sample rates.
      // The values were obtained by listening tests.
      // The values are now recalculated for different sample rates,
      // maybe some prime checking would be in order though

      combtuningL1    := NextPrime(( aSampleRate * 1116 div 44100) + stereospread_l);
      combtuningR1    := NextPrime(( aSampleRate * 1116 div 44100) + stereospread_r);
      combtuningL2    := NextPrime(( aSampleRate * 1188 div 44100) + stereospread_l);
      combtuningR2    := NextPrime(( aSampleRate * 1188 div 44100) + stereospread_r);
      combtuningL3    := NextPrime(( aSampleRate * 1277 div 44100) + stereospread_l);
      combtuningR3    := NextPrime(( aSampleRate * 1277 div 44100) + stereospread_r);
      combtuningL4    := NextPrime(( aSampleRate * 1356 div 44100) + stereospread_l);
      combtuningR4    := NextPrime(( aSampleRate * 1356 div 44100) + stereospread_r);
      combtuningL5    := NextPrime(( aSampleRate * 1422 div 44100) + stereospread_l);
      combtuningR5    := NextPrime(( aSampleRate * 1422 div 44100) + stereospread_r);
      combtuningL6    := NextPrime(( aSampleRate * 1491 div 44100) + stereospread_l);
      combtuningR6    := NextPrime(( aSampleRate * 1491 div 44100) + stereospread_r);
      combtuningL7    := NextPrime(( aSampleRate * 1557 div 44100) + stereospread_l);
      combtuningR7    := NextPrime(( aSampleRate * 1557 div 44100) + stereospread_r);
      combtuningL8    := NextPrime(( aSampleRate * 1617 div 44100) + stereospread_l);
      combtuningR8    := NextPrime(( aSampleRate * 1617 div 44100) + stereospread_r);
      allpasstuningL1 := NextPrime(( aSampleRate *  556 div 44100) + stereospread_l);
      allpasstuningR1 := NextPrime(( aSampleRate *  556 div 44100) + stereospread_r);
      allpasstuningL2 := NextPrime(( aSampleRate *  441 div 44100) + stereospread_l);
      allpasstuningR2 := NextPrime(( aSampleRate *  441 div 44100) + stereospread_r);
      allpasstuningL3 := NextPrime(( aSampleRate *  341 div 44100) + stereospread_l);
      allpasstuningR3 := NextPrime(( aSampleRate *  341 div 44100) + stereospread_r);
      allpasstuningL4 := NextPrime(( aSampleRate *  225 div 44100) + stereospread_l);
      allpasstuningR4 := NextPrime(( aSampleRate *  225 div 44100) + stereospread_r);

      combL   [ 0]    := Tcomb   .Create( combtuningL1   );
      combR   [ 0]    := Tcomb   .Create( combtuningR1   );
      combL   [ 1]    := Tcomb   .Create( combtuningL2   );
      combR   [ 1]    := Tcomb   .Create( combtuningR2   );
      combL   [ 2]    := Tcomb   .Create( combtuningL3   );
      combR   [ 2]    := Tcomb   .Create( combtuningR3   );
      combL   [ 3]    := Tcomb   .Create( combtuningL4   );
      combR   [ 3]    := Tcomb   .Create( combtuningR4   );
      combL   [ 4]    := Tcomb   .Create( combtuningL5   );
      combR   [ 4]    := Tcomb   .Create( combtuningR5   );
      combL   [ 5]    := Tcomb   .Create( combtuningL6   );
      combR   [ 5]    := Tcomb   .Create( combtuningR6   );
      combL   [ 6]    := Tcomb   .Create( combtuningL7   );
      combR   [ 6]    := Tcomb   .Create( combtuningR7   );
      combL   [ 7]    := Tcomb   .Create( combtuningL8   );
      combR   [ 7]    := Tcomb   .Create( combtuningR8   );

      allpassL[ 0]    := TAllpass.Create( allpasstuningL1);
      allpassR[ 0]    := TAllpass.Create( allpasstuningR1);
      allpassL[ 1]    := TAllpass.Create( allpasstuningL2);
      allpassR[ 1]    := TAllpass.Create( allpasstuningR2);
      allpassL[ 2]    := TAllpass.Create( allpasstuningL3);
      allpassR[ 2]    := TAllpass.Create( allpasstuningR3);
      allpassL[ 3]    := TAllpass.Create( allpasstuningL4);
      allpassR[ 3]    := TAllpass.Create( allpasstuningR4);

      // Set default values

      allpassL[ 0].setfeedback( 0.5);
      allpassR[ 0].setfeedback( 0.5);
      allpassL[ 1].setfeedback( 0.5);
      allpassR[ 1].setfeedback( 0.5);
      allpassL[ 2].setfeedback( 0.5);
      allpassR[ 2].setfeedback( 0.5);
      allpassL[ 3].setfeedback( 0.5);
      allpassR[ 3].setfeedback( 0.5);

      setwet     ( initialwet  );
      setroomsize( initialroom );
      setdry     ( initialdry  );
      setdamp    ( initialdamp );
      setwidth   ( initialwidth);
      setmode    ( initialmode );
      mute;
    end;

    destructor TReverb.Destroy;
    var
      i : integer;
    begin
      for i := Low( allPassL) to High( allPassL) do allpassL[ i].Free;
      for i := Low( allPassR) to High( allPassR) do allpassR[ i].Free;
      for i := Low( combL   ) to High( combL   ) do combL   [ i].Free;
      for i := Low( combR   ) to High( combR   ) do combR   [ i].Free;

      inherited;
    end;

    function TReverb.getdamp: single;
    begin
      Result := damp / scaledamp;
    end;

    function TReverb.getdry: single;
    begin
      Result := dry / scaledry;
    end;

    function TReverb.getmode: single;
    begin
      if mode >= freezemode
      then Result := 1
      else Result := 0;
    end;

    function TReverb.getroomsize : single;
    begin
      Result := (roomsize - offsetroom) / scaleroom;
    end;

    function TReverb.getwet : single;
    begin
      Result := wet / scalewet;
    end;

    function TReverb.getwidth : single;
    begin
      Result := Width;
    end;

    procedure TReverb.mute;
    var
      i : integer;
    begin
      if getmode >= freezemode
      then Exit;

      for i := 0 to numcombs - 1
      do begin
        combL[ i].mute;
        combR[ i].mute;
      end;
      for i := 0 to numallpasses - 1
      do begin
        allpassL[ i].mute;
        allpassR[ i].mute;
      end;
    end;

    procedure TReverb.processmix(const inputL : single; const inputR : single; var outputL : single; var outputR : single);
    var
      outL, outR, input : single;
      j                 : integer;
    begin
      outL  := inputL;
      outR  := inputR;
      input := ( inputL + inputR) * gain;
      // Accumulate comb filters in parallel
      for j := 0 to numcombs - 1
      do begin
        outL := outL + combL[ j].process( input);
        outR := outR + combR[ j].process( input);
      end;
      // Feed through allpasses in series
      for j := 0 to numallpasses - 1
      do begin
        outL := allpassL[ j].process( outL);
        outR := allpassR[ j].process( outR);
      end;
      // Calculate output MIXING with anything already there
      outputL := outputL + outL * wet1 + outR * wet2 + inputL * dry;
      outputR := outputR + outR * wet1 + outL * wet2 + inputR * dry;
    end;

    procedure TReverb.processreplace(const inputL : single; const inputR : single; var outputL : single; var outputR : single);
    var
      outL, outR, input : single;
      j                 : integer;
    begin
      outL  := 0;
      outR  := 0;
      input := ( inputL + inputR) * gain;
      // Accumulate comb filters in parallel
      for j := 0 to numcombs - 1
      do begin
        outL := outL + combL[ j].process( input);
        outR := outR + combR[ j].process( input);
      end;
      // Feed through allpasses in series
      for j := 0 to numallpasses - 1
      do begin
        outL := allpassL[ j].process( outL);
        outR := allpassR[ j].process( outR);
      end;
      // Calculate output REPLACING anything already there
      outputL := outL * wet1 + outR * wet2 + inputL * dry;
      outputR := outR * wet1 + outL * wet2 + inputR * dry;
    end;

    procedure TReverb.setdamp( Value : single);
    var
      aNewValue: Single;
    begin
      aNewValue := Value * scaledamp;
      if aNewValue <> damp
      then begin
        damp := aNewValue;
        Update;
      end;
    end;

    procedure TReverb.setdry( Value : single);
    var
      aNewValue: Single;
    begin
      aNewValue := Value * scaledry;
      if aNewValue <> dry
      then begin
        dry := aNewValue;
        Update;
      end;
    end;

    procedure TReverb.setmode( Value : single);
    begin
      if Value <> mode
      then begin
        mode := Value;
        Update;
      end;
    end;

    procedure TReverb.setroomsize( Value : single);
    var
      aNewValue: Single;
    begin
      aNewValue := ( Value * scaleroom) + offsetroom;
      if aNewValue <> roomsize
      then begin
        roomsize := aNewValue;
        Update;
      end;
    end;

    procedure TReverb.setwet( Value : single);
    var
      aNewValue: Single;
    begin
      aNewValue := Value * scalewet;
      if aNewValue <> wet
      then begin
        wet := aNewValue;
        Update;
      end;
    end;

    procedure TReverb.setwidth( Value : single);
    begin
      if Value <> width
      then begin
        Width := Value;
        Update;
      end;
    end;


end.
