unit knobs2013;

{

  (C) COPYRIGHT 1999 .. 2015 Blue Hell / Jan Punter

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation;

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  For all listed email addresses :

    _dot. to be substituted by a dot      '.'
    2@t2  to be substituted by an at sign '@'


  Blue Hell is a trade mark owned by

    Jan Punter
    Oogstplein 6
    7545 HP Enschede
    the Netherlands
    http://www.iaf.nl/Users/BlueHell/
    http://bluehell.electro-music.com/
    j_dot.punter2@t2iaf_dot.nl

  All rights attributed to Blue Hell are owned by Jan Punter.
}

interface

uses

  WinApi.Windows, WinApi.Messages, System.SysUtils, System.Types, System.StrUtils, System.Math, System.Classes,
  System.Contnrs, System.UITypes, Vcl.Graphics, Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.ExtCtrls, Vcl.Buttons,
  Vcl.StdCtrls, Vcl.ComCtrls, WinApi.MMSystem,

  KnobsUtils, ComplexMath, KnobsConversions;

const

  DO_RESTACK                = True;
  NO_RESTACK                = not DO_RESTACK;
  DO_DRAG                   = True;
  NO_DRAG                   = not DO_DRAG;
  DO_REDRAW_WIRES           = True;
  NO_REDRAW_WIRES           = not DO_REDRAW_WIRES;
  DO_NOTIFY                 = True;
  NO_NOTIFY                 = not DO_NOTIFY;

  KNOBS_UM_CONNECTIONADD    = WM_USER + 0;
  KNOBS_UM_CONNECTIONREMOVE = WM_USER + 1;
  KNOBS_UM_COLORCHANGE      = WM_USER + 2;
  KNOBS_UM_HIGHLIGHT        = WM_USER + 3;

  CL_NO_OUTPUT              = clAqua;
  CL_HIGHLIGHT              = clWhite;

const

  KnobsColorDisplay = clGray;    // Default BackgroundColor for TKnobsDisplay.
  KnobsPatchVersion = 2;

type

  TKnobsLedShape = (
    lsRound,
    lsSquare
  );

  TKnobsModuleFlag = (
    mfSelected,
    mfCapturing
  );
  TKnobsModuleFlags = set of TKnobsModuleFlag;

  TKnobsConnectOp = (
    coConnect,
    coDisconnect
  );

  TKnobsReadMode = (
    rmAppend,                    // Adds the read mudules to an editor patch
    rmReplace                    // Removes the existing modules and replaces them with the read modules
  );

  TKnobsWriteMode = (
    wmAll,
    wmSelected
  );

  TKnobsDisplayType = (
    dtSingleFromLow,
    dtSingleFromHigh,
    dtFilledFromLow,
    dtFilledFromHigh
  );

  TKnobsDisplayOrientation = (
    doHorizontal,
    doVertical
  );

  TKnobsButtonsDisplayMode = (
    bdmHorizontal,
    bdmVertical
  );

  TKnobsDVStyle = (        // DataViewer design time styles
    dvsBlank,              // Paint nothing
    dvsSine,               // Paint a sine
    dvsTri,                // triangle
    dvsSaw,                // saw
    dvsSquare,             // pulse / square
    dvsNoise,              // noise / random
    dvsEnvAr,              // AR   type envelope
    dvsEnvAhd,             // AHD  type envelope
    dvsEnvAdsr,            // ADSR type envelope
    dvsXYScope             // XY scope display type
  );

  TKnobsPairingMode = (    // Pairing mode for paired knobs
    pmOff,                 // Pairing is off
    pmNormal,              // Pairing is normal, both knobs have the same position
    pmMirrored             // Pairing is anti normal, knobs have mirrored positions
  );

  EKnobs              = class( Exception);
  EKnobsCorruptPatch  = class( EKnobs);
  EKnobsNoDuplicates  = class( EKnobs);
  EKnobsShortCircuit  = class( EKnobs);
  EKnobsDuplicateItem = class( EKnobs);

  TKnobsBezierPoints = array[ 0 .. 3] of TPoint;
  TKnobsBezierParams = record
    Angle1    : Integer;
    Angle2    : Integer;
    Strength1 : Integer;
    Strength2 : Integer;
  end;

  TKnobsConnector            = class;
  TKnobsValuedControl        = class;
  TKnobsDisplay              = class;
  TKnobsEditLabel            = class;
  TKnobsCustomModule         = class;
  TKnobsWirePanel            = class;
  TKnobsModuleSelector       = class;
  TKnobsModuleButton         = class;

  TKnobsModuleClass          = class of TKnobsCustomModule;

  TKnobsOnValueChange        = procedure( const aSender: TObject; const aPath, aControlType: string; aValue: TSignal; IsFinal: Boolean) of object;
  TKnobsOnModuleButtonClick  = procedure( const aSender: TObject; aModuleclass: TKnobsModuleType)                                 of object;
  TKnobsOnGetGlyph           = function ( const aSender: TObject; const aName: string): TBitmap                                   of object;
  TKnobsOnLog                = procedure( const aSender: TObject; const aMsg: string)                                             of object;
  TKnobsOnHistoryChange      = procedure( const aSender: TObject; anUndoCount, aRedoCount: Integer)                               of object;
  TKnobsOnPatchChanged       = procedure( const aSender: TKnobsWirePanel)                                                         of object;
  TKnobsOnWiresChanged       = procedure( const aSender: TKnobsWirePanel)                                                         of object;
  TKnobsOnEditLabelChanged   = procedure( const aSender: TObject; const aValue: string)                                           of object;
  TKnobsOnTextChanged        = procedure( const aSender: TObject; const aPath, aValue: string)                                    of object;
  TOnCreateModuleBitmap      = function ( const aModuleType: TKnobsModuleType; UseCache: Boolean): TBitmap                        of object;
  TOnReadModuleComment       = function ( const aModuleType: TKnobsModuleType): string                                            of object;
  TOnShowPopup               = procedure( const aSender: TKnobsWirePanel; const aControl: TControl)                               of object;
  TOnUnfocus                 = procedure( const aSender: TObject)                                                                 of object;


  TKnobsUMConnectionChange = record
    Msg         : Cardinal;
    Source      : TKnobsConnector;
    Destination : TKnobsConnector;
    Result      : LRESULT;
  end;
  TKnobsUMConnectionAdd   = TKnobsUMConnectionChange;
  TKnobUMConnectionRemove = TKnobsUMConnectionChange;


  TKnobsUMColorChange = record
    Msg    : Cardinal;
    Source : TKnobsConnector;
    Color  : TColor;
    Result : LRESULT;
  end;


  TKnobsUMHighlight = record
    Msg         : Cardinal;
    Source      : TKnobsConnector;
    NotUsed2    : LPARAM;
    Result      : LRESULT;
  end;


  TKnobsUMButtonUp = record
    Msg         : Cardinal;
    Source      : TKnobsModuleButton;
    NotUsed     : LPARAM;
    Result      : LRESULT;
  end;


  IKnobsPatchReader = interface
    procedure   ReadString(
      const aData      : string;
      const aWirePanel : TKnobsWirePanel;
      aDistance        : TPoint;
      aReadMode        : TKnobsReadMode = rmReplace;
      const aFileName  : string         = ''
    );

    procedure   ReadFile(
      const aFileName  : string;
      const aWirePanel : TKnobsWirePanel;
      aReadMode        : TKnobsReadMode = rmReplace
    );

    procedure   ReadFromClipboard(
      const aWirePanel : TKnobsWirePanel;
      aReadMode        : TKnobsReadMode = rmReplace
    );

    procedure   ParamsFromClipboard(
      const aWirePanel : TKnobsWirePanel
    );
  end;


  IKnobsPatchWriter = interface
    function    WriteString(
      const aWirePanel : TKnobsWirePanel;
      const aFileName  : string = '';
      aWriteMode       : TKnobsWriteMode = wmAll
    ): string;

    procedure   WriteFile(
      const aWirePanel : TKnobsWirePanel;
      const aFileName  : string;
      aWriteMode       : TKnobsWriteMode = wmAll
    );

    procedure   WriteToClipboard(
      const aWirePanel : TKnobsWirePanel;
      aWriteMode       : TKnobsWriteMode = wmAll
    );
  end;


  TKnobsLed = class( TGraphicControl)
  private
    FTime       : Cardinal;
    FTimer      : TTimer;
    FColorOff   : TColor;
    FColorOn    : TColor;
    FHighLight  : Boolean;
    FActive     : Boolean;
    FShape      : TKnobsLedShape;
  private
    procedure   TimerFired( aSender: TObject);
    procedure   StartTimer( aValue: Cardinal);
    procedure   StopTimer;
    procedure   SetTime     ( aValue: Cardinal      );
    procedure   SetActive   ( aValue: Boolean       );
    procedure   SetColorOn  ( aValue: TColor        );
    procedure   SetColorOff ( aValue: TColor        );
    procedure   SetHighLight( aValue: Boolean       );
    procedure   SetShape    ( aValue: TKnobsLedShape);
  protected
    procedure   Paint;                       override;
  public
    constructor Create( aOwner: TComponent); override;
    destructor  Destroy;                     override;
    procedure   Toggle;
  published
    property    Time       : Cardinal       read FTime       write SetTime       default 0;
    property    Active     : Boolean        read FActive     write SetActive     default True;
    property    ColorOn    : TColor         read FColorOn    write SetColorOn    default clLime;
    property    ColorOff   : TColor         read FColorOff   write SetColorOff   default clGreen;
    property    HighLight  : Boolean        read FHighLight  write SetHighLight  default True;
    property    Shape      : TKnobsLedShape read FShape      write SetShape      default lsRound;
    property    Height                                                           default 15;
    property    Width                                                            default 15;
    property    Hint;
    property    ShowHint;
    property    ParentShowHint                                                   default True;
    property    Dragcursor;
    property    Dragmode;
    property    OnDragDrop;
    property    OnDragOver;
    property    OnEndDrag;
    property    OnMouseDown;
    property    OnMouseMove;
    property    OnMouseUp;
  end;


  TKnobsClicker = class( TSpeedButton)
  // Repeating speed button, semi flat ...
  private
    FRepeatTimer : TTimer;
    FRepeatCount : Integer;
    FCanRepeat   : Boolean;
    FMouseDown   : Boolean;
    FShowFocus   : Boolean;
    FBitmap      : TBitmap;
  private
    procedure   TimerExpired( aSender: TObject);
    procedure   SetShowFocus( aValue: Boolean);
  private
    function    FindWirePanel: TKnobsWirePanel;
  protected
    procedure   MouseDown( Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure   MouseUp  ( Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure   AssignBitmap( aBitmap: TBitmap);
  public
    constructor Create( anOwner: TComponent);                                        override;
    destructor  Destroy;                                                             override;
    procedure   Click;                                                               override;
  private
    property    ShowFocus: Boolean read FShowFocus write SetShowFocus;
  public
    property    CanRepeat: Boolean read FCanRepeat write FCanRepeat default True;
  end;


  TKnobsValuedControl = class( TCustomControl)
  // Base type for valued controls, ones that can have an effect on an external model
  private
    FFriendlyName     : string;
    FStepCount        : Integer;             // Number of steps in the total knob travel
    FLocked           : Boolean;
    FKnobPosition     : Integer;             // Current knob position
    FDefaultPosition  : Integer;             // default knob position;
    FControlType      : string;              // The control's behaviour
    FOnValueChanged   : TKnobsOnValueChange; // Called with the new value
    FDisplay          : TKnobsDisplay;
    FPopupCount       : Integer;
    FPopupWindow      : THintWindow;
    FWheelSupport     : Boolean;
    FWheelSensitivity : Integer;
    FWheelCounter     : Integer;
    FOnUnFocus        : TOnUnFocus;
  private
    procedure   CommonKnobPosition;
    procedure   FixKnobPosition;
    procedure   FinalizeKnobPosition;
  private
    procedure   SetKnobPosition   ( aValue: Integer);                   virtual;
    procedure   SetDefaultPosition( aValue: Integer);                   virtual;
    procedure   SetStepCount      ( aValue: Integer);                   virtual;
    procedure   SetLocked         ( aValue: Boolean);
    procedure   SetControlType    ( const aValue: string);
    function    GetAsHint         : string;
    function    GetAsDisplay      : string;
    function    GetAsValue        : TSignal;
    procedure   SetOnValueChanged ( aValue: TKnobsOnValueChange);
    procedure   SetDisplay        ( aValue: TKnobsDisplay);
    procedure   NotifyPeers;                                            virtual;
    procedure   DoWheel( aSender: TObject; aShift: TShiftState; aWheelDelta: Integer; aMousePos: TPoint; var Handled: Boolean);
  protected
    procedure   ValueChanged( IsFinal: Boolean);                        virtual;
    function    FindModule   : TKnobsCustomModule;
    function    FindWirePanel: TKnobsWirePanel;
    procedure   BeginStateChange;
    procedure   EndStateChange;
    procedure   ActivatePopup;
    procedure   ShowPopup;
    procedure   HidePopup;
    procedure   ForceHidePopup;
    procedure   HandleWheelUp  ( anAmount: Integer);                    virtual;
    procedure   HandleWheelDown( anAmount: Integer);                    virtual;
    procedure   UnFocus;                                                virtual;
  public
    constructor Create( anOwner: TComponent);                          override;
    procedure   SetDefault;
    procedure   FixDisplay;
    procedure   Dump( var aFile: TextFile; anInd: Integer);
  public
    property    AsDisplay        : string          read GetAsDisplay;
    property    AsHint           : string          read GetAsHint;
    property    AsValue          : TSignal         read GetAsValue;
    property    WheelSupport     : Boolean         read FWheelSupport     write FWheelSupport;
    property    WheelSensitivity : Integer         read FWheelSensitivity write FWheelSensitivity;
    property    WheelCounter     : Integer         read FWheelCounter     write FWheelCounter;
  published
    property    Action;
    property    FriendlyName     : string          read FFriendlyName     write FFriendlyName;
    // NOTE: declaration order is important here to avoid loss of knob resolution on a load action ...
    property    Display          : TKnobsDisplay   read FDisplay          write SetDisplay;
    property    StepCount        : Integer         read FStepCount        write SetStepCount       default 1;
    property    KnobPosition     : Integer         read FKnobPosition     write SetKnobPosition    default 0;
    property    DefaultPosition  : Integer         read FDefaultPosition  write SetDefaultPosition default 0;
    property    ControlType      : string          read FControlType      write SetControlType              ;
    property    Locked           : Boolean         read FLocked           write SetLocked                   ;
    // END NOTE.
    property    TabOrder;
    property    TabStop;
    property    Visible;
    property    OnValueChanged  : TKnobsOnValueChange  read FOnValueChanged  write SetOnValueChanged;
    property    OnUnFocus       : TOnUnFocus           read FOnUnFocus       write FOnUnFocus;
  end;


  TKnobsXYControl = class( TCustomControl)
  // Base type for two valued controls, ones that can have an effect on an external model
  private
    FFriendlyName     : string;
    FStepCount        : Integer;             // Number of steps in the total knob travel, same for X and Y directions
    FKnobPositionX    : Integer;             // Current knob X position
    FKnobPositionY    : Integer;             // Current knob Y position
    FDefaultPositionX : Integer;             // default knob X position;
    FDefaultPositionY : Integer;             // default knob Y position;
    FControlType      : string;              // The control's behaviour
    FOnValueChanged   : TKnobsOnValueChange; // Called with the new value - twice, once for X, once for Y
    FPopupCount       : Integer;
    FPopupWindow      : THintWindow;
    FOnUnFocus        : TOnUnFocus;
  private
    procedure   CommonKnobPosition;
    procedure   FixKnobPosition;
    procedure   FinalizeKnobPosition;
  private
    procedure   SetKnobPositionX   ( aValue: Integer);                   virtual;
    procedure   SetKnobPositionY   ( aValue: Integer);                   virtual;
    procedure   SetDefaultPositionX( aValue: Integer);                   virtual;
    procedure   SetDefaultPositionY( aValue: Integer);                   virtual;
    procedure   SetStepCount       ( aValue: Integer);                   virtual;
    procedure   SetControlType     ( const aValue: string);
    function    GetAsHint          : string;
    function    GetAsValueX        : TSignal;
    function    GetAsValueY        : TSignal;
    procedure   SetOnValueChanged  ( aValue: TKnobsOnValueChange);
    procedure   NotifyPeers;                                            virtual;
  protected
    procedure   ValueChanged( IsFinal: Boolean);                        virtual;
    function    FindModule   : TKnobsCustomModule;
    function    FindWirePanel: TKnobsWirePanel;
    procedure   BeginStateChange;
    procedure   EndStateChange;
    procedure   ActivatePopup;
    procedure   ShowPopup;
    procedure   HidePopup;
    procedure   ForceHidePopup;
    procedure   UnFocus;                                                virtual;
  public
    constructor Create( anOwner: TComponent);                          override;
    procedure   SetDefault;
    procedure   Dump( var aFile: TextFile; anInd: Integer);
  public
    property    AsHint           : string          read GetAsHint;
    property    AsValueX         : TSignal         read GetAsValueX;
    property    AsValueY         : TSignal         read GetAsValueY;
  published
    property    Action;
    property    FriendlyName     : string          read FFriendlyName      write FFriendlyName;
    // NOTE: declaration order is important here to avoid loss of resolution on a load action ...
    property    StepCount        : Integer         read FStepCount         write SetStepCount        default 1;
    property    KnobPositionX    : Integer         read FKnobPositionX     write SetKnobPositionX    default 0;
    property    KnobPositionY    : Integer         read FKnobPositionY     write SetKnobPositionY    default 0;
    property    DefaultPositionX : Integer         read FDefaultPositionX  write SetDefaultPositionX default 0;
    property    DefaultPositionY : Integer         read FDefaultPositionY  write SetDefaultPositionY default 0;
    property    ControlType      : string          read FControlType       write SetControlType              ;
    // END NOTE.
    property    TabOrder;
    property    TabStop;
    property    Visible;
    property    OnValueChanged  : TKnobsOnValueChange  read FOnValueChanged  write SetOnValueChanged;
    property    OnUnFocus       : TOnUnFocus           read FOnUnFocus       write FOnUnFocus;
  end;


  TKnobsSelector = class( TKnobsValuedControl)
  private
    FCaptions       : TStringList;
    FGlyphs         : TImageList;
    FBorderColor    : TColor;
    FMouseInControl : Boolean;
    FOlDFontColor   : TColor;
    FOldFontStyle   : TFontStyles;
    FOldColor       : TColor;
  private
    procedure   SetCaptions   ( const aValue: TStringList);
    procedure   SetGlyphs     ( const aValue: TImageList );
    procedure   SetBorderColor( aValue: TColor);
  private
    procedure   WMSetFocus  ( var aMessage: TWMSetFocus  );  message WM_SETFOCUS;
    procedure   WMKillFocus ( var aMessage: TWMKillFocus );  message WM_KILLFOCUS;
    procedure   WMGetDlgCode( var aMessage: TWMGetDlgCode);  message WM_GETDLGCODE;
    procedure   CMMouseEnter( var aMessage: TMessage);       message CM_MOUSEENTER;
    procedure   CMMouseLeave( var aMessage: TMessage);       message CM_MOUSELEAVE;
  protected
    procedure   DoClick( aSender: TObject);
    procedure   MouseDown( Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure   Paint;                                       override;
  public
    constructor Create( anOwner: TComponent);                override;
    destructor  Destroy;                                     override;
    procedure   FixPeers;
  published
    property    Captions    : TStringList read FCaptions    write SetCaptions;
    property    Glyphs      : TImageList  read FGlyphs      write SetGlyphs;
    property    Color                                                             default clGray;
    property    BorderColor : TColor      read FBorderColor write SetBorderColor  default clWhite;
    property    Font;
  end;


  TKnobsKnob = class( TKnobsValuedControl)
  // A rotary knob, or linear horizontal or vertical control, mouse wheel works too.
  private
    FDownClicker       : TKnobsClicker;
    FUpClicker         : TKnobsClicker;
    FMouseDown         : Boolean;
    FGraphic           : TBitmap;
    FMouseInControl    : Boolean;
    FTransparent       : Boolean;
    FPairedWith        : TKnobsKnob;
    FPairingMode       : TKnobsPairingMode;
    FPairingLock       : Integer;
    FControlMode       : TDistanceMode;
    FFinalizationTimer : TTimer;
    FTresholdPast      : Boolean;
  private
    procedure   InitializeFinalizationTimer;
    procedure   FinalizeFinalizationTimer;
    procedure   StartFinalizationTimer;
    procedure   StopFinalizationTimer;
    procedure   FinalizationTimerFired( aSender: TObject);
  private
    procedure   WMSetFocus  ( var aMessage: TWMSetFocus  );  message WM_SETFOCUS;
    procedure   WMKillFocus ( var aMessage: TWMKillFocus );  message WM_KILLFOCUS;
    procedure   WMGetDlgCode( var aMessage: TWMGetDlgCode);  message WM_GETDLGCODE;
    procedure   CMMouseEnter( var aMessage: TMessage);       message CM_MOUSEENTER;
    procedure   CMMouseLeave( var aMessage: TMessage);       message CM_MOUSELEAVE;
    procedure   OnDownClick( aSender: TObject);
    procedure   OnUpClick  ( aSender: TObject);
    procedure   NotifyPeers;                                           override;
    procedure   LockPairing;
    procedure   UnlockPairing;
    function    PairingLocked: Boolean;
    procedure   FixPairing;
  private
    procedure   SetTransparent    ( aValue: Boolean          );
    procedure   SetKnobPosition   ( aValue: Integer          );        override;
    procedure   SetPairedWith     ( const aValue: TKnobsKnob );
    procedure   SetPairingMode    ( aValue: TKnobsPairingMode);
    procedure   SetControlMode    ( aValue: TDistanceMode    );         virtual;
  private
    function    CenterPoint: TPoint;
    procedure   BeginMove ( aPoint: TPoint);                           virtual;
    procedure   EndMove   ( aPoint: TPoint);                           virtual;
    procedure   UpdateMove( aPoint: TPoint);                           virtual;
  protected
    function    CreateClicker( aLeft, aTop: Integer; aNotifyEvent: TNotifyEvent; aBitmap: TBitmap): TKnobsClicker; virtual;
    procedure   Paint;                                                 override;
    procedure   MouseDown( Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure   MouseUp  ( Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure   MouseMove( Shift: TShiftState; X, Y: Integer);         override;
    procedure   HandleWheelUp  ( anAmount: Integer);                   override;
    procedure   HandleWheelDown( anAmount: Integer);                   override;
    procedure   KeyDown( var Key: Word; Shift: TShiftState);           override;
    procedure   AssignGraphic;                                          virtual;
    procedure   Loaded;                                                override;
    function    WheelAmount: Integer;                                   virtual;
    procedure   SetSize;                                                virtual;
    procedure   CreateClickers;                                         virtual;
    procedure   InvalidateDisplay;
  public
    constructor Create( anOwner: TComponent);                          override;
    destructor  Destroy;                                               override;
  published
    property    Action;
    property    Transparent  : Boolean           read FTransparent   write SetTransparent default False;
    property    Color;
    property    TabOrder;
    property    TabStop;
    property    ParentColor                                                               default True;
    property    PairedWith   : TKnobsKnob        read FPairedWith    write SetPairedWith;
    property    PairingMode  : TKnobsPairingMode read FPairingMode   write SetPairingMode;
    property    ControlMode  : TDistanceMode     read FControlMode   write SetControlMode;
  end;


  TKnobsSmallKnob = class( TKnobsKnob)
  // A bit smaller than a TKnobsKnob
  protected
    procedure   AssignGraphic;                                         override;
  end;


  TKnobsNoKnob = class( TKnobsKnob)
  // No knob at all, just the clickers
  protected
    procedure   AssignGraphic;                                         override;
  public
    constructor Create( anOwner: TComponent);                          override;
  end;


  TKnobsSlider = class( TKnobsKnob)
  // A vertical slider type of control
  private
    procedure   SetControlMode( aValue: TDistanceMode);                override;
    procedure   DoResize( aSender: TObject);
  protected
    procedure   AssignGraphic;                                         override;
    function    CreateClicker( aLeft, aTop: Integer; aNotifyEvent: TNotifyEvent; aBitmap: TBitmap): TKnobsClicker; override;
    procedure   Paint;                                                 override;
    procedure   SetSize;                                               override;
    procedure   CreateClickers;                                        override;
  public
    constructor Create( anOwner: TComponent);                          override;
  end;


  TKnobsPad = class( TKnobsXYControl)
  // A pad control horizontal and vertical control, no mouse wheel
  private
    FBackGround        : TBitmap;
    FBorderColor       : TColor;
    FMouseDown         : Boolean;
    FMouseInControl    : Boolean;
    FFinalizationTimer : TTimer;
    FTresholdPast      : Boolean;
  private
    procedure   SetBackGround ( aValue: TBitmap);
    procedure   SetBorderColor( aValue: TColor );
  private
    procedure   InitializeFinalizationTimer;
    procedure   FinalizeFinalizationTimer;
    procedure   StartFinalizationTimer;
    procedure   StopFinalizationTimer;
    procedure   FinalizationTimerFired( aSender: TObject);
  private
    procedure   WMSetFocus  ( var aMessage: TWMSetFocus  );                          message WM_SETFOCUS;
    procedure   WMKillFocus ( var aMessage: TWMKillFocus );                          message WM_KILLFOCUS;
    procedure   WMGetDlgCode( var aMessage: TWMGetDlgCode);                          message WM_GETDLGCODE;
    procedure   CMMouseEnter( var aMessage: TMessage);                               message CM_MOUSEENTER;
    procedure   CMMouseLeave( var aMessage: TMessage);                               message CM_MOUSELEAVE;
  private
    function    CenterPoint: TPoint;
    procedure   BeginMove ( aPoint: TPoint);                                         virtual;
    procedure   EndMove   ( aPoint: TPoint);                                         virtual;
    procedure   UpdateMove( aPoint: TPoint);                                         virtual;
  protected
    procedure   Paint;                                                               override;
    procedure   MouseDown( Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure   MouseUp  ( Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure   MouseMove( Shift: TShiftState; X, Y: Integer);                       override;
    procedure   KeyDown( var Key: Word; Shift: TShiftState);                         override;
    procedure   Loaded;                                                              override;
  public
    constructor Create( anOwner: TComponent);                                        override;
    destructor  Destroy;                                                             override;
  published
    property    Background  : TBitmap read FBackGround  write SetBackGround;
    property    BorderColor : TColor  read FBorderColor write SetBorderColor;
    property    Action;
    property    Color;
    property    TabOrder;
    property    TabStop;
    property    ParentColor default True;
  end;


  TKnobsValuedButton = class( TKnobsValuedControl)
  private
    FButton     : TSpeedButton;
    FColorFocus : TColor;
    FUserColor  : TColor;
    FUserFlat   : Boolean;
  private
    procedure   SetKnobPosition    ( aValue: Integer);                 override;
    function    GetCaption         : string;
    procedure   SetCaption         ( const aValue: string);
    function    GetGlyph           : TBitmap;
    procedure   SetGlyph           ( const aValue: TBitmap);
    function    GetHint            : string;
    procedure   SetHint            ( const aValue: string);
    function    GetGroupIndex      : Integer;
    procedure   SetGroupIndex      ( aValue: Integer);
    function    GetAllowAllUp      : Boolean;
    procedure   SetAllowAllUp      ( aValue: Boolean);
  private
    function    GetAlignWithMargins: Boolean;
    procedure   SetAlignWithMargins( aValue: Boolean);
    function    GetFlat            : Boolean;
    procedure   SetFlat            ( aValue: Boolean);
    function    GetFont            : TFont;
    procedure   SetFont            ( const aValue: TFont);
    function    GetLayout          : TButtonLayout;
    procedure   SetLayout          ( aValue: TButtonLayout);
    function    GetMargin          : Integer;
    procedure   SetMargin          ( aValue: Integer);
    function    GetMargins         : TMargins;
    procedure   SetMargins         ( aValue: TMargins);
    function    GetNumGlyphs       : TNumGlyphs;
    procedure   SetNumGlyphs       ( aValue: TNumGlyphs);
    function    GetParentFont      : Boolean;
    procedure   SetParentFont      ( aValue: Boolean);
    function    GetParentShowHint  : Boolean;
    procedure   SetParentShowHint  ( aValue: Boolean);
    function    GetShowHint        : Boolean;
    procedure   SetShowHint        ( aValue: Boolean);
    function    Gettransparent     : Boolean;
    procedure   SetTransparent     ( aValue: Boolean);
    function    GetVisible         : Boolean;
    procedure   SetVisible         ( aValue: Boolean);
  private
    procedure   OnButtonClick( aSender: TObject);
  protected
    procedure   WMSetFocus  ( var aMessage: TWMSetFocus  );            message WM_SETFOCUS;
    procedure   WMKillFocus ( var aMessage: TWMKillFocus );            message WM_KILLFOCUS;
    procedure   WMGetDlgCode( var aMessage: TWMGetDlgCode);            message WM_GETDLGCODE;
    procedure   KeyDown     ( var Key: Word; Shift: TShiftState);      override;
    procedure   SetBiDiMode        ( aValue: TBiDiMode);               override;
    procedure   SetParentBiDiMode  ( aValue: Boolean);                 override;
    function    GetEnabled         : Boolean;                          override;
    procedure   SetEnabled         ( aValue: Boolean);                 override;
  protected
    procedure   Resize;                                                override;
    procedure   Loaded;                                                override;
    procedure   SetDefaults;                                            virtual;
  public
    constructor Create( anOwner: TComponent);                          override;
    destructor  Destroy;                                               override;
  published
    property    Caption          : string        read GetCaption          write SetCaption                               ;
    property    Glyph            : TBitmap       read GetGlyph            write SetGlyph                                 ;
    property    Hint             : string        read GetHint             write SetHint                                  ;
    property    GroupIndex       : Integer       read GetGroupIndex       write SetGroupIndex       default             1;
    property    AllowAllUp       : Boolean       read GetAllowAllUp       write SetAllowAllUp       default         false;
    property    Height                                                                              default            18;
    property    Width                                                                               default            18;
    property    AlignWithMargins : Boolean       read GetAlignWithMargins write SetAlignWithMargins default         False;
    property    Color                                                                               default      clSilver;
    property    ColorFocus       : TColor        read FColorFocus         write FColorFocus         default     clSkyBlue;
    property    Flat             : Boolean       read GetFlat             write SetFlat             default         False;
    property    Font             : TFont         read GetFont             write SetFont                                  ;
    property    Layout           : TButtonLayout read GetLayout           write SetLayout           default   blGlyphLeft;
    property    Margin           : Integer       read GetMargin           write SetMargin           default            -1;
    property    Margins          : TMargins      read GetMargins          write SetMargins                               ;
    property    NumGlyphs        : TNumGlyphs    read GetNumGlyphs        write SetNumGlyphs        default             1;
    property    ParentFont       : Boolean       read GetParentFont       write SetParentFont       default          True;
    property    ParentShowHint   : Boolean       read GetParentShowHint   write SetParentShowHint   default          True;
    property    ShowHint         : Boolean       read GetShowHint         write SetShowHint         default         False;
    property    Transparent      : Boolean       read GetTransparent      write SetTransparent      default          True;
    property    Visible          : Boolean       read GetVisible          write SetVisible          default          True;
  end;


  TKnobsValuedButtons = class( TKnobsValuedControl)
  private
    FButtons          : array of TSpeedButton;
    FCaptions         : TStrings;
    FImageIndices     : TStrings;
    FGlyphs           : TImageList;
    FHints            : TStrings;
    FGroupIndex       : Integer;
    FAlignWithMargins : Boolean;
    FAllowAllUp       : Boolean;
    FFlat             : Boolean;
    FFont             : TFont;
    FLayout           : TButtonLayout;
    FDisplayMode      : TKnobsButtonsDisplayMode;
    FMargin           : Integer;
    FMargins          : TMargins;
    FNumGlyphs        : TNumGlyphs;
    FParentFont       : Boolean;
    FParentShowHint   : Boolean;
    FShowHint         : Boolean;
    FTransparent      : Boolean;
    FVisible          : Boolean;
    FColorFocus       : TColor;
    FUserColor        : TColor;
    FUserFlat         : Boolean;
  private
    procedure   FixDimensions;
    procedure   FixGlyph( anIndex: Integer; aBitmap: TBitmap);
    function    FindGlyphIndex( anIndex: Integer): Integer;
  private
    procedure   FixAlignWithMargins;
    procedure   FixBiDiMode;
    procedure   FixCaptions;
    procedure   FixEnabled;
    procedure   FixFlat;
    procedure   FixFont;
    procedure   FixGlyphs;
    procedure   FixHints;
    procedure   FixLayout;
    procedure   FixMargin;
    procedure   FixMargins;
    procedure   FixNumGlyphs;
    procedure   FixParentFont;
    procedure   FixParentShowHint;
    procedure   FixParentBiDiMode;
    procedure   FixShowHint;
    procedure   FixTransparent;
    procedure   FixVisible;
    procedure   FixMode( aNewStepCount: Integer);
  private
    procedure   SetKnobPosition    ( aValue: Integer);                 override;
    procedure   SetStepCount       ( aValue: Integer);                 override;
    function    GetCaptions        : TStrings;
    procedure   SetCaptions        ( const aValue: TStrings);
    function    GetImageIndices    : TStrings;
    procedure   SetImageIndices    ( const aValue: TStrings);
    function    GetGlyphs          : TImageList;
    procedure   SetGlyphs          ( const aValue: TImageList);
    function    GetHints           : TStrings;
    procedure   SetHints           ( const aValue: TStrings);
    procedure   SetGroupIndex      ( aValue: Integer);
    procedure   SetAllowAllUp      ( aValue: Boolean);
  private
    procedure   SetAlignWithMargins( aValue: Boolean);
    procedure   SetFlat            ( aValue: Boolean);
    procedure   SetFont            ( const aValue: TFont);
    procedure   SetLayout          ( aValue: TButtonLayout);
    procedure   SetDisplayMode     ( aValue: TKnobsButtonsDisplayMode);
    procedure   SetMargin          ( aValue: Integer);
    procedure   SetMargins         ( aValue: TMargins);
    procedure   SetNumGlyphs       ( aValue: TNumGlyphs);
    procedure   SetParentFont      ( aValue: Boolean);
    procedure   SetParentShowHint  ( aValue: Boolean);
  private
    procedure   SetShowHint        ( aValue: Boolean);
    procedure   SetTransparent     ( aValue: Boolean);
    procedure   SetVisible         ( aValue: Boolean);
  private
    procedure   OnButtonClick( aSender: TObject);
    procedure   CreateButton;
  protected
    procedure   WMSetFocus  ( var aMessage: TWMSetFocus  );            message WM_SETFOCUS;
    procedure   WMKillFocus ( var aMessage: TWMKillFocus );            message WM_KILLFOCUS;
    procedure   WMGetDlgCode( var aMessage: TWMGetDlgCode);            message WM_GETDLGCODE;
    procedure   KeyDown     ( var Key: Word; Shift: TShiftState);      override;
    procedure   SetBiDiMode        ( aValue: TBiDiMode);               override;
    procedure   SetParentBiDiMode  ( aValue: Boolean);                 override;
    procedure   SetEnabled         ( aValue: Boolean);                 override;
  protected
    procedure   Resize;                                                override;
    procedure   Loaded;                                                override;
    procedure   FixButtons;                                             virtual;
    procedure   SetDefaults;                                            virtual;
  public
    constructor Create( anOwner: TComponent);                          override;
    destructor  Destroy;                                               override;
  published
    property    StepCount                                                                                    default             2;
    property    Captions         : TStrings                 read GetCaptions       write SetCaptions                              ;
    property    ImageIndices     : TStrings                 read GetImageIndices   write SetImageIndices                          ;
    property    Glyphs           : TImageList               read GetGlyphs         write SetGlyphs                                ;
    property    Hints            : TStrings                 read GetHints          write SetHints                                 ;
    property    GroupIndex       : Integer                  read FGroupIndex       write SetGroupIndex       default             1;
    property    AllowAllUp       : Boolean                  read FAllowAllUp       write SetAllowAllUp       default         false;
    property    Height                                                                                       default            25;
    property    Width                                                                                        default            51;
    property    AlignWithMargins : Boolean                  read FAlignWithMargins write SetAlignWithMargins default         False;
    property    Color                                                                                        default      clSilver;
    property    ColorFocus       : TColor                   read FColorFocus       write FColorFocus         default     clSkyBlue;
    property    Flat             : Boolean                  read FFlat             write SetFlat             default         False;
    property    Font             : TFont                    read FFont             write SetFont                                  ;
    property    Layout           : TButtonLayout            read FLayout           write SetLayout           default   blGlyphLeft;
    property    DisplayMode      : TKnobsButtonsDisplayMode read FDisplayMode      write SetDisplayMode      default bdmHorizontal;
    property    Margin           : Integer                  read FMargin           write SetMargin           default            -1;
    property    Margins          : TMargins                 read FMargins          write SetMargins                               ;
    property    NumGlyphs        : TNumGlyphs               read FNumGlyphs        write SetNumGlyphs        default             1;
    property    ParentFont       : Boolean                  read FParentFont       write SetParentFont       default          True;
    property    ParentShowHint   : Boolean                  read FParentShowHint   write SetParentShowHint   default          True;
    property    ShowHint         : Boolean                  read FShowHint         write SetShowHint         default         False;
    property    Transparent      : Boolean                  read FTransparent      write SetTransparent      default          True;
    property    Visible          : Boolean                  read FVisible          write SetVisible          default          True;
  end;


  TKnobsDisplayEditor = class( TMemo)
  // An multi line editor for TDisplay
  private
    FDisplay : TKnobsDisplay;
  private
    procedure   WMKillFocus( var aMessage: TWMKillFocus );  message WM_KILLFOCUS;
    procedure   Do_Exit   ( aSender: TObject);
    procedure   Do_KeyDown( aSender: TObject; var aKey: Word; aShift: TShiftState);
  protected
    procedure   SetDisplay( aDisplay: TKnobsDisplay);
  public
    constructor Create( anOwner: TComponent);                          override;
    destructor  Destroy;                                               override;
  end;


  TKnobsTextControl = class( TStaticText)
  private
    FMustSave   : Boolean;
    FMustNotify : Boolean;
    FOnChanged  : TKnobsOnTextChanged;
    FSavedColor : TColor;
  private
    function    FindWirePanel: TKnobsWirePanel;
    function    GetModule: TKnobsCustomModule;
    procedure   Do_KeyDown( aSender: TObject; var aKey: Word; aShift: TShiftState);
  protected
    procedure   HandleReturnKey;                                                      virtual; abstract;
    procedure   HandleDoubleClick;                                                    virtual; abstract;
    function    GetTextValue: string;                                                 virtual; abstract;
    procedure   SetTextValue( const aValue: string);                                  virtual; abstract;
    procedure   SetText( const aValue: string);
    procedure   TextChanged( const aNewValue: string; DoNotify: Boolean);             virtual;
    procedure   HandleTextChange( const aValue: string);                              virtual;
    procedure   MouseDown( Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure   WMSetFocus ( var aMessage: TWMSetFocus );                             message WM_SETFOCUS;
    procedure   WMKillFocus( var aMessage: TWMKillFocus);                             message WM_KILLFOCUS;
  public
    constructor Create( anOwner: TComponent);                                        override;
  public
    property    Module: TKnobsCustomModule  read GetModule;
    property    TextValue: string           read GetTextValue write SetTextValue;
  published
    property    Ctl3D                                                                  default False;
    property    ParentColor                                                            default False;
    property    Color                                                                  default KnobsColorDisplay;
    property    Width                                                                  default 33;
    property    Height                                                                 default 15;
    property    AutoSize                                                               default False;
    property    BorderStyle                                                            default sbsSingle;
    property    TabStop                                                                default True;
    property    MustSave    : Boolean             read FMustSave     write FMustSave   default False;
    property    MustNotify  : Boolean             read FMustNotify   write FMustNotify default False;
    property    OnChanged   : TKnobsOnTextChanged read FOnChanged    write FOnChanged;
  end;


  TKnobsDisplay = class( TKnobsTextControl)
  // Can be linked to aTKnob for auto update
  // Can have an editor to manually edit the value
  private
    FEditor          : TKnobsDisplayEditor;
    FController      : TKnobsValuedControl;
    FHasEditor       : Boolean;
    FHasParentEditor : Boolean;
    FHasScrollbar    : Boolean;
  protected
    procedure   HandleReturnKey;                                                     override;
    procedure   HandleDoubleClick;                                                   override;
    function    GetTextValue: string;                                                override;
    procedure   SetTextValue( const aValue: string);                                 override;
    procedure   TextChanged( const aNewValue: string; DoNotify: Boolean);            override;
    procedure   WMNCHitTest( var aMessage: TWMNCHitTest );                           message WM_NCHITTEST;
    procedure   LockKnob;
  public
    constructor Create( anOwner: TComponent);                                        override;
    destructor  Destroy;                                                             override;
  public
    property    Editor: TKnobsDisplayEditor read FEditor   write FEditor;
  published
    property    HasEditor       : Boolean read FHasEditor       write FHasEditor       default False;
    property    HasParentEditor : Boolean read FHasParentEditor write FHasParentEditor default False;
    property    HasScrollbar    : Boolean read FHasScrollbar    write FHasScrollbar    default False;
  end;


  TKnobsFileSelector = class( TKnobsTextControl)
  private
    FFileSelector : TOpenDialog;
    FDataFileName : string;
    FDataFileExt  : string;
    FDataFileMsg  : string;
  private
    procedure   SetDataFileName( const aValue: string);
    procedure   FixCaption;
    procedure   SelectFile;
  protected
    procedure   HandleReturnKey;                                                     override;
    procedure   HandleDoubleClick;                                                   override;
    function    GetTextValue: string;                                                override;
    procedure   SetTextValue( const aValue: string);                                 override;
  public
    constructor Create( anOwner: TComponent);                                        override;
    destructor  Destroy;                                                             override;
  published
    property    DataFileName : string read FDataFileName write SetDataFileName;
    property    DataFileExt  : string read FDataFileExt  write FDataFileExt;
    property    DataFileMsg  : string read FDataFileMsg  write FDataFileMsg;
  end;


  TKnobsIndicator = class( TGraphicControl)
  // an LED
  private
    FActive     : Boolean;
    FActiveBm   : TBitmap;
    FInActiveBm : TBitmap;
  private
    procedure   SetActive        ( aValue: Boolean);
    procedure   SetBitmapActive  ( aValue: TBitmap);
    procedure   SetBitmapInactive( aValue: TBitmap);
  protected
    procedure   AssignBitmaps;                                          virtual;
    procedure   Paint;                                                 override;
  public
    constructor Create( anOwner: TComponent);                          override;
  public
    property    BitmapActive   : TBitmap read FActiveBm   write SetBitmapActive;
    property    BitmapInactive : TBitmap read FInactiveBm write SetBitmapInactive;
  published
    property    Active         : Boolean read FActive     write SetActive;
    property    Visible;
    property    ShowHint;
  end;


  TKnobsIndicatorBar = class( TGraphicControl)
  // A bar of indicators, horizontal or vertical
  private
    FIndicatorCount : Integer;
    FValue          : TSignal;
    FPeakValue      : TSignal;
    FValeyValue     : TSignal;
    FShowPeakValue  : Boolean;
    FShowValeyValue : Boolean;
    FValueLowMid    : TSignal;
    FValueMidHigh   : TSignal;
    FDisplayType    : TKnobsDisplayType;
    FOrientation    : TKnobsDisplayOrientation;
    FSpacing        : Integer;
    FActiveLowBm    : TBitmap;
    FInActiveLowBm  : TBitmap;
    FActiveMidBm    : TBitmap;
    FInActiveMidBm  : TBitmap;
    FActiveHighBm   : TBitmap;
    FInActiveHighBm : TBitmap;
    FPeakColor      : TColor;
    FValeyColor     : TColor;
  private
    function    GetMaxBmWidth : Integer;
    function    GetMaxBmHeight: Integer;
    procedure   FixDimensions;
  private
    procedure   SetIndicatorCount( aValue: Integer            );
    procedure   SetValue         ( aValue: TSignal            );
    procedure   SetPeakValue     ( aValue: TSignal            );
    procedure   SetValeyValue    ( aValue: TSignal            );
    procedure   SetShowPeakValue ( aValue: Boolean            );
    procedure   SetShowValeyValue( aValue: Boolean            );
    procedure   SetValueLowMid   ( aValue: TSignal            );
    procedure   SetValueMidHigh  ( aValue: TSignal            );
    procedure   SetDisplayType   ( aValue: TKnobsDisplayType       );
    procedure   SetOrientation   ( aValue: TKnobsDisplayOrientation);
    procedure   SetSpacing       ( aValue: Integer            );
    procedure   SetActiveLowBm   ( aValue: TBitmap            );
    procedure   SetInActiveLowBm ( aValue: TBitmap            );
    procedure   SetActiveMidBm   ( aValue: TBitmap            );
    procedure   SetInActiveMidBm ( aValue: TBitmap            );
    procedure   SetActiveHighBm  ( aValue: TBitmap            );
    procedure   SetInActiveHighBm( aValue: TBitmap            );
    procedure   SetPeakColor     ( aValue: TColor             );
    procedure   SetValeyColor    ( aValue: TColor             );
  protected
    procedure   AssignBitmaps;                                          virtual;
    function    PaintActive( anIndex: Integer): Boolean;                virtual;
    procedure   Paint;                                                 override;
  public
    constructor Create( anOwner: TComponent);                          override;
  public
    property    ActiveLowBm    : TBitmap                  read FActiveLowBm    write SetActiveLowBm;
    property    InActiveLowBm  : TBitmap                  read FInActiveLowBm  write SetInActiveLowBm;
    property    ActiveMidBm    : TBitmap                  read FActiveMidBm    write SetActiveMidBm;
    property    InActiveMidBm  : TBitmap                  read FInActiveMidBm  write SetInActiveMidBm;
    property    ActiveHighBm   : TBitmap                  read FActiveHighBm   write SetActiveHighBm;
    property    InActiveHighBm : TBitmap                  read FInActiveHighBm write SetInActiveHighBm;
  published
    property    IndicatorCount : Integer                  read FIndicatorCount write SetIndicatorCount default 10;
    property    Value          : TSignal                  read FValue          write SetValue          nodefault;
    property    PeakValue      : TSignal                  read FPeakValue      write SetPeakValue      nodefault;
    property    ValeyValue     : TSignal                  read FValeyValue     write SetValeyValue     nodefault;
    property    ShowPeakValue  : Boolean                  read FShowPeakValue  write SetShowPeakValue  default False;
    property    ShowValeyValue : Boolean                  read FShowValeyValue write SetShowValeyValue default False;
    property    ValueLowMid    : TSignal                  read FValueLowMid    write SetValueLowMid    nodefault;
    property    ValueMidHigh   : TSignal                  read FValueMidHigh   write SetValueMidHigh   nodefault;
    property    DisplayType    : TKnobsDisplayType        read FDisplayType    write SetDisplayType    default dtFilledFromLow;
    property    Orientation    : TKnobsDisplayOrientation read FOrientation    write SetOrientation    default doHorizontal;
    property    Spacing        : Integer                  read FSpacing        write SetSpacing        default 2;
    property    PeakColor      : TColor                   read FPeakColor      write SetPeakColor      default clWhite;
    property    ValeyColor     : TColor                   read FValeyColor     write SetValeyColor     default clBlack;
    property    Visible;
    property    Enabled;
  published
    property    OnClick;
    property    OnDblClick;
  end;


  TKnobsIndicatorText = class( TCustomLabel)
  private
    FValue : TSignal;
  protected
    procedure   SetValue( aValue: TSignal);
    procedure   DoDrawText( var aRect: TRect; aFlags: Longint);        override;
  public
    constructor Create( anOwner: TComponent);                          override;
  published
    property Value : TSignal read FValue write SetValue nodefault;
    property Width                                      default 10;
    property Height                                     default 10;
    property AutoSize                                   default false;
    property Align                                      default alNone;
    property Alignment                                  default taCenter;
    property Color                                      default KnobsColorDisplay;
    property Transparent                                default True;
    property Visible                                    default True;
    property WordWrap                                   default False;
    property ParentColor                                default False;
    property ParentFont                                 default False;
    property ShowHint                                   default False;
    property ParentShowHint                             default False;
    property ShowAccelChar                              default False;
    property Layout;
    property Anchors;
    property BiDiMode;
    property Caption;
    property Constraints;
    property Enabled;
    property FocusControl;
    property Font;
    property ParentBiDiMode;
    property PopupMenu;
  end;


  TKnobsDataViewer = class( TGraphicControl)
  // a viewer
  private
    FBackGround  : TBitmap;
    FYData       : TSignalArray;
    FXYData      : TSignalPairFifo;
    FXYPoints    : Integer;
    FXOffset     : TSignal;
    FYOffset     : TSignal;
    FMaxValue    : TSignal;
    FMinValue    : TSignal;
    FXZoom       : TSignal;
    FYZoom       : TSignal;
    FBorderColor : TColor;
    FLineColor   : TColor;
    FFillColor   : TColor;
    FStyle       : TKnobsDVStyle;
  private
    function    GetModule     : TKnobsCustomModule;
    procedure   SetBackGround ( aValue: TBitmap);
    procedure   SetYData      ( const aValue: TSignalArray);
    procedure   SetXYData     ( const aValue: TSignalPairFifo);
    procedure   SetXOffset    ( aValue: TSignal);
    procedure   SetYOffset    ( aValue: TSignal);
    procedure   SetMaxValue   ( aValue: TSignal);
    procedure   SetMinValue   ( aValue: TSignal);
    procedure   SetXZoom      ( aValue: TSignal);
    procedure   SetYZoom      ( aValue: TSignal);
    procedure   SetLineColor  ( aValue: TColor );
    procedure   SetFillColor  ( aValue: TColor );
    procedure   SetBorderColor( aValue: TColor );
    procedure   SetStyle      ( aValue: TKnobsDVStyle);
    procedure   SetXYPoints   ( aValue: Integer);
  private
    procedure   FillBlank;
    procedure   FillSine;
    procedure   FillTri;
    procedure   FillSaw;
    procedure   FillSquare;
    procedure   FillNoise;
    procedure   FillEnvAr;
    procedure   FillEnvAhd;
    procedure   FillEnvAdsr;
    procedure   FillXYScope;
  protected
    procedure   WMNCHitTest( var aMessage: TWMNCHitTest);              message WM_NCHITTEST;
    procedure   WMEraseBackground( var aMsg: TWMEraseBkgnd);           message WM_ERASEBKGND;
    procedure   MouseDown( Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    function    HasNegativeData: Boolean;
    function    ScaleData        ( anIndex: Integer): Integer;
    function    ScaleDataNegative( anIndex: Integer): Integer;
    function    ScaleX           ( anIndex: Integer): Integer;
    procedure   Paint;                                                 override;
  public
    constructor Create( anOwner: TComponent);                          override;
    destructor  Destroy;                                               override;
  public
    property    YData       : TSignalArray       read FYData     write SetYData;
    property    XYData      : TSignalPairFifo    read FXYData    write SetXYData;
    property    Module      : TKnobsCustomModule read GetModule;
  published
    property    Background  : TBitmap       read FBackGround  write SetBackGround;
    property    XOffset     : TSignal       read FXOffset     write SetXOffset;    // default  0;
    property    YOffset     : TSignal       read FYOffset     write SetYOffset;    // default 75;
    property    MaxValue    : TSignal       read FMaxValue    write SetMaxValue;   // default High( SmallInt);
    property    MinValue    : TSignal       read FMinValue    write SetMinValue;   // default Low ( SmallInt);
    property    XZoom       : TSignal       read FXZoom       write SetXZoom;      // default 1;
    property    YZoom       : TSignal       read FYZoom       write SetYZoom;      // default 1;
    property    LineColor   : TColor        read FLineColor   write SetLineColor   default clWhite;
    property    FillColor   : TColor        read FFillColor   write SetFillColor   default clSilver;
    property    BorderColor : TColor        read FBorderColor write SetBorderColor default KnobsColorDisplay;
    property    Style       : TKnobsDVStyle read FStyle       write SetStyle       default dvsNoise;
    property    XYPoints    : Integer       read FXYPoints    write SetXYPoints    default 255;
    property    Width                                                              default 150;
    property    Height                                                             default 60;
    property    Color                                                              default KnobsColorDisplay;
  end;


  TKnobsConnector = class( TGraphicControl)
  // to connect wires to
  private
    FBitmap     : TBitmap;
    FSignalType : TSignalType;
    FWireColor  : TColor;
    FLinks      : TList;   // Connections, not wires
    FMouseDown  : Boolean;
    FConnected  : Boolean; // True when at least one connection exists
    FHighLight  : Boolean; // Highlighted connections
    FVisited    : Boolean;
  private
    procedure   SetConnected( aValue: Boolean);
    function    GetModule: TKnobsCustomModule;
    function    GetModuleName: string;
    procedure   SetHighLight( aValue: Boolean);
    procedure   SetWireColor( aValue: TColor );
    procedure   SetSignalType( aValue: TSignalType);
  private
    function    FindWirePanel: TKnobsWirePanel;
    procedure   AddLink   ( aConnector: TKnobsConnector);
    procedure   DeleteLink( aConnector: TKnobsConnector);
    function    HasDirectlinkTo( aConnector: TKnobsConnector): Boolean;
    function    FindConnector( aScreenPoint: TPoint): TKnobsConnector;
    function    CenterPoint: TPoint;
    procedure   DrawLine( X, Y: Integer);
    procedure   SendWireNotification( Msg: Cardinal; aConnector: TKnobsConnector);
    procedure   SendColorNotification( aColor: TColor);
    procedure   SendHighlightNotification;
    procedure   ScrollMouseInView( anX, anY: LongInt);
  protected
    procedure   MouseDown( Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure   MouseUp  ( Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure   MouseMove( Shift: TShiftState; X, Y: Integer);         override;
    function    FindOutput: TKnobsConnector;
    function    IsConnectedToOutput: Boolean;
    procedure   FindDistanceToOutput( var aDistance: Integer);
    function    IsConnectedTo( aConnector: TKnobsConnector): Boolean;
    function    CanAccept( aSource: TKnobsConnector; isMove: Boolean): Boolean;
    procedure   AssignBitmap;                                 virtual; Abstract;
    procedure   MoveLinksTo   ( aConnector: TKnobsConnector);
    procedure   ConnectTo     ( aConnector: TKnobsConnector);
    procedure   DisconnectFrom( aConnector: TKnobsConnector);
    function    DisConnectAll: Boolean;
    procedure   DisConnectDownStream;
    procedure   DisConnectDownStreamRecursive;
    procedure   Paint;                                                 override;
  public
    constructor Create( anOwner: TComponent);                          override;
    destructor  Destroy;                                               override;
    procedure   Dump     ( var aFile: TextFile; anInd: Integer);
    procedure   DumpFlat ( var aFile: TextFile; anInd: Integer);
    procedure   DumpLinks( var aFile: TextFile; anInd: Integer);
    procedure   Log( const aMsg: string);
    function    DistanceToOutput: Integer;
  protected
    property    HighLight  : Boolean            read FHighLight    write SetHighLight;
  public
    property    Connected  : Boolean            read FConnected    write SetConnected;
    property    Module     : TKnobsCustomModule read GetModule;
    property    ModuleName : string             read GetModuleName;
    property    Signaltype : TSignalType        read FSignalType   write SetSignalType  default stControl;
  published
    property    WireColor  : TColor             read FWireColor    write SetWireColor   default clBlue;
    property    Visible;
  end;

  TKnobsConnectors = array of TKnobsConnector;


  TKnobsInput = class( TKnobsConnector)
  // Specialized connector with it's own bitmap
  protected
    procedure   AssignBitmap;                                          override;
  end;


  TKnobsOutput = class( TKnobsConnector)
  // Specialized connector with it's own bitmap
  protected
    procedure   AssignBitmap;                                          override;
  end;


  TKnobsTextLabel = class( TCustomLabel)
  // Just a small print TLabel
  protected
    procedure   DoDrawText( var aRect: TRect; aFlags: Longint);        override;
  public
    constructor Create( anOwner: TComponent);                          override;
  published
    property Align;
    property Alignment;
    property Anchors;
    property AutoSize;
    property BiDiMode;
    property Caption;
    property Color;
    property Constraints;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property FocusControl;
    property Font;
    property ParentBiDiMode;
    property ParentColor;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowAccelChar;
    property ShowHint;
    property Transparent;
    property Layout;
    property Visible;
    property WordWrap;
  end;


  TKnobsLabelEditor = class( TEdit)
  // An editor for TEditLabel
  private
    FLabel : TKnobsEditLabel;
  private
    procedure   WMKillFocus( var aMessage: TWMKillFocus );  message WM_KILLFOCUS;
    procedure   Do_Exit   ( aSender: TObject);
    procedure   Do_KeyDown( aSender: TObject; var aKey: Word; aShift: TShiftState);
  protected
    procedure   SetLabel( aLabel: TKnobsEditLabel);
  public
    constructor Create( anOwner: TComponent);                          override;
    destructor  Destroy;                                               override;
  end;


  TKnobsEditLabel = class( TKnobsTextLabel)
  // a label with a popup editor so it's caption can be edited.
  private
    FOnChanged : TKnobsOnEditLabelChanged;
  private
    function    GetModule: TKnobsCustomModule;
  protected
    procedure   SetText( const aValue: string);
    procedure   MouseDown( Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
  public
    constructor Create( anOwner: TComponent);                          override;
  public
    property    Module: TKnobsCustomModule read GetModule;
  published
    property    OnChanged: TKnobsOnEditLabelChanged read FOnChanged write FOnChanged;
  end;


  TKnobsBox = class( TBevel)
  // A bevel that can be disabled
  public
    constructor Create( anOwner: TComponent);                          override;
  published
    property    Enabled default False;
  end;


  TKnobsWire = class
  // Wire, visual aspect of a connection
  private
    FSignalType   : TSignalType;
    FSource       : TKnobsConnector;
    FDestination  : TKnobsConnector;
    FHighlight    : Boolean;
    FBezierParams : TKnobsBezierParams;
  private
    function    GetSourceName: string;
    function    GetDestinationName: string;
    function    GetFullName: string;
  public
    constructor Create( aSource, aDestination: TKnobsConnector);
    procedure   Wiggle;
    procedure   Dump( var aFile: TextFile; anInd: Integer);
    procedure   FixWire;
    procedure   SwapConnectors;
  public
    property    Source          : TKnobsConnector    read FSource;
    property    Destination     : TKnobsConnector    read FDestination;
    property    SourceName      : string        read GetSourceName;
    property    DestinationName : string        read GetDestinationName;
    property    BezierParams    : TKnobsBezierParams read FBezierParams;
    property    FullName        : string        read GetFullName;
  end;


  TKnobsWireOverlay = class( TCustomControl)
  //
  // Responsibilty to maintain connections is assumed by the TConnectors
  // themselves. Wires are maintained here, but it's a cache really to ease
  // wire painting. The wires are maintained by the connectors theselves that
  // send their info to us through the WirePanel (by means of UM_CONNEXTIONxx
  // and UM_HIGHLIGHT messages). The WirePanel imforms us by calling the
  // AddWire, DelWire and Highlight methodes.
  //
  // All Wires are owned by FWires, so our destructor will eventually destroy
  // all dangling wires. Nevertheless, TConnector still must, on destruction,
  // clean up the wires it is responsible for .. it wouldn't look good other
  // wise in the user interface.
  //
  private
    FWires            : TObjectList;
    FCurvedLines      : Boolean;
    FWireThickness    : Integer;
    FdefaultWireColor : TColor;
    FBlockCount       : Integer;
    FPaintCount       : Integer;
    FPaintClocks      : Int64;
  private
    procedure   WMNCHitTest( var aMessage: TWMNCHitTest);  message WM_NCHITTEST;
  private
    function    GetWire         ( anIndex: Integer): TKnobsWire;
    function    GetSource       ( anIndex: Integer): TKnobsConnector;
    function    GetDestination  ( anIndex: Integer): TKnobsConnector;
    procedure   SetCurvedLines  ( aValue: Boolean );
    procedure   SetWireThickNess( aValue: Integer );
  protected
    procedure   Paint;                                                 override;
    procedure   PaintWire( aDC: HDC; aWire: TKnobsWire);
    function    FindWire ( aSource, aDestination: TKnobsConnector): Integer;
    procedure   AddWire  ( aSource, aDestination: TKnobsConnector);
    procedure   DelWire  ( aSource, aDestination: TKnobsConnector);
    procedure   Connect  ( aSource, aDestination: TKnobsConnector);
    procedure   HandleHighlight;
    procedure   HandleColorChange( aSource: TKnobsConnector; aColor: TColor);
    function    WireCount: Integer;
    procedure   WiggleWires;
  public
    constructor Create( anOwner: TComponent);                          override;
    destructor  Destroy;                                               override;
    procedure   InvalidateWires;
    function    Disconnect( aConnector: TKnobsConnector): Boolean;
    procedure   Dump( var aFile: TextFile; anInd: Integer);
    procedure   Log( const aMsg: string);
    procedure   BlockScreenUpdates;
    procedure   UnBlockScreenUpdates;
    procedure   FixAllWires;
  public
    property    Wire       [ anIndex: Integer]: TKnobsWire      read GetWire;
    property    Source     [ anIndex: Integer]: TKnobsConnector read GetSource;
    property    Destination[ anIndex: Integer]: TKnobsConnector read GetDestination;
  public
    property    CurvedLines      : Boolean read FCurvedLines      write SetCurvedLines    default True;
    property    WireThickness    : Integer read FWireThickness    write SetWireThickness  nodefault;
    property    defaultWireColor : TColor  read FdefaultWireColor write FdefaultWireColor default clBlue;
    property    BlockCount       : Integer read FBlockCount;
    property    PaintCount       : Integer read FPaintCount;
    property    PaintClocks      : Int64   read FPaintClocks;
  published
    property    Left   stored False;
    property    Top    stored False;
    property    Width  stored False;
    property    Height stored False;
  end;


  TKnobsModuleList = class( TList)
  // List of TCustomModule - some comon ops defined here
  private
    FonLog : TKnobsOnLog;
  private
    function    GetModule  ( anIndex: Integer): TKnobsCustomModule;
    function    GetSelected( anIndex: Integer): Boolean;
    procedure   SetSelected( anIndex: Integer; aValue: Boolean);
    function    GetAnySelected: Boolean;
    function    GetNoneSelected: Boolean;
  public
    procedure   FreeModule( anIndex: Integer);
    function    FreeModules: Boolean;
    function    FreeSelectedModules: Boolean;
    procedure   SelectAll;
    procedure   UnselectAll;
    procedure   SelectUnique( aModule: TKnobsCustomModule);
    procedure   FixSelection( aModule: TKnobsCustomModule);
    procedure   SetSelectedModuleColor( aValue: TColor);
    procedure   InvertSelection;
    procedure   InvalidateDisplays;
  public
    //* Some helpers for selector population and component registration
    function    FindType( aModuleType: TKnobsModuleType): TKnobsCustomModule;
    procedure   PopulateSelector( aModuleSelector: TKnobsModuleSelector);
    function    CreateModule( aWirePanel: TKnobsWirePanel; aModuleType: TKnobsModuleType; MustDrag: Boolean): TKnobsCustomModule;
    procedure   RegisterComponents( doRegister: Boolean);
    procedure   Dump( var aFile: TextFile; anInd: Integer);
    procedure   Log( const aMsg: string);
  public
    property    Module  [ anIndex: Integer]: TKnobsCustomModule read GetModule;                        default;
    property    Selected[ anIndex: Integer]: Boolean            read GetSelected     write SetSelected;
    property    AnySelected                : Boolean            read GetAnySelected;
    property    NoneSelected               : Boolean            read GetNoneSelected;
    property    OnLog                      : TKnobsOnLog        read FOnLog          write FOnLog;
  end;


  TKnobsHistoryStack = class( TStringList)
  private
    FMaxHistory : Integer;
  private
    procedure   SetMaxHistory( aValue: Integer);
  protected
    procedure   PushString( const S: string);
    function    PopString: string;
  public
    procedure   Push( const aWirePanel: TKnobsWirePanel);
    procedure   Pop ( const aWirePanel: TKnobsWirePanel);
  public
    property    MaxHistory : Integer read FMaxHistory write SetMaxHistory;
  end;


  TKnobsUndoRedoHistory = class( TObject)
  private
    FOnHistoryChange : TKnobsOnHistoryChange;
    FMaxHistory      : Integer;
    FLockCount       : Integer;
    FUndos           : TKnobsHistoryStack;
    FRedos           : TKnobsHistoryStack;
  private
    function    GetUndoCount: Integer;
    function    GetRedoCount: Integer;
    procedure   SetmaxHistory( aValue: Integer);
  public
    constructor Create;
    destructor  Destroy;                                               override;
    procedure   HistoryChanged;
    procedure   SaveState( const aWirePanel: TKnobsWirePanel);
    procedure   Undo     ( const aWirePanel: TKnobsWirePanel);
    procedure   Redo     ( const aWirePanel: TKnobsWirePanel);
    procedure   Lock;
    procedure   UnLock;
    procedure   Clear;
  public
    property    UndoCount       : Integer          read GetUndoCount;
    property    RedoCount       : Integer          read GetRedoCount;
    property    MaxHistory      : Integer          read FMaxHistory  write SetmaxHistory;
  public
    property    OnHistoryChange : TKnobsOnHistoryChange read FOnHistoryChange write FOnHistoryChange;
  end;


  TKnobsWirePanel = class( TScrollingWinControl)
  // Contains the modules and the wire overlay - scrollable.
  private
    FOffsetLeft           : Integer;
    FOffsetTop            : Integer;
    FTitle                : string;
    FVersion              : Integer;
    FMouseDown            : Boolean;
    FWireOverlay          : TKnobsWireOverlay;
    FModules              : TKnobsModuleList;
    FFilename             : string;
    FSnapX                : Integer;
    FSnapY                : Integer;
    FControlMode          : TDistanceMode;
    FWheelSupportOnKnobs  : Boolean;
    FWheelSensitivity     : Integer;
    FModuleOpacity        : Byte;
    FModuleColor          : TColor;
    FSnapActive           : Boolean;
    FPatchReader          : IKnobsPatchReader;
    FPatchWriter          : IKnobsPatchWriter;
    FOnLog                : TKnobsOnLog;
    FOnHistoryChange      : TKnobsOnHistoryChange;
    FHistory              : TKnobsUndoRedoHistory;
    FOnValueChanged       : TKnobsOnValueChange;
    FOnTextChanged        : TKnobsOnTextChanged;
    FOnPatchChanged       : TKnobsOnPatchChanged;
    FOnWiresChanged       : TKnobsOnWiresChanged;
    FOnShowPopup          : TOnShowPopup;
    FOnUnFocus            : TOnUnFocus;
  private
    procedure   UMConnectionAdd   ( var aMessage: TKnobsUMConnectionAdd  ); message KNOBS_UM_CONNECTIONADD;
    procedure   UMConnectionRemove( var aMessage: TKnobUMConnectionRemove); message KNOBS_UM_CONNECTIONREMOVE;
    procedure   UMColorChange     ( var aMessage: TKnobsUMColorChange    ); message KNOBS_UM_COLORCHANGE;
    procedure   UMHighlight       ( var aMessage: TKnobsUMHighlight      ); message KNOBS_UM_HIGHLIGHT;
  private
    function    GetModule     ( anIndex: Integer): TKnobsCustomModule;
    function    GetWire       ( anIndex: Integer): TKnobsWire;
    function    GetSource     ( anIndex: Integer): TKnobsConnector;
    function    GetDestination( anIndex: Integer): TKnobsConnector;
  private
    procedure   SetOffsetLeft         ( aValue: Integer);
    procedure   SetOffsetTop          ( aValue: Integer);
    function    GetCurvedLines        : Boolean;
    procedure   SetCurvedLines        ( aValue: Boolean);
    function    GetWireThickness      : Integer;
    procedure   SetWireThickness      ( aValue: Integer);
    function    GetdefaultWireColor   : TColor;
    procedure   SetdefaultWireColor   ( aValue: TColor);
    function    GetModuleCount        : Integer;
    function    GetWireCount          : Integer;
    function    GetMaxHistory         : Integer;
    procedure   SetMaxHistory         ( aValue: Integer);
    function    GetBlockCount         : Integer;
    function    GetPaintCount         : Integer;
    function    GetPaintClocks        : Int64;
    function    GetScrollOffset       : TPoint;
    procedure   SetControlMode        ( aValue: TDistanceMode);
    procedure   SetModuleOpacity      ( aValue: Byte);
    procedure   SetModuleColor        ( aValue: TColor);
    procedure   SetWheelSupportOnKnobs( aValue: Boolean);
    procedure   SetWheelSensitivity   ( aValue: Integer);
  private
    procedure   StrToConnection( const S: string);
  protected
    procedure   HandleRightClick( const aSender: TControl);
    procedure   CreateParams(var Params: TCreateParams);                             override;
    procedure   MouseDown( Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure   MouseUp  ( Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure   MouseMove( Shift: TShiftState; X, Y: Integer);                       override;
    procedure   DragStart( aPoint: TPoint);                                           virtual;
    procedure   DragEnd  ( aPoint: TPoint);                                           virtual;
    procedure   DragMove ( aPoint: TPoint);                                           virtual;
    procedure   SelectModulesInRect( aRect: TRect);
    function    IndexOfModule( aModule: TKnobsCustomModule): Integer;
    procedure   Notification( aComponent: TComponent; anOperation: TOperation);      override;
    procedure   Connect( aSource, aDestination: TKnobsConnector);
    procedure   Restack( DoRestack, RedrawWires: Boolean);
    procedure   Resnap;
    procedure   FixNames;
    procedure   SolveNamingConflictsWith( aWirePanel: TKnobsWirePanel);
    function    Snap( const aPoint: TPoint): TPoint;
    procedure   DoHistoryChanged( const aSender: TObject; anUndoCount, aRedoCount: Integer);
    procedure   ValueChanged  ( aSender: TObject; const aPath, aControlType: string; aValue: TSignal; IsFinal: Boolean); virtual;
    procedure   TextChanged( aSender: TObject; const aPath, aValue: string);
    procedure   PatchChanged;
    procedure   WiresChanged;
    procedure   ScrollPointInView( aPoint: TPoint);                                  overload;
    procedure   ScrollPointInView( anX, anY: LongInt);                               overload;
  public
    constructor Create( anOwner: TComponent);                                        override;
    destructor  Destroy;                                                             override;
    procedure   FixAllSynthParams;
    procedure   ShowAllPopups;
    procedure   HideAllPopups;
    procedure   ForceHideAllPopups;
    procedure   BeginStateChange;
    procedure   EndStateChange( PatchWasChanged: Boolean);
    procedure   Undo;
    procedure   Redo;
    procedure   ClearUndoRedo;
    procedure   LockUndoRedo;
    procedure   UnLockUndoRedo;
    procedure   Dump( const aFileName: string; DoAppend: Boolean);                   overload;
    procedure   Dump( var aFile: TextFile);                                          overload;
    procedure   DumpModules( var aFile: TextFile; anInd: Integer);
    procedure   DumpWires( var aFile: TextFile; anInd: Integer);
    procedure   Log( const aMsg: string);
    function    FindModule( const aName: string): TKnobsCustomModule;
    function    FindFirstSelectedModule: TKnobsCustomModule;
  public
    procedure   InvalidateWires;
    procedure   InvalidateDisplays;
    procedure   Disconnect( aConnector: TKnobsConnector);
    procedure   FreeModules;
    function    HasEditPopup: Boolean;
    procedure   SetSelectedModuleColor( aValue: TColor);
    procedure   FreeSelectedModules;
    procedure   SelectAllModules;
    procedure   UnSelectAllModules;
    procedure   SelectUnique( aModule: TKnobsCustomModule);
    procedure   FixSelection( aModule: TKnobsCustomModule);
    procedure   InvertModuleSelection;
    function    CopyToString( aWriteMode: TKnobsWriteMode): string;
    procedure   CopyFromString( const S: string; aReadMode: TKnobsReadMode);
    procedure   CopyTo( aPanel: TKnobsWirePanel; aWriteMode: TKnobsWriteMode; aReadMode: TKnobsReadMode);
    function    SameStructure( aPanel: TKnobsWirePanel): Boolean;
    procedure   CopyParamsFrom( aSrc: TKnobsWirePanel);
    procedure   AddModules( aSrc: TKnobsWirePanel; aDistance: TPoint; DoRestack, RedrawWires, MustDrag: Boolean); virtual;
    procedure   AddWire( const aSrc, aDst: string);
    procedure   WiggleWires;
    procedure   FixAllWires;
    procedure   FixControlMode;
    procedure   BlockScreenUpdates;
    procedure   UnblockScreenUpdates;
    procedure   Search( const aValue: string);
    procedure   FindUnconnectedModules;
    procedure   UnFocus( const aSender: TObject);
    procedure   FindGhosts( const aGhosts: TStringList);
    function    CountGhosts: Integer;
  public
    procedure   ConnectorDelete    ( const aConnector: TKnobsConnector);
    procedure   ConnectorBreak     ( const aConnector: TKnobsConnector);
    procedure   ConnectorDisconnect( const aConnector: TKnobsConnector);
  public
    property    PatchReader  : IKnobsPatchReader   read FPatchReader   write FPatchReader;
    property    PatchWriter  : IKnobsPatchWriter   read FPatchWriter   write FPatchWriter;
  public
    property    Module     [ anIndex: Integer]: TKnobsCustomModule  read GetModule;
    property    Wire       [ anIndex: Integer]: TKnobsWire          read GetWire;
    property    Source     [ anIndex: Integer]: TKnobsConnector     read GetSource;
    property    Destination[ anIndex: Integer]: TKnobsConnector     read GetDestination;
  public
    property    OnHistoryChange     : TKnobsOnHistoryChange  read FOnHistoryChange     write FOnHistoryChange;
    property    OnLog               : TKnobsOnLog            read FOnLog               write FonLog;
    property    Filename            : string                 read FFilename            write FFilename;
    property    ModuleCount         : Integer                read GetModuleCount;
    property    WireCount           : Integer                read GetWireCount;
    property    BlockCount          : Integer                read GetBlockCount;
    property    PaintCount          : Integer                read GetPaintCount;
    property    PaintClocks         : Int64                  read GetPaintClocks;
    property    ScrollOffset        : TPoint                 read GetScrollOffset;
  published
    property    MaxHistory          : Integer                read GetMaxHistory        write SetMaxHistory          default 100;
    property    OnValueChanged      : TKnobsOnValueChange    read FOnValueChanged      write FOnValueChanged;
    property    OnTextChanged       : TKnobsOnTextChanged    read FOnTextChanged       write FOnTextChanged;
    property    OnPatchChanged      : TKnobsOnPatchChanged   read FOnPatchChanged      write FOnPatchChanged;
    property    OnWiresChanged      : TKnobsOnWiresChanged   read FOnWiresChanged      write FOnWiresChanged;
    property    OnShowPoup          : TOnShowPopup           read FOnShowPopup         write FOnShowPopup;
    property    OnUnFocus           : TOnUnFocus             read FOnUnFocus           write FOnUnFocus;
    property    OffsetLeft          : Integer                read FOffsetLeft          write SetOffsetLeft          default 4;
    property    OffsetTop           : Integer                read FOffsetTop           write SetOffsetTop           default 4;
    property    CurvedLines         : Boolean                read GetCurvedLines       write SetCurvedLines         default True;
    property    WireThickness       : Integer                read GetWireThickness     write SetWireThickness       default 1;
    property    defaultWireColor    : TColor                 read GetdefaultWireColor  write SetdefaultWireColor    default clBlue;
    property    Title               : string                 read FTitle               write FTitle;
    property    Version             : Integer                read FVersion             write FVersion;
    property    SnapX               : Integer                read FSnapX               write FSnapX                 default 250;
    property    SnapY               : Integer                read FSnapY               write FSnapY                 default 15;
    property    SnapActive          : Boolean                read FSnapActive          write FSnapActive            default True;
    property    ControlMode         : TDistanceMode          read FControlMode         write SetControlMode         default dmCircular;
    property    WheelSupportOnKnobs : Boolean                read FWheelSupportOnKnobs write SetWheelSupportOnKnobs;
    property    WheelSensitivity    : Integer                read FWheelSensitivity    write SetWheelSensitivity    default 50;
    property    ModuleOpacity       : Byte                   read FModuleOpacity       write SetModuleOpacity       default 223;
    property    ModuleColor         : TColor                 read FModuleColor         write SetModuleColor         default clWhite;
    property    Align;
    property    Anchors;
    property    Constraints;
    property    ParentColor;
    property    ParentShowHint;
    property    PopupMenu;
    property    ShowHint;
    property    TabOrder;
    property    TabStop;
    property    Visible;
    property    Color;
    property    Top;
    property    Left;
    property    Width;
    property    Height;
    property    HorzScrollBar;
    property    VertScrollBar;
  end;


  TKnobsCustomModule = class( TCustomPanel)
  // Base type for TModule, it has a bitmap and a title mainly
  private
    FAnchor               : TPoint;                 // Drag origin
    FDistance             : TPoint;                 // Drag distance
    FTresholdPast         : Boolean;                // Set when dragged over a minimum distance
    FFlags                : TKnobsModuleFlags;      // Selected / Capturing
    FTitleLabel           : TKnobsEditLabel;        // All editor modules can have a title label, it may be nil
    FTitle                : string;                 // User editible module title
    FComment              : string;                 // Comment as displayed on mouseover in the module selector
    FModuleType           : TKnobsModuleType;       // The type, used to couple editor modules to synth modules
    FPageName             : string;                 // The page in the module selector this module will be on,
                                                    // when the parent is a TTabSheet and that TTabSheet has a non empty
                                                    // Caption that value will be used instead.
    FBitmap               : TBitmap;                // Background bitmap for module
    FOpacity              : Byte;                   // Opacity for background bitmap
    FControlMode          : TDistanceMode;          // Knobs control mode, circular, horizontal or vertical
    FWheelSupportOnKnobs  : Boolean;
    FWheelSensitivity     : Integer;
    FOnValueChanged       : TKnobsOnValueChange;    // Called with the new value
    FOnTextChanged        : TKnobsOnTextChanged;    // Called with the new text
    FOnLog                : TKnobsOnLog;            // To atach an external logger
    FOnUnFocus            : TOnUnFocus;             // Called when a control could be unfocussed
  private
    procedure   IncludeFlags( aFlags: TKnobsModuleFlags);
    procedure   ExcludeFlags( aFlags: TKnobsModuleFlags);
    function    HasFlags( aFlags: TKnobsModuleFlags): Boolean;
  private
    function    GetSelected   : Boolean;
    procedure   SetSelected   ( aValue: Boolean);
    function    GetCapturing  : Boolean;
    procedure   SetCapturing  ( aValue: Boolean);
  private
    function    GetPageName           : string;
    procedure   SetPageName           ( const aValue: string);
    procedure   SetModuleType         ( aValue : TKnobsModuleType);
    procedure   SetTitle              ( const aValue: string);
    procedure   SetTitleLabel         ( const aValue: TKnobsEditLabel);
    procedure   SetOpacity            ( aValue: Byte);
    procedure   SetControlMode        ( aValue: TDistanceMode);
    procedure   SetWheelSupportOnKnobs( aValue: Boolean);
    procedure   SetWheelSensitivity   ( aValue: Integer);
  protected
    procedure   WMEraseBackground( var aMsg: TWMEraseBkgnd);                         message WM_ERASEBKGND;
    procedure   MouseDown( Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure   MouseMove( Shift: TShiftState; X, Y: Integer);                       override;
    procedure   MouseUp  ( Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
  protected
    procedure   InvalidateKnobs;
    procedure   InvalidateDisplays;
    procedure   ValueChanged  ( aSender: TObject; const aPath, aControlType: string; aValue: TSignal; IsFinal: Boolean); virtual;
    procedure   TextChanged( aSender: TObject; const aPath, aValue: string);
    function    FindWirePanel: TKnobsWirePanel;
    function    Snap( const aPoint: TPoint): TPoint;
    procedure   Paint;                                                 override;
    procedure   Loaded;                                                override;
    procedure   AssignBitmap;                                           virtual;
    procedure   ScrollSelfInView;
    procedure   HandleSelection( CtrlKeyActive: Boolean);
    procedure   BeginMove        ( anAnchor : TPoint);
    procedure   Move             ( aDistance: TPoint);
    procedure   MoveSelected     ( aDistance: TPoint);
    procedure   FinalMoveSelected( aDistance: TPoint; CtrlKeyActive: Boolean);
    procedure   ReleasePeerCapture;
    procedure   Notification( aComponent: TComponent; anOperation: TOperation); override;
    procedure   FixOwnership( aRoot, aControl: TWinControl);
    procedure   FixNames;
    function    FindValuedControl( const aName: string): TKnobsValuedControl;
    function    FindXYControl    ( const aName: string): TKnobsXYControl;
    function    FindDisplay      ( const aName: string): TKnobsDisplay;
    function    FindFileSelector ( const aName: string): TKnobsFileSelector;
  public
    constructor Create( anOwner: TComponent);                          override;
    destructor  Destroy;                                               override;
    procedure   ShowAllPopups;
    procedure   HideAllPopups;
    procedure   ForceHideAllPopups;
    procedure   FixAllSynthParams;
    function    Clone( anOwner: TComponent; aParent: TWinControl; anOffset: TPoint; MustDrag: Boolean): TKnobsCustomModule;
    procedure   SetKnobValue  ( const aType, aName: string; aValue: Integer);
    procedure   SetDetailValue( const aType, aName: string; const aValue: string);
    procedure   SetLockOn( const aType, aname: string);
    procedure   SetLight ( const aName: string; aValue: TSignal);
    procedure   SetData  ( const aName: string; const aData: TSignalArray);
    procedure   SetXYData( const aName: string; const aData: TSignalPairFifo);
    procedure   SetInfo  ( const aName: string; const aValue: string);
    procedure   Dump( var aFile: TextFile; anInd: Integer);
    procedure   Log( const aMsg: string);
    procedure   LogFmt( const aFmt: string; const anArgs: array of const);
    function    AsBitmap: TBitmap;
  public
    procedure   Restack( DoRestack, RedrawWires: Boolean);
    procedure   FixupInsertion;
    procedure   FixDisplays;
    procedure   Disconnect;
    procedure   SelectUnique;
    procedure   FixSelection;
    procedure   CollectInputsWithBaseName ( var anInputs : TKnobsConnectors; const aBaseName: string);
    procedure   CollectOutputsWithBaseName( var anOutputs: TKnobsConnectors; const aBaseName: string);
    procedure   UnFocus( const aSender: TObject);                       virtual;
    function    IsConnected: Boolean;
    procedure   CopyParamsFrom( aModule: TKnobsCustomModule);
  public
    property    Selected            : Boolean                read GetSelected          write SetSelected;
    property    Capturing           : Boolean                read GetCapturing         write SetCapturing;
    property    Title               : string                 read FTitle               write SetTitle;
    property    Comment             : string                 read FComment             write FComment;
    property    PageName            : string                 read GetPageName          write SetPageName;
    property    ModuleType          : TKnobsModuleType       read FModuleType          write SetModuleType;
    property    OnLog               : TKnobsOnLog            read FOnLog               write FOnLog;
  published
    property    Color;
    property    TitleLabel          : TKnobsEditLabel        read FTitleLabel          write SetTitleLabel;
    property    Opacity             : Byte                   read FOpacity             write SetOpacity             default 223;
    property    ControlMode         : TDistanceMode          read FControlMode         write SetControlMode;
    property    WheelSupportOnKnobs : Boolean                read FWheelSupportOnKnobs write SetWheelSupportOnKnobs;
    property    WheelSensitivity    : Integer                read FWheelSensitivity    write SetWheelSensitivity;
    property    OnValueChanged      : TKnobsOnValueChange    read FOnValueChanged      write FOnValueChanged;
    property    OnTextChanged       : TKnobsOnTextChanged    read FOnTextChanged       write FOnTextChanged;
    property    OnUnFocus           : TOnUnFocus             read FOnUnFocus           write FOnUnFocus;
  end;


  TKnobsModule = class( TKnobsCustomModule)
  // exports stuff of TCustomModule
  published
    property    BiDiMode;
    property    Constraints;
    property    Enabled;
    property    Color;
    property    Ctl3D;
    property    Font;
    property    ParentBiDiMode;
    property    ParentColor;
    property    ParentCtl3D;
    property    ParentFont;
    property    ParentShowHint;
    property    PopupMenu;
    property    ShowHint;
    property    TabOrder;
    property    TabStop;
    property    Visible;
    property    OnContextPopup;
    property    OnEnter;
    property    OnExit;
    property    OnMouseDown;
    property    OnMouseMove;
    property    OnMouseUp;
    property    OnMouseWheel;
    property    OnMouseWheelDown;
    property    OnMouseWheelUp;
    property    OnResize;
    property    Width;
    property    Height;
    property    Tag;
    property    Title;
    property    Comment;
    property    PageName;
    property    ModuleType;
    property    BevelInner;
    property    BevelOuter;
    property    BevelWidth;
  end;


  TKnobsModuleButton = class( TSpeedButton)
  // Button as used in module selector, on click it creates a module
  // of the type that got registered with it.
  private
    FPageName   : string;
    FModuleType : TKnobsModuleType;
    FModuleName : string;
  public
    procedure   LoadGlyph( const aGlyph: TBitmap);
    procedure   Initialize( const aPageName: string; aModuleType: TKnobsModuleType; const aModuleName: string; const aGlyph: TBitmap);
  public
    property    PageName   : string           read FPageName;
    property    ModuleType : TKnobsModuleType read FModuleType;
    property    ModuleName : string           read FModuleName;
  end;

  TKnobsModuleSeparator = class( TKnobsModuleButton);


  TKnobsModuleSelector = class( TPageControl)
  // A tabbed thingy having TModuleButtons on it's pages.
  private
    FRegisteredModules   : TKnobsModuleList;
    FOnModuleButtonClick : TKnobsOnModuleButtonClick;
    FOnGetGlyph          : TKnobsOnGetGlyph;
    FButtonsEnabled      : Boolean;
    FAllowDuplicates     : Boolean;
    FTimer               : TTimer;
  private
    procedure   StartTimeout;
    procedure   StopTimeout;
    procedure   TimeOutFired( aSender: TObject);
    procedure   AllButtonsUp;
  private
    procedure   SetButtonsEnabled( aValue: Boolean);
    function    AddPage          ( const aName: string): TTabSheet;
    function    PageButtonCount  ( const aName: string): Integer;
    function    PageExists       ( const aName: string): Boolean;
    function    PageByName       ( const aName: string): TTabSheet;
    function    FindModulePage   ( aModuleType: TKnobsModuleType): TTabSheet;
    function    FindModuleButton ( aModuleType: TKnobsModuleType): TKnobsModuleButton;
    procedure   DoModuleMouseDown( aSender: TObject; aButton: TMouseButton; aShift: TShiftState; X, Y: Integer);
    procedure   FormatPage( aPage: TTabSheet);
  protected
    procedure   Notification( aComponent: TComponent; anOperation: TOperation); override;
  public
    constructor Create( anOwner: TComponent);                          override;
    destructor  Destroy;                                               override;
    procedure   RegisterModuleType  ( const aPageName: string; aModuleType: TKnobsModuleType; const aModuleName: string);
    procedure   UnregisterModuleType( aModuleType: TKnobsModuleType);
  published
    property    OnModuleButtonClick : TKnobsOnModuleButtonClick read FonModuleButtonClick write FOnModuleButtonClick;
    property    OnGetGlyph          : TKnobsOnGetGlyph          read FOnGetGlyph          write FonGetGlyph;
    property    ButtonsEnabled      : Boolean                   read FButtonsEnabled      write SetButtonsEnabled;
    property    AllowDuplicates     : Boolean                   read FAllowDuplicates     write FAllowDuplicates     default True;
  end;


  TKnobsHintWindow = class( THintWindow)
  private
    FModuleType : string;
    FBitmap     : TBitmap;
    FComment    : string;
  protected
    function    MakeBitmap( const aModuleType: string): TBitmap;
    procedure   ParseHint( var aHintStr: string);
    procedure   Paint;                                            override;
  public
    constructor Create( anOwner: TComponent);                     override;
    destructor  Destroy;                                          override;
    procedure   ActivateHint( aRect: TRect; const aHint: string); override;
  published
    property    Caption;
  end;

var

  GOnCreateModuleBitmap : TOnCreateModuleBitmap;
  GOnReadModuleComment  : TOnReadModuleComment;


implementation



{$R KnobsBitmaps.res}


const

  InitRepeatPause     = 200;          // Initial pause before repeat kicks in
  SlowRepeatPause     = 100;          // RepeatPause slow repeat
  FastRepeatPause     =  30;          // RepeatPause fast repeat;
  SlowRepeats         =  15;          // After SlowRepeats repeats we start the fast repeat
  ClickerSize         =  10;          // Height for clickers

  // Bitmap Id's as used in the compiled in resource 'Bitmaps.res'

  idbmDn              =   1;          // Down pointing triangle
  idbmUp              =   2;          // Up pointing triangle
  idbmConIn           =   3;          // Input connector - round
  idbmConOut          =   4;          // Output connector - square
  idbmIndActive       =   5;          // Active LED
  idbmIndInactive     =   6;          // Inactivee LED
  idbmKnob            =   7;          // Large knob
  idbmSmallKnob       =   8;          // Small knob
  idbmSkin            =   9;          // Texture overlay
  idbmIndLowActive    =  10;          // Active LED for low values (green)
  idbmIndLowInactive  =  11;          // Inactive LED for low values (green)
  idbmIndMidActive    =  12;          // Active LED for mid values (yellow)
  idbmIndMidInactive  =  13;          // Inactive LED for mid values (yellow)
  idbmIndHighActive   =  14;          // Active LED for high values (red)
  idbmIndHighInactive =  15;          // Inactive LED for high values (red)
  idbmNoKnob          =  16;          // Bitmap for invisible knobs - clickers only
  idbmSkin2           =  17;          // Alternate texture overlay
  idbmKnobNew         =  18;          // Alternate large knob
  idbmSmallKnobNew    =  19;          // Alternate small knob
  idbmConInTrans      =  20;          // Alternate input connector (round)
  idbmConOutTrans     =  21;          // Alternate output connector (square)


var

  GAnchor             : TPoint              = ( x: 0; y: 0);   // Ussed for mouse drag operaations
  GStartPoint         : TPoint              = ( x: 0; y: 0);   // Ussed for mouse drag operaations
  GDistance           : TPoint              = ( x: 0; y: 0);   // Ussed for mouse drag operaations
  GLabelEditor        : TKnobsLabelEditor   = nil;             // The label editor popup
  GDisplayEditor      : TKnobsDisplayEditor = nil;             // The display (multiline) editor popup

   // Shared bitmaps, loaded from 'Bitmaps.res' at startup

  bmDn              : TBitmap = nil;  // Down pointing triangle
  bmUp              : TBitmap = nil;  // Up pointing triangle
  bmConIn           : TBitmap = nil;  // Input connector - round
  bmConOut          : TBitmap = nil;  // Output connector - square
  bmIndActive       : TBitmap = nil;  // Active LED
  bmIndInactive     : TBitmap = nil;  // Inactive LED
  bmIndLowActive    : TBitmap = nil;  // Active LED for low values (green)
  bmIndLowInactive  : TBitmap = nil;  // Inactive LED for low values (green)
  bmIndMidActive    : TBitmap = nil;  // Active LED for mid values (yellow)
  bmIndMidInactive  : TBitmap = nil;  // Inactive LED for mid values (yellow)
  bmIndHighActive   : TBitmap = nil;  // Active LED for high values (red)
  bmIndHighInactive : TBitmap = nil;  // Inactive LED for high values (red)
  bmKnob            : TBitmap = nil;  // Large knob
  bmSmallKnob       : TBitmap = nil;  // Small knob
  bmNoKnob          : TBitmap = nil;  // Bitmap for invisible knobs - clickers only
  bmSkin            : TBitmap = nil;  // Texture overlays


  procedure NeedsLabelEditor;
  begin
    if GLabelEditor = nil
    then TKnobsLabelEditor.Create( Application);
  end;

  procedure NeedsDisplayEditor;
  begin
    if GDisplayEditor = nil
    then TKnobsDisplayEditor.Create( Application);
  end;

  function  CreateModuleBitmap( aModuleType: TKnobsModuleType; UseCache: Boolean): TBitmap;
  begin
    if Assigned( GOnCreateModuleBitmap)
    then Result := GOnCreateModuleBitmap( aModuleType, UseCache)
    else Result := nil;
  end;

  function  ReadModuleComment( aModuleType: TKnobsModuleType): string;
  begin
    if Assigned( GOnReadModuleComment)
    then Result := GOnReadModuleComment( aModuleType)
    else Result := '';
  end;

  procedure   LoadBitmap( var aBitmap: TBitmap; anIndex: Integer);
  // Loads one bitmap from 'Bitmaps.res'
  begin
    if not Assigned( aBitmap)
    then aBitmap := TBitmap.Create;
    with aBitmap
    do begin
      LoadFromResourceID( HInstance, anIndex);
      TransparentMode := tmAuto;
      Transparent     := True;
    end;
  end;

  procedure CreateBitmaps;
  // Loads all bitmaps from 'Bitmaps.res'
  begin
    LoadBitmap( bmDn             , idbmDn             );
    LoadBitmap( bmUp             , idbmUp             );
    LoadBitmap( bmConIn          , idbmConIn          );
    LoadBitmap( bmConOut         , idbmConOut         );
    LoadBitmap( bmIndActive      , idbmIndActive      );
    LoadBitmap( bmIndInactive    , idbmIndInactive    );
    LoadBitmap( bmIndLowActive   , idbmIndLowActive   );
    LoadBitmap( bmIndLowInactive , idbmIndLowInactive );
    LoadBitmap( bmIndMidActive   , idbmIndMidActive   );
    LoadBitmap( bmIndMidInactive , idbmIndMidInactive );
    LoadBitmap( bmIndHighActive  , idbmIndHighActive  );
    LoadBitmap( bmIndHighInactive, idbmIndHighInactive);
    LoadBitmap( bmKnob           , idbmKnobNew        );
    LoadBitmap( bmSmallKnob      , idbmSmallKnobNew   );
    LoadBitmap( bmNoKnob         , idbmNoKnob         );
    LoadBitmap( bmSkin           , idbmSkin2          );
  end;

  procedure FreeBitmaps;
  // Free the bitmaps loaded at program startup
  begin
    FreeAndnil( bmKnob           );
    FreeAndnil( bmSmallKnob      );
    FreeAndnil( bmNoKnob         );
    FreeAndnil( bmIndActive      );
    FreeAndnil( bmIndInactive    );
    FreeAndnil( bmIndLowActive   );
    FreeAndnil( bmIndLowInactive );
    FreeAndnil( bmIndMidActive   );
    FreeAndnil( bmIndMidInactive );
    FreeAndnil( bmIndHighActive  );
    FreeAndnil( bmIndHighInactive);
    FreeAndnil( bmConIn          );
    FreeAndnil( bmConOut         );
    FreeAndnil( bmDn             );
    FreeAndnil( bmUp             );
    FreeAndnil( bmSkin           );
  end;

  procedure   CalcBezierCurve( var Points: TKnobsBezierPoints; const Params: TKnobsBezierParams);
  // Used to calculate wire bending - for rotate etc. see 'knobutils.pas'.
  var
    Dir1 : TPoint;
    Dir2 : TPoint;
  begin
    Dir1       := Rotate( Points[ 0] - Points[ 3], Params.Angle1);
    Dir2       := Rotate( Points[ 0] - Points[ 3], Params.Angle2);
    Points[ 1] := Points[ 0] - DivPoint( MultPoint( Dir1, Params.Strength1), 100);
    Points[ 2] := Points[ 3] + DivPoint( MultPoint( Dir2, Params.Strength2), 100);
  end;

  function  ConnectorName( aConnector: TKnobsConnector): string;
  begin
    if Assigned( aConnector)
    then begin
      if aConnector.HasParent
      then Result := Format( '%s.%s', [ aConnector.Parent.Name, aConnector.Name], AppLocale)
      else Result := Format( '<nil>.%s', [ aConnector.Name], AppLocale);
    end
    else Result := '<nil>.<nil>';
  end;

  function  WireName( aSource, aDestination: TKnobsConnector): string;
  begin
    Result := Format( '%s -> %s', [ ConnectorName( aSource), ConnectorName( aDestination)], AppLocale);
  end;

  function  CleanControlName( const aName: string): string;
  var
    aParts : TStringList;
  begin
    aParts := Explode( aName, '_');
    try
      if aParts.Count >= 0
      then Result := aParts[ aParts.Count - 1]
      else Result := '';
    finally
      aParts.Free;
    end;
  end;


{ ========
  TKnobsLed = class( TGraphicControl)
  private
    FTime       : Cardinal;
    FTimer      : TTimer;
    FColorOff   : TColor;
    FColorOn    : TColor;
    FHighLight  : Boolean;
    FActive     : Boolean;
    FShape      : TKnobsLedShape;
  published
    property    Time       : Cardinal       read FTime       write SetTime       default 0;
    property    Active     : Boolean        read FActive     write SetActive     default True;
    property    ColorOn    : TColor         read FColorOn    write SetColorOn    default clLime;
    property    ColorOff   : TColor         read FColorOff   write SetColorOff   default clGreen;
    property    HighLight  : Boolean        read FHighLight  write SetHighLight  default True;
    property    Shape      : TKnobsLedShape read FShape      write SetShape      default lsRound;
    property    Height                                                           default 15;
    property    Width                                                            default 15;
    property    Hint;
    property    ShowHint;
    property    ParentShowHint                                                   default True;
    property    Dragcursor;
    property    Dragmode;
    property    OnDragDrop;
    property    OnDragOver;
    property    OnEndDrag;
    property    OnMouseDown;
    property    OnMouseMove;
    property    OnMouseUp;
  private
}

    procedure   TKnobsLed.TimerFired( aSender: TObject);
    begin
      if FTime <> 0
      then begin
        Active := False;
        Stoptimer;
      end;
    end;

    procedure   TKnobsLed.StartTimer( aValue: Cardinal);
    begin
      if not ( csDesigning in ComponentState)
      then begin
        if not Assigned( FTimer)
        then FTimer := TTimer.Create( nil);
        if Assigned( FTimer)
        then begin
          with FTimer
          do begin
            OnTimer  := nil;
            Enabled  := False;
            Interval := FTime;
            Enabled  := True;
            OnTimer  := TimerFired;
          end;
        end;
      end;
    end;

    procedure   TKnobsLed.StopTimer;
    begin
      if Assigned( FTimer)
      then begin
        with FTimer
        do begin
          OnTimer := nil;
          Enabled := False;
        end;
        FreeAndNil( FTimer);
      end;
    end;

    procedure   TKnobsLed.SetTime( aValue: Cardinal);
    begin
      if aValue <> FTime
      then begin
        FTime := aValue;
        if ( FTime > 0) and Active
        then Starttimer( FTime)
      end;
    end;

    procedure   TKnobsLed.SetActive( aValue: Boolean);
    begin
      if csDestroying in ComponentState
      then Exit;

      if aValue <> FActive
      then begin
        FActive := aValue;
        Invalidate;
      end;
      if FActive and ( FTime > 0)
      then StartTimer( FTime)
      else StopTimer;
    end;

    procedure   TKnobsLed.SetColorOn( aValue: TColor);
    begin
      if csDestroying in ComponentState
      then Exit;

      if aValue <> FColorOn
      then begin
        FColorOn := aValue;
        Invalidate;
      end;
    end;

    procedure   TKnobsLed.SetColorOff( aValue: TColor);
    begin
      if csDestroying in ComponentState
      then Exit;

      if aValue <> FColorOff
      then begin
        FColorOff := aValue;
        Invalidate;
      end;
    end;

    procedure   TKnobsLed.SetHighLight( aValue: Boolean);
    begin
      if csDestroying in ComponentState
      then Exit;

      if aValue <> FHighLight
      then begin
        FHighLight := aValue;
        Invalidate;
      end;
    end;

    procedure   TKnobsLed.SetShape( aValue: TKnobsLedShape);
    begin
      if csDestroying in ComponentState
      then Exit;

      if aValue <> FShape
      then begin
        FShape := aValue;
        Invalidate;
      end;
    end;

//  protected

    procedure   TKnobsLed.Paint; // override;
    begin
      with Canvas
      do begin
        Pen.Width := 1;
        Pen.Color := clBlack;

        if Active
        then Brush.Color := FColorOn
        else Brush.Color := FColorOff;

        case Shape of

          lsRound :
            begin
              Ellipse( 0, 0, Width, Height);

              if FHighLight
              then begin
                Pen.Color := clWhite;
                Pen.Width := 2;
                Arc(
                  Width  * 3 div 4         ,
                  Height * 3 div 4         ,
                  Width  - Width  * 3 div 4,
                  Height - Height * 3 div 4,
                  Width  div 2             ,
                  0                        ,
                  0                        ,
                  Height div 2
                );
              end;
            end;

          lsSquare :
            begin
              Rectangle( 0, 0, Width, Height);
            end;

        end;
      end;
    end;

//  public

    constructor TKnobsLed.Create( aOwner: TComponent); // override;
    begin
      inherited;
      Height    := 15;
      Width     := 15;
      Shape     := lsRound;
      ColorOn   := clLime;
      ColorOff  := clGreen;
      HighLight := True;
      Active    := True;
    end;

    destructor  TKnobsLed.Destroy; // override;
    begin
      StopTimer;
      inherited;
    end;

    procedure   TKnobsLed.Toggle;
    begin
      if csDestroying in ComponentState
      then Exit;

      Active := not Active;
    end;


{ ========
  TKnobsClicker = class( TSpeedButton)
  // Repeating speed button, semi flat ...
  private
    FRepeatTimer : TTimer;
    FRepeatCount : Integer;
    FCanRepeat   : Boolean;
    FMouseDown   : Boolean;
    FShowFocus   : Boolean;
    FBitmap      : TBitmap;
  private
    property    ShowFocus: Boolean read FShowFocus write SetShowFocus;
  public
    property    CanRepeat: Boolean read FCanRepeat write FCanRepeat default True;
  private
}

    procedure   TKnobsClicker.TimerExpired( aSender: TObject);
    begin
      with FRepeatTimer
      do begin
        if FRepeatCount > SlowRepeats
        then Interval := FastRepeatPause
        else Interval := SlowRepeatPause;
      end;

      if FMouseDown and MouseCapture
      then begin
        try
          Click;
        except
          FRepeatTimer.Enabled := False;
          raise;
        end;
      end;

      Inc( FRepeatCount);
    end;

    procedure   TKnobsClicker.SetShowFocus( aValue: Boolean);
    begin
      if ( Parent is TKnobsNoKnob) or ( Parent is TKnobsSlider)
      then begin
        Flat   := False;
        Glyph  := FBitmap;
        Invalidate;
      end
      else begin
        if aValue <> FShowFocus
        then begin
          FShowFocus  := aValue;
          Flat        := not FShowFocus;
          if ShowFocus
          then Glyph := FBitmap
          else Glyph := nil;
          Invalidate;
        end;
      end;
    end;

//  private

    function    TKnobsClicker.FindWirePanel: TKnobsWirePanel;
    var
      aParent : TWinControl;
    begin
      Result  := nil;
      aParent := Parent;

      while Assigned( aParent) and not ( aParent is TKnobsWirePanel)
      do aParent :=  aParent.Parent;

      if aParent is TKnobsWirePanel
      then Result := TKnobsWirePanel( aParent);
    end;

//  protected

    procedure   TKnobsClicker.MouseDown( Button: TMouseButton; Shift: TShiftState; X, Y: Integer); // override;
    var
      aWP : TKnobsWirePanel;
    begin
      inherited;

      if FMouseDown
      then Exit;

      if CanRepeat and ( Button = mbLeft)
      then begin
        FMouseDown   := True;
        FRepeatCount := 0;

        if not Assigned( FRepeatTimer)
        then FRepeatTimer := TTimer.Create( Self);

        with FRepeatTimer
        do begin
          OnTimer  := TimerExpired;
          Interval := InitRepeatPause;
          Enabled  := True;
        end;
      end
      else if Button = mbRight
      then begin
        aWP := FindWirePanel;
        if Assigned( aWP)
        then aWP.HandleRightClick( Self);
      end;
    end;

    procedure   TKnobsClicker.MouseUp( Button: TMouseButton; Shift: TShiftState; X, Y: Integer); // override;
    begin
      inherited;

      if Assigned( FRepeatTimer)
      then FRepeatTimer.Enabled  := False;

      FMouseDown := False;
    end;

    procedure   TKnobsClicker.AssignBitmap( aBitmap: TBitmap);
    begin
      FBitmap := aBitmap;
    end;

//  public

    constructor TKnobsClicker.Create( anOwner: TComponent); // override;
    begin
      inherited;

      CanRepeat := True;
      if ( anOwner is TKnobsNoKnob) or ( anOwner is TKnobsSlider)
      then begin
        Transparent := False;
        Flat        := False;
      end
      else begin
        Transparent := True;
        Flat        := True;
      end;
    end;

    destructor  TKnobsClicker.Destroy; // override;
    begin
      if Assigned( FRepeatTimer)
      then FRepeatTimer.Free;

      inherited;
    end;

    procedure   TKnobsClicker.Click; // override;
    begin
      inherited;

      if HasParent
      then Parent.SetFocus;
    end;


{ ========
  TKnobsValuedControl = class( TCustomControl)
  // Base type for valued controls, ones that can have an effect on an external model
  private
    FFriendlyName     : string;
    FStepCount        : Integer;             // Number of steps in the total knob travel
    FLocked           : Boolean;
    FKnobPosition     : Integer;             // Current knob position
    FDefaultPosition  : Integer;             // default knob position;
    FControlType      : string;              // The control's behaviour
    FOnValueChanged   : TKnobsOnValueChange; // Called with the new value
    FDisplay          : TKnobsDisplay;
    FPopupCount       : Integer;
    FPopupWindow      : THintWindow;
    FWheelSupport     : Boolean;
    FWheelSensitivity : Integer;
    FWheelCounter     : Integer;
    FOnUnFocus        : TOnUnFocus;
  public
    property    AsDisplay        : string          read GetAsDisplay;
    property    AsHint           : string          read GetAsHint;
    property    AsValue          : TSignal         read GetAsValue;
    property    WheelSupport     : Boolean         read FWheelSupport     write FWheelSupport;
    property    WheelSensitivity : Integer         read FWheelSensitivity write FWheelSensitivity;
    property    WheelCounter     : Integer         read FWheelCounter     write FWheelCounter;
  published
    property    Action;
    property    FriendlyName     : string          read FFriendlyName     write FFriendlyName;
    // NOTE: declaration order is important here to avoid loss of resolution on a load action ...
    property    Display          : TKnobsDisplay   read FDisplay          write SetDisplay;
    property    StepCount        : Integer         read FStepCount        write SetStepCount       default 1;
    property    KnobPosition     : Integer         read FKnobPosition     write SetKnobPosition    default 0;
    property    DefaultPosition  : Integer         read FDefaultPosition  write SetDefaultPosition default 0;
    property    ControlType      : string          read FControlType      write SetControlType              ;
    property    Locked           : Boolean         read FLocked           write SetLocked                   ;
    // END NOTE.
    property    TabOrder;
    property    TabStop;
    property    Visible;
    property    OnValueChanged  : TKnobsOnValueChange  read FOnValueChanged  write SetOnValueChanged;
    property    OnUnFocus       : TOnUnFocus           read FOnUnFocus       write FOnUnFocus;
  private
}

    procedure   TKnobsValuedControl.CommonKnobPosition;
    begin
      KnobPosition    := KnobPosition;
      DefaultPosition := DefaultPosition;
      Invalidate;
      Hint := AsHint;
    end;

    procedure   TKnobsValuedControl.FixKnobPosition;
    begin
      CommonKnobPosition;
      ValueChanged( False);
    end;

    procedure   TKnobsValuedControl.FinalizeKnobPosition;
    begin
      CommonKnobPosition;
      ValueChanged( True);
    end;

//  private

    procedure   TKnobsValuedControl.SetDefaultPosition( aValue: Integer); // virtual;
    begin
      aValue := Clip( aValue, 0, StepCount);
      if aValue <> FDefaultPosition
      then FDefaultPosition := aValue;
    end;

    procedure   TKnobsValuedControl.SetStepCount( aValue: Integer); // virtual;
    begin
      if aValue <= 0
      then aValue := 1;
      if aValue <> FStepCount
      then begin
        FStepCount   := aValue;
        FinalizeKnobPosition;
      end;
    end;

    procedure   TKnobsValuedControl.SetKnobPosition( aValue: Integer); // virtual;
    begin
      aValue := Clip( aValue, 0, StepCount - 1);
      if aValue <> FKnobPosition
      then begin
        FKnobPosition := aValue;
        if csDesigning in ComponentState
        then DefaultPosition := FKnobPosition;
        FixKnobPosition;
        ActivatePopup;
      end;
    end;

    procedure   TKnobsValuedControl.SetLocked( aValue: Boolean);
    begin
      if aValue <> FLocked
      then begin
        FLocked := aValue;
        Invalidate;
      end;
    end;

    procedure   TKnobsValuedControl.SetControlType( const aValue: string);
    begin
      if aValue <> FControlType
      then begin
        FControlType := aValue;
        FinalizeKnobPosition;
      end;
    end;

    function    TKnobsValuedControl.GetAsHint: string;
    var
      aParts : TStringList;
      S      : string;
    begin
      if csDesigning in ComponentState
      then Result := Converters.PosToDisplay( ControlType, KnobPosition, StepCount)
      else begin
        aParts := Explode( Name, '_');
        try
          if FriendlyName = ''
          then S := aParts[ aParts.Count - 1]
          else S := FriendlyName;
          Result := Format( '%s'^M^M'%s', [ S, Converters.PosToHint( ControlType, KnobPosition, StepCount)], AppLocale);
        finally
          aParts.Free;
        end;
      end;
    end;

    function    TKnobsValuedControl.GetAsDisplay: string;
    begin
      Result := Converters.PosToDisplay( ControlType, KnobPosition, StepCount);
    end;

    function    TKnobsValuedControl.GetAsValue: TSignal;
    begin
      Result := 0;
      if Locked and Assigned( FDisplay)
      then ParseTextValue( FDisplay.Text, Result, Converters.Rate( ControlType))
      else Result := Converters.PosToValue( ControlType, KnobPosition, StepCount);
    end;

    procedure   TKnobsValuedControl.SetOnValueChanged( aValue: TKnobsOnValueChange); // virtual;
    begin
      FOnValueChanged := aValue;
      ValueChanged( True);
    end;

    procedure   TKnobsValuedControl.SetDisplay( aValue: TKnobsDisplay);
    begin
      FDisplay := aValue;

      if Assigned( FDisplay)
      then begin
        if not Locked
        then FDisplay.SetText( AsDisplay);
        FDisplay.FController := Self;
      end;
    end;

    procedure   TKnobsValuedControl.NotifyPeers; // virtual;
    begin
      ActivatePopup;
    end;

    procedure   TKnobsValuedControl.DoWheel( aSender: TObject; aShift: TShiftState; aWheelDelta: Integer; aMousePos: TPoint; var Handled: Boolean);
    begin
      if WheelSupport and not ( csDesigning in ComponentState) and not Locked
      then begin
        {$Q-R-}
        FWheelCounter := FWheelCounter + aWheelDelta;
        {$Q+R+}
        if aWheelDelta > 0
        then HandleWheelUp( aWheelDelta)
        else if aWheelDelta < 0
        then HandleWheelDown( - aWheelDelta)
        else FWheelCounter := 0;
        Handled := aWheelDelta <> 0;
      end
      else Handled := False;
    end;

//  protected

    procedure   TKnobsValuedControl.ValueChanged( IsFinal: Boolean);
    var
      aModule : TKnobsCustomModule;
    begin
      NotifyPeers;

      if Assigned( FOnValueChanged)
      then FOnValueChanged( Self, Name, ControlType, AsValue, IsFinal)
      else begin
        aModule := FindModule;
        if Assigned( aModule)
        then aModule.ValueChanged( Self, MakePath( [ aModule.Name, Name]), ControlType, AsValue, IsFinal);
      end;
    end;

    function    TKnobsValuedControl.FindModule: TKnobsCustomModule;
    var
      aParent : TWinControl;
    begin
      Result  := nil;
      aParent := Parent;

      while Assigned( aParent) and not ( aParent is TKnobsCustomModule)
      do aParent :=  aParent.Parent;

      if aParent is TKnobsCustomModule
      then Result := TKnobsCustomModule( aParent);
    end;

    function    TKnobsValuedControl.FindWirePanel: TKnobsWirePanel;
    var
      aParent : TWinControl;
    begin
      Result  := nil;
      aParent := Parent;

      while Assigned( aParent) and not ( aParent is TKnobsWirePanel)
      do aParent :=  aParent.Parent;

      if aParent is TKnobsWirePanel
      then Result := TKnobsWirePanel( aParent);
    end;

    procedure   TKnobsValuedControl.BeginStateChange;
    var
      aWp : TKnobsWirePanel;
    begin
      aWp := FindWirePanel;
      if Assigned( aWp)
      then aWp.BeginStateChange;
    end;

    procedure   TKnobsValuedControl.EndStateChange;
    var
      aWp : TKnobsWirePanel;
    begin
      aWp := FindWirePanel;
      if Assigned( aWp)
      then aWp.EndStateChange( False);
    end;

    procedure   TKnobsValuedControl.ActivatePopup;
    var
      P : TPoint;
      R : TRect;
    begin
      if Assigned( FPopupWindow) and ( ControlType <> '')
      then begin
        R := FPopupWindow.CalcHintRect( 200, Hint, nil);
        P := ClientToScreen( Point( Width, Height));
        R.Offset( P.X, P.Y);
        FPopupWindow.ActivateHint( R, Hint);
      end;
    end;

    procedure   TKnobsValuedControl.ShowPopup;
    begin
      if ( FPopupCount = 0) and ( ControlType <> '')
      then begin
        FPopupWindow := THintWindow.Create( self);
        FPopupWindow.DoubleBuffered := True;
        ActivatePopup;
      end;
      Inc( FPopupCount);
    end;

    procedure   TKnobsValuedControl.HidePopup;
    begin
      if FPopupCount > 0
      then Dec( FPopupCount);
      if FPopupCount = 0
      then FreeAndNil( FPopupWindow);
    end;

    procedure   TKnobsValuedControl.ForceHidePopup;
    begin
      FPopupCount := 0;
      FreeAndNil( FPopupWindow);
    end;

    procedure   TKnobsValuedControl.HandleWheelUp( anAmount: Integer); // virtual;
    begin
      // Nothing here - abstract ...
    end;

    procedure   TKnobsValuedControl.HandleWheelDown( anAmount: Integer); // virtual;
    begin
      // Nothing here - abstract ...
    end;

    procedure   TKnobsValuedControl.UnFocus; // virtual;
    var
      aModule : TKnobsCustomModule;
    begin
      if Assigned( FOnUnFocus)
      then FOnUnFocus( Self)
      else begin
        aModule := FindModule;
        if Assigned( aModule)
        then aModule.UnFocus( Self);
      end;
    end;

//  public

    constructor TKnobsValuedControl.Create( anOwner: TComponent); // override;
    begin
      inherited;
      FWheelSupport     := False;
      FWheelSensitivity := 50;
      StepCount         := 1;
      KnobPosition      := 0;
      DefaultPosition   := KnobPosition;
      OnMouseWheel      := DoWheel;
    end;

    procedure   TKnobsValuedControl.SetDefault;
    begin
      Locked       := False;
      KnobPosition := DefaultPosition;
    end;

    procedure   TKnobsValuedControl.FixDisplay;
    var
      Old : Integer;
    begin
      Old := KnobPosition;
      if KnobPosition > 0
      then KnobPosition := KnobPosition - 1
      else KnobPosition := KnobPosition + 1;
      KnobPosition := Old;
    end;

    procedure   TKnobsValuedControl.Dump( var aFile: TextFile; anInd: Integer);
    begin
      WriteLnInd( aFile, anInd, Format( 'ValuedControl "%s" = %f', [ Name, AsValue], AppLocale));
    end;


{ ========
  TKnobsXYControl = class( TCustomControl)
  // Base type for two valued controls, ones that can have an effect on an external model
  private
    FFriendlyName     : string;
    FStepCount        : Integer;             // Number of steps in the total knob travel, same for X and Y directions
    FKnobPositionX    : Integer;             // Current knob X position
    FKnobPositionY    : Integer;             // Current knob Y position
    FDefaultPositionX : Integer;             // default knob X position;
    FDefaultPositionY : Integer;             // default knob Y position;
    FControlType      : string;              // The control's behaviour
    FOnValueChanged   : TKnobsOnValueChange; // Called with the new value - twice, once for X, once for Y
    FPopupCount       : Integer;
    FPopupWindow      : THintWindow;
    FOnUnFocus        : TOnUnFocus;
  public
    property    AsHint           : string          read GetAsHint;
    property    AsValueX         : TSignal         read GetAsValueX;
    property    AsValueY         : TSignal         read GetAsValueY;
  published
    property    Action;
    property    FriendlyName     : string          read FFriendlyName      write FFriendlyName;
    // NOTE: declaration order is important here to avoid loss of resolution on a load action ...
    property    StepCount        : Integer         read FStepCount         write SetStepCount        default 1;
    property    KnobPositionX    : Integer         read FKnobPositionX     write SetKnobPositionX    default 0;
    property    KnobPositionY    : Integer         read FKnobPositionY     write SetKnobPositionY    default 0;
    property    DefaultPositionX : Integer         read FDefaultPositionX  write SetDefaultPositionX default 0;
    property    DefaultPositionY : Integer         read FDefaultPositionY  write SetDefaultPositionY default 0;
    property    ControlType      : string          read FControlType       write SetControlType              ;
    // END NOTE.
    property    TabOrder;
    property    TabStop;
    property    Visible;
    property    OnValueChanged  : TKnobsOnValueChange  read FOnValueChanged  write SetOnValueChanged;
    property    OnUnFocus       : TOnUnFocus           read FOnUnFocus       write FOnUnFocus;
  private
}

    procedure   TKnobsXYControl.CommonKnobPosition;
    begin
      KnobPositionX    := KnobPositionX;
      KnobPositionY    := KnobPositionY;
      DefaultPositionX := DefaultPositionX;
      DefaultPositionY := DefaultPositionY;
      Invalidate;
      Hint := AsHint;
    end;

    procedure   TKnobsXYControl.FixKnobPosition;
    begin
      CommonKnobPosition;
      ValueChanged( False);
    end;

    procedure   TKnobsXYControl.FinalizeKnobPosition;
    begin
      CommonKnobPosition;
      ValueChanged( True);
    end;

//  private

    procedure   TKnobsXYControl.SetDefaultPositionX( aValue: Integer); // virtual;
    begin
      aValue := Clip( aValue, 0, StepCount);
      if aValue <> FDefaultPositionX
      then FDefaultPositionX := aValue;
    end;

    procedure   TKnobsXYControl.SetDefaultPositionY( aValue: Integer); // virtual;
    begin
      aValue := Clip( aValue, 0, StepCount);
      if aValue <> FDefaultPositionY
      then FDefaultPositionY := aValue;
    end;

    procedure   TKnobsXYControl.SetKnobPositionX( aValue: Integer); // virtual;
    begin
      aValue := Clip( aValue, 0, StepCount - 1);
      if aValue <> FKnobPositionX
      then begin
        FKnobPositionX := aValue;
        if csDesigning in ComponentState
        then DefaultPositionX := FKnobPositionX;
        FixKnobPosition;
        ActivatePopup;
      end;
    end;

    procedure   TKnobsXYControl.SetKnobPositionY( aValue: Integer); // virtual;
    begin
      aValue := Clip( aValue, 0, StepCount - 1);
      if aValue <> FKnobPositionY
      then begin
        FKnobPositionY := aValue;
        if csDesigning in ComponentState
        then DefaultPositionY := FKnobPositionY;
        FixKnobPosition;
        ActivatePopup;
      end;
    end;

    procedure   TKnobsXYControl.SetStepCount( aValue: Integer); // virtual;
    begin
      if aValue <= 0
      then aValue := 1;
      if aValue <> FStepCount
      then begin
        FStepCount   := aValue;
        FinalizeKnobPosition;
      end;
    end;

    procedure   TKnobsXYControl.SetControlType( const aValue: string);
    begin
      if aValue <> FControlType
      then begin
        FControlType := aValue;
        FinalizeKnobPosition;
      end;
    end;

    function    TKnobsXYControl.GetAsHint: string;
    var
      aParts : TStringList;
      S      : string;
    begin
      if csDesigning in ComponentState
      then Result := Format( '%s, %s', [ Converters.PosToDisplay( ControlType, KnobPositionX, StepCount), Converters.PosToDisplay( ControlType, KnobPositionY, StepCount)], AppLocale)
      else begin
        aParts := Explode( Name, '_');
        try
          if FriendlyName = ''
          then S := aParts[ aParts.Count - 1]
          else S := FriendlyName;
          Result := Format( '%s'^M^M'%s, %s', [ S, Converters.PosToHint( ControlType, KnobPositionX, StepCount), Converters.PosToHint( ControlType, KnobPositionY, StepCount)], AppLocale);
        finally
          aParts.Free;
        end;
      end;
    end;

    function    TKnobsXYControl.GetAsValueX: TSignal;
    begin
      Result := Converters.PosToValue( ControlType, KnobPositionX, StepCount);
    end;

    function    TKnobsXYControl.GetAsValueY: TSignal;
    begin
      Result := Converters.PosToValue( ControlType, KnobPositionY, StepCount);
    end;

    procedure   TKnobsXYControl.SetOnValueChanged( aValue: TKnobsOnValueChange);
    begin
      FOnValueChanged := aValue;
      ValueChanged( True);
    end;

    procedure   TKnobsXYControl.NotifyPeers; // virtual;
    begin
      ActivatePopup;
    end;

//  protected

    procedure   TKnobsXYControl.ValueChanged( IsFinal: Boolean); // virtual;
    var
      aModule : TKnobsCustomModule;
    begin
      NotifyPeers;

      if Assigned( FOnValueChanged)
      then begin
        FOnValueChanged( Self, Name + 'x', ControlType, AsValueX, IsFinal);
        FOnValueChanged( Self, Name + 'y', ControlType, AsValueY, IsFinal);
      end
      else begin
        aModule := FindModule;
        if Assigned( aModule)
        then begin
          aModule.ValueChanged( Self, MakePath( [ aModule.Name, Name + 'x']), ControlType, AsValueX, IsFinal);
          aModule.ValueChanged( Self, MakePath( [ aModule.Name, Name + 'y']), ControlType, AsValueY, IsFinal);
        end;
      end;
    end;

    function    TKnobsXYControl.FindModule: TKnobsCustomModule;
    var
      aParent : TWinControl;
    begin
      Result  := nil;
      aParent := Parent;

      while Assigned( aParent) and not ( aParent is TKnobsCustomModule)
      do aParent :=  aParent.Parent;

      if aParent is TKnobsCustomModule
      then Result := TKnobsCustomModule( aParent);
    end;


    function    TKnobsXYControl.FindWirePanel: TKnobsWirePanel;
    var
      aParent : TWinControl;
    begin
      Result  := nil;
      aParent := Parent;

      while Assigned( aParent) and not ( aParent is TKnobsWirePanel)
      do aParent :=  aParent.Parent;

      if aParent is TKnobsWirePanel
      then Result := TKnobsWirePanel( aParent);
    end;

    procedure   TKnobsXYControl.BeginStateChange;
    var
      aWp : TKnobsWirePanel;
    begin
      aWp := FindWirePanel;
      if Assigned( aWp)
      then aWp.BeginStateChange;
    end;

    procedure   TKnobsXYControl.EndStateChange;
    var
      aWp : TKnobsWirePanel;
    begin
      aWp := FindWirePanel;
      if Assigned( aWp)
      then aWp.EndStateChange( False);
    end;

    procedure   TKnobsXYControl.ActivatePopup;
    var
      P : TPoint;
      R : TRect;
    begin
      if Assigned( FPopupWindow) and ( ControlType <> '')
      then begin
        R := FPopupWindow.CalcHintRect( 200, Hint, nil);
        P := ClientToScreen( Point( Width, Height));
        R.Offset( P.X, P.Y);
        FPopupWindow.ActivateHint( R, Hint);
      end;
    end;

    procedure   TKnobsXYControl.ShowPopup;
    begin
      if ( FPopupCount = 0) and ( ControlType <> '')
      then begin
        FPopupWindow := THintWindow.Create( self);
        FPopupWindow.DoubleBuffered := True;
        ActivatePopup;
      end;
      Inc( FPopupCount);
    end;

    procedure   TKnobsXYControl.HidePopup;
    begin
      if FPopupCount > 0
      then Dec( FPopupCount);
      if FPopupCount = 0
      then FreeAndNil( FPopupWindow);
    end;

    procedure   TKnobsXYControl.ForceHidePopup;
    begin
      FPopupCount := 0;
      FreeAndNil( FPopupWindow);
    end;

    procedure   TKnobsXYControl.UnFocus; // virtual;
    var
      aModule : TKnobsCustomModule;
    begin
      if Assigned( FOnUnFocus)
      then FOnUnFocus( Self)
      else begin
        aModule := FindModule;
        if Assigned( aModule)
        then aModule.UnFocus( Self);
      end;
    end;

//  public

    constructor TKnobsXYControl.Create( anOwner: TComponent); // override;
    begin
      inherited;
      StepCount         := 1;
      KnobPositionX     := 0;
      KnobPositionY     := 0;
      DefaultPositionX  := KnobPositionX;
      DefaultPositionY  := KnobPositionY;
    end;

    procedure   TKnobsXYControl.SetDefault;
    begin
      KnobPositionX := DefaultPositionX;
      KnobPositionY := DefaultPositionY;
    end;

    procedure   TKnobsXYControl.Dump( var aFile: TextFile; anInd: Integer);
    begin
      WriteLnInd( aFile, anInd, Format( 'XYControl "%s" = %f, %f', [ Name, AsValueX, AsValueY], AppLocale));
    end;


{ ========
  TKnobsSelector = class( TKnobsValuedControl)
  private
    FCaptions       : TStringList;
    FGlyphs         : TImageList;
    FBorderColor    : TColor;
    FMouseInControl : Boolean;
    FOlDFontColor   : TColor;
    FOldFontStyle   : TFontStyles;
    FOldColor       : TColor;
  published
    property    Captions    : TStringList read FCaptions    write SetCaptions;
    property    Glyphs      : TImageList  read FGlyphs      write SetGlyphs;
    property    Color                                                             default clGray;
    property    BorderColor : TColor      read FBorderColor write SetBorderColor  default clWhite;
    property    Font;
  private
}

    procedure   TKnobsSelector.SetCaptions( const aValue: TStringList);
    begin
      if Assigned( FCaptions)
      then begin
        FCaptions.Assign( aValue);
        StepCount := FCaptions.Count;
        Invalidate;
      end;
    end;

    procedure   TKnobsSelector.SetGlyphs( const aValue: TImageList);
    begin
      if Assigned( aValue)
      then begin
        FGlyphs   := aValue;
        if Assigned( FGlyphs) and Assigned( FCaptions)
        then StepCount := Min( FGlyphs.Count, FCaptions.Count)
        else if Assigned( FGlyphs)
        then StepCount := FGlyphs.Count
        else if Assigned( FCaptions)
        then StepCount := FCaptions.Count
        else StepCount := 0;
        Invalidate;
      end
    end;

    procedure   TKnobsSelector.SetBorderColor( aValue: TColor);
    begin
      if aValue <> FBorderColor
      then begin
        FBorderColor := aValue;
        Invalidate;
      end;
    end;

//  private

    procedure   TKnobsSelector.WMSetFocus( var aMessage: TWMSetFocus  ); // message WM_SETFOCUS;
    begin
      FOldColor := Color;
      Color     := clBlack;
      inherited;
    end;

    procedure   TKnobsSelector.WMKillFocus( var aMessage: TWMKillFocus ); // message WM_KILLFOCUS;
    begin
      Color := FOldColor;
      inherited;
    end;

    procedure   TKnobsSelector.WMGetDlgCode( var aMessage: TWMGetDlgCode); // message WM_GETDLGCODE;
    begin
      aMessage.Result := DLGC_WANTARROWS;
    end;

    procedure   TKnobsSelector.CMMouseEnter( var aMessage: TMessage); // message CM_MOUSEENTER;
    begin
      if not FMouseInControl
      then begin
        FMouseInControl := True;
        FOldFontColor   := Font.Color;
        FOldFontStyle   := Font.Style;
        Font.Color      := BorderColor;
        Font.Style      := [ fsBold];
        ShowPopup;
      end;
      inherited;
    end;

    procedure   TKnobsSelector.CMMouseLeave( var aMessage: TMessage); // message CM_MOUSELEAVE;
    begin
      if FMouseInControl
      then begin
        FMouseInControl := False;
        Font.Color      := FOldFontColor;
        Font.Style      := FOldFontStyle;
        HidePopup;
      end;
      inherited;
    end;

//  protected

    procedure   TKnobsSelector.DoClick( aSender: TObject);
    begin
      if KnobPosition < StepCount - 1
      then KnobPosition := KnobPosition + 1
      else KnobPosition := 0;
      FinalizeKnobPosition;
      Invalidate;
    end;

    procedure   TKnobsSelector.MouseDown( Button: TMouseButton; Shift: TShiftState; X, Y: Integer); // override;
    var
      aWP: TKnobsWirePanel;
    begin
      if Button = mbRight
      then begin
        aWP := FindWirePanel;
        if Assigned( aWP)
        then aWP.HandleRightClick( Self);
      end
      else inherited;
    end;

    procedure   TKnobsSelector.Paint; // override;
    var
      aTextSize : TSize;
      aLeft     : Integer;
      aTop      : Integer;
    begin
      if ( KnobPosition >= 0) and ( KnobPosition < Captions.Count)
      then Caption := Captions[ KnobPosition]
      else Caption := 'invalid';

      with Canvas
      do begin
        Pen.Color   := BorderColor;
        Pen.Width   := 1;
        Brush.Style := bsSolid;
        Brush.Color := Color;
        Font        := Self.Font;
        Rectangle( 0, 0, Width, Height);
        if Assigned( FGlyphs)
        then FGlyphs.Draw( Canvas, ( Width - FGlyphs.Width) div 2, ( Height - FGlyphs.Height) div 2, KnobPosition, True)
        else begin
          aTextSize := TextExtent( Caption);
          aLeft     := ( Self.Width  - aTextSize.cx) div 2;
          aTop      := ( Self.Height - aTextSize.cy) div 2;
          TextOut( aLeft, aTop, Caption);
        end;
        Brush.Style := bsClear;
        Rectangle( 0, 0, Width, Height);
      end;
    end;

//  public

    constructor TKnobsSelector.Create( anOwner: TComponent); // override;
    begin
      inherited;
      ControlStyle    := ControlStyle - [ csOpaque] + [ csSetCaption];
      FCaptions       := TStringList.Create;
      KnobPosition    :=  0;
      Height          := 16;
      Width           := 35;
      Color           := clGray;
      Font.Color      := clWhite;
      BorderColor     := clWhite;
      TabStop         := True;
      OnClick         := DoClick;
    end;

    destructor  TKnobsSelector.Destroy; // override;
    begin
      FreeAndNil( FCaptions);
      inherited;
    end;

    procedure   TKnobsSelector.FixPeers;
    begin
      FinalizeKnobPosition;
    end;


{ ========
  TKnobsKnob = class( TKnobsValuedControl)
  // A rotary knob, or linear horizontal or vertical control, mouse wheel works too.
  private
    FDownClicker       : TKnobsClicker;
    FUpClicker         : TKnobsClicker;
    FMouseDown         : Boolean;
    FGraphic           : TBitmap;
    FMouseInControl    : Boolean;
    FTransparent       : Boolean;
    FPairedWith        : TKnobsKnob;
    FPairingMode       : TKnobsPairingMode;
    FPairingLock       : Integer;
    FControlMode       : TDistanceMode;
    FFinalizationTimer : TTimer;
    FTresholdPast      : Boolean;
  published
    property    Action;
    property    Transparent  : Boolean           read FTransparent   write SetTransparent default False;
    property    Color;
    property    TabOrder;
    property    TabStop;
    property    ParentColor                                                               default True;
    property    PairedWith   : TKnobsKnob        read FPairedWith    write SetPairedWith;
    property    PairingMode  : TKnobsPairingMode read FPairingMode   write SetPairingMode;
    property    ControlMode  : TDistanceMode     read FControlMode   write SetControlMode;
  private
}

    procedure   TKnobsKnob.InitializeFinalizationTimer;
    begin
      FFinalizationTimer := TTimer.Create( nil);
      StopFinalizationTimer;
      FFinalizationTimer.OnTimer := FinalizationTimerFired;
    end;

    procedure   TKnobsKnob.FinalizeFinalizationTimer;
    begin
      FFinalizationTimer.OnTimer := nil;
      StopFinalizationTimer;
      FreeAndNil( FFinalizationTimer);
    end;

    procedure   TKnobsKnob.StartFinalizationTimer;
    begin
      StopFinalizationTimer;
      FFinalizationTimer.Interval := 739;
      FFinalizationTimer.Enabled  := True;
    end;

    procedure   TKnobsKnob.StopFinalizationTimer;
    begin
      FFinalizationTimer.Enabled := False;
    end;

    procedure   TKnobsKnob.FinalizationTimerFired( aSender: TObject);
    begin
      StopFinalizationTimer;
      FinalizeKnobPosition;
    end;

//  private

    procedure   TKnobsKnob.WMSetFocus( var aMessage: TWMSetFocus ); // message WM_SETFOCUS;
    begin
      FUpClicker  .ShowFocus := True;
      FDownClicker.ShowFocus := True;
      Invalidate;

      inherited;
    end;

    procedure   TKnobsKnob.WMKillFocus( var aMessage: TWMKillFocus); // message WM_KILLFOCUS;
    begin
      FUpClicker  .ShowFocus := False;
      FDownClicker.ShowFocus := False;
      Invalidate;

      inherited;
    end;

    procedure   TKnobsKnob.WMGetDlgCode( var aMessage: TWMGetDlgCode); // message WM_GETDLGCODE;
    begin
      aMessage.Result := DLGC_WANTARROWS;
    end;

    procedure TKnobsKnob.CMMouseEnter( var aMessage: TMessage);
    begin
      inherited;

      ShowPopup;
      if not FMouseInControl and Enabled and ( GetCapture = 0)
      then begin
        if WheelSupport
        then SetFocus;
        FMouseInControl := True;
        Invalidate;
      end;
    end;

    procedure TKnobsKnob.CMMouseLeave( var aMessage: TMessage);
    begin
      inherited;

      HidePopup;
      if FMouseInControl and Enabled and not Dragging
      then begin
        FMouseInControl := False;
        if WheelSupport
        then UnFocus;
        Invalidate;
      end;
    end;

    procedure   TKnobsKnob.OnDownClick( aSender: TObject);
    begin
      if not Locked
      then begin
        StartFinalizationTimer;
        try
          KnobPosition := KnobPosition - 1;
        except
          on E: Exception
          do KilledException( E);
        end;
      end;
    end;

    procedure   TKnobsKnob.OnUpClick( aSender: TObject);
    begin
      if not Locked
      then begin
        StartFinalizationTimer;
        try
          KnobPosition := KnobPosition + 1;
        except
          on E: Exception
          do KilledException( E);
        end;
      end;
    end;

    procedure   TKnobsKnob.NotifyPeers; // Override;
    begin
      inherited;
      if Assigned( FDisplay) and not Locked
      then FDisplay.SetText( AsDisplay);
    end;

    procedure   TKnobsKnob.LockPairing;
    begin
      Inc( FPairingLock);
    end;

    procedure   TKnobsKnob.UnlockPairing;
    begin
      if PairingLocked
      then Dec( FPairingLock);
    end;

    function    TKnobsKnob.PairingLocked: Boolean;
    begin
      Result := FPairingLock > 0;
    end;

    procedure   TKnobsKnob.FixPairing;
    begin
      if not PairingLocked and Assigned( PairedWith)
      then begin
        LockPairing;
        try
          case PairingMode of
            pmOff      : ;
            pmNormal   : PairedWith.KnobPosition := KnobPosition;
            pmMirrored : PairedWith.KnobPosition := PairedWith.StepCount - KnobPosition - 1;
          end;
        finally
          UnlockPairing;
        end;
      end;
    end;

    procedure   TKnobsKnob.SetTransparent( aValue: Boolean);
    begin
      if aValue <> FTransparent
      then begin
        FTransparent := aValue;
        Invalidate;
      end;
    end;

    procedure   TKnobsKnob.SetKnobPosition( aValue: Integer); // override;
    var
      OldVal : TSignal;
    begin
      OldVal := KnobPosition;
      inherited;
      if OldVal <> KnobPosition
      then FixPairing;
    end;

    procedure   TKnobsKnob.SetPairedWith( const aValue: TKnobsKnob);
    begin
      if aValue <> FPairedWith
      then begin
        if not PairingLocked
        then begin
          LockPairing;
          try
            if Assigned( aValue)
            then aValue.PairedWith := Self
            else if Assigned( FPairedWith)
            then begin
              FPairedWith.PairingMode := pmOff;
              FPairedWith.PairedWith  := nil;
            end;
            FPairedWith := aValue;
          finally
            UnlockPairing;
          end;
          FixPairing;
        end;
      end;
    end;

    procedure   TKnobsKnob.SetPairingMode( aValue: TKnobsPairingMode);
    begin
      if aValue <> FPairingMode
      then begin
        if not PairingLocked
        then begin
          LockPairing;
          try
            FPairingMode := aValue;
            if Assigned( FPairedWith)
            then FPairedWith.PairingMode := aValue;
          finally
            UnlockPairing;
          end;
          FixPairing;
        end;
      end;
    end;

    procedure   TKnobsKnob.SetControlMode( aValue: TDistanceMode); // virtual;
    begin
      FControlMode := aValue;
    end;

    function    TKnobsKnob.CenterPoint: TPoint;
    begin
      with Result
      do begin
        X := Width         div 2;
        Y := ( Height - 1) div 2;
      end;
    end;

    procedure   TKnobsKnob.BeginMove( aPoint: TPoint); // override;
    var
      P : TPoint;
      A : Extended;
    begin
      BeginStateChange;               // This will lock UndoRedo as well

      case ControlMode of

        dmCircular :

          begin
            P := PolarToPoint( CenterPoint, ValuesToAngle( 0, KnobPosition, StepCount - 1), Width / 2);
            with ClientToScreen( P)
            do SetCursorPos( X, Y);
          end;

        dmHorizontal :

          begin
            if StepCount > 0
            then begin
              A := Clip( KnobPosition / StepCount, 0, 1);
              P := CenterPoint + Point( Round( -100 + A * 200), GAnchor.Y);
            end
            else P := CenterPoint;
            with ClientToScreen( P)
            do SetCursorPos( X, Y);
          end;

        dmVertical :

          begin
            if StepCount > 0
            then begin
              A := Clip( KnobPosition / StepCount, 0, 1);
              P := CenterPoint + Point( GAnchor.X, Round( 100 -  A * 200));
            end
            else P := CenterPoint;
            with ClientToScreen( P)
            do SetCursorPos( X, Y);
          end;

      end;

      GDistance   := CenterPoint - P;
      GStartPoint := GAnchor;
      GAnchor     := CenterPoint;
      Application.ActivateHint( ClientToScreen( CenterPoint));
    end;

    procedure   TKnobsKnob.EndMove( aPoint: TPoint); // override;
    var
      aVal : TSignal;
    begin
      with ClientToScreen( GStartPoint)
      do SetCursorPos( X, Y);
      Application.HideHint;
      aVal := DistanceToValue( 0, StepCount - 1, aPoint, ControlMode);
      KnobPosition := Round( aVal);
      FinalizeKnobPosition;
      EndStateChange;                 // This will unlock UndoRedo as well
    end;

    procedure   TKnobsKnob.UpdateMove( aPoint: TPoint); // override;
    var
      aVal : TSignal;
    begin
      aVal := DistanceToValue( 0, StepCount - 1, aPoint, ControlMode);
      KnobPosition := Round( aVal);
    end;

//  protected

    function    TKnobsKnob.CreateClicker( aLeft, aTop: Integer; aNotifyEvent: TNotifyEvent; aBitmap: TBitmap): TKnobsClicker; // virtual
    begin
      Result := TKnobsClicker.Create( Self);

      with Result
      do begin
        Left    := aLeft;
        Top     := aTop;
        Width   := Self.Width div 2 + 1;
        Height  := ClickerSize;
        OnClick := aNotifyEvent;
        Parent  := Self;
        AssignBitmap( aBitmap);
        SetShowFocus( ClassType = TKnobsNoKnob);
      end;
    end;

    procedure   TKnobsKnob.Paint; // override;

      function GetParentColor : TColor;
      begin
        Result := clSilver;
        if Assigned( Parent)
        then begin
          if Parent is TKnobsCustomModule
          then Result := TKnobsCustomModule( Parent).Color
          else if Parent is TPanel
          then Result := TPanel( Parent).Color;
        end;
      end;

      function GetParentOpacity : Byte;
      begin
        Result := 255;
        if Assigned( Parent) and ( Parent is TKnobsCustomModule)
        then Result := TKnobsCustomModule( Parent).Opacity;
      end;

      function GetParentBitmap : TBitmap;
      begin
        Result := nil;
        if Assigned( Parent) and ( Parent is TKnobsCustomModule)
        then Result := TKnobsCustomModule( Parent).FBitmap;
      end;

      function GetParentWidth : Integer;
      begin
        Result := Self.Width;
        if Assigned( Parent) and ( Parent is TControl)
        then Result := TControl( Parent).Width;
      end;

      function GetParentHeight : Integer;
      begin
        Result := Self.Height;
        if Assigned( Parent) and ( Parent is TControl)
        then Result := TControl( Parent).Height;
      end;

    const
      LockSpotSize = 3;
    var
      aReplica           : TBitmap;
      aParentBmp     : TBitmap;
      aLineStart     : TPoint;
      aLineEnd       : TPoint;
      anAngle        : TSignal;
      aParentColor   : TColor;
      aParentOpacity : Byte;
      aParentWidth   : Integer;
      aParentHeight  : Integer;
    begin
      aParentColor   := GetParentColor;
      aParentOpacity := GetParentOpacity;
      aParentBmp     := GetParentBitmap;
      aParentWidth   := GetParentWidth;
      aParentHeight  := GetParentHeight;
      aReplica           := nil;
      try
        // When the Parent has an overlay make a local replica.
        if Assigned( aParentBmp)
        then begin
          aReplica := TBitmap.Create;
          aReplica.Width  := aParentWidth;
          aReplica.Height := aParentHeight;
          aReplica.Canvas.StretchDraw( Rect( 0, 0, aReplica.Width, aReplica.Height), aParentBmp);
        end;

        with Canvas
        do begin
          // Paint background in Parent's color

          Pen.Color   := aParentColor;
          Pen.Width   := 1;
          Brush.Style := bsSolid;
          Brush.Color := aParentColor;
          Rectangle( 0, 0, width, height);

          // When the Parent has an overlay, paint it here too
          if Assigned( aReplica)
          then Draw( -Left, -Top, aReplica, aParentOpacity);

          // When the knob has a graphic for itself, paint it
          if Assigned( FGraphic)
          then begin
            FGraphic.AlphaFormat := afDefined;
            FGraphic.Transparent := False;
            Draw( 1, 1, FGraphic);
          end;

          // Paint the position indictor
          if Focused and not Locked
          then begin
            Pen.Color := clRed;
            Pen.Width := 3;
          end
          else begin
            Pen.Color := clBlack;
            Pen.Width := 1;
          end;

          anAngle    := ValuesToAngle( 0, KnobPosition, StepCount - 1);
          aLineStart := PolarToPoint( CenterPoint, anAngle, 1);
          aLineEnd   := PolarToPoint( CenterPoint, anAngle, Width div 2 - 2);
          MoveTo( aLineStart.x, aLineStart.y);
          LineTo( aLineEnd.x, aLineEnd.y);

          // Paint he locked indicator or the selected indicator
          if Locked
          then begin
            Pen.Color   := clRed;
            Pen.Width   := 1;
            Brush.Color := clRed;
            Ellipse(
              CenterPoint.X - LockSpotSize,
              CenterPoint.Y - LockSpotSize,
              CenterPoint.X + LockSpotSize + 1,
              CenterPoint.Y + LockSpotSize + 1
            );
          end
          else if focused
          then begin
            Pen.Color   := clRed;
            Pen.Width   := 1;
            Brush.Style := bsClear;
            Rectangle( 0, 0, width, height);
          end;
        end;
      finally
        FreeAndNil( aReplica);
      end;
    end;

    procedure   TKnobsKnob.MouseDown( Button: TMouseButton; Shift: TShiftState; X, Y: Integer); // override;
    var
      aWP: TKnobsWirePanel;
    begin
      inherited;

      if FMouseDown
      then Exit;

      if ( Button = mbLeft) and not Locked
      then begin
        if TabStop and ( GetFocus <> Handle) and CanFocus
        then SetFocus;

        GAnchor       := Point( X, Y);
        FTresholdPast := False;
        BeginMove( GAnchor);
        FMouseDown := True;
      end
      else if Button = mbRight
      then begin
        aWP := FindWirePanel;
        if Assigned( aWP)
        then aWP.HandleRightClick( Self);
      end;
    end;

    procedure   TKnobsKnob.MouseUp( Button: TMouseButton; Shift: TShiftState; X, Y: Integer); // override;
    begin
      inherited;

      if FMouseDown
      then begin
        FMouseDown  := False;
        if FTresholdPast
        then EndMove( Point( X - GAnchor.x, Y - GAnchor.y))
        else EndMove( GAnchor);
      end;
    end;

    procedure   TKnobsKnob.MouseMove( Shift: TShiftState; X, Y: Integer); // override;
    var
      aDistance : TPoint;
    begin
      inherited;

      if FMouseDown
      then begin
        aDistance := Point( X, Y) - GAnchor;

        if not FTresholdPast
        then FTresholdPast := Distance( aDistance) > 10;

        if FTresholdPast
        then UpdateMove( Point( X - GAnchor.x, Y - GAnchor.y));
      end;
    end;

    procedure   TKnobsKnob.HandleWheelUp( anAmount: Integer); // override;
    var
      WA : Integer;
    begin
      StartFinalizationTimer;
      try
        WA := WheelAmount;
        while WheelCounter > WA
        do begin
          KnobPosition := KnobPosition + 1;
          WheelCounter := WheelCounter - WA;
        end;
      except
        on E: Exception
        do KilledException( E);
      end;
    end;

    procedure   TKnobsKnob.HandleWheelDown( anAmount: Integer); // override;
    var
      WA : Integer;
    begin
      StartFinalizationTimer;
      try
        WA := WheelAmount;
        while WheelCounter < -WA
        do begin
          KnobPosition := KnobPosition -  1;
          WheelCounter := WheelCounter + WA;
        end;
      except
        on E: Exception
        do KilledException( E);
      end;
    end;

    procedure   TKnobsKnob.KeyDown( var Key: Word; Shift: TShiftState); // override;
    begin
      if not Locked
      then begin
        case Key of
          VK_UP   : begin FUpClicker  .Click; Key := 0; end;
          VK_DOWN : begin FDownClicker.Click; Key := 0; end;
          else      inherited;
        end;
      end;
    end;

    procedure   TKnobsKnob.AssignGraphic; // virtual;
    begin
      FGraphic := bmKnob;
    end;

    procedure   TKnobsKnob.Loaded; // override;
    begin
      inherited;

      NotifyPeers;
      Hint := AsHint;
    end;

    function    TKnobsKnob.WheelAmount: Integer; // virtual;
    begin
      Result := Max( Trunc((( 100 - Clip( WheelSensitivity, 0, 99)) * 256) / Max( StepCount, 1)), 1);
    end;

    procedure   TKnobsKnob.SetSize; // virtual;
    begin
      Width  := FGraphic.Width  + 2;
      Height := FGraphic.Height + 3;
    end;

    procedure   TKnobsKnob.CreateClickers; // virtual;
    begin
      FDownClicker := CreateClicker( 0          , Height - ClickerSize, OnDownClick, bmDn);
      FUpClicker   := CreateClicker( Width div 2, Height - ClickerSize, OnUpClick  , bmUp);
    end;

    procedure   TKnobsKnob.InvalidateDisplay;
    begin
      NotifyPeers;
    end;

//  public

    constructor TKnobsKnob.Create( anOwner: TComponent); // override;
    begin
      inherited;

      InitializeFinalizationTimer;

      ControlStyle := ControlStyle - [ csSetCaption, csOpaque];
      AssignGraphic;
      SetSize;
      CreateClickers;
      DoubleBuffered  := True;
      TabStop         := True;
      ShowHint        := False;
      Color           := clSilver;
      Transparent     := False;
      ParentColor     := True;
    end;

    destructor  TKnobsKnob.Destroy; // override;
    begin
      FinalizeFinalizationTimer;
      if Assigned( FDisplay)
      then FDisplay.FController := nil;
      inherited;
    end;

{ ========
  TKnobsSmallKnob = class( TKnobsKnob)
  // just a bit smaller than TKnob
  protected
}

    procedure   TKnobsSmallKnob.AssignGraphic; // Overide;
    begin
      FGraphic := bmSmallKnob;
    end;


{ ========
  TKnobsNoKnob = class( TKnobsKnob)
  // No knob at all, just the clickers
  protected
}

    procedure   TKnobsNoKnob.AssignGraphic; // Overide;
    begin
      FGraphic := bmNoKnob;
    end;

//  public

    constructor TKnobsNoKnob.Create( anOwner: TComponent); // override;
    begin
      inherited;

      FDownClicker.Top  := -1;        // todo : this sucks a bit ...
      FUpClicker  .Top  := -1;
      FUpClicker  .Left := FUpClicker.Left - 1;
    end;


{ ========
  TKnobsSlider = class( TKnobsKnob)
  // A vertical slider type of control
  private
}

    procedure   TKnobsSlider.SetControlMode( aValue: TDistanceMode); // override;
    begin
      FControlMode := dmVertical;
    end;

    procedure   TKnobsSlider.DoResize( aSender: TObject);
    begin
      FDownClicker.Top := Height - ClickerSize;
    end;

//  protected

    procedure   TKnobsSlider.AssignGraphic; // override;
    begin
      FGraphic := nil;
    end;

    function    TKnobsSlider.CreateClicker( aLeft, aTop: Integer; aNotifyEvent: TNotifyEvent; aBitmap: TBitmap): TKnobsClicker; // override;
    begin
      Result := TKnobsClicker.Create( Self);

      with Result
      do begin
        Left    := aLeft;
        Top     := aTop;
        Width   := ClickerSize;
        Height  := ClickerSize;
        OnClick := aNotifyEvent;
        Parent  := Self;
        Transparent := True;
        AssignBitmap( aBitmap);
        SetShowFocus( True);
      end;
    end;

    procedure   TKnobsSlider.Paint; // override;
    var
      HPos : Integer;
    begin
      with Canvas
      do begin
        if Locked or not Focused
        then Pen.Color := clGray
        else Pen.Color := clRed;
        Pen.Width   := 1;
        Brush.Style := bsClear;
        Rectangle( 0, 0, width, height);

        if StepCount > 1
        then begin
          try
            HPos := SignalToInt( RangeMap( KnobPosition, 0, StepCount - 1, Height - ClickerSize, ClickerSize))
          except
            HPos := 15;
          end;
        end
        else HPos := 15;

        if Focused and not Locked
        then Pen.Color := clRed
        else Pen.Color := clBlack;
        Pen.Width := 2;
        MoveTo( 0        , HPos);
        LineTo( width - 1, HPos);

        if locked
        then begin
          Pen.Color   := clRed;
          Pen.Width   := 2;
          Brush.Style := bsClear;
          Rectangle( 2, 2, width - 1, height - 1);
        end;
      end;
    end;

    procedure   TKnobsSlider.SetSize; // override;
    begin
      Width  :=     ClickerSize;
      Height := 2 * ClickerSize + 30;
    end;

    procedure   TKnobsSlider.CreateClickers; // override;
    begin
      FDownClicker := CreateClicker( 0, Height - ClickerSize, OnDownClick, bmDn);
      FUpClicker   := CreateClicker( 0, 0                   , OnUpClick  , bmUp);
    end;

//  public

    constructor TKnobsSlider.Create( anOwner: TComponent); // override;
    begin
      inherited;
      FControlMode := dmVertical;
      OnResize     := DoResize;
    end;


{ ========
  TKnobsPad = class( TKnobsXYControl)
  // A pad control horizontal and vertical control, no mouse wheel
  private
    FBackGround        : TBitmap;
    FBorderColor       : TColor;
    FMouseDown         : Boolean;
    FMouseInControl    : Boolean;
    FFinalizationTimer : TTimer;
    FTresholdPast      : Boolean;
  published
    property    Background  : TBitmap read FBackGround  write SetBackGround;
    property    BorderColor : TColor  read FBorderColor write SetBorderColor;
    property    Action;
    property    Color;
    property    TabOrder;
    property    TabStop;
    property    ParentColor default True;
  private
}

    procedure   TKnobsPad.SetBackGround( aValue: TBitmap);
    begin
      FBackGround.Assign( aValue);
      Invalidate;
    end;

    procedure   TKnobsPad.SetBorderColor( aValue: TColor);
    begin
      if aValue <> FBorderColor
      then begin
        FBorderColor := aValue;
        invalidate;
      end;
    end;

//  private

    procedure   TKnobsPad.InitializeFinalizationTimer;
    begin
      FFinalizationTimer := TTimer.Create( nil);
      StopFinalizationTimer;
      FFinalizationTimer.OnTimer := FinalizationTimerFired;
    end;

    procedure   TKnobsPad.FinalizeFinalizationTimer;
    begin
      FFinalizationTimer.OnTimer := nil;
      StopFinalizationTimer;
      FreeAndNil( FFinalizationTimer);
    end;

    procedure   TKnobsPad.StartFinalizationTimer;
    begin
      StopFinalizationTimer;
      FFinalizationTimer.Interval := 739;
      FFinalizationTimer.Enabled  := True;
    end;

    procedure   TKnobsPad.StopFinalizationTimer;
    begin
      FFinalizationTimer.Enabled := False;
    end;

    procedure   TKnobsPad.FinalizationTimerFired( aSender: TObject);
    begin
      StopFinalizationTimer;
      FinalizeKnobPosition;
    end;

//  private

    procedure   TKnobsPad.WMSetFocus( var aMessage: TWMSetFocus); // message WM_SETFOCUS;
    begin
      Invalidate;
      inherited;
    end;

    procedure   TKnobsPad.WMKillFocus( var aMessage: TWMKillFocus); // message WM_KILLFOCUS;
    begin
      Invalidate;
      inherited;
    end;

    procedure   TKnobsPad.WMGetDlgCode( var aMessage: TWMGetDlgCode); // message WM_GETDLGCODE;
    begin
      aMessage.Result := DLGC_WANTARROWS;
    end;

    procedure   TKnobsPad.CMMouseEnter( var aMessage: TMessage); // message CM_MOUSEENTER;
    begin
      inherited;

      ShowPopup;
      if not FMouseInControl and Enabled and ( GetCapture = 0)
      then begin
        FMouseInControl := True;
        Invalidate;
      end;
    end;

    procedure   TKnobsPad.CMMouseLeave( var aMessage: TMessage); // message CM_MOUSELEAVE;
    begin
      inherited;

      HidePopup;
      if FMouseInControl and Enabled and not Dragging
      then begin
        FMouseInControl := False;
        Invalidate;
      end;
    end;

//  private

    function    TKnobsPad.CenterPoint: TPoint;
    begin
      with Result
      do begin
        X := Width         div 2;
        Y := ( Height - 1) div 2;
      end;
    end;

    procedure   TKnobsPad.BeginMove( aPoint: TPoint); // virtual;
    var
      P : TPoint;
      A : Extended;
      B : Extended;
    begin
      BeginStateChange;               // This will lock UndoRedo as well

      if StepCount > 0
      then begin
        A := Clip( KnobPositionX / StepCount, 0, 1);
        B := Clip( KnobPositionY / StepCount, 0, 1);
        P := CenterPoint + Point( Round( -100 + A * 200), Round( 100 -  B * 200));
      end
      else P := CenterPoint;
      with ClientToScreen( P)
      do SetCursorPos( X, Y);

      GDistance   := CenterPoint - P;
      GStartPoint := GAnchor;
      GAnchor     := CenterPoint;
      Application.ActivateHint( ClientToScreen( CenterPoint));
    end;

    procedure   TKnobsPad.EndMove( aPoint: TPoint); // virtual;
    var
      aVal : TSignal;
    begin
      with ClientToScreen( GStartPoint)
      do SetCursorPos( X, Y);
      Application.HideHint;
      aVal := DistanceToValue( 0, StepCount - 1, aPoint, dmHorizontal);
      KnobPositionX := Round( aVal);
      aVal := DistanceToValue( 0, StepCount - 1, aPoint, dmVertical);
      KnobPositionY := Round( aVal);
      FinalizeKnobPosition;
      EndStateChange;                 // This will unlock UndoRedo as well
    end;

    procedure   TKnobsPad.UpdateMove( aPoint: TPoint); // virtual;
    var
      aVal : TSignal;
    begin
      aVal := DistanceToValue( 0, StepCount - 1, aPoint, dmHorizontal);
      KnobPositionX := Round( aVal);
      aVal := DistanceToValue( 0, StepCount - 1, aPoint, dmVertical);
      KnobPositionY := Round( aVal);
    end;

//  protected

    procedure   TKnobsPad.Paint; // override;
    const
      SpotSize = 3;
    var
      HPos : Integer;
      VPos : Integer;
    begin
      with Canvas
      do begin

        if Assigned( FBackGround) and not FBackGround.Empty
        then StretchDraw( Rect( 0, 0, Height, Width), FBackGround)
        else begin
          Pen.Width   := 1;
          Pen.Color   := BorderColor;
          Brush.Color := Color;
          Brush.Style := bsClear;
          Rectangle( 0, 0, Width, Height);
        end;

        HPos := CenterPoint.X;
        VPos := CenterPoint.Y;

        if StepCount > 1
        then begin
          try
            HPos := SignalToInt( RangeMap( KnobPositionX, 0, StepCount - 1, 2, Height - 4));
            VPos := SignalToInt( RangeMap( KnobPositionY, 0, StepCount - 1, Width - 4, 2 ));
          except
            on E: Exception
            do KilledException( E);
          end;
        end;

        Pen.Width := 1;
        if Focused
        then Pen.Color := clRed
        else Pen.Color := clBlack;
        Brush.Style := bsSolid;
        Brush.Color := clWhite;
        MoveTo( HPos, VPos);
        Ellipse(
          HPos - SpotSize,
          VPos - SpotSize,
          HPos + SpotSize + 1,
          VPos + SpotSize + 1
        );

      end;
    end;

    procedure   TKnobsPad.MouseDown( Button: TMouseButton; Shift: TShiftState; X, Y: Integer); // override;
    var
      aWP: TKnobsWirePanel;
    begin
      inherited;

      if FMouseDown
      then Exit;

      if ( Button = mbLeft)
      then begin
        if TabStop and ( GetFocus <> Handle) and CanFocus
        then SetFocus;

        GAnchor       := Point( X, Y);
        FTresholdPast := False;
        BeginMove( GAnchor);
        FMouseDown := True;
      end
      else if Button = mbRight
      then begin
        aWP := FindWirePanel;
        if Assigned( aWP)
        then aWP.HandleRightClick( Self);
      end;
    end;

    procedure   TKnobsPad.MouseUp( Button: TMouseButton; Shift: TShiftState; X, Y: Integer); // override;
    begin
      inherited;

      if FMouseDown
      then begin
        FMouseDown  := False;
        if FTresholdPast
        then EndMove( Point( X - GAnchor.x, Y - GAnchor.y))
        else EndMove( GAnchor);
      end;
    end;

    procedure   TKnobsPad.MouseMove( Shift: TShiftState; X, Y: Integer); // override;
    var
      aDistance : TPoint;
    begin
      inherited;

      if FMouseDown
      then begin
        aDistance := Point( X, Y) - GAnchor;

        if not FTresholdPast
        then FTresholdPast := Distance( aDistance) > 10;

        if FTresholdPast
        then UpdateMove( Point( X - GAnchor.x, Y - GAnchor.y));
      end;
    end;

    procedure   TKnobsPad.KeyDown( var Key: Word; Shift: TShiftState); // override;
    begin
      case Key of
        VK_RIGHT : begin StartFinalizationTimer; KnobPositionX := KnobPositionX + 1; Key := 0; end;
        VK_LEFT  : begin StartFinalizationTimer; KnobPositionX := KnobPositionX - 1; Key := 0; end;
        VK_UP    : begin StartFinalizationTimer; KnobPositionY := KnobPositionY + 1; Key := 0; end;
        VK_DOWN  : begin StartFinalizationTimer; KnobPositionY := KnobPositionY - 1; Key := 0; end;
        else      inherited;
      end;
    end;

    procedure   TKnobsPad.Loaded; // override;
    begin
      inherited;

      NotifyPeers;
      Hint := AsHint;
    end;

//  public

    constructor TKnobsPad.Create( anOwner: TComponent); // override;
    begin
      inherited;

      ControlStyle := ControlStyle + [ csOpaque] - [ csSetCaption];
      FBackGround  := TBitmap.Create;

      InitializeFinalizationTimer;

      DoubleBuffered  := True;
      TabStop         := True;
      ShowHint        := False;
      Color           := clSilver;
      BorderColor     := clGray;
      ParentColor     := True;
      Width           := 155;
      Height          := 155;
    end;

    destructor  TKnobsPad.Destroy; // override;
    begin
      FinalizeFinalizationTimer;

      inherited;
    end;


{ ========
  TKnobsValuedButton = class( TKnobsValuedControl)
  private
    FButton     : TSpeedButton;
    FColorFocus : TColor;
    FUserColor  : TColor;
    FUserFlat   : Boolean;
  published
    property    Caption          : string        read GetCaption          write SetCaption                               ;
    property    Glyph            : TBitmap       read GetGlyph            write SetGlyph                                 ;
    property    Hint             : string        read GetHint             write SetHint                                  ;
    property    GroupIndex       : Integer       read GetGroupIndex       write SetGroupIndex       default             1;
    property    AllowAllUp       : Boolean       read GetAllowAllUp       write SetAllowAllUp       default         false;
    property    Height                                                                              default            18;
    property    Width                                                                               default            18;
    property    AlignWithMargins : Boolean       read GetAlignWithMargins write SetAlignWithMargins default         False;
    property    Color                                                                               default      clSilver;
    property    ColorFocus       : TColor        read FColorFocus         write FColorFocus         default     clSkyBlue;
    property    Flat             : Boolean       read GetFlat             write SetFlat             default         False;
    property    Font             : TFont         read GetFont             write SetFont                                  ;
    property    Layout           : TButtonLayout read GetLayout           write SetLayout           default   blGlyphLeft;
    property    Margin           : Integer       read GetMargin           write SetMargin           default            -1;
    property    Margins          : TMargins      read GetMargins          write SetMargins                               ;
    property    NumGlyphs        : TNumGlyphs    read GetNumGlyphs        write SetNumGlyphs        default             1;
    property    ParentFont       : Boolean       read GetParentFont       write SetParentFont       default          True;
    property    ParentShowHint   : Boolean       read GetParentShowHint   write SetParentShowHint   default          True;
    property    ShowHint         : Boolean       read GetShowHint         write SetShowHint         default         False;
    property    Transparent      : Boolean       read GetTransparent      write SetTransparent      default          True;
    property    Visible          : Boolean       read GetVisible          write SetVisible          default          True;
  private
}

    procedure   TKnobsValuedButton.SetKnobPosition( aValue: Integer); // override;
    begin
      if Assigned( FButton)
      then FButton.Down := SignalToLogic( aValue);
      inherited;
    end;

    function    TKnobsValuedButton.GetCaption: string;
    begin
      if Assigned( FButton)
      then Result := FButton.Caption
      else Result := '';
    end;

    procedure   TKnobsValuedButton.SetCaption( const aValue: string);
    begin
      if Assigned( FButton)
      then FButton.Caption := aValue;
    end;

    function    TKnobsValuedButton.GetGlyph: TBitmap;
    begin
      if Assigned( FButton)
      then Result := FButton.Glyph
      else Result := nil;
    end;

    procedure   TKnobsValuedButton.SetGlyph( const aValue: TBitmap);
    begin
      if Assigned( FButton)
      then FButton.Glyph.Assign( aValue);
    end;

    function    TKnobsValuedButton.GetHint: string;
    begin
      if Assigned( FButton)
      then Result := FButton.Hint
      else Result := '';
    end;

    procedure   TKnobsValuedButton.SetHint( const aValue: string);
    begin
      if Assigned( FButton)
      then FButton.Hint := aValue;
    end;

    function    TKnobsValuedButton.GetGroupIndex: Integer;
    begin
      if Assigned( FButton)
      then Result := FButton.GroupIndex
      else Result := 0;
    end;

    procedure   TKnobsValuedButton.SetGroupIndex( aValue: Integer);
    begin
      if Assigned( FButton)
      then FButton.GroupIndex := aValue;
    end;

    function    TKnobsValuedButton.GetAllowAllUp: Boolean;
    begin
      if Assigned( FButton)
      then Result := FButton.AllowAllUp
      else Result := False;
    end;

    procedure   TKnobsValuedButton.SetAllowAllUp( aValue: Boolean);
    begin
      if Assigned( FButton)
      then FButton.AllowAllUp := aValue;
    end;

//  private

    function    TKnobsValuedButton.GetAlignWithMargins: Boolean;
    begin
      if Assigned( FButton)
      then Result := FButton.AlignWithMargins
      else Result := False;
    end;

    procedure   TKnobsValuedButton.SetAlignWithMargins( aValue: Boolean);
    begin
      if Assigned( FButton)
      then FButton.AlignWithMargins := aValue;
    end;

    function    TKnobsValuedButton.GetFlat: Boolean;
    begin
      if Assigned( FButton)
      then Result := FButton.Flat
      else Result := False;
    end;

    procedure   TKnobsValuedButton.SetFlat( aValue: Boolean);
    begin
      if Assigned( FButton)
      then FButton.Flat := aValue;
    end;

    function    TKnobsValuedButton.GetFont: TFont;
    begin
      if Assigned( FButton)
      then Result := FButton.Font
      else Result := nil;
    end;

    procedure   TKnobsValuedButton.SetFont( const aValue: TFont);
    begin
      if Assigned( FButton)
      then FButton.Font.Assign( aValue);
    end;

    function    TKnobsValuedButton.GetLayout: TButtonLayout;
    begin
      if Assigned( FButton)
      then Result := FButton.Layout
      else Result := blGlyphLeft;
    end;

    procedure   TKnobsValuedButton.SetLayout( aValue: TButtonLayout);
    begin
      if Assigned( FButton)
      then FButton.Layout := aValue;
    end;

    function    TKnobsValuedButton.GetMargin: Integer;
    begin
      if Assigned( FButton)
      then Result := FButton.Margin
      else Result := -1;
    end;

    procedure   TKnobsValuedButton.SetMargin( aValue: Integer);
    begin
      if Assigned( FButton)
      then FButton.Margin := aValue;
    end;

    function    TKnobsValuedButton.GetMargins: TMargins;
    begin
      if Assigned( FButton)
      then Result := FButton.Margins
      else Result := nil;
    end;

    procedure   TKnobsValuedButton.SetMargins( aValue: TMargins);
    begin
      if Assigned( FButton)
      then FButton.Margins := aValue;
    end;

    function    TKnobsValuedButton.GetNumGlyphs: TNumGlyphs;
    begin
      if Assigned( FButton)
      then Result := FButton.NumGlyphs
      else Result := 1;
    end;

    procedure   TKnobsValuedButton.SetNumGlyphs( aValue: TNumGlyphs);
    begin
      if Assigned( FButton)
      then FButton.NumGlyphs := aValue;
    end;

    function    TKnobsValuedButton.GetParentFont: Boolean;
    begin
      if Assigned( FButton)
      then Result := FButton.ParentFont
      else Result := False;
    end;

    procedure   TKnobsValuedButton.SetParentFont( aValue: Boolean);
    begin
      if Assigned( FButton)
      then FButton.ParentFont := aValue;
    end;

    function    TKnobsValuedButton.GetParentShowHint: Boolean;
    begin
      if Assigned( FButton)
      then Result := FButton.ParentShowHint
      else Result := False;
    end;

    procedure   TKnobsValuedButton.SetParentShowHint( aValue: Boolean);
    begin
      if Assigned( FButton)
      then FButton.ParentShowHint := aValue;
    end;

    function    TKnobsValuedButton.GetShowHint: Boolean;
    begin
      if Assigned( FButton)
      then Result := FButton.ShowHint
      else Result := False;
    end;

    procedure   TKnobsValuedButton.SetShowHint( aValue: Boolean);
    begin
      if Assigned( FButton)
      then FButton.ShowHint := aValue;
    end;

    function    TKnobsValuedButton.Gettransparent: Boolean;
    begin
      if Assigned( FButton)
      then Result := FButton.Transparent
      else Result := False;
    end;

    procedure   TKnobsValuedButton.SetTransparent( aValue: Boolean);
    begin
      if Assigned( FButton)
      then FButton.Transparent := aValue;
    end;

    function    TKnobsValuedButton.GetVisible: Boolean;
    begin
      if Assigned( FButton)
      then Result := FButton.Visible
      else Result := True;
    end;

    procedure   TKnobsValuedButton.SetVisible( aValue: Boolean);
    begin
      if Assigned( FButton)
      then FButton.Visible := aValue;
    end;

//  private

    procedure   TKnobsValuedButton.OnButtonClick( aSender: TObject);
    begin
      if Assigned( FButton)
      then begin
        if StepCount = 1
        then FButton.Down := False
        else KnobPosition := Round( LogicToSignal( FButton.Down));
      end;
      FinalizeKnobPosition;
    end;

//  protected

    procedure   TKnobsValuedButton.WMSetFocus( var aMessage: TWMSetFocus); // message WM_SETFOCUS;
    begin
      inherited;
      FUserColor := Color;
      FUserFlat  := Flat;
      Color      := ColorFocus;
      Flat       := not Flat;
    end;

    procedure   TKnobsValuedButton.WMKillFocus( var aMessage: TWMKillFocus); // message WM_KILLFOCUS;
    begin
      inherited;
      Color := FUserColor;
      Flat  := FUserFlat;
    end;

    procedure   TKnobsValuedButton.WMGetDlgCode( var aMessage: TWMGetDlgCode); // message WM_GETDLGCODE;
    begin
      aMessage.Result := DLGC_WANTARROWS;
    end;

    procedure   TKnobsValuedButton.KeyDown( var Key: Word; Shift: TShiftState); // override;
    begin
      if not Assigned( FButton)
      then Exit;

      case Key of
        VK_UP  , VK_RIGHT : begin FButton.Down := False; Key := 0; end;
        VK_DOWN, VK_LEFT  : begin FButton.Down := True ; Key := 0; end;
        else                inherited;
      end;
    end;

    procedure   TKnobsValuedButton.SetBiDiMode( aValue: TBiDiMode); // override;
    begin
      if Assigned( FButton)
      then FButton.BiDiMode := aValue;
    end;

    procedure   TKnobsValuedButton.SetParentBiDiMode( aValue: Boolean); // override;
    begin
      if Assigned( FButton)
      then FButton.ParentBiDiMode := aValue;
    end;

    function    TKnobsValuedButton.GetEnabled: Boolean; // override;
    begin
      if Assigned( FButton)
      then Result := FButton.Enabled
      else Result := False;
    end;

    procedure   TKnobsValuedButton.SetEnabled( aValue: Boolean); // override;
    begin
      if Assigned( FButton)
      then FButton.Enabled := aValue;
    end;

//  protected

    procedure   TKnobsValuedButton.Resize; // override;
    begin
      inherited;
      if Assigned( FButton)
      then FButton.SetBounds( 0, 0, Width, Height);
    end;

    procedure   TKnobsValuedButton.Loaded; // override;
    begin
      inherited;
      Width := Width - 1; // A resize is needed to make the button visible ... weird
      Width := Width + 1;
    end;

    procedure   TKnobsValuedButton.SetDefaults; // virtual;
    begin
      StepCount       := 1;
      KnobPosition    := 0;
      GroupIndex      := 1;
      AllowAllUp      := false;
      Height          := 18;
      Width           := 18;
      Color           := clSilver;
      Flat            := False;
      Layout          := blGlyphLeft;
      Margin          := -1;
      NumGlyphs       := 1;
      ParentFont      := True;
      ParentShowHint  := True;
      ShowHint        := False;
      Transparent     := True;
      Visible         := True;
    end;

//  public

    constructor TKnobsValuedButton.Create( anOwner: TComponent); // override;
    begin
      inherited;
      ControlStyle := ControlStyle - [ csSetCaption] + [ csOpaque];

      FButton := TSpeedButton.Create( Self);
      FButton.Parent := Self;
      FButton.SetBounds( 0, 0, Width, Height);
      FButton.OnClick := OnButtonClick;
      SetDefaults;
      DoubleBuffered := True;
    end;

    destructor  TKnobsValuedButton.Destroy; // override;
    begin
      FButton.OnClick := nil;
      FreeAndNil( FButton);
      inherited;
    end;


{ ========
  TKnobsValuedButtons = class( TKnobsValuedControl)
  private
    FButtons          : array of TSpeedButton;
    FCaptions         : TStrings;
    FImageIndices     : TStrings;
    FGlyphs           : TImageList;
    FHints            : TStrings;
    FGroupIndex       : Integer;
    FAlignWithMargins : Boolean;
    FAllowAllUp       : Boolean;
    FFlat             : Boolean;
    FFont             : TFont;
    FLayout           : TButtonLayout;
    FDisplayMode      : TKnobsButtonsDisplayMode;
    FMargin           : Integer;
    FMargins          : TMargins;
    FNumGlyphs        : TNumGlyphs;
    FParentFont       : Boolean;
    FParentShowHint   : Boolean;
    FShowHint         : Boolean;
    FTransparent      : Boolean;
    FVisible          : Boolean;
    FColorFocus       : TColor;
    FUserColor        : TColor;
    FUserFlat         : Boolean;
  published
    property    StepCount                                                                                    default             2;
    property    Captions         : TStrings                 read GetCaptions       write SetCaptions                              ;
    property    ImageIndices     : TStrings                 read GetImageIndices   write SetImageIndices                          ;
    property    Glyphs           : TImageList               read GetGlyphs         write SetGlyphs                                ;
    property    Hints            : TStrings                 read GetHints          write SetHints                                 ;
    property    GroupIndex       : Integer                  read FGroupIndex       write SetGroupIndex       default             1;
    property    AllowAllUp       : Boolean                  read FAllowAllUp       write SetAllowAllUp       default         false;
    property    Height                                                                                       default            25;
    property    Width                                                                                        default            51;
    property    AlignWithMargins : Boolean                  read FAlignWithMargins write SetAlignWithMargins default         False;
    property    Color                                                                                        default      clSilver;
    property    ColorFocus       : TColor                   read FColorFocus       write FColorFocus         default     clSkyBlue;
    property    Flat             : Boolean                  read FFlat             write SetFlat             default         False;
    property    Font             : TFont                    read FFont             write SetFont                                  ;
    property    Layout           : TButtonLayout            read FLayout           write SetLayout           default   blGlyphLeft;
    property    DisplayMode      : TKnobsButtonsDisplayMode read FDisplayMode      write SetDisplayMode      default bdmHorizontal;
    property    Margin           : Integer                  read FMargin           write SetMargin           default            -1;
    property    Margins          : TMargins                 read FMargins          write SetMargins                               ;
    property    NumGlyphs        : TNumGlyphs               read FNumGlyphs        write SetNumGlyphs        default             1;
    property    ParentFont       : Boolean                  read FParentFont       write SetParentFont       default          True;
    property    ParentShowHint   : Boolean                  read FParentShowHint   write SetParentShowHint   default          True;
    property    ShowHint         : Boolean                  read FShowHint         write SetShowHint         default         False;
    property    Transparent      : Boolean                  read FTransparent      write SetTransparent      default          True;
    property    Visible          : Boolean                  read FVisible          write SetVisible          default          True;
  private
}

    procedure   TKnobsValuedButtons.FixDimensions;
    var
      i : Integer;
      w : Integer;
      h : Integer;
    begin
      case DisplayMode of

        bdmHorizontal :

          if StepCount > 0
          then begin
            w := Width div StepCount;
            h := Height;
            if AlignWithMargins
            then begin
              w := w - Margins.Left - Margins.Right;
              h := h - Margins.Top  - Margins.Bottom;
            end;
            for i := 0 to StepCount - 1
            do begin
              FButtons[ i].Width  := w;
              FButtons[ i].Height := h;
            end;
          end;

        bdmVertical :

          if StepCount > 0
          then begin
            w := Width;
            h := Height div StepCount;
            if AlignWithMargins
            then begin
              w := w - Margins.Left - Margins.Right;
              h := h - Margins.Top  - Margins.Bottom;
            end;
            for i := 0 to StepCount - 1
            do begin
              FButtons[ i].Width  := w;
              FButtons[ i].Height := h;
            end;
          end;
      end;
    end;

    procedure   TKnobsValuedButtons.FixGlyph( anIndex: Integer; aBitmap: TBitmap);
    begin
      FButtons[ anIndex].Glyph.Assign( aBitmap);
    end;

    function    TKnobsValuedButtons.FindGlyphIndex( anIndex: Integer): Integer;
    begin
      if ( anIndex >= 0) and ( anIndex < FImageIndices.Count)
      then Result := StrToIntDef( FImageIndices[ anIndex], -1)
      else Result := -1;
    end;

//  private

    procedure   TKnobsValuedButtons.FixAlignWithMargins;
    var
      i : Integer;
    begin
      for i := 0 to StepCount - 1
      do FButtons[ i].AlignWithMargins := AlignWithMargins;
      FixDimensions;
    end;

    procedure   TKnobsValuedButtons.FixBiDiMode;
    var
      i : Integer;
    begin
      for i := 0 to StepCount - 1
      do FButtons[ i].BiDiMode := BiDiMode;
    end;

    procedure   TKnobsValuedButtons.FixCaptions;
    var
      i : Integer;
    begin
      if Assigned( FCaptions)
      then begin
        for i := 0 to StepCount - 1
        do begin
          if i < FCaptions.Count
          then FButtons[ i].Caption := FCaptions[ i]
          else FButtons[ i].Caption := '';
        end;
      end;
    end;

    procedure   TKnobsValuedButtons.FixEnabled;
    var
      i : Integer;
    begin
      for i := 0 to StepCount - 1
      do FButtons[ i].Enabled := Enabled;
    end;

    procedure   TKnobsValuedButtons.FixFlat;
    var
      i : Integer;
    begin
      for i := 0 to StepCount - 1
      do FButtons[ i].Flat := Flat;
    end;

    procedure   TKnobsValuedButtons.FixFont;
    var
      i : Integer;
    begin
      for i := 0 to StepCount - 1
      do FButtons[ i].Font.Assign( Font);
    end;

    procedure   TKnobsValuedButtons.FixGlyphs;
    var
      i   : Integer;
      aBm : TBitmap;
      idx : Integer;
    begin
      if Assigned( FGlyphs)
      then begin
        aBm := TBitmap.Create;
        try
          for i := 0 to StepCount - 1
          do begin
            idx := FindGlyphIndex( i);
            if ( idx >= 0) and ( idx < FGlyphs.Count)
            then begin
              FGlyphs.GetBitmap( idx, aBm);
              FixGlyph( i, aBm);
            end
            else FixGlyph( i, nil);
          end;
        finally
          aBm.Free;
        end;
      end;
    end;

    procedure   TKnobsValuedButtons.FixHints;
    var
      i : Integer;
    begin
      if Assigned( FHints)
      then begin
        for i := 0 to StepCount - 1
        do begin
          if i < FHints.Count
          then FButtons[ i].Hint := FHints[ i]
          else FButtons[ i].Hint := '';
        end;
      end;
    end;

    procedure   TKnobsValuedButtons.FixLayout;
    var
      i : Integer;
    begin
      for i := 0 to StepCount - 1
      do FButtons[ i].Layout := Layout;
    end;

    procedure   TKnobsValuedButtons.FixMargin;
    var
      i : Integer;
    begin
      for i := 0 to StepCount - 1
      do FButtons[ i].Margin := Margin;
    end;

    procedure   TKnobsValuedButtons.FixMargins;
    var
      i : Integer;
    begin
      for i := 0 to StepCount - 1
      do FButtons[ i].Margins.Assign( FMargins);
      FixDimensions;
    end;

    procedure   TKnobsValuedButtons.FixNumGlyphs;
    var
      i : Integer;
    begin
      for i := 0 to StepCount - 1
      do FButtons[ i].NumGlyphs := NumGlyphs;
    end;

    procedure   TKnobsValuedButtons.FixParentFont;
    var
      i : Integer;
    begin
      for i := 0 to StepCount - 1
      do FButtons[ i].ParentFont := ParentFont;
    end;

    procedure   TKnobsValuedButtons.FixParentShowHint;
    var
      i : Integer;
    begin
      for i := 0 to StepCount - 1
      do FButtons[ i].ParentShowHint := ParentShowHint;
    end;

    procedure   TKnobsValuedButtons.FixParentBiDiMode;
    var
      i : Integer;
    begin
      for i := 0 to StepCount - 1
      do FButtons[ i].ParentBiDiMode := ParentBiDiMode;
    end;

    procedure   TKnobsValuedButtons.FixShowHint;
    var
      i : Integer;
    begin
      for i := 0 to StepCount - 1
      do FButtons[ i].ShowHint := ShowHint;
    end;

    procedure   TKnobsValuedButtons.FixTransparent;
    var
      i : Integer;
    begin
      for i := 0 to StepCount - 1
      do FButtons[ i].Transparent := Transparent;
    end;

    procedure   TKnobsValuedButtons.FixVisible;
    var
      i : Integer;
    begin
      for i := 0 to StepCount - 1
      do FButtons[ i].Visible := Visible;
    end;

//  private

    procedure   TKnobsValuedButtons.FixMode( aNewStepCount: Integer);
    var
      aComponent: TComponent;
    begin
      while aNewStepCount < Length( FButtons)
      do begin
        aComponent := FButtons[ Length( FButtons) - 1];
        SetLength( FButtons, Length( FButtons) - 1);
        Owner.RemoveComponent( aComponent);
        aComponent.Free;
      end;
      FStepCount := aNewStepCount;
      while aNewStepCount > Length( FButtons)
      do CreateButton;
      FixDimensions;
    end;

    procedure   TKnobsValuedButtons.SetKnobPosition( aValue: Integer); // override;
    begin
      if aValue <> FKnobPosition
      then begin
        inherited;
        if ( aValue >= 0) and ( aValue < StepCount)
        then FButtons[ aValue].Down := True;
        FinalizeKnobPosition;
      end;
    end;

    procedure   TKnobsValuedButtons.SetStepCount( aValue: Integer);
    begin
      if aValue <= 0
      then aValue := 1;
      if aValue <> StepCount
      then begin
        FixMode( aValue);
        FixGlyphs;
        FixCaptions;
        FixHints;
        FixDimensions;
        FinalizeKnobPosition;
      end;
    end;

    function    TKnobsValuedButtons.GetCaptions: TStrings;
    begin
      Result := FCaptions;
    end;

    procedure   TKnobsValuedButtons.SetCaptions( const aValue: TStrings);
    begin
      FCaptions.Assign( aValue);
      FixCaptions;
    end;

    function    TKnobsValuedButtons.GetImageIndices: TStrings;
    begin
      Result := FImageIndices;
    end;

    procedure   TKnobsValuedButtons.SetImageIndices( const aValue: TStrings);
    begin
      FImageIndices.Assign( aValue);
      FixGlyphs;
    end;

    function    TKnobsValuedButtons.GetGlyphs: TImageList;
    begin
      Result := FGlyphs;
    end;

    procedure   TKnobsValuedButtons.SetGlyphs( const aValue: TImageList);
    begin
      FGlyphs := aValue;
      FixGlyphs;
    end;

    function    TKnobsValuedButtons.GetHints: TStrings;
    begin
      Result := FHints;
    end;

    procedure   TKnobsValuedButtons.SetHints( const aValue: TStrings);
    begin
      FHints.Assign( aValue);
      FixHints;
    end;

    procedure   TKnobsValuedButtons.SetGroupIndex( aValue: Integer);
    var
      i : Integer;
    begin
      if aValue <> FGroupIndex
      then begin
        FGroupIndex := aValue;
        for i := 0 to StepCount - 1
        do FButtons[ i].GroupIndex := FGroupIndex;
      end;
    end;

    procedure   TKnobsValuedButtons.SetAllowAllUp( aValue: Boolean);
    var
      i : Integer;
    begin
      if aValue <> FAllowAllUp
      then begin
        FAllowAllUp := aValue;
        for i := 0 to StepCount - 1
        do FButtons[ i].AllowAllUp := FAllowAllUp;
      end;
    end;

    procedure   TKnobsValuedButtons.SetAlignWithMargins( aValue: Boolean);
    begin
      if aValue <> FAlignWithMargins
      then begin
        FAlignWithMargins := aValue;
        FixAlignWithMargins;
      end;
    end;

    procedure   TKnobsValuedButtons.SetFlat( aValue: Boolean);
    begin
      if avalue <> FFlat
      then begin
        FFlat := aValue;
        FixFlat;
      end;
    end;

    procedure   TKnobsValuedButtons.SetFont( const aValue: TFont);
    begin
      FFont.Assign( aValue);
      FixFont;
    end;

    procedure   TKnobsValuedButtons.SetLayout( aValue: TButtonLayout);
    begin
      if aValue <> FLayout
      then begin
        FLayout := aValue;
        FixLayout;
      end;
    end;

    procedure   TKnobsValuedButtons.SetDisplayMode( aValue: TKnobsButtonsDisplayMode);
    begin
      if aValue <> FDisplayMode
      then begin
        FDisplayMode := aValue;
        FixMode( StepCount);
      end;
    end;

    procedure   TKnobsValuedButtons.SetMargin( aValue: Integer);
    begin
      if avalue <> FMargin
      then begin
        FMargin := aValue;
        FixMargin;
      end;
    end;

    procedure   TKnobsValuedButtons.SetMargins( aValue: TMargins);
    begin
      FMargins.Assign( aValue);
      FixMargins;
    end;

    procedure   TKnobsValuedButtons.SetNumGlyphs( aValue: TNumGlyphs);
    begin
      if avalue <> FNumGlyphs
      then begin
        FNumGlyphs := aValue;
        FixNumGlyphs;
      end;
    end;

    procedure   TKnobsValuedButtons.SetParentFont( aValue: Boolean);
    begin
      if aValue <> FParentFont
      then begin
        FParentFont := aValue;
        FixParentFont;
      end;
    end;

    procedure   TKnobsValuedButtons.SetParentShowHint( aValue: Boolean);
    begin
      if aValue <> FParentShowHint
      then begin
        FParentShowHint := aValue;
        FixParentShowHint;
        if ParentShowHint
        then ShowHint := False;
      end;
    end;

    procedure   TKnobsValuedButtons.setShowHint( aValue: Boolean);
    begin
      if aValue <> FShowHint
      then begin
        FShowHint := aValue;
        FixShowHint;
        if ShowHint
        then ParentShowHint := False;
      end;
    end;

    procedure   TKnobsValuedButtons.SetTransparent( aValue: Boolean);
    begin
      if aValue <> FTransparent
      then begin
        FTransparent := aValue;
        FixTransparent;
      end;
    end;

    procedure   TKnobsValuedButtons.SetVisible( aValue: Boolean);
    begin
      if aValue <> FVisible
      then begin
        FVisible := aValue;
        FixVisible;
      end;
    end;

//  private

    procedure   TKnobsValuedButtons.OnButtonClick( aSender: TObject);
    var
      aTag: Integer;
    begin
      aTag := TSpeedButton( aSender).Tag;
      KnobPosition := aTag;
      FinalizeKnobPosition;
    end;

    procedure   TKnobsValuedButtons.CreateButton;
    var
      aButton : TSpeedButton;
    begin
      aButton := TSpeedButton.Create( Self);
      with aButton do
      begin
        Parent     := Self;
        AllowAllUp := Self.AllowAllUp;
        GroupIndex := Self.GroupIndex;
        Name       := '';
        Tag        := Length( FButtons);
        OnClick    := OnButtonClick;
        Align      := alLeft;
        SetLength( FButtons, Length( FButtons) + 1);
        FButtons[ Length( FButtons) - 1] := aButton;
      end;
      StepCount := Length( FButtons);
    end;

//  protected

    procedure   TKnobsValuedButtons.WMSetFocus( var aMessage: TWMSetFocus); // message WM_SETFOCUS;
    begin
      inherited;
      FUserColor := Color;
      FUserFlat  := Flat;
      Color      := ColorFocus;
      Flat       := not Flat;
    end;

    procedure   TKnobsValuedButtons.WMKillFocus( var aMessage: TWMKillFocus); // message WM_KILLFOCUS;
    begin
      inherited;
      Color := FUserColor;
      Flat  := FUserFlat;
    end;

    procedure   TKnobsValuedButtons.WMGetDlgCode( var aMessage: TWMGetDlgCode); // message WM_GETDLGCODE;
    begin
      aMessage.Result := DLGC_WANTARROWS;
    end;

    procedure   TKnobsValuedButtons.KeyDown( var Key: Word; Shift: TShiftState); // override;
    begin
      case Key of
        VK_UP  , VK_RIGHT : begin if KnobPosition < StepCount - 1 then KnobPosition := KnobPosition + 1; Key := 0; end;
        VK_DOWN, VK_LEFT  : begin if KnobPosition > 0             then KnobPosition := KnobPosition - 1; Key := 0; end;
        else                inherited;
      end;
    end;

    procedure   TKnobsValuedButtons.SetBiDiMode( aValue: TBiDiMode); // Override;
    begin
      if aValue <> BiDiMode
      then begin
        inherited;
        FixBiDiMode;
      end;
    end;

    procedure   TKnobsValuedButtons.SetEnabled( aValue: Boolean); // Override;
    begin
      if aValue <> Enabled
      then begin
        inherited;
        FixEnabled;
      end;
    end;

    procedure   TKnobsValuedButtons.SetParentBiDiMode( aValue: Boolean); // Override;
    begin
      if aValue <> ParentBiDiMode
      then begin
        inherited;
        FixParentBiDiMode;
      end;
    end;

//  protected

    procedure   TKnobsValuedButtons.Resize; // override;
    begin
      inherited;
      FixDimensions;
    end;

    procedure   TKnobsValuedButtons.Loaded; // override;
    begin
      // FixCaptions;
      FixButtons;
    end;

    procedure   TKnobsValuedButtons.FixButtons; // virtual;
    begin
      FixAlignWithMargins;
      FixBiDiMode;
      FixCaptions;
      FixEnabled;
      FixFlat;
      FixFont;
      FixGlyphs;
      FixHints;
      FixLayout;
      FixMargin;
      FixMargins;
      FixNumGlyphs;
      FixParentFont;
      FixParentShowHint;
      FixParentBiDiMode;
      FixShowHint;
      FixTransparent;
      FixVisible;
    end;

    procedure   TKnobsValuedButtons.SetDefaults; // virtual;
    begin
      StepCount       := 2;
      GroupIndex      := 1;
      AllowAllUp      := False;
      Height          := 25;
      Width           := 51;
      Color           := clSilver;
      ColorFocus      := clSkyBlue;
      Flat            := False;
      Layout          := blGlyphLeft;
      Margin          := -1;
      NumGlyphs       := 1;
      ParentFont      := True;
      ParentShowHint  := True;
      ShowHint        := False;
      Transparent     := True;
      Visible         := True;
    end;

//  public

    constructor TKnobsValuedButtons.Create( anOwner: TComponent); // override;
    begin
      inherited;
      ControlStyle  := ControlStyle + [ csOpaque];
      FCaptions     := TStringList.Create;
      FHints        := TStringList.Create;
      FImageIndices := TStringList.Create;
      FFont         := TFont      .Create;
      FMargins      := TMargins   .Create( Self);
      SetDefaults;
    end;

    destructor  TKnobsValuedButtons.Destroy; // override;
    begin
      FFont        .Free;
      FImageIndices.Free;
      FHints       .Free;
      FCaptions    .Free;
      inherited;
    end;


{ ========
  TKnobsDisplayEditor = class( TMemo)
  // An multi line editor for TDisplay
  private
    FDisplay : TKnobsDisplay;
  private
}

    procedure   TKnobsDisplayEditor.WMKillFocus( var aMessage: TWMKillFocus ); // message WM_KILLFOCUS;
    begin
      inherited;

      Visible := False;
    end;

    procedure   TKnobsDisplayEditor.Do_Exit( aSender: TObject);
    begin
      Visible := False;
      if Assigned( FDisplay)
      then FDisplay.SetFocus;
    end;

    procedure   TKnobsDisplayEditor.Do_KeyDown( aSender: TObject; var aKey: Word; aShift: TShiftState);
    begin
      if aShift = []
      then
        case aKey Of

          VK_ESCAPE :
            begin
              Visible := False;
              aKey    := 0;
            end;

        end
      else if aShift = [ ssCtrl]
      then
        case aKey Of

          VK_RETURN:
            begin
              FDisplay.SetText( Text);
              Visible := False;
              aKey    := 0;
              if Assigned( FDisplay)
              then FDisplay.LockKnob;
            end;
        end;
    end;

//  protected

    procedure   TKnobsDisplayEditor.SetDisplay( aDisplay: TKnobsDisplay);
    var
      aLeft   : Integer;
      aTop    : Integer;
      aWidth  : Integer;
      aHeight : Integer;
    begin
      FDisplay := aDisplay;
      Parent   := aDisplay.Parent;
      Text     := aDisplay.Caption;

      if aDisplay.HasParentEditor
      then SetBounds( 0, 0, Parent.Width, Parent.Height)
      else begin
        aLeft   := 0;
        aTop    := aDisplay.Top -  2;
        aWidth  := Parent.Width;
        aHeight := aDisplay.Height +  4;
        if aTop < 0
        then aTop := 0;
        if aTop + aHeight > Parent.Height
        then aTop := Parent.Height - aHeight;
        SetBounds( aLeft, aTop, aWidth, aHeight);
      end;

      if aDisplay.HasScrollbar
      then ScrollBars := ssVertical
      else ScrollBars := ssNone;

      Font       := aDisplay.Font;
      Color      := clWhite;
      Font.Color := clBlack;
      Visible    := True;
      SetFocus;
    end;

//  public

    constructor TKnobsDisplayEditor.Create( anOwner: TComponent); // override;
    begin
      if Assigned( GDisplayEditor)
      then Exit;

      inherited;

      ControlStyle   := ControlStyle + [ csOpaque];

      Visible        := False;
      Left           := 1;
      Top            := 1;
      OnExit         := Do_Exit;
      OnKeyDown      := Do_KeyDown;
      GDisplayEditor := Self;
    end;

    destructor  TKnobsDisplayEditor.Destroy; // override;
    begin
      inherited;
      if Assigned( Owner)
      then Owner.RemoveComponent( Self);
    end;

{ ========
  TKnobsTextControl = class( TStaticText)
  private
    FMustSave   : Boolean;
    FMustNotify : Boolean;
    FOnChanged  : TKnobsOnTextChanged;
    FSavedColor : TColor;
  public
    property    Module: TKnobsCustomModule  read GetModule;
    property    TextValue: string           read GetTextValue write SetTextValue;
  published
    property    Ctl3D                                                                  default False;
    property    ParentColor                                                            default False;
    property    Color                                                                  default KnobsColorDisplay;
    property    Width                                                                  default 33;
    property    Height                                                                 default 15;
    property    AutoSize                                                               default False;
    property    BorderStyle                                                            default sbsSingle;
    property    TabStop                                                                default True;
    property    MustSave    : Boolean             read FMustSave     write FMustSave   default False;
    property    MustNotify  : Boolean             read FMustNotify   write FMustNotify default False;
    property    OnChanged   : TKnobsOnTextChanged read FOnChanged    write FOnChanged;
  private
 }

    function    TKnobsTextControl.FindWirePanel: TKnobsWirePanel;
    var
      aParent : TWinControl;
    begin
      Result  := nil;
      aParent := Parent;

      while Assigned( aParent) and not ( aParent is TKnobsWirePanel)
      do aParent :=  aParent.Parent;

      if aParent is TKnobsWirePanel
      then Result := TKnobsWirePanel( aParent);
    end;

    function    TKnobsTextControl.GetModule: TKnobsCustomModule;
    begin
      if Owner is TKnobsCustomModule
      then Result := TKnobsCustomModule( Owner)
      else Result := nil;
    end;

    procedure   TKnobsTextControl.Do_KeyDown( aSender: TObject; var aKey: Word; aShift: TShiftState);
    begin
      if aShift = []
      then begin
        case aKey of
          VK_RETURN:
            begin
              HandleReturnKey;
              aKey := 0;
            end;
        end;
      end;
    end;

 // protected

    procedure   TKnobsTextControl.SetText( const aValue: string);
    begin
      if TextValue <> aValue
      then begin
        try
          HandleTextChange( aValue);
          TextValue := aValue;
        except
          on E: EAbort
          do KilledException( E)  // Absorb aborts
          else raise;             // Re-raise errors
        end;
      end;
    end;

    procedure   TKnobsTextControl.TextChanged( const aNewValue: string; DoNotify: Boolean); // virtual;
    var
      aModule : TKnobsCustomModule;
    begin
      if Assigned( FOnChanged)
      then FOnChanged( Self, Name, aNewValue)
      else begin
        aModule := Module;
        if Assigned( aModule) and ( DoNotify or MustNotify)
        then Module.TextChanged( Self, MakePath( [ aModule.Name, Name]), aNewValue);
      end;
    end;

    procedure   TKnobsTextControl.HandleTextChange( const aValue: string); // virtual;
    begin
      TextChanged( aValue, DO_NOTIFY);
    end;

    procedure   TKnobsTextControl.MouseDown( Button: TMouseButton; Shift: TShiftState; X, Y: Integer); // override;
    // Double click should fire up some text editor
    // Right clicks ripple up to the wire panel for right mouse menu support
    // other events should be handled by the containing module.
    var
      aModule : TKnobsCustomModule;
      aWP     : TKnobsWirePanel;
    begin
      SetFocus;
      if csDesigning in ComponentState
      then inherited
      else begin
        if ( ssDouble in Shift) and ( Button = mbLeft)
        then HandleDoubleClick
        else if Button = mbRight
        then begin
          aWP := FindWirePanel;
          if Assigned( aWP)
          then aWP.HandleRightClick( Self);
        end
        else begin
          // Pass it on to the containing module
          aModule := Module;
          if Assigned( aModule)
          then begin
            aModule.MouseCapture := True;
            aModule.MouseDown( Button, Shift, X + Left, Y + Top);
          end;
        end;
      end;
    end;

    procedure   TKnobsTextControl.WMSetFocus( var aMessage: TWMSetFocus); // message WM_SETFOCUS;
    begin
      inherited;
      FSavedColor := Color;
      Color       := clBlack;
    end;

    procedure   TKnobsTextControl.WMKillFocus( var aMessage: TWMKillFocus); // message WM_KILLFOCUS;
    begin
      inherited;
      Color := FSavedColor;
    end;

//  public

    constructor TKnobsTextControl.Create( anOwner: TComponent); // override;
    begin
      inherited;

      ControlStyle   := ControlStyle - [ csSetCaption] + [ csOpaque];
      DoubleBuffered := True;
      Ctl3D          := False;
      AutoSize       := False;
      BorderStyle    := sbsSingle;
      Width          := 33;
      Height         := 15;
      Font.Color     := clWhite;
      ParentColor    := False;
      Color          := KnobsColorDisplay;
      FSavedColor    := Color;
      TabStop        := True;
      MustSave       := False;
      MustNotify     := False;
      OnKeyDown      := Do_KeyDown;
    end;


{ ========
  TKnobsDisplay = class( TStaticText)
  // Can be linked to aTKnob for auto update
  // Can have an editor to manually edit the value
  private
    FEditor          : TKnobsDisplayEditor;
    FController      : TKnobsValuedControl;
    FHasEditor       : Boolean;
    FHasParentEditor : Boolean;
    FHasScrollbar    : Boolean;
  public
    property    Editor: TKnobsDisplayEditor read FEditor   write FEditor;
  published
    property    HasEditor       : Boolean read FHasEditor       write FHasEditor       default False;
    property    HasParentEditor : Boolean read FHasParentEditor write FHasParentEditor default False;
    property    HasScrollbar    : Boolean read FHasScrollbar    write FHasScrollbar    default False;
  protected

}

    procedure   TKnobsDisplay.HandleReturnKey; // override;
    begin
      NeedsDisplayEditor;
      GDisplayEditor.SetDisplay( Self);
    end;

    procedure   TKnobsDisplay.HandleDoubleClick; // override;
    begin
      NeedsDisplayEditor;
      GDisplayEditor.SetDisplay( Self);
    end;

    function    TKnobsDisplay.GetTextValue: string; // override;
    begin
      Result := Caption;
    end;

    procedure   TKnobsDisplay.SetTextValue( const aValue: string); // override;
    begin
      Caption := aValue;
    end;

    procedure   TKnobsDisplay.TextChanged( const aNewValue: string; DoNotify: Boolean); // override;
    begin
      if not Assigned( FController) or ( Assigned( FController) and FController.Locked)
      then inherited;
    end;

    procedure   TKnobsDisplay.WMNCHitTest( var aMessage: TWMNCHitTest); // message WM_NCHITTEST;
    begin
      if ( csDesigning in ComponentState) or HasEditor
      then inherited
      else aMessage.Result := HTTRANSPARENT;
    end;

    procedure   TKnobsDisplay.LockKnob;
    begin
      if Assigned( FController)
      then FController.Locked := True;
    end;

//  public

    constructor TKnobsDisplay.Create( anOwner: TComponent); // override;
    begin
      inherited;

      TextValue       := '0.57';
      HasEditor       := False;
      HasParentEditor := False;
      HasScrollbar    := False;
    end;

    destructor  TKnobsDisplay.Destroy; // override;
    begin
      if FController is TKnobsKnob
      then TKnobsKnob( FController).Display := nil;
      inherited;
    end;


{ ========
  TKnobsFileSelector = class( TStaticText)
  private
    FFileSelector : TOpenDialog;
    FDataFileName : string;
    FDataFileExt  : string;
    FDataFileMsg  : string;
  published
    property    DataFileName : string read FDataFileName write SetDataFileName;
    property    DataFileExt  : string read FDataFileExt  write FDataFileExt;
    property    DataFileMsg  : string read FDataFileMsg  write FDataFileMsg;
  private
}

    procedure   TKnobsFileSelector.SetDataFileName( const aValue: string);
    begin
      FDataFileName := aValue;
      FixCaption;
    end;

    procedure   TKnobsFileSelector.FixCaption;
    begin
      Caption := ' ' + ExtractFileName( FDataFileName);
    end;

    procedure   TKnobsFileSelector.SelectFile;
    begin
      with FFileSelector
      do begin
        DefaultExt := DataFileExt;
        Title      := DataFileMsg;
        if SameText( DataFileExt, '.wav')
        then Filter := 'wave files (*.wav)|*.wav|All files (*.*)|*.*'
        else if SameText( DataFileExt, '.mid')
        then Filter := 'midi files (*.mid)|*.mid|All files (*.*)|*.*'
        else if SameText( DataFileExt, '.kbm')
        then Filter := 'scala key mapping files (*.kbm)|*.kbm|All files (*.*)|*.*'
        else if SameText( DataFileExt, '.scl')
        then Filter := 'scala scale files (*.scl)|*.scl|All files (*.*)|*.*'
        else if SameText( DataFileExt, '.4th')
        then Filter := 'forth files (*.4th)|*.4th|All files (*.*)|*.*'
        else Filter := Format( '%0:s files (*.%0:s)|*.%0:s|All files (*.*)|*.*', [ Copy( DataFileExt, 2, Length( DataFileExt))], AppLocale);
        FileName := DataFileName;
        if Execute
        then SetText( FileName);
      end;
    end;

//  protected

    procedure   TKnobsFileSelector.HandleReturnKey; // override;
    begin
      SelectFile;
    end;

    procedure   TKnobsFileSelector.HandleDoubleClick; // override;
    begin
      SelectFile;
    end;

    function    TKnobsFileSelector.GetTextValue: string; // override;
    begin
      Result := DataFileName;
    end;

    procedure   TKnobsFileSelector.SetTextValue( const aValue: string); // override;
    begin
      DataFileName := aValue;
    end;

//  public

    constructor TKnobsFileSelector.Create( anOwner: TComponent); // override;
    begin
      inherited;

      FFileSelector := TOpenDialog.Create( nil);
      Caption       := ' double click to load a wave file';
      TextValue     := Caption;

      with FFileSelector
      do begin
        DefaultExt := '.wav';
        Title      := 'select wave file';
        Filter     := 'wave files (*.wav)|*.wav|All files (*.*)|*.*';
      end;
    end;

    destructor  TKnobsFileSelector.Destroy; // override;
    begin
      FreeAndNil( FFileSelector);
      inherited;
    end;


{ ========
  TKnobsIndicator = class( TGraphicControl)
  // an LED
  private
    FActive     : Boolean;
    FActiveBm   : TBitmap;
    FInActiveBm : TBitmap;
  public
    property    BitmapActive   : TBitmap read FActiveBm   write SetBitmapActive;
    property    BitmapInactive : TBitmap read FInactiveBm write SetBitmapInactive;
  published
    property    Active         : Boolean read FActive     write SetActive;
    property    Visible;
    property    ShowHint;
  private
}

    procedure   TKnobsIndicator.SetActive( aValue: Boolean);
    begin
      if aValue <> FActive
      then begin
        FActive := aValue;
        Invalidate;
      end;
    end;

    procedure   TKnobsIndicator.SetBitmapActive( aValue: TBitmap);
    begin
      FActiveBm.Assign( aValue);
      if FActive
      then Invalidate;
    end;

    procedure   TKnobsIndicator.SetBitmapInactive( aValue: TBitmap);
    begin
      FInactiveBm.Assign( aValue);
      if not FActive
      then Invalidate;
    end;

//  protected

    procedure   TKnobsIndicator.AssignBitmaps; // virtual;
    begin
      FInactiveBm := bmIndInactive;
      FActiveBm   := bmIndActive  ;
      FInActiveBm.Transparent := False;
      FActiveBm  .Transparent := False;
    end;

    procedure   TKnobsIndicator.Paint; // override;
    begin
      inherited;

      with Canvas
      do begin
        if Active
        then Draw( 0, 0, FActiveBm  )
        else Draw( 0, 0, FInactiveBm);
      end;
    end;

//  public

    constructor TKnobsIndicator.Create( anOwner: TComponent); // override;
    begin
      inherited;

      ControlStyle := ControlStyle + [ csOpaque];
      AssignBitmaps;
      Width       := Max( FActiveBm.Width , FInactiveBm.Width );
      Height      := Max( FActiveBm.Height, FInactiveBm.Height);
      Enabled     := False;
    end;


{ ========
  TKnobsIndicatorBar = class( TGraphicControl)
  // A bar of indicators, horizontal or vertical
  private
    FIndicatorCount : Integer;
    FValue          : TSignal;
    FPeakValue      : TSignal;
    FValeyValue     : TSignal;
    FShowPeakValue  : Boolean;
    FShowValeyValue : Boolean;
    FValueLowMid    : TSignal;
    FValueMidHigh   : TSignal;
    FDisplayType    : TKnobsDisplayType;
    FOrientation    : TKnobsDisplayOrientation;
    FSpacing        : Integer;
    FActiveLowBm    : TBitmap;
    FInActiveLowBm  : TBitmap;
    FActiveMidBm    : TBitmap;
    FInActiveMidBm  : TBitmap;
    FActiveHighBm   : TBitmap;
    FInActiveHighBm : TBitmap;
    FPeakColor      : TColor;
    FValeyColor     : TColor;
  public
    property    ActiveLowBm    : TBitmap                  read FActiveLowBm    write SetActiveLowBm;
    property    InActiveLowBm  : TBitmap                  read FInActiveLowBm  write SetInActiveLowBm;
    property    ActiveMidBm    : TBitmap                  read FActiveMidBm    write SetActiveMidBm;
    property    InActiveMidBm  : TBitmap                  read FInActiveMidBm  write SetInActiveMidBm;
    property    ActiveHighBm   : TBitmap                  read FActiveHighBm   write SetActiveHighBm;
    property    InActiveHighBm : TBitmap                  read FInActiveHighBm write SetInActiveHighBm;
  published
    property    IndicatorCount : Integer                  read FIndicatorCount write SetIndicatorCount default 10;
    property    Value          : TSignal                  read FValue          write SetValue          nodefault;
    property    PeakValue      : TSignal                  read FPeakValue      write SetPeakValue      nodefault;
    property    ValeyValue     : TSignal                  read FValeyValue     write SetValeyValue     nodefault;
    property    ShowPeakValue  : Boolean                  read FShowPeakValue  write SetShowPeakValue  default False;
    property    ShowValeyValue : Boolean                  read FShowValeyValue write SetShowValeyValue default False;
    property    ValueLowMid    : TSignal                  read FValueLowMid    write SetValueLowMid    nodefault;
    property    ValueMidHigh   : TSignal                  read FValueMidHigh   write SetValueMidHigh   nodefault;
    property    DisplayType    : TKnobsDisplayType        read FDisplayType    write SetDisplayType    default dtFilledFromLow;
    property    Orientation    : TKnobsDisplayOrientation read FOrientation    write SetOrientation    default doHorizontal;
    property    Spacing        : Integer                  read FSpacing        write SetSpacing        default 2;
    property    PeakColor      : TColor                   read FPeakColor      write SetPeakColor      default clWhite;
    property    ValeyColor     : TColor                   read FValeyColor     write SetValeyColor     default clBlack;
    property    Visible;
    property    Enabled;
  published
    property    OnClick;
    property    OnDblClick;
  private
}

    function    TKnobsIndicatorBar.GetMaxBmWidth: Integer;
    begin
      Result := 0;
      Result := Max( Result,   ActiveLowBm .Width);
      Result := Max( Result, InActiveLowBm .Width);
      Result := Max( Result,   ActiveMidBm .Width);
      Result := Max( Result, InActiveMidBm .Width);
      Result := Max( Result,   ActiveHighBm.Width);
      Result := Max( Result, InActiveHighBm.Width);
    end;

    function    TKnobsIndicatorBar.GetMaxBmHeight: Integer;
    begin
      Result := 0;
      Result := Max( Result,   ActiveLowBm .Height);
      Result := Max( Result, InActiveLowBm .Height);
      Result := Max( Result,   ActiveMidBm .Height);
      Result := Max( Result, InActiveMidBm .Height);
      Result := Max( Result,   ActiveHighBm.Height);
      Result := Max( Result, InActiveHighBm.Height);
    end;

    procedure   TKnobsIndicatorBar.FixDimensions;
    begin
      if Orientation = doHorizontal
      then begin
        Width  := IndicatorCount * ( GetMaxBmWidth + Spacing);
        Height :=                    GetMaxBmHeight;
      end
      else begin
        Width  :=                    GetMaxBmWidth;
        Height := IndicatorCount * ( GetMaxBmHeight + Spacing);
      end;
    end;

//  private

    procedure   TKnobsIndicatorBar.SetIndicatorCount( aValue: Integer);
    begin
      if aValue <> FIndicatorCount
      then begin
        FIndicatorCount := aValue;

        if ( Value < 0) or ( Value > aValue)
        then Value := 3 * aValue div 2;

        FixDimensions;
        Invalidate;
      end;
    end;

    procedure   TKnobsIndicatorBar.SetValue( aValue: TSignal);
    begin
      if aValue <> FValue
      then begin
        FValue := aValue;
        Invalidate;
      end;
    end;

    procedure   TKnobsIndicatorBar.SetPeakValue( aValue: TSignal);
    begin
      if aValue <> FPeakValue
      then begin
        FPeakValue := aValue;
        Invalidate;
      end;
    end;

    procedure   TKnobsIndicatorBar.SetValeyValue( aValue: TSignal);
    begin
      if aValue <> FValeyValue
      then begin
        FValeyValue := aValue;
        Invalidate;
      end;
    end;

    procedure   TKnobsIndicatorBar.SetShowPeakValue( aValue: Boolean);
    begin
      if aValue <> FShowPeakValue
      then begin
        FShowPeakValue := aValue;
        Invalidate;
      end;
    end;

    procedure   TKnobsIndicatorBar.SetShowValeyValue( aValue: Boolean);
    begin
      if aValue <> FShowValeyValue
      then begin
        FShowValeyValue := aValue;
        Invalidate;
      end;
    end;

    procedure   TKnobsIndicatorBar.SetValueLowMid( aValue: TSignal);
    begin
      if aValue <> FValueLowMid
      then begin
        FValueLowMid := aValue;
        Invalidate;
      end;
    end;

    procedure   TKnobsIndicatorBar.SetValueMidHigh( aValue: TSignal);
    begin
      if aValue <> FValueMidHigh
      then begin
        FValueMidHigh := aValue;
        Invalidate;
      end;
    end;

    procedure   TKnobsIndicatorBar.SetDisplayType( aValue: TKnobsDisplayType);
    begin
      if aValue <> FDisplayType
      then begin
        FDisplayType := aValue;
        Invalidate;
      end;
    end;

    procedure   TKnobsIndicatorBar.SetOrientation( aValue: TKnobsDisplayOrientation);
    begin
      if aValue <> FOrientation
      then begin
        FOrientation := aValue;
        FixDimensions;
        Invalidate;
      end;
    end;

    procedure   TKnobsIndicatorBar.SetSpacing( aValue: Integer);
    begin
      if aValue <> FSpacing
      then begin
        FSpacing := aValue;
        FixDimensions;
        Invalidate;
      end;
    end;

    procedure   TKnobsIndicatorBar.SetActiveLowBm( aValue: TBitmap);
    begin
      FActiveLowBm.Assign( aValue);
      FixDimensions;
      Invalidate;
    end;

    procedure   TKnobsIndicatorBar.SetInActiveLowBm( aValue: TBitmap);
    begin
      FInActiveLowBm.Assign( aValue);
      FixDimensions;
      Invalidate;
    end;

    procedure   TKnobsIndicatorBar.SetActiveMidBm( aValue: TBitmap);
    begin
      FActiveMidBm.Assign( aValue);
      FixDimensions;
      Invalidate;
    end;

    procedure   TKnobsIndicatorBar.SetInActiveMidBm( aValue: TBitmap);
    begin
      FInActiveMidBm.Assign( aValue);
      FixDimensions;
      Invalidate;
    end;

    procedure   TKnobsIndicatorBar.SetActiveHighBm( aValue: TBitmap);
    begin
      FActiveHighBm.Assign( aValue);
      FixDimensions;
      Invalidate;
    end;

    procedure   TKnobsIndicatorBar.SetInActiveHighBm( aValue: TBitmap);
    begin
      FInActiveHighBm.Assign( aValue);
      FixDimensions;
      Invalidate;
    end;

    procedure   TKnobsIndicatorBar.SetPeakColor( aValue: TColor);
    begin
      if aValue <> FPeakColor
      then begin
        FPeakColor := aValue;
        Invalidate;
      end;
    end;

    procedure   TKnobsIndicatorBar.SetValeyColor( aValue: TColor);
    begin
      if aValue <> FValeyColor
      then begin
        FValeyColor := aValue;
        Invalidate;
      end;
    end;

//  protected

    procedure   TKnobsIndicatorBar.AssignBitmaps; // virtual;
    begin
      FInactiveLowBm  := bmIndLowInactive;
      FActiveLowBm    := bmIndLowActive;
      FInactiveMidBm  := bmIndMidInactive;
      FActiveMidBm    := bmIndMidActive;
      FInactiveHighBm := bmIndHighInactive;
      FActiveHighBm   := bmIndHighActive;
      FInactiveLowBm .Transparent := False;
      FActiveLowBm   .Transparent := False;
      FInactiveMidBm .Transparent := False;
      FActiveMidBm   .Transparent := False;
      FInactiveHighBm.Transparent := False;
      FActiveHighBm  .Transparent := False;
      FixDimensions;
    end;

    function    TKnobsIndicatorBar.PaintActive( anIndex: Integer): Boolean;
    begin
      Result := False;
      case DisplayType of
        dtSingleFromLow, dtSingleFromHigh : Result := anIndex = Value - 1;
        dtFilledFromLow, dtFilledFromHigh : Result := anIndex < Value;
      end;
    end;

    procedure   TKnobsIndicatorBar.Paint; // override;
    var
      i  : Integer;
      L  : Integer;
      T  : Integer;
      W  : Integer;
      H  : Integer;
      LA : Integer;
      TA : Integer;
    begin
      inherited;

      W := GetMaxBmWidth  + Spacing;
      H := GetMaxBmHeight + Spacing;

      if Orientation = doHorizontal
      then begin
        TA := 0;

        if DisplayType in [ dtSingleFromHigh, dtFilledFromHigh]
        then begin
          W  := - W;
          LA := Width + W;
        end
        else LA := 0;
      end
      else begin
        LA := 0;

        if DisplayType in [ dtSingleFromLow, dtFilledFromLow]
        then begin
          H  := - H;
          TA := Height + H;
        end
        else TA := 0;
      end;

      L := LA;
      T := TA;

      with Canvas
      do begin
        Pen.Color := clBlack;
        Brush.Color := clBlack;
        Rectangle( 0, 0, Width, Height);

        for i := 0 to IndicatorCount - 1
        do begin
          if PaintActive( i)
          then begin
            if i < ValueLowMid
            then Draw( L, T, FActiveLowBm  )
            else if i >= ValueMidHigh
            then Draw( L, T, FActiveHighBm)
            else Draw( L, T, FActiveMidBm);
          end
          else begin
            if i < ValueLowMid
            then Draw( L, T, FInActiveLowBm  )
            else if i >= ValueMidHigh
            then Draw( L, T, FInActiveHighBm)
            else Draw( L, T, FInActiveMidBm);
          end;

          if Orientation = doHorizontal
          then L := L + W
          else T := T + H;
        end;

        if ShowPeakValue
        then begin
          Pen.Color := PeakColor;
          Pen.Width := 3;
          if Orientation = doHorizontal
          then begin
            L := LA + ( Ceil( PeakValue) - 1) * W;
            MoveTo( L, 0     );
            LineTo( L, Height);
          end
          else begin
            T := TA + ( Ceil( PeakValue) - 1) * H;
            MoveTo( 0    , T);
            LineTo( Width, T);
          end;
        end;

        if ShowValeyValue
        then begin
          Pen.Color := ValeyColor;
          Pen.Width := 3;
          if Orientation = doHorizontal
          then begin
            L := LA + ( Ceil( ValeyValue) - 1) * W;
            MoveTo( L, 0     );
            LineTo( L, Height);
          end
          else begin
            T := TA + ( Ceil( ValeyValue) - 1) * H;
            MoveTo( 0    , T);
            LineTo( Width, T);
          end;
        end;

      end;
    end;

//  public

    constructor TKnobsIndicatorBar.Create( anOwner: TComponent); // override;
    begin
      inherited;
      AssignBitmaps;
      ShowPeakValue  := False;
      ShowValeyValue := False;
      IndicatorCount := 10;
      Value          :=  6;
      PeakValue      :=  0;
      ValueLowMid    :=  5;
      ValueMidHigh   :=  8;
      DisplayType    := dtFilledFromLow;
      Orientation    := doHorizontal;
      Spacing        := 2;
      FPeakColor     := clWhite;
      FValeyColor    := clBlack;

      FixDimensions;
    end;


{ ========
  TKnobsIndicatorText = class( TCustomLabel)
  private
    FValue : TSignal;
  published
    property Value : TSignal read FValue write SetValue nodefault;
    property Width                                      default 10;
    property Height                                     default 10;
    property AutoSize                                   default false;
    property Align                                      default alNone;
    property Alignment                                  default taCenter;
    property Color                                      default KnobsColorDisplay;
    property Transparent                                default True;
    property Visible                                    default True;
    property WordWrap                                   default False;
    property ParentColor                                default False;
    property ParentFont                                 default False;
    property ShowHint                                   default False;
    property ParentShowHint                             default False;
    property ShowAccelChar                              default False;
    property Layout;
    property Anchors;
    property BiDiMode;
    property Caption;
    property Constraints;
    property Enabled;
    property FocusControl;
    property Font;
    property ParentBiDiMode;
    property PopupMenu;
  protected
}

    procedure   TKnobsIndicatorText.SetValue( aValue: TSignal);
    begin
      if aValue <> FValue
      then begin
        FValue := aValue;
        Invalidate;
      end;
    end;

    procedure   TKnobsIndicatorText.DoDrawText( var aRect: TRect; aFlags: Longint); // override;
    var
      Text: string;
    begin
      Text   := IntToStr( Round( FValue));
      aFlags := aFlags or DT_NOPREFIX;
      aFlags := DrawTextBiDiModeFlags( aFlags);
      Canvas.Font := Font;
      DrawText( Canvas.Handle, PChar( Text), Length( Text), aRect, aFlags);
    end;

//  public

    constructor TKnobsIndicatorText.Create( anOwner: TComponent); // override;
    begin
      inherited;

      Width          := 10;
      Height         := 10;
      AutoSize       := false;
      Align          := alNone;
      Alignment      := taCenter;
      Color          := KnobsColorDisplay;
      Transparent    := True;
      Visible        := True;
      WordWrap       := False;
      ParentColor    := False;
      ParentFont     := False;
      ShowHint       := False;
      ParentShowHint := False;
      ShowAccelChar  := False;

      with Font
      do begin
        Name  := 'Small Fonts';
        Size  := 7;
        Color := clWhite;
      end;
    end;


{ ========
  TKnobsDataViewer = class( TGraphicControl)
  // a viewer
  private
    FBackGround  : TBitmap;
    FYData       : TSignalArray;
    FXYData      : TSignalPairFifo;
    FXYPoints    : Integer;
    FXOffset     : TSignal;
    FYOffset     : TSignal;
    FMaxValue    : TSignal;
    FMinValue    : TSignal;
    FXZoom       : TSignal;
    FYZoom       : TSignal;
    FBorderColor : TColor;
    FLineColor   : TColor;
    FFillColor   : TColor;
    FStyle       : TKnobsDVStyle;
  public
    property    YData       : TSignalArray       read FYData     write SetYData;
    property    XYData      : TSignalPairFifo    read FXYData    write SetXYData;
    property    Module      : TKnobsCustomModule read GetModule;
  published
    property    Background  : TBitmap       read FBackGround  write SetBackGround;
    property    XOffset     : TSignal       read FXOffset     write SetXOffset;    // default  0;
    property    YOffset     : TSignal       read FYOffset     write SetYOffset;    // default 75;
    property    MaxValue    : TSignal       read FMaxValue    write SetMaxValue;   // default High( SmallInt);
    property    MinValue    : TSignal       read FMinValue    write SetMinValue;   // default Low ( SmallInt);
    property    XZoom       : TSignal       read FXZoom       write SetXZoom;      // default 1;
    property    YZoom       : TSignal       read FYZoom       write SetYZoom;      // default 1;
    property    LineColor   : TColor        read FLineColor   write SetLineColor   default clWhite;
    property    FillColor   : TColor        read FFillColor   write SetFillColor   default clSilver;
    property    BorderColor : TColor        read FBorderColor write SetBorderColor default KnobsColorDisplay;
    property    Style       : TKnobsDVStyle read FStyle       write SetStyle       default dvsNoise;
    property    XYPoints    : Integer       read FXYPoints    write SetXYPoints    default 255;
    property    Width                                                              default 150;
    property    Height                                                             default 60;
    property    Color                                                              default KnobsColorDisplay;
  private
}

    function    TKnobsDataViewer.GetModule: TKnobsCustomModule;
    begin
      if Owner is TKnobsCustomModule
      then result := TKnobsCustomModule( Owner)
      else result := nil;
    end;

    procedure   TKnobsDataViewer.SetBackGround( aValue: TBitmap);
    begin
      FBackGround.Assign( aValue);
      Invalidate;
    end;

    procedure   TKnobsDataViewer.SetYData( const aValue: TSignalArray);
    begin
      FYData := aValue;
      Invalidate;
    end;

    procedure   TKnobsDataViewer.SetXYData( const aValue: TSignalPairFifo);
    begin
      if Assigned( aValue) and ( aValue.FillCount > 0)
      then begin
        while aValue.FillCount > 0
        do FXYData.Append( aValue.GetData);
        Invalidate;
      end;
    end;

    procedure   TKnobsDataViewer.SetXOffset( aValue: TSignal);
    begin
      if aValue <> FXOffset
      then begin
        FXOffset := aValue;
        Invalidate;
      end;
    end;

    procedure   TKnobsDataViewer.SetYOffset( aValue: TSignal);
    begin
      if aValue <> FYOffset
      then begin
        FYOffset := aValue;
        Invalidate;
      end;
    end;

    procedure   TKnobsDataViewer.SetMaxValue( aValue: TSignal);
    begin
      if aValue <> FMaxValue
      then begin
        FMaxValue := aValue;
        Invalidate;
      end;
    end;

    procedure   TKnobsDataViewer.SetMinValue( aValue: TSignal);
    begin
      if aValue <> FMinValue
      then begin
        FMinValue := aValue;
        Invalidate;
      end;
    end;

    procedure   TKnobsDataViewer.SetXZoom( aValue: TSignal);
    begin
      if aValue <> FXZoom
      then begin
        FXZoom := aValue;
        Invalidate;
      end;
    end;

    procedure   TKnobsDataViewer.SetYZoom( aValue: TSignal);
    begin
      if aValue <> FYZoom
      then begin
        FYZoom := aValue;
        Invalidate;
      end;
    end;

    procedure   TKnobsDataViewer.SetLineColor( aValue: TColor );
    begin
      if aValue <> FLineColor
      then begin
        FLineColor := aValue;
        Invalidate;
      end;
    end;

    procedure   TKnobsDataViewer.SetFillColor( aValue: TColor);
    begin
      if aValue <> FFillColor
      then begin
        FFillColor := aValue;
        Invalidate;
      end;
    end;

    procedure   TKnobsDataViewer.SetBorderColor( aValue: TColor );
    begin
      if aValue <> FBorderColor
      then begin
        FBorderColor := aValue;
        Invalidate;
      end;
    end;

    procedure   TKnobsDataViewer.SetStyle( aValue: TKnobsDVStyle);
    begin
      if aValue <> FStyle
      then begin
        FStyle := aValue;
          case aValue of
            dvsBlank   : FillBlank;
            dvsSine    : FillSine;
            dvsTri     : FillTri;
            dvsSaw     : FillSaw;
            dvsSquare  : FillSquare;
            dvsNoise   : FillNoise;
            dvsEnvAr   : FillEnvAr;
            dvsEnvAhd  : FillEnvAhd;
            dvsEnvAdsr : FillEnvAdsr;
            dvsXYScope : FillXYScope;
            else         FillBlank;
        end;
        Invalidate;
      end;
    end;

    procedure   TKnobsDataViewer.SetXYPoints( aValue: Integer);
    begin
      if aValue <> FXYPoints
      then begin
        FXYPoints := aValue;
      end;
    end;

//  private

    procedure   TKnobsDataViewer.FillBlank;
    begin
      SetLength( FYData, 0);
    end;

    procedure   TKnobsDataViewer.FillSine;
    var
      i : Integer;
    begin
      SetLength( FYData, Width);
      for i := 0 to Width
      do FYData[ i] := ( MaxValue - MinValue) * ( Sin( 2 * pi * i / width) + 1) / 2;
    end;

    procedure   TKnobsDataViewer.FillTri;
    var
      i         : Integer;
      PhaseAccu : TSignal;
      HalfValue : TSignal;
    begin
      SetLength( FYData, Width);
      HalfValue := ( MaxValue - MinValue) / 2;
      for i := 0 to Width
      do begin
        PhaseAccu := i / Width;
        if PhaseAccu <= 0.5
        then FYData[ i] := HalfValue * 4 * PhaseAccu
        else FYData[ i] := HalfValue * 4 * ( 1 - PhaseAccu);
      end;
    end;

    procedure   TKnobsDataViewer.FillSaw;
    var
      i         : Integer;
      PhaseAccu : TSignal;
      HalfValue : TSignal;
    begin
      SetLength( FYData, Width);
      HalfValue := ( MaxValue - MinValue) / 2;
      for i := 0 to Width
      do begin
        PhaseAccu := i / Width;
        FYData[ i] := HalfValue * 2 * PhaseAccu;
      end;
    end;

    procedure   TKnobsDataViewer.FillSquare;
    var
      i         : Integer;
      PhaseAccu : TSignal;
    begin
      SetLength( FYData, Width);
      for i := 0 to Width
      do begin
        PhaseAccu := i / Width;
        if PhaseAccu < 0.5
        then FYData[ i] := MinValue
        else FYData[ i] := MaxValue;
      end;
    end;

    procedure   TKnobsDataViewer.FillNoise;
    var
      i : Integer;
    begin
      SetLength( FYData, Width div 5);
      for i := 0 to Length( FYData) - 1
      do FYData[ i] := MinValue + ( MaxValue - MinValue) * Random;
    end;

    procedure   TKnobsDataViewer.FillEnvAr;
    var
      i    : Integer;
      Knee : Integer;
    begin
      SetLength( FYData, Width);
      Knee := Width div 5;
      for i := 0 to Length( FYData) - 1
      do begin
        if i < Knee
        then FYData[ i] := LinAttack( i, Knee, MinValue, MaxValue)
        else FYData[ i] := ExpDecay( i - Knee, Width - Knee, MaxValue, MinValue);
      end;
    end;

    procedure   TKnobsDataViewer.FillEnvAhd;
    var
      i     : Integer;
      Knee1 : Integer;
      Knee2 : Integer;
    begin
      SetLength( FYData, Width);
      Knee1 := Width div 6;
      Knee2 := Width div 3;
      for i := 0 to Length( FYData) - 1
      do begin
        if i < Knee1
        then FYData[ i] := LinAttack( i, Knee1, MinValue, MaxValue)
        else if i < Knee2
        then FYData[ i] := MaxValue
        else FYData[ i] := ExpDecay( i - Knee2, Width - Knee2, MaxValue, MinValue);
      end;
    end;

    procedure   TKnobsDataViewer.FillEnvAdsr;
    var
      i     : Integer;
      Knee1 : Integer;
      Knee2 : Integer;
      Knee3 : Integer;
      Sust  : TSignal;
    begin
      SetLength( FYData, Width);
      Knee1 := Width div 8;
      Knee2 := Width div 2;
      Knee3 := 2 * Width div 3;
      Sust  := MinValue + ( MaxValue - MinValue) / 3;
      for i := 0 to Length( FYData) - 1
      do begin
        if i < Knee1
        then FYData[ i] := LinAttack( i, Knee1, MinValue, MaxValue)
        else if i < Knee2
        then FYData[ i] := ExpDecay( i - Knee1, Knee3 - Knee2, MaxValue, Sust)
        else if i < Knee3
        then FYData[ i] := Sust
        else FYData[ i] := ExpDecay( i - Knee3, Width - Knee3, Sust, MinValue);
      end;
    end;

    procedure   TKnobsDataViewer.FillXYScope;
    const
      AngleCount = 5;
      Angles : array[ 0 .. AngleCount - 1] of TSignal = (
        1.5, 2.5, 3.5, 4.5, 5.5
      );
    var
      Angle : Integer;
      Rad   : TSignal;
      X     : TSignal;
      Y     : TSignal;
      Typ   : Integer;
    begin
      FXYData.Clear;
      Typ := Random( AngleCount);
      for Angle := 0 to 5 * 360
      do begin
        Rad := DegToRad( Angle);
        X   := Sin( Angles[ Typ] * Rad);
        Y   := Cos( 3.01         * Rad);
        FXYData.Append( 0.95 * X, 0.95 * Y);
      end;
    end;

//  protected

    procedure   TKnobsDataViewer.WMNCHitTest( var aMessage: TWMNCHitTest);  // message WM_NCHITTEST;
    begin
      if csDesigning in ComponentState
      then inherited
      else aMessage.Result := HTTRANSPARENT;
    end;

    procedure   TKnobsDataViewer.WMEraseBackground( var aMsg: TWMEraseBkgnd); // message WM_ERASEBKGND;
    begin
      aMsg.Result := -1;
    end;

    procedure   TKnobsDataViewer.MouseDown( Button: TMouseButton; Shift: TShiftState; X, Y: Integer); // override;
    var
      aModule: TKnobsCustomModule;
    begin
      if csDesigning in ComponentState
      then inherited
      else begin
        // Pass it on
        aModule := Module;

        if Assigned( aModule)
        then begin
          aModule.MouseCapture := True;
          aModule.MouseDown( Button, Shift, X + Left, Y + Top);
        end;
      end;
    end;

    function    TKnobsDataViewer.HasNegativeData: Boolean;
    var
      i : Integer;
    begin
      Result := False;
      for i := 0 to Length( FYData) - 1
      do begin
        if FYData[ i] < 0
        then begin
          Result := True;
          Break;
        end;
      end;
    end;

    function    TKnobsDataViewer.ScaleData( anIndex: Integer): Integer;
    var
      aVal: TSignal;
    begin
      aVal := Clip( FYData[ anIndex] , MinValue, MaxValue);
      Result := Height - Round( Clip( aVal * YZoom * ( Height / ( MaxValue - MinValue)) + YOffset, 0, Height));
    end;

    function    TKnobsDataViewer.ScaleDataNegative( anIndex: Integer): Integer;
    var
      aVal: TSignal;
    begin
      aVal := Clip( 1 - Abs( FYData[ anIndex]), MinValue, MaxValue);
      Result := Height - Round( Clip( aVal * YZoom * ( Height / ( MaxValue - MinValue)) + YOffset, 0, Height));
    end;

    function    TKnobsDataViewer.ScaleX( anIndex: Integer): Integer;
    begin
      Result := Round( Clip( anIndex * XZoom * ( Width / Length( FYData)) - XOffset, 0, Width));
    end;

    procedure   TKnobsDataViewer.Paint; // override;
    var
      i     : Integer;
      X     : Integer;
      Y     : Integer;
      LastX : Integer;
      LastY : Integer;
      aData : TSignalPair;
    begin
      inherited;

      with Canvas
      do begin
        if csDesigning in ComponentState
        then begin
          Pen.Width   := 1;
          Pen.Style   := psDot;
          Brush.Style := bsClear;
          Rectangle( 0, 0, Width, Height);
          Pen.Style   := psSolid;
          Brush.Style := bsSolid;
        end;

        if Assigned( FBackGround) and not FBackGround.Empty
        then StretchDraw( Rect( 0, 0, Height, Width), FBackGround)
        else begin
          Pen.Width   := 1;
          Pen.Color   := BorderColor;
          Brush.Color := Color;
          Rectangle( 0, 0, Width, Height);
        end;

        if ( Length( FYData) > 0) or ( FXYData.Size > 0)
        then begin
          Pen.Width := 1;
          Pen.Color := FillColor;

          if ( Style in [ dvsEnvAr, dvsEnvAhd, dvsEnvAdsr]) and HasNegativeData
          then begin
            for i := 1 to Length( FYData) - 1
            do begin
              MoveTo( ScaleX( i), 0);
              LineTo( ScaleX( i), ScaleDataNegative( i));
            end;
            Pen.Color := LineColor;
            Pen.Width := 1;
            MoveTo( ScaleX( 0), ScaleDataNegative( 0));
            for i := 1 to Length( FYData) - 1
            do LineTo( ScaleX( i), ScaleDataNegative( i));
          end
          else if ( Style = dvsXYScope)
          then begin
            ClipRect.Inflate( -2, -2);
            Pen.Color := LineColor;
            Pen.Width := 1;
            LastX := 0;
            LastY := 0;
            for i := 0 to FXYData.Size - 1
            do begin
              aData := FXYData.GetData;
              X := Round( ( 0.5 * ( Width  - 5) * ( 1 + aData.Left ))) + 2;
              Y := Round( ( 0.5 * ( Height - 5) * ( 1 - aData.Right))) + 2;
              if i = 0
              then begin
                MoveTo( X, Y);
              end
              else begin
                if ( X <> LastX) or ( Y <> LastY)
                then LineTo( X, Y);
              end;
              LastX := X;
              LastY := Y;
            end;
            if csDesigning in ComponentState
            then FillXYScope;
          end
          else begin
            for i := 1 to Length( FYData) - 1
            do begin
              MoveTo( ScaleX( i), Height - 1);
              LineTo( ScaleX( i), ScaleData( i));
            end;
            Pen.Color := LineColor;
            Pen.Width := 1;
            MoveTo( ScaleX( 0), ScaleData( 0));
            for i := 1 to Length( FYData) - 1
            do LineTo( ScaleX( i), ScaleData( i));
          end;

        end;
      end;
    end;

//  public

    constructor TKnobsDataViewer.Create( anOwner: TComponent); // override;
    begin
      inherited;

      FXYData := TSignalPairFifo.Create( System_Rate);
      FXYData.CountOverflows := False;

      ControlStyle := ControlStyle + [ csOpaque];
      FBackGround  := TBitmap.Create;
      XOffset      := 0;
      YOffset      := 30;
      MaxValue     := High( SmallInt);
      MinValue     := Low ( SmallInt);
      XZoom        := 1;
      YZoom        := 1;
      Width        := 150;
      Height       := 60;
      LineColor    := clWhite;
      FillColor    := clSilver;
      BorderColor  := KnobsColorDisplay;
      Color        := KnobsColorDisplay;
      Style        := dvsNoise;
    end;

    destructor  TKnobsDataViewer.Destroy; //                                               override;
    begin
      FreeAndNil( FXYData);
      inherited;
    end;


{ ========
  TKnobsConnector = class( TGraphicControl)
  // to connect wires to
  private
    FBitmap     : TBitmap;
    FSignalType : TSignalType;
    FWireColor  : TColor;
    FLinks      : TList;   // Connections, not wires
    FMouseDown  : Boolean;
    FConnected  : Boolean; // True when at least one connection exists
    FHighLight  : Boolean; // Highlighted connections
    FVisited    : Boolean;
  protected
    property    HighLight  : Boolean            read FHighLight    write SetHighLight;
  public
    property    Connected  : Boolean            read FConnected    write SetConnected;
    property    Module     : TKnobsCustomModule read GetModule;
    property    ModuleName : string             read GetModuleName;
    property    Signaltype : TSignalType        read FSignalType   write SetSignalType  default stControl;
  published
    property    WireColor  : TColor             read FWireColor    write SetWireColor   default clBlue;
    property    Visible;
  private
}

    procedure   TKnobsConnector.SetConnected( aValue: Boolean);
    begin
      if aValue <> FConnected
      then begin
        FConnected := aValue;
        Invalidate;
      end;
    end;

    function    TKnobsConnector.GetModule: TKnobsCustomModule;
    begin
      if Owner is TKnobsCustomModule
      then result := TKnobsCustomModule( Owner)
      else result := nil;
    end;

    function    TKnobsConnector.GetModuleName: string;
    var
      aModule: TKnobsCustomModule;
    begin
      aModule := Module;

      if Assigned( aModule)
      then Result := aModule.Name
      else Result := '';
    end;

    procedure   TKnobsConnector.SetHighLight( aValue: Boolean);
    var
      i : Integer;
    const
      Depth : Integer = 0;
    begin
      if aValue <> FHighLight
      then begin
        if not FVisited
        then begin
          try
            FVisited   := True;
            FHighLight := aValue;
            Invalidate;
            Inc( Depth);

            try
              with FLinks
              do
                for i := 0 to Count - 1
                do TKnobsConnector( Items[ i]).HighLight := aValue;
            finally
              Dec( Depth);
            end;

          finally
            FVisited := False;
          end;
        end;
      end;
      if Depth = 0
      then SendHighlightNotification;
    end;

    procedure   TKnobsConnector.SetWireColor( aValue: TColor);
    begin
      if aValue <> FWireColor
      then begin
        FWireColor  := aValue;
        FSignalType := ColorToSignalType( aValue);
        SendColorNotification( FWireColor);
        Invalidate;
      end;
    end;

    procedure   TKnobsConnector.SetSignalType( aValue: TSignalType);
    begin
      if aValue <> FSignalType
      then begin
        FSignalType := aValue;
        WireColor   := SignalTypeToColor( aValue);
      end;
    end;

    function    TKnobsConnector.FindWirePanel: TKnobsWirePanel;
    var
      aParent : TWinControl;
    begin
      Result  := nil;
      aParent := Parent;

      while Assigned( aParent) and not ( aParent is TKnobsWirePanel)
      do aParent :=  aParent.Parent;

      if aParent is TKnobsWirePanel
      then Result := TKnobsWirePanel( aParent);
    end;

    procedure   TKnobsConnector.AddLink( aConnector: TKnobsConnector);
    begin
      if not HasDirectlinkTo( aConnector)
      then begin
        FLinks.Add( aConnector);
        Connected := True;
      end;
    end;

    procedure   TKnobsConnector.DeleteLink( aConnector: TKnobsConnector);
    begin
      Connected := False;

      with FLinks
      do begin
        Remove( aConnector);
        Connected := Count > 0;
      end;
    end;

    function    TKnobsConnector.HasDirectlinkTo( aConnector: TKnobsConnector): Boolean;
    begin
      Result := FLinks.IndexOf( aConnector) >= 0;
    end;

    function    TKnobsConnector.FindConnector( aScreenPoint: TPoint): TKnobsConnector;
    var
      aControl : TControl;
    begin
      Result := nil;
      aControl := FindDragTarget( aScreenPoint, True);

      if Assigned( aControl) and ( aControl is TKnobsConnector)
      then Result := TKnobsConnector( aControl);
    end;

    procedure   TKnobsConnector.DrawLine( X, Y: Integer);
    var
      aDC    : HDC;
      aPen   : HPen;
      OldObj : HGDIOBJ;
      aColor : TColor;
      aWp    : TKnobsWirePanel;
    begin
      ScrollMouseInView( X, Y);
      aColor := clWhite;
      aDC    := GetDC( 0);

      try
        SetRop2( aDC, R2_XORPEN);
        aPen   := CreatePen( PS_SOLID, 2, ColorToRGB( aColor));
        OldObj := SelectObject( aDC, aPen);

        try
          aWp := FindWirePanel;
          if Assigned( aWp)
          then
            with aWp, ClientOrigin
            do IntersectClipRect( aDC, X, Y, X + ClientWidth, Y + ClientHeight);
          with ClientToScreen( CenterPoint)
          do MoveToEx( aDC, X, Y, nil);
          with ClientToScreen( Point( X, Y))
          do LineTo( aDC, X, Y);
        finally
          SelectObject( aDC, OldObj);
          DeleteObject( aPen);
        end;

      finally
        ReleaseDC( 0, aDC);
      end;
    end;

    procedure   TKnobsConnector.SendWireNotification( Msg: Cardinal; aConnector: TKnobsConnector);
    var
      aWp : TKnobsWirePanel;
    begin
      aWp := FindWirePanel;

      if Assigned( aWp)
      then aWp.Perform( Msg, WPARAM( Self), LPARAM( aConnector));
    end;

    procedure   TKnobsConnector.SendColorNotification( aColor: TColor);
    var
      aWp : TKnobsWirePanel;
    begin
      aWp := FindWirePanel;

      if Assigned( aWp)
      then aWp.Perform( KNOBS_UM_COLORCHANGE, WPARAM( Self), LPARAM( aColor));
    end;

    procedure   TKnobsConnector.SendHighlightNotification;
    var
      aWp : TKnobsWirePanel;
    begin
      aWp := FindWirePanel;

      if Assigned( aWp)
      then aWp.Perform( KNOBS_UM_HIGHLIGHT, WPARAM( Self), 0);
    end;

    procedure   TKnobsConnector.ScrollMouseInView( anX, anY: LongInt);
    var
      aWp : TKnobsWirePanel;
    begin
      aWp := FindWirePanel;

      if Assigned( aWp)
      then aWp.ScrollPointInView( aWp.ScreenToClient( ClientToScreen( Point( anX, anY))));
    end;

//  protected

    function    TKnobsConnector.CenterPoint: TPoint;
    begin
      with Result
      do begin
        x := Width  div 2;
        y := Height div 2;
      end;
    end;

    procedure   TKnobsConnector.MouseDown( Button: TMouseButton; Shift: TShiftState; X, Y: Integer); // override;
    var
      aWP : TKnobsWirePanel;
    begin
      inherited;

      if FMouseDown
      then Exit;

      if Button = mbLeft
      then begin
        FMouseDown := True;
        GDistance  := Point( X, Y);
        DrawLine( X, Y);
        HighLight := True;
      end
      else if Button = mbRight
      then begin
        aWP := FindWirePanel;
        if Assigned( aWP)
        then aWP.HandleRightClick( Self);
      end;
    end;

    procedure   TKnobsConnector.MouseUp( Button: TMouseButton; Shift: TShiftState; X, Y: Integer); // override;
    var
      aConnector : TKnobsConnector;
      isMove     : Boolean;
    begin
      inherited;

      if FMouseDown
      then begin
        with GDistance
        do DrawLine( X, Y);

        FMouseDown := False;
        isMove        := ( ssDouble in Shift) or ( ssCtrl in Shift);
        HighLight     := False;
        Screen.Cursor := crdefault;
        aConnector    := FindConnector( ClientToScreen( Point( X, Y)));

        if Assigned( aConnector) and aConnector.CanAccept( Self, isMove)
        then begin
          if isMove
          then MoveLinksTo( aConnector)
          else ConnectTo  ( aConnector);
        end
        else begin
          if not Assigned( aConnector) and isMove
          then DisconnectAll;
        end;
      end;
    end;

    procedure   TKnobsConnector.MouseMove( Shift: TShiftState; X, Y: Integer); // override;
    var
      aConnector : TKnobsConnector;
      isMove     : Boolean;
    begin
      inherited;

      if FMouseDown
      then begin
        with GDistance
        do DrawLine( X, Y);

        GDistance := Point( X, Y);

        with GDistance
        do DrawLine( X, Y);

        isMove     := ( ssDouble in Shift) or ( ssCtrl in Shift);
        aConnector := FindConnector( ClientToScreen( Point( X, Y)));

        if Assigned( aConnector)
        then begin
          if aConnector.CanAccept( Self, isMove)
          then begin
            if isMove
            then Screen.Cursor := crMultiDrag
            else Screen.Cursor := crHandPoint;
          end
          else Screen.Cursor := crDefault;
        end
        else Screen.Cursor := crdefault;
      end;
    end;

    function    TKnobsConnector.FindOutput: TKnobsConnector;
    var
      i : Integer;
    begin
      if Self is TKnobsOutput
      then Result := Self
      else begin
        Result := nil;

        if not FVisited
        then begin
          FVisited := True;

          try
            with FLinks
            do
              for i := 0 to Count - 1
              do begin
                Result := TKnobsConnector( Items[ i]).FindOutput;
                if Assigned( Result)
                then Break;
              end;
          finally
            FVisited := False;
          end;

        end;
      end;
    end;

    function    TKnobsConnector.IsConnectedToOutput: Boolean;
    begin
      Result := FindOutput <> nil;
    end;

    procedure   TKnobsConnector.FindDistanceToOutput( var aDistance: Integer);
    // Should at the top level be called with aDistance set to -1
    // so that when no originating output can be found the final
    // value of aDistance will still be -1.
    // When this module is an output, just set aDistance to 0 and return,
    // otherwise iterate over all links to find one that ends on an output.
    // If so, that recursive call will return 0, which then will be
    // inceremented before this function returns .. i.e it is on further away
    // for every return executed.
    var
      i : Integer;
    begin
      if Self is TKnobsOutput
      then aDistance := 0
      else begin
        if not FVisited
        then begin
          FVisited := True;
          try
            with FLinks
            do
              for i := 0 to Count - 1
              do begin
                TKnobsConnector( Items[ i]).FindDistanceToOutput( aDistance);
                if aDistance >= 0
                then begin
                  Inc( aDistance);
                  Break;
                end;
              end;
          finally
            FVisited := False;
          end;
        end;
      end;
    end;

    function    TKnobsConnector.IsConnectedTo( aConnector: TKnobsConnector): Boolean;
    var
      i : Integer;
    begin
      Result := False;

      if not FVisited
      then begin
        FVisited := True;

        try
          with FLinks
          do
            for i := 0 to Count - 1
            do begin
              Result :=
                HasDirectlinkTo( aConnector) or
                TKnobsConnector( Items[ i]).IsConnectedTo( aConnector);

              if Result
              then Break;
            end;
        finally
          FVisited := False;
        end;
      end;
    end;

    function    TKnobsConnector.CanAccept( aSource: TKnobsConnector; isMove: Boolean): Boolean;
    begin
      Result := False;

      if Assigned( aSource) and ( aSource <> Self)
      then begin
        if isMove
        then begin
          Result :=
            ( aSource is TKnobsOutput) or
            not ( IsConnectedToOutput and aSource.IsConnectedToOutput);
        end
        else begin
          if IsConnectedToOutput
          then Result := not aSource.IsConnectedToOutput
          else Result := not IsConnectedTo( aSource);
        end;
      end;
    end;

    procedure   TKnobsConnector.MoveLinksTo( aConnector: TKnobsConnector);
    var
      aMoved  : TKnobsConnector;
      aWp     : TKnobsWirePanel;
      Changed : Boolean;
    begin
      if Assigned( aConnector) and ( aConnector <> Self)
      then begin
        Changed := False;
        aWp     := FindWirePanel;
        if Assigned( aWp)
        then begin
          with FLinks
          do begin
            if Count > 0
            then begin
              aWp.BeginStateChange;
              try
                while Count > 0
                do begin
                  aMoved := TKnobsConnector( Items[ Count - 1]);
                  DisConnectFrom( aMoved);
                  aConnector.ConnectTo( aMoved);
                  Changed := True;
                end;
              finally
                aWp.EndStateChange( Changed);
              end;
            end;
          end;
        end;
      end;
    end;

    procedure   TKnobsConnector.ConnectTo( aConnector: TKnobsConnector);
    var
      aWp     : TKnobsWirePanel;
      Changed : Boolean;
    begin
      if Assigned( aConnector) and aConnector.CanAccept( Self, False)
      then begin
        Changed := False;
        aWp     := FindWirePanel;
        if Assigned( aWp)
        then begin
          aWp.BeginStateChange;
          try
            AddLink( aConnector);
            aConnector.AddLink( Self);
            SendWireNotification( KNOBS_UM_CONNECTIONADD, aConnector);
            Changed := True;
          finally
            aWp.EndStateChange( Changed);
          end;
        end;
      end;
    end;

    procedure   TKnobsConnector.DisconnectFrom( aConnector: TKnobsConnector);
    var
      aWp     : TKnobsWirePanel;
      Changed : Boolean;
    begin
      if HasDirectlinkTo( aConnector)
      then begin
        Changed := False;
        aWp     := FindWirePanel;
        if Assigned( aWp)
        then begin
          aWp.BeginStateChange;
          try
            DeleteLink( aConnector);
            aConnector.DeleteLink( Self);
            SendWireNotification( KNOBS_UM_CONNECTIONREMOVE, aConnector);
            Changed := True;
          finally
            aWp.EndStateChange( Changed);
          end;
        end;
      end;
    end;

    function   TKnobsConnector.DisConnectAll: Boolean;
    var
      i        : Integer;
      OldLinks : TList;
      aWp      : TKnobsWirePanel;
      Changed  : Boolean;
    begin
      Changed := False;
      if FLinks.Count > 0
      then begin
        Changed := False;
        aWp     := FindWirePanel;
        if Assigned( aWp)
        then begin
          aWp.BeginStateChange;
          try
            OldLinks := TList.Create;

            try
              with FLinks do
              begin
                for i := 0 to Count - 1
                do OldLinks.Add( Items[ i]);

                while Count > 0
                do begin
                  DisconnectFrom( FLinks[ Count - 1]);
                  Changed := True;
                end;
              end;

              with OldLinks
              do begin
                if Count > 1
                then begin
                  for i := 1 to Count - 1
                  do begin
                    if TKnobsConnector( Items[ i - 1]).HasParent and TKnobsConnector( Items[ i]).HasParent
                    then begin
                      TKnobsConnector( Items[ i - 1]).ConnectTo( TKnobsConnector( Items[ i]));
                      Changed := True;
                    end;
                  end;
                end;
              end;
            finally
              OldLinks.Free;
            end;
          finally
            aWp.EndStateChange( Changed);
          end;
        end;
      end;
      Result := Changed;
    end;

    procedure   TKnobsConnector.DisConnectDownStream;
    // todo : Should disconnect the downstream links, currently breaks the upstream one too
    var
      aWp     : TKnobsWirePanel;
      Changed : Boolean;
    begin
      if FLinks.Count > 0
      then begin
        Changed := False;
        aWp     := FindWirePanel;
        if Assigned( aWp)
        then begin
          aWp.BeginStateChange;
          try
            with FLinks do
            begin
              while Count > 0
              do begin
                DisconnectFrom( FLinks[ Count - 1]);
                Changed := True;
              end;
            end;
          finally
            aWp.EndStateChange( Changed);
          end;
        end;
      end;
    end;

    procedure   TKnobsConnector.DisConnectDownStreamRecursive;
    var
      aWp     : TKnobsWirePanel;
      Changed : Boolean;
    begin
      if ( FLinks.Count > 0) and not FVisited
      then begin
        Changed := False;
        aWp     := FindWirePanel;
        if Assigned( aWp)
        then begin
          aWp.BeginStateChange;
          FVisited := True;
          try
            with FLinks do
            begin
              while Count > 0
              do begin
                if TKnobsConnector( FLinks[ Count - 1]).DisConnectAll
                then Changed := True;
              end;
            end;
          finally
            FVisited := False;
            aWp.EndStateChange( Changed);
          end;
        end;
      end;
    end;

    procedure   TKnobsConnector.Paint; // override;
    begin
      with Canvas
      do begin
        if FHighLight or FConnected
        then begin
          Pen.Width := 1;

          if FHighLight
          then begin
            Brush.Color := CL_HIGHLIGHT;
            Pen  .Color := CL_HIGHLIGHT;
          end
          else if FConnected
          then begin
            Brush.Color := WireColor;
            Pen  .Color := clBlack;
          end;

          if Self is TKnobsOutput
          then Rectangle( 0, 0, Width, Height)
          else Ellipse( 0, 0, Width, Height);

          Pen.Color := clWhite;

          Arc(
            Width  * 3 div 4         ,
            Height * 3 div 4         ,
            Width  - Width  * 3 div 4,
            Height - Height * 3 div 4,
            Width div 2              ,
            0                        ,
            0                        ,
            Height
          );

          Pen.Color := clGray;

          Arc(
            Width  * 3 div 4         ,
            Height * 3 div 4         ,
            Width  - Width  * 3 div 4,
            Height - Height * 3 div 4,
            0                        ,
            Width div 2              ,
            Height                   ,
            0
          );
        end
        else begin
          Draw( 0, 0, FBitmap);
          Brush.Color := WireColor;
          Pen  .Color := WireColor;
          Ellipse( 3, 3, Width - 3, Height - 3);
        end;
      end;
    end;

//  public

    constructor TKnobsConnector.Create( anOwner: TComponent); // override;
    begin
      inherited;

      AssignBitmap;
      Width       := FBitmap.Width;
      Height      := FBitmap.Height;
      FLinks      := TList.Create;
      FWireColor  := clBlue;
      FSignalType := stControl;
    end;

    destructor  TKnobsConnector.Destroy; // override;
    begin
      DisConnectAll;
      FLinks.Free;
      inherited;
    end;

    procedure   TKnobsConnector.Dump( var aFile: TextFile; anInd: Integer);
    begin
      WriteLnInd( aFile, anInd, Format( 'connector     : %s', [ ConnectorName( self)], AppLocale));
      DumpLinks( aFile, anInd + 1);
      WriteLnInd( aFile, anInd,         'end connector');
    end;

    procedure   TKnobsConnector.DumpFlat( var aFile: TextFile; anInd: Integer);
    begin
      WriteLnInd( aFile, anInd, Format( 'links to : %s', [ ConnectorName( Self)], AppLocale));
    end;

    procedure   TKnobsConnector.DumpLinks( var aFile: TextFile; anInd: Integer);
    var
      i : Integer;
    begin
      for i := 0 to FLinks.Count - 1
      do TKnobsConnector( FLinks[ i]).DumpFlat( aFile, anInd);
    end;

    procedure   TKnobsConnector.Log( const aMsg: string);
    begin
      if Parent is TKnobsCustomModule
      then TKnobsCustomModule( Parent).Log( Format( 'connector %s( %s)', [ ConnectorName( Self), aMsg], AppLocale));
    end;

    function    TKnobsConnector.DistanceToOutput: Integer;
    // Return the distance to the output, there should be one at most
    // (as code elswhere should prohibit output to output connections,
    // raising an EShortCircuit on that).
    // When an output can not be reached return -1
    // When this connector is an output return 0
    begin
      Result := -1;                     // Assume no connection to an output ...
      FindDistanceToOutput( Result);
    end;


{ ========
  TKnobsInput = class( TKnobsConnector)
  // Specialized connector with it's own bitmap
  protected
}

    procedure   TKnobsInput.AssignBitmap; // override;
    begin
      FBitmap := bmConIn;
      FBitmap.Transparent := True;
    end;


{ ========
  TKnobsOutput = class( TKnobsConnector)
  // Specialized connector with it's own bitmap
}

    procedure   TKnobsOutput.AssignBitmap; // override;
    begin
      FBitmap := bmConOut;
      FBitmap.Transparent := False;
    end;


{ ========
  TKnobsTextLabel = class( TCustomLabel)
  // Just a small print TLabel
  protected
    property Align;
    property Alignment;
    property Anchors;
    property AutoSize;
    property BiDiMode;
    property Caption;
    property Color;
    property Constraints;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property FocusControl;
    property Font;
    property ParentBiDiMode;
    property ParentColor;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowAccelChar;
    property ShowHint;
    property Transparent;
    property Layout;
    property Visible;
    property WordWrap;
  protected
}

    procedure   TKnobsTextLabel.DoDrawText( var aRect: TRect; aFlags: Longint); // override;
    var
      Text: string;
    begin
      Text   := GetLabelText;
      aFlags := aFlags or DT_NOPREFIX;
      aFlags := DrawTextBiDiModeFlags( aFlags);
      Canvas.Font := Font;
      DrawText( Canvas.Handle, PChar( Text), Length( Text), aRect, aFlags);
    end;

//  public

    constructor TKnobsTextLabel.Create( anOwner: TComponent); // override;
    begin
      inherited;

      with Font
      do begin
        Name := 'Small Fonts';
        Size := 7;
      end;

      Enabled     := False;
      Transparent := True;
    end;


{ ========
  TKnobsLabelEditor = class( TEdit)
  // An editor for TEditLabel
  private
    FLabel : TKnobsEditLabel;
  private
}

    procedure   TKnobsLabelEditor.WMKillFocus( var aMessage: TWMKillFocus ); //  message WM_KILLFOCUS;
    begin
      inherited;

      Visible := False;
    end;

    procedure   TKnobsLabelEditor.Do_Exit( aSender: TObject);
    begin
      Visible := False;
    end;

    procedure   TKnobsLabelEditor.Do_KeyDown( aSender: TObject; var aKey: Word; aShift: TShiftState);
    begin
      if aShift = []
      then
        case aKey Of

          VK_RETURN:
            begin
              FLabel.SetText( Text);
              Visible := False;
              aKey    := 0;
            end;

          VK_ESCAPE :
            begin
              Visible := False;
              aKey    := 0;
            end;

        end;
    end;

//  protected

    procedure   TKnobsLabelEditor.SetLabel( aLabel: TKnobsEditLabel);
    begin
      FLabel  := aLabel;
      Parent  := aLabel.Parent;
      Text    := aLabel.Caption;
      Visible := True;
      SetFocus;
    end;

//  public

    constructor TKnobsLabelEditor.Create( anOwner: TComponent); // override;
    begin
      if Assigned( GLabelEditor)
      then Exit;

      inherited;

      ControlStyle := ControlStyle + [ csOpaque];
      Visible      := False;
      Left         := 1;
      Top          := 1;
      OnExit       := Do_Exit;
      OnKeyDown    := Do_KeyDown;
      GLabelEditor := Self;
    end;

    destructor  TKnobsLabelEditor.Destroy; // override;
    begin
      if Assigned( Owner)
      then Owner.RemoveComponent( Self);
      inherited;
    end;


{ ========
  TKnobsEditLabel = class( TKnobsTextLabel)
  // a label with a popup editor so it's caption can be edited.
  private
    FOnChanged : TKnobsOnEditLabelChanged;
  public
    property    Module: TKnobsCustomModule read GetModule;
  published
    property    OnChanged: TKnobsOnEditLabelChanged read FOnChanged write FOnChanged;
  private
}

    function    TKnobsEditLabel.GetModule: TKnobsCustomModule;
    begin
      if Parent is TKnobsCustomModule
      then Result := TKnobsCustomModule( Parent)
      else Result := nil;
    end;

//  protected

    procedure   TKnobsEditLabel.SetText( const aValue: string);
    begin
      Caption := aValue;

      if Assigned( Module)
      then Module.Title := aValue;

      if Assigned( FOnChanged)
      then FOnChanged( Self, aValue);
    end;

    procedure   TKnobsEditLabel.MouseDown( Button: TMouseButton; Shift: TShiftState; X, Y: Integer); // override;

    // We only need the double click here to start a label editor, all other
    // events should be handled by the containing module.

    var
      aModule : TKnobsCustomModule;
    begin
      if not ( csDesigning in ComponentState)
      then begin
        if ( ssDouble in Shift) and ( Button = mbLeft)
        then begin
          // Handle the double click
          NeedsLabelEditor;
          GLabelEditor.SetLabel( Self);
        end
        else begin
          // Pass it on
          aModule := Module;

          if Assigned( aModule)
          then begin
            aModule.MouseCapture := True;
            aModule.MouseDown( Button, Shift, X + Left, Y + Top);
          end;
        end;
      end;
    end;

//  public

    constructor TKnobsEditLabel.Create( anOwner: TComponent); // override;
    begin
      inherited;

      AutoSize  := False;
      Width     := 75;
      Enabled   := True;
      ShowHint  := True;
      if Owner is TKnobsCustomModule
      then Hint := TKnobsCustomModule( Owner).Title;
    end;


{ ========
  TKnobsBox = class( TBevel)
  // A disablable bevel
  published
    property    Enabled default False;
  public
}

    constructor TKnobsBox.Create( anOwner: TComponent); // override;
    begin
      inherited;

      Enabled := False;
    end;


{
  TKnobsWire = class
  // Wire, visual aspect of a connection
  private
    FSignalType   : TSignalType;
    FSource       : TKnobsConnector;
    FDestination  : TKnobsConnector;
    FHighlight    : Boolean;
    FBezierParams : TKnobsBezierParams;
  public
    property    Source          : TKnobsConnector    read FSource;
    property    Destination     : TKnobsConnector    read FDestination;
    property    SourceName      : string        read GetSourceName;
    property    DestinationName : string        read GetDestinationName;
    property    BezierParams    : TKnobsBezierParams read FBezierParams;
    property    FullName        : string        read GetFullName;
  private
}

    function    TKnobsWire.GetSourceName: string;
    begin
      Result := ConnectorName( Source.FindOutput); // Trace to ultimate source (an output)
    end;

    function    TKnobsWire.GetDestinationName: string;
    begin
      Result := ConnectorName( Destination); // Use factual destination
    end;

    function    TKnobsWire.GetFullName: string;
    begin
      Result := Format( '%s -> %s', [ ConnectorName( Source), ConnectorName( Destination)], AppLocale);
    end;

//  public

    constructor TKnobsWire.Create( aSource, aDestination: TKnobsConnector);
    var
      anOutput: TKnobsConnector;
    begin
      inherited Create;

      FSource      := aSource;
      FDestination := aDestination;
      anOutput     := aSource.FindOutput;
      if Assigned( anOutput)
      then FSignalType := anOutput.Signaltype
      else FSignalType := stUnknown;

      Wiggle;
    end;

    procedure   TKnobsWire.Wiggle;
    begin
      with FBezierParams
      do begin
        Angle1    := 15 + Random( 21);
        Strength1 := 20 + Random( 21);
        Angle2    := 20 + Random( 21);
        Strength2 := 20 + Random( 21);
      end;
    end;

    procedure   TKnobsWire.Dump( var aFile: TextFile; anInd: Integer);
    begin
      WriteLnInd( aFile, anInd, Format( 'wire : %s', [ WireName( Source, Destination)], AppLocale));
    end;

    procedure   TKnobsWire.FixWire;
    // When the wire is input to input fix it such such that the Source connector is closest
    // to the originating output connector. When there is no output in the chain  just
    // do something arbitrary - which must then be fixed later once an output gets to
    // be connected.
    begin
      if ( Source is TKnobsInput) and ( Destination is TKnobsInput)
      then begin
        if Source.DistanceToOutput > Destination.DistanceToOutput
        then SwapConnectors;
      end;
    end;

    procedure   TKnobsWire.SwapConnectors;
    var
      aTmp: TKnobsConnector;
    begin
      aTmp         := FSource;
      FSource      := FDestination;
      FDestination := aTmp;
    end;


{ ========
  TKnobsWireOverlay = class( TCustomControl)
  //
  // Responsibilty to maintain connections is assumed by the TConnectors
  // themselves. Wires are maintained here, but it's a cache really to ease
  // wire painting. The wires are maintained by the connectors theselves that
  // send their info to us through the WirePanel (by means of UM_CONNEXTIONxx
  // and UM_HIGHLIGHT messages). The WirePanel imforms us by calling the
  // AddWire, DelWire and Highlight methodes.
  //
  // All Wires are owned by FWires, so our destructor will eventually destroy
  // all dangling wires. Nevertheless, TConnector still must, on destruction,
  // clean up the wires it is responsible for .. it wouldn't look good other
  // wise.
  //
  private
    FWires            : TObjectList;
    FCurvedLines      : Boolean;
    FWireThickness    : Integer;
    FdefaultWireColor : TColor;
    FBlockCount       : Integer;
    FPaintCount       : Integer;
    FPaintClocks      : Int64;
  public
    property    Wire       [ anIndex: Integer]: TKnobsWire      read GetWire;
    property    Source     [ anIndex: Integer]: TKnobsConnector read GetSource;
    property    Destination[ anIndex: Integer]: TKnobsConnector read GetDestination;
  public
    property    CurvedLines      : Boolean read FCurvedLines      write SetCurvedLines    default True;
    property    WireThickness    : Integer read FWireThickness    write SetWireThickness  nodefault;
    property    defaultWireColor : TColor  read FdefaultWireColor write FdefaultWireColor default clBlue;
    property    BlockCount       : Integer read FBlockCount;
    property    PaintCount       : Integer read FPaintCount;
    property    PaintClocks      : Int64   read FPaintClocks;
  published
    property    Left   stored False;
    property    Top    stored False;
    property    Width  stored False;
    property    Height stored False;
  private
}

    procedure   TKnobsWireOverlay.WMNCHitTest( var aMessage: TWMNCHitTest); // message WM_NCHITTESTq;
    begin
      if csDesigning in ComponentState
      then inherited
      else aMessage.Result := HTTRANSPARENT;
    end;

    function    TKnobsWireOverlay.GetWire( anIndex: Integer): TKnobsWire;
    begin
      Result := TKnobsWire( FWires[ anIndex]);
    end;

    function    TKnobsWireOverlay.GetSource( anIndex: Integer): TKnobsConnector;
    begin
      Result := Wire[ anIndex].FSource;
    end;

    function    TKnobsWireOverlay.GetDestination( anIndex: Integer): TKnobsConnector;
    begin
      Result := Wire[ anIndex].FDestination;
    end;

    procedure   TKnobsWireOverlay.SetCurvedLines( aValue: Boolean);
    begin
      if aValue <> FCurvedLines
      then begin
        FCurvedLines := aValue;
        InvalidateWires;
      end;
    end;

    procedure   TKnobsWireOverlay.SetWireThickNess( aValue: Integer);
    begin
      if aValue <> FWireThickness
      then begin
        FWireThickness := aValue;
        InvalidateWires;
      end;
    end;

//   protected

    procedure   TKnobsWireOverlay.Paint; // override;
    var
      i : Integer;
      O : TKnobsConnector;
    begin
      if FBlockCount = 0
      then begin
        with Canvas
        do begin
          Pen.Width := FWireThickness + 1;

          for i := 0 to WireCount - 1
          do begin
            with Wire[ i]
            do begin
              O := FSource.FindOutput;
              if not Assigned( O)
              then Pen.Color := CL_NO_OUTPUT
              else if FHighlight
              then Pen.Color := CL_HIGHLIGHT
              else Pen.Color := O.WireColor;
            end;
            PaintWire( Handle, Wire[ i]);
          end;
        end;
      end;
    end;

    procedure   TKnobsWireOverlay.PaintWire( aDC: HDC; aWire: TKnobsWire);
    var
      Points : TKnobsBezierPoints;
    begin
      with aWire
      do begin
        if FSource.HasParent and FDestination.HasParent and ( WireThickness > 0)
        then begin
          with FSource
          do Points[ 0] := ClientToScreen( CenterPoint);

          Points[ 0] := ScreenToClient( Points[ 0]);

          with FDestination
          do Points[ 3] := ClientToScreen( CenterPoint);

          Points[ 3] := ScreenToClient( Points[ 3]);

          if CurvedLines
          then begin
            CalcBezierCurve( Points, aWire.BezierParams);
            PolyBezier( aDC, Points, 4);
          end
          else begin
            with Points[ 0]
            do MoveToEx( aDC, X, Y, nil);

            with Points[ 3]
            do LineTo  ( aDC, X, Y     );
          end;
        end;
      end;
    end;

    function    TKnobsWireOverlay.FindWire( aSource, aDestination: TKnobsConnector): Integer;
    var
      i : Integer;
    begin
      Result := -1;
      for i := 0 to WireCount - 1
      do begin
        if
          (( Wire[ i].FSource = aSource     ) and ( Wire[ i].FDestination = aDestination)) or
          (( Wire[ i].FSource = aDestination) and ( Wire[ i].FDestination = aSource     ))
        then begin
          Result := i;
          Break;
        end;
      end;
    end;

    procedure   TKnobsWireOverlay.AddWire( aSource, aDestination: TKnobsConnector);
    var
      anIndex      : Integer;
      aWire        : TKnobsWire;
      aDistanceSrc : Integer;
      aDistanceDst : Integer;
    begin
      anIndex := FindWire( aSource, aDestination);

      if anIndex < 0
      then begin
        if ( aSource is TKnobsOutput) and ( aDestination is TKnobsOutput)
        then raise EKnobsShortCircuit.Create( 'output to output connection detected ... should not be there')
        else if ( aSource is TKnobsInput) and ( aDestination is TKnobsOutput)
        then aWire := TKnobsWire.Create( aDestination, aSource)
        else if ( aSource is TKnobsOutput) and ( aDestination is TKnobsInput)
        then aWire := TKnobsWire.Create( aSource, aDestination)
        else begin
          // For input to input connections the source should be closest to originating output ..
          // when there is no output the wire might still be added with the connectors in the
          // wrong order, this must be fixed before patch compilation with a call to
          // FixAllWires - we will not do that here as the order is irrelevant for the editor
          // itself. So just add the wire here in an arbitrary way.
          aDistanceSrc := aSource     .DistanceToOutput;
          aDistanceDst := aDestination.DistanceToOutput;

          if aDistanceSrc > aDistanceDst
          then aWire := TKnobsWire.Create( aDestination, aSource)
          else aWire := TKnobsWire.Create( aSource, aDestination);

        end;
        FWires.Add( aWire);
        InvalidateWires;
      end;
    end;

    procedure   TKnobsWireOverlay.DelWire( aSource, aDestination: TKnobsConnector);
    var
      anIndex : Integer;
    begin
      anIndex := FindWire( aSource, aDestination);

      if anIndex >= 0
      then begin
        FWires.Delete( anIndex);
        InvalidateWires;
      end;
    end;

    procedure   TKnobsWireOverlay.Connect( aSource, aDestination: TKnobsConnector);
    begin
      if Assigned( aSource)
      then aSource.ConnectTo( aDestination);
    end;

    procedure   TKnobsWireOverlay.HandleHighlight;
    var
      i : Integer;
    begin
      for i := 0 to WireCount - 1
      do begin
        with Wire[ i]
        do FHighlight := FSource.HighLight or FDestination.HighLight;
      end;

      Invalidate;
    end;

    procedure   TKnobsWireOverlay.HandleColorChange( aSource: TKnobsConnector; aColor: TColor);
    begin
      InvalidateWires;
    end;

    function    TKnobsWireOverlay.WireCount: Integer;
    begin
      if Assigned( FWires)
      then Result := FWires.Count
      else Result := 0;
    end;

    procedure   TKnobsWireOverlay.WiggleWires;
    var
      i : Integer;
    begin
      if CurvedLines and Assigned( FWires)
      then begin
        for i := 0 to WireCount - 1
        do Wire[ i].Wiggle;
        InvalidateWires;
      end;
    end;

//   public

    constructor TKnobsWireOverlay.Create( anOwner: TComponent); // override;
    begin
      inherited;

      FBlockCount        := 0;
      FWires             := TObjectList.Create;
      Align              := alClient;
      FCurvedLines       := True;
      FWireThickness     := 1;
      FdefaultWireColor  := clBlue;
      FWires.OwnsObjects := True;
    end;

    destructor  TKnobsWireOverlay.Destroy; // override;
    begin
      FreeAndnil( FWires);

      inherited;
    end;

    procedure   TKnobsWireOverlay.InvalidateWires;

      function CreateRgn( aDC: HDC; aWire: TKnobsWire): HRGN;
      begin
        BeginPath ( aDC);
        PaintWire ( aDC, aWire);
        EndPath   ( aDC);
        WidenPath ( aDC);
        Result := PathToRegion( aDC);
      end;

    var
      DC       : HDC;
      i        : Integer;
      HRegion1 : HRGN;
      HRegion2 : HRGN;
      SavedPen : HPEN;
      Clocks   : Int64;
    begin
      if HasParent and Parent.HasParent and ( FBlockCount = 0)
      then begin
        Inc( FPaintCount);
        Clocks := GetCpuClockCycleCount;

        with Canvas
        do begin
          Pen.Width := FWireThickness;
          DC        := CreateCompatibleDC( Handle);
          SavedPen  := SelectObject( DC, Pen.Handle);
        end;

        try
          if ( WireCount > 0) and ( WireThickness > 0)
          then begin
            HRegion1 := CreateRgn( DC, Wire[ 0]);

            for i := 1 to WireCount - 1
            do begin
              HRegion2 := CreateRgn( DC, Wire[ i]);
              CombineRgn( HRegion2, HRegion2, HRegion1, RGN_OR);
              DeleteObject( HRegion1);
              HRegion1 := HRegion2;
            end;
          end
          else HRegion1 := CreateRectRgn( 0, 0, 1, 1);

          SetWindowRgn( Handle, HRegion1, True);
          BringToFront; // see other __1 remarks - it will paint behind .. otherwise
        finally
          SelectObject( DC, SavedPen);
          DeleteDC( DC);
        end;

        Invalidate; // And tell windows we want a repaint.

        FPaintClocks := GetCpuClockCycleCount - Clocks;
      end;
    end;

    function    TKnobsWireOverlay.Disconnect( aConnector: TKnobsConnector): Boolean;
    var
      i : Integer;
    begin
      Result := False;
      for i := WireCount - 1 downto 0
      do begin
        with Wire[ i]
        do begin
          if ( FSource = aConnector) or ( FDestination = aConnector)
          then begin
            FWires.Delete( i);
            Result := True;
          end;
        end;
      end;
    end;

    procedure   TKnobsWireOverlay.Dump( var aFile: TextFile; anInd: Integer);
    var
      i : Integer;
    begin
      for i := 0 to FWires.Count - 1
      do TKnobsWire( FWires[ i]).Dump( aFile, anInd);
    end;

    procedure   TKnobsWireOverlay.Log( const aMsg: string);
    begin
      if Parent is TKnobsWirePanel
      then TKnobsWirePanel( Parent).Log( Format( 'wire overlay %s( %s)', [ Name, aMsg], AppLocale));
    end;

    procedure   TKnobsWireOverlay.BlockScreenUpdates;
    begin
      Inc( FBlockCount);
    end;

    procedure   TKnobsWireOverlay.UnBlockScreenUpdates;
    begin
      if FBlockCount > 0
      then begin
        Dec( FBlockCount);
        if FBlockCount = 0
        then InvalidateWires;
      end;
    end;

    procedure   TKnobsWireOverlay.FixAllWires;
    // Fix all wires that are input to input such that the Source connector is closest
    // to the originating output connector. When there is no output in the chain  just
    // do something arbitrary - which must then be fixed later once an output gets to
    // be connected. Must be called before a patch is translated into a synth patch.
    var
      i : Integer;
    begin
      for i := 0 to WireCount - 1
      do Wire[ i].FixWire;
    end;



{ =======
  TModuleList = class( TList)
  private
    FonLog : TKnobsOnLog;
  public
    property    Module  [ anIndex: Integer]: TKnobsCustomModule read GetModule;                        default;
    property    Selected[ anIndex: Integer]: Boolean            read GetSelected     write SetSelected;
    property    AnySelected                : Boolean            read GetAnySelected;
    property    NoneSelected               : Boolean            read GetNoneSelected;
    property    OnLog                      : TKnobsOnLog        read FOnLog          write FOnLog;
  private
}

    function    TKnobsModuleList.GetModule( anIndex: Integer): TKnobsCustomModule;
    begin
      Result := TKnobsCustomModule( Items[ anIndex]);
    end;

    function    TKnobsModuleList.GetSelected( anIndex: Integer): Boolean;
    begin
      Result := Module[ anIndex].Selected;
    end;

    procedure   TKnobsModuleList.SetSelected( anIndex: Integer; aValue: Boolean);
    begin
      Module[ anIndex].Selected := aValue;
    end;

    function    TKnobsModuleList.GetAnySelected: Boolean;
    var
      i : Integer;
    begin
      Result := False;
      for i := 0 to Count - 1
      do begin
        if Selected[ i]
        then begin
          Result := True;
          Break;
        end;
      end;
    end;

    function    TKnobsModuleList.GetNoneSelected: Boolean;
    var
      i : Integer;
    begin
      Result := True;
      for i := 0 to Count - 1
      do begin
        if Selected[ i]
        then begin
          Result := False;
          Break;
        end;
      end;
    end;

//  public

    procedure   TKnobsModuleList.FreeModule( anIndex: Integer);
    var
      aModule : TKnobsCustomModule;
    begin
      aModule := Module[ anIndex];
      Delete( anIndex);
      aModule.Free;
    end;

    function    TKnobsModuleList.FreeModules: Boolean;
    var
      i : Integer;
    begin
      Result := False;
      for i := Count - 1 downto 0
      do begin
        FreeModule( i);
        Result := True;
      end;
    end;

    function    TKnobsModuleList.FreeSelectedModules: Boolean;
    var
      i : Integer;
    begin
      Result := False;
      for i := Count - 1 downto 0
      do begin
        if Selected[ i]
        then begin
          FreeModule( i);
          Result := True;
        end;
      end;
    end;

    procedure   TKnobsModuleList.SelectAll;
    var
      i : Integer;
    begin
      for i := Count - 1 downto 0
      do Selected[ i] := True;
    end;

    procedure   TKnobsModuleList.UnselectAll;
    var
      i : Integer;
    begin
      for i := Count - 1 downto 0
      do Selected[ i] := False;
    end;

    procedure   TKnobsModuleList.SelectUnique( aModule: TKnobsCustomModule);
    var
      i : Integer;
    begin
      for i := 0 to Count - 1
      do Module[ i].Selected := Module[ i] = aModule;
    end;

    procedure   TKnobsModuleList.FixSelection( aModule: TKnobsCustomModule);
    // When no modules are selected select aModule uniquely
    begin
      if Assigned( aModule) and NoneSelected
      then SelectUnique( aModule);
    end;

    procedure   TKnobsModuleList.SetSelectedModuleColor( aValue: TColor);
    var
      i : Integer;
    begin
      for i := 0 to Count - 1
      do begin
        if Selected[ i]
        then Module[ i].Color := aValue;
      end;
    end;

    procedure   TKnobsModuleList.InvertSelection;
    var
      i : Integer;
    begin
      for i := Count - 1 downto 0
      do Selected[ i] := not Selected[ i];
    end;

    procedure   TKnobsModuleList.InvalidateDisplays;
    var
      i : Integer;
    begin
      for i := 0 to Count - 1
      do Module[ i].InvalidateDisplays;
    end;


//  public

    //* Some helpers for selector population and component registration

    function    TKnobsModuleList.FindType( aModuleType: TKnobsModuleType): TKnobsCustomModule;
    //* Finds the first enabled module of the ModuleType specified
    var
      i : Integer;
    begin
      Result := nil;

      for i := 0 to Count - 1
      do begin
        if ( Module[ i].ModuleType = aModuleType) and ( Module[ i]).Enabled
        then begin
          Result := Module[ i];
          Break;
        end;
      end;
    end;

    function TabIndexSorter( Item1, Item2: Pointer): Integer;

      function CompareModules( m1, m2: TKnobsCustomModule): Integer;
      begin
        if m1.TabOrder > m2.TabOrder
        then Result := 1
        else if m1.TabOrder < m2.TabOrder
        then Result := -1
        else Result := 0;
      end;

    var
      m1 : TKnobsCustomModule;
      m2 : TKnobsCustomModule;
      t1 : TTabSheet;
      t2 : TTabSheet;
    begin
      m1 := TKnobsCustomModule( Item1);
      m2 := TKnobsCustomModule( Item2);
      t1 := nil;
      if Assigned( m1.Parent) and ( m1.Parent is TTabSheet)
      then t1 := TTabSheet( m1.Parent);
      t2 := nil;
      if Assigned( m2.Parent) and ( m2.Parent is TTabSheet)
      then t2 := TTabSheet( m2.Parent);
      if not Assigned( t1) or not Assigned( t2)
      then Result := CompareModules( m1, m2)
      else begin
        if t1.TabIndex > t2.TabIndex
        then Result := 1
        else if t1.TabIndex < t2.TabIndex
        then Result := -1
        else Result := CompareModules( m1, m2);
      end;
    end;

    procedure   TKnobsModuleList.PopulateSelector( aModuleSelector: TKnobsModuleSelector);
    //* Populate a module selector using all the modules contained in the list
    var
      i : Integer;
    begin
      Sort( TabIndexSorter);
      if Assigned( aModuleSelector)
      then begin
        for i := 0 to Count - 1
        do begin
          if Module[ i].Enabled
          then begin
            with Module[ i]
            do aModuleSelector.RegisterModuleType( PageName, ModuleType, Title);
          end;
        end;
      end;
    end;

    function    TKnobsModuleList.CreateModule( aWirePanel: TKnobsWirePanel; aModuleType: TKnobsModuleType; MustDrag: Boolean): TKnobsCustomModule;
    //* Factory method, create a clone from the specified module type when an
    //* enabled module of that type can be found in the list.
    //* This one is called from the module selector only, it should set
    //* the one and only module into drag mode. See __2 for other cases
    //* (also in KnobParser)
    //* See Clone method .. that one fixes it
    var
      aModule : TKnobsCustomModule;
    begin
      Result  := nil;
      aModule := FindType( aModuleType);

      if Assigned( aModule)
      then begin
        Result := aModule.Clone( aWirepanel, aWirePanel, Point( 0, 0), MustDrag);
        Result.Color               := aWirePanel.ModuleColor;
        Result.Opacity             := aWirePanel.ModuleOpacity;
        Result.WheelSupportOnKnobs := aWirePanel.WheelSupportOnKnobs;
        Result.WheelSensitivity    := aWirePanel.WheelSensitivity;
      end;
    end;

    procedure   TKnobsModuleList.RegisterComponents( doRegister: Boolean);
    //* Register or unregister all module classes and all component's classes
    //* used on all modules with Delphi so that they can be used with the
    //* component streaming mechanism.
    var
      aList : TClassList;
      i     : Integer;
      j     : Integer;
      p     : Integer;
    begin
      aList := TClassList.Create;

      try
        // Build a list having each class only once
        for i := 0 to Count - 1
        do begin
          with Module[ i]
          do begin
            // Register the module class itself
            p := aList.IndexOf( ClassType);
            if p < 0
            then aList.Add( ClassType);

            // Register component classes for all components direcly parented by
            // the module.
            for j := 0 to ControlCount - 1
            do begin
              if Controls[ j] is TPersistent
              then begin
                p := aList.IndexOf( Controls[ j].ClassType);
                if p < 0
                then aList.Add( Controls[ j].ClassType);
              end;
            end;
          end;
        end;

        // Register the classes found
        with aList
        do begin
          for i := 0 to Count - 1
          do begin
            if doRegister
            then Registerclass  ( TPersistentClass( Items[ i]))
            else UnRegisterClass( TPersistentClass( Items[ i]));
          end;
        end;
      finally
        aList.Free;
      end;
    end;

    procedure   TKnobsModuleList.Dump( var aFile: TextFile; anInd: Integer);
    var
      i : Integer;
    begin
      WriteLnInd( afile, anInd, '.');
      for i := 0 to Count - 1
      do Module[ i].Dump( aFile, anInd);
      WriteLnInd( afile, anInd, '.');
    end;

    procedure   TKnobsModuleList.Log( const aMsg: string);
    begin
      if Assigned( FOnLog)
      then FOnLog( Self, Format( 'TKnobsModuleList( %s)', [ aMsg], AppLocale));
    end;


{ ========
  TKnobsHistoryStack = class( TStringList)
  private
    FMaxHistory : Integer;
  public
    property    MaxHistory : Integer read FMaxHistory write SetMaxHistory;
  private
}

    procedure   TKnobsHistoryStack.SetMaxHistory( aValue: Integer);
    begin
      if aValue <> FMaxHistory
      then begin
        FMaxHistory := aValue;
        while Count > MaxHistory
        do Delete( 0);
      end;
    end;

//  protected

    procedure   TKnobsHistoryStack.PushString( const S: string);
    begin
      if ( Count = 0) or ( S <> Strings[ Count - 1])
      then begin
        Add( S);
        while Count > MaxHistory
        do Delete( 0);
      end;
    end;

    function    TKnobsHistoryStack.PopString: string;
    begin
      Result := '';
      if Count > 0
      then begin
        Result := Strings[ Count - 1];
        Delete( Count - 1);
      end;
    end;

//  public

    procedure   TKnobsHistoryStack.Push( const aWirePanel: TKnobsWirePanel);
    begin
      if Assigned( aWirePanel)
      then PushString( aWirePanel.PatchWriter.WriteString( aWirePanel, '', wmAll));
    end;

    procedure   TKnobsHistoryStack.Pop( const aWirePanel: TKnobsWirePanel);
    begin
      if Count > 0
      then aWirePanel.PatchReader.ReadString( PopString, aWirePanel, Point( 0, 0), rmReplace, '');
    end;


{ ========
  TKnobsUndoRedoHistory = class( TObject)
  private
    FOnHistoryChange : TKnobsOnHistoryChange;
    FMaxHistory      : Integer;
    FLockCount       : Integer;
    FUndos           : TKnobsHistoryStack;
    FRedos           : TKnobsHistoryStack;
  public
    property    UndoCount       : Integer          read GetUndoCount;
    property    RedoCount       : Integer          read GetRedoCount;
    property    MaxHistory      : Integer          read FMaxHistory  write SetmaxHistory;
  public
    property    OnHistoryChange : TKnobsOnHistoryChange read FOnHistoryChange write FOnHistoryChange;
  private
}

    function    TKnobsUndoRedoHistory.GetUndoCount: Integer;
    begin
      Result := FUndos.Count;
    end;

    function    TKnobsUndoRedoHistory.GetRedoCount: Integer;
    begin
      Result := FRedos.Count;
    end;

    procedure   TKnobsUndoRedoHistory.SetmaxHistory( aValue: Integer);
    begin
      if aValue <> FMaxHistory
      then begin
        FMaxHistory := aValue;
        FUndos.MaxHistory := aValue;
        FRedos.MaxHistory := aValue;
      end;
    end;

//  public

    constructor TKnobsUndoRedoHistory.Create;
    begin
      inherited Create;
      FMaxHistory       := 1;
      FUndos            := TKnobsHistoryStack.Create;
      FRedos            := TKnobsHistoryStack.Create;
      FUndos.MaxHistory := 1;
      FRedos.MaxHistory := 1;
    end;

    destructor  TKnobsUndoRedoHistory.Destroy; // override;
    begin
      FreeAndNil( FUndos);
      FreeAndNil( FRedos);
      inherited;
    end;

    procedure   TKnobsUndoRedoHistory.HistoryChanged;
    begin
      if Assigned( FOnHistoryChange)
      then FOnHistoryChange( Self, UndoCount, RedoCount);
    end;

    procedure   TKnobsUndoRedoHistory.SaveState( const aWirePanel: TKnobsWirePanel);
    begin
      if FLockCount = 0
      then begin
        FRedos.Clear;
        FUndos.Push( aWirePanel);
        HistoryChanged;
      end;
    end;

    procedure   TKnobsUndoRedoHistory.Undo( const aWirePanel: TKnobsWirePanel);
    begin
      if UndoCount > 0
      then begin
        FRedos.Push( aWirePanel);
        try
          Lock;                    // Prevent undo from changing history
          FUndos.Pop( aWirePanel);
        finally
          UnLock;
        end;
        HistoryChanged;
      end;
    end;

    procedure   TKnobsUndoRedoHistory.Redo( const aWirePanel: TKnobsWirePanel);
    begin
      if RedoCount > 0
      then begin
        FUndos.Push( aWirePanel);
        try
          Lock;                    // Prevent redo from changing history
          FRedos.Pop( aWirePanel);
        finally
          UnLock;
        end;
        HistoryChanged;
      end;
    end;

    procedure   TKnobsUndoRedoHistory.Lock;
    begin
      Inc( FLockCount);
    end;

    procedure   TKnobsUndoRedoHistory.UnLock;
    begin
      if FLockCount > 0
      then Dec( FLockCount);
    end;

    procedure   TKnobsUndoRedoHistory.Clear;
    begin
      FUndos.Clear;
      FRedos.Clear;
      HistoryChanged;
    end;


{ ========
  TKnobsWirePanel = class( TScrollingWinControl)
  // Contains the modules and the wire overlay - scrollable.
  private
    FOffsetLeft           : Integer;
    FOffsetTop            : Integer;
    FTitle                : string;
    FVersion              : Integer;
    FMouseDown            : Boolean;
    FWireOverlay          : TKnobsWireOverlay;
    FModules              : TKnobsModuleList;
    FFilename             : string;
    FSnapX                : Integer;
    FSnapY                : Integer;
    FControlMode          : TDistanceMode;
    FWheelSupportOnKnobs  : Boolean;
    FWheelSensitivity     : Integer;
    FModuleOpacity        : Byte;
    FModuleColor          : TColor;
    FSnapActive           : Boolean;
    FPatchReader          : IKnobsPatchReader;
    FPatchWriter          : IKnobsPatchWriter;
    FOnLog                : TKnobsOnLog;
    FOnHistoryChange      : TKnobsOnHistoryChange;
    FHistory              : TKnobsUndoRedoHistory;
    FOnValueChanged       : TKnobsOnValueChange;
    FOnTextChanged        : TKnobsOnTextChanged;
    FOnPatchChanged       : TKnobsOnPatchChanged;
    FOnWiresChanged       : TKnobsOnWiresChanged;
    FOnShowPopup          : TOnShowPopup;
    FOnUnFocus            : TOnUnFocus;
  public
    property    PatchReader  : IKnobsPatchReader   read FPatchReader   write FPatchReader;
    property    PatchWriter  : IKnobsPatchWriter   read FPatchWriter   write FPatchWriter;
  public
    property    Module     [ anIndex: Integer]: TKnobsCustomModule  read GetModule;
    property    Wire       [ anIndex: Integer]: TKnobsWire          read GetWire;
    property    Source     [ anIndex: Integer]: TKnobsConnector     read GetSource;
    property    Destination[ anIndex: Integer]: TKnobsConnector     read GetDestination;
  public
    property    OnHistoryChange     : TKnobsOnHistoryChange  read FOnHistoryChange     write FOnHistoryChange;
    property    OnLog               : TKnobsOnLog            read FOnLog               write FonLog;
    property    Filename            : string                 read FFilename            write FFilename;
    property    ModuleCount         : Integer                read GetModuleCount;
    property    WireCount           : Integer                read GetWireCount;
    property    BlockCount          : Integer                read GetBlockCount;
    property    PaintCount          : Integer                read GetPaintCount;
    property    PaintClocks         : Int64                  read GetPaintClocks;
    property    ScrollOffset        : TPoint                 read GetScrollOffset;
  published
    property    MaxHistory          : Integer                read GetMaxHistory        write SetMaxHistory          default 100;
    property    OnValueChanged      : TKnobsOnValueChange    read FOnValueChanged      write FOnValueChanged;
    property    OnTextChanged       : TKnobsOnTextChanged    read FOnTextChanged       write FOnTextChanged;
    property    OnPatchChanged      : TKnobsOnPatchChanged   read FOnPatchChanged      write FOnPatchChanged;
    property    OnWiresChanged      : TKnobsOnWiresChanged   read FOnWiresChanged      write FOnWiresChanged;
    property    OnShowPoup          : TOnShowPopup           read FOnShowPopup         write FOnShowPopup;
    property    OnUnFocus           : TOnUnFocus             read FOnUnFocus           write FOnUnFocus;
    property    OffsetLeft          : Integer                read FOffsetLeft          write SetOffsetLeft          default 4;
    property    OffsetTop           : Integer                read FOffsetTop           write SetOffsetTop           default 4;
    property    CurvedLines         : Boolean                read GetCurvedLines       write SetCurvedLines         default True;
    property    WireThickness       : Integer                read GetWireThickness     write SetWireThickness       default 1;
    property    defaultWireColor    : TColor                 read GetdefaultWireColor  write SetdefaultWireColor    default clBlue;
    property    Title               : string                 read FTitle               write FTitle;
    property    Version             : Integer                read FVersion             write FVersion;
    property    SnapX               : Integer                read FSnapX               write FSnapX                 default 250;
    property    SnapY               : Integer                read FSnapY               write FSnapY                 default 15;
    property    SnapActive          : Boolean                read FSnapActive          write FSnapActive            default True;
    property    ControlMode         : TDistanceMode          read FControlMode         write SetControlMode         default dmCircular;
    property    WheelSupportOnKnobs : Boolean                read FWheelSupportOnKnobs write SetWheelSupportOnKnobs;
    property    WheelSensitivity    : Integer                read FWheelSensitivity    write SetWheelSensitivity    default 50;
    property    ModuleOpacity       : Byte                   read FModuleOpacity       write SetModuleOpacity       default 223;
    property    ModuleColor         : TColor                 read FModuleColor         write SetModuleColor         default clWhite;
    property    Align;
    property    Anchors;
    property    Constraints;
    property    ParentColor;
    property    ParentShowHint;
    property    PopupMenu;
    property    ShowHint;
    property    TabOrder;
    property    TabStop;
    property    Visible;
    property    Color;
    property    Top;
    property    Left;
    property    Width;
    property    Height;
    property    HorzScrollBar;
    property    VertScrollBar;
  protected
}

    procedure   TKnobsWirePanel.UMConnectionAdd( var aMessage: TKnobsUMConnectionAdd); // message UM_CONNECTIONADD;
    begin
      with aMessage
      do FWireOverlay.AddWire( Source, Destination);
    end;

    procedure   TKnobsWirePanel.UMConnectionRemove( var aMessage: TKnobUMConnectionRemove); // message UM_CONNECTIONREMOVE;
    begin
      with aMessage
      do FWireOverlay.DelWire( Source, Destination);
    end;

    procedure   TKnobsWirePanel.UMColorChange( var aMessage: TKnobsUMColorChange); // message UM_COLORCHANGE;
    begin
      with aMessage
      do FWireOverlay.HandleColorChange( Source, Color);
    end;

    procedure   TKnobsWirePanel.UMHighlight( var aMessage: TKnobsUMHighlight); // message UM_HIGHLIGHT;
    begin
      FWireOverlay.HandleHighlight;
    end;

//  private

    function    TKnobsWirePanel.GetModule( anIndex: Integer): TKnobsCustomModule;
    begin
      Result := FModules[ anIndex];
    end;

    function    TKnobsWirePanel.GetWire( anIndex: Integer): TKnobsWire;
    begin
      if Assigned( FWireOverlay)
      then Result := FWireOverlay.Wire[ anIndex]
      else Result := nil;
    end;

    function    TKnobsWirePanel.GetSource( anIndex: Integer): TKnobsConnector;
    begin
      if Assigned( FWireOverlay)
      then Result := FWireOverlay.Source[ anIndex]
      else Result := nil;
    end;

    function    TKnobsWirePanel.GetDestination( anIndex: Integer): TKnobsConnector;
    begin
      if Assigned( FWireOverlay)
      then Result := FWireOverlay.Destination[ anIndex]
      else Result := nil;
    end;

    procedure   TKnobsWirePanel.SetOffsetLeft( aValue: Integer);
    begin
      if aValue <> FOffsetLeft
      then begin
        FOffsetLeft := aValue;
        Resnap;
        Restack( DO_RESTACK, DO_REDRAW_WIRES);
      end;
    end;

    procedure   TKnobsWirePanel.SetOffsetTop( aValue: Integer);
    begin
      if aValue <> FOffsetTop
      then begin
        FOffsetTop := aValue;
        Resnap;
        Restack( DO_RESTACK, DO_REDRAW_WIRES);
      end;
    end;

    function    TKnobsWirePanel.GetCurvedLines: Boolean;
    begin
      Result := FWireOverlay.CurvedLines;
    end;

    procedure   TKnobsWirePanel.SetCurvedLines( aValue: Boolean);
    begin
      FWireOverlay.CurvedLines := aValue;
    end;

    function    TKnobsWirePanel.GetWireThickness: Integer;
    begin
      Result := FWireOverlay.WireThickness;
    end;

    procedure   TKnobsWirePanel.SetWireThickness( aValue: Integer);
    begin
      FWireOverlay.WireThickness := aValue;
    end;

    function    TKnobsWirePanel.GetdefaultWireColor: TColor;
    begin
      Result := FWireOverlay.defaultWireColor;
    end;

    procedure   TKnobsWirePanel.SetdefaultWireColor( aValue: TColor);
    begin
      FWireOverlay.defaultWireColor := aValue;
    end;

    function    TKnobsWirePanel.GetModuleCount: Integer;
    begin
      if Assigned( FModules)
      then Result := FModules.Count
      else Result := 0;
    end;

    function    TKnobsWirePanel.GetWireCount: Integer;
    begin
      if Assigned( FWireOverlay)
      then Result := FWireOverlay.WireCount
      else Result := 0;
    end;

    function    TKnobsWirePanel.GetMaxHistory: Integer;
    begin
      Result := FHistory.MaxHistory;
    end;

    procedure   TKnobsWirePanel.SetMaxHistory( aValue: Integer);
    begin
      FHistory.MaxHistory := aValue;
    end;

    function    TKnobsWirePanel.GetBlockCount: Integer;
    begin
      if Assigned( FWireOverlay)
      then Result := FWireOverlay.BlockCount
      else Result := -1;
    end;

    function    TKnobsWirePanel.GetPaintCount: Integer;
    begin
      if Assigned( FWireOverlay)
      then Result := FWireOverlay.PaintCount
      else Result := -1;
    end;

    function    TKnobsWirePanel.GetPaintClocks: Int64;
    begin
      if Assigned( FWireOverlay)
      then Result := FWireOverlay.PaintClocks
      else Result := -1;
    end;

    function    TKnobsWirePanel.GetScrollOffset: TPoint;
    begin
      Result := Point( HorzScrollBar.Position, VertScrollBar.Position);
    end;

    procedure   TKnobsWirePanel.SetControlMode( aValue: TDistanceMode);
    begin
      if aValue <> FControlMode
      then begin
        FControlMode := aValue;
        FixControlMode;
      end;
    end;

    procedure   TKnobsWirePanel.SetModuleOpacity( aValue: Byte);
    var
      i : Integer;
    begin
      if aValue <> FModuleOpacity
      then begin
        FModuleOpacity := aValue;
        for i := 0 to ModuleCount - 1
        do Module[ i].Opacity := aValue;
      end;
    end;

    procedure   TKnobsWirePanel.SetModuleColor( aValue: TColor);
    var
      i : Integer;
    begin
      if aValue <> FModuleColor
      then begin
        FModuleColor := aValue;
        for i := 0 to ModuleCount - 1
        do Module[ i].Color := aValue;
      end;
    end;

    procedure   TKnobsWirePanel.SetWheelSupportOnKnobs( aValue: Boolean);
    begin
      if aValue <> FWheelSupportOnKnobs
      then begin
        FWheelSupportOnKnobs := aValue;
        FixControlMode;
      end;
    end;

    procedure   TKnobsWirePanel.SetWheelSensitivity( aValue: Integer);
    begin
      if aValue <> FWheelSensitivity
      then begin
        FWheelSensitivity := aValue;
        FixControlMode;
      end;
    end;

    procedure   TKnobsWirePanel.StrToConnection( const S: string);
    const
      Nums : set of AnsiChar = [ '0' .. '9'];
      Ascs : set of AnsiChar = [ '0' .. '9', 'A' .. 'Z', 'a' .. 'z', '_'];
    var
      p        : Integer;
      i        : Integer;
      SrcModNm : string;
      SrcConNm : string;
      DstModNm : string;
      DstConNm : string;
      SrcMod   : TKnobsCustomModule;
      SrcCon   : TKnobsConnector;
      DstMod   : TKnobsCustomModule;
      DstCon   : TKnobsConnector;
    begin

      p := 1;
      i := 1;

      if Length( S) > 0
      then begin

        while ( p <= Length( S)) and CharInSet( S[ p], Ascs)
        do Inc( p);
        SrcModNm := Copy( S, i, p - i);
        if ( p > Length( S)) or ( S[ p] <> '.')
        then Exit
        else Inc( p);

        i := p;
        while ( p <= Length( S)) and CharInSet( S[ p], Ascs)
        do Inc( p);
        SrcConNm := Copy( S, i, p - i);
        if ( p > Length( S)) or ( S[ p] <> ' ')
        then Exit
        else Inc( p);

        i := p;
        while ( p <= Length( S)) and CharInSet( S[ p], Ascs)
        do Inc( p);
        DstModNm := Copy( S, i, p - i);
        if ( p > Length( S)) or ( S[ p] <> '.')
        then Exit
        else Inc( p);

        i := p;
        while ( p <= Length( S)) and CharInSet( S[ p], Ascs)
        do Inc( p);
        DstConNm := Copy( S, i, p - i);

        try
          SrcMod :=        FindComponent( SrcModNm) as TKnobsCustomModule;
          SrcCon := SrcMod.FindComponent( SrcConNm) as TKnobsConnector;
          DstMod :=        FindComponent( DstModNm) as TKnobsCustomModule;
          DstCon := DstMod.FindComponent( DstConNm) as TKnobsConnector;
          Connect( SrcCon, DstCon);
        except
          on Exception
          do
            raise
              EKnobsCorruptPatch.
                Create(
                  'The patch is corrupted or incompatible:'^M +
                  'can''t make all the connections.'
                );
        end;

      end
      else
        raise
          EKnobsCorruptPatch.
            Create(
              'The patch is corrupted, incompatible or not a patch.'^M
            );
    end;

//  protected

    procedure   TKnobsWirePanel.HandleRightClick( const aSender: TControl);
    begin
      if Assigned( FOnShowPopup)
      then FOnShowPopup( Self, aSender);
    end;

    procedure   TKnobsWirePanel.CreateParams(var Params: TCreateParams); // override
    begin
      inherited CreateParams( Params);
      with Params.WindowClass
      do Style := Style and not ( CS_HREDRAW or CS_VREDRAW);
    end;

    procedure   TKnobsWirePanel.MouseDown( Button: TMouseButton; Shift: TShiftState; X, Y: Integer); // override;
    begin
      if FMouseDown
      then Exit;

      if Button = mbRight
      then HandleRightClick( Self)
      else if not FMouseDown and ( Button = mbLeft)
      then begin
        FMouseDown := True;
        GAnchor    := Point( X, Y);
        GDistance  := Point( 0, 0);
        DragStart( GAnchor  );
        DragMove ( GDistance);
      end;
    end;

    procedure   TKnobsWirePanel.MouseUp( Button: TMouseButton; Shift: TShiftState; X, Y: Integer); // override;
    begin
      if FMouseDown
      then begin
        DragMove( GDistance);
        GDistance.x := X - GAnchor.x;
        GDistance.y := Y - GAnchor.y;
        FMouseDown := False;
        DragEnd( GDistance);
      end;
    end;

    procedure   TKnobsWirePanel.MouseMove( Shift: TShiftState; X, Y: Integer); // override;
    begin
      if FMouseDown
      then begin
        DragMove( GDistance);
        GDistance.x := X - GAnchor.x;
        GDistance.y := Y - GAnchor.y;
        DragMove( GDistance);
      end
    end;

    procedure   TKnobsWirePanel.DragStart( aPoint: TPoint); // virtual;
    begin
      // Nothing
    end;

    procedure   TKnobsWirePanel.DragEnd( aPoint: TPoint); // virtual;
    begin
      SelectModulesInRect(
        Rect(
          GAnchor.x,
          GAnchor.y,
          GAnchor.x + aPoint.x,
          GAnchor.y + aPoint.y
        )
      );
    end;

    procedure   TKnobsWirePanel.DragMove( aPoint: TPoint); // virtual;
    var
      aDC   : HDC;
      aRect : TRect;
    begin
      aDC := GetDC( 0);

      with ClientOrigin
      do IntersectClipRect( aDC, X, Y, X + ClientWidth, Y + ClientHeight);

      try
        with aRect
        do begin
          Left   := Self.ClientOrigin.x + GAnchor.x;
          Top    := Self.ClientOrigin.y + GAnchor.y;
          Right  := Left + aPoint.x;
          Bottom := Top  + aPoint.y;
          if Left > Right
          then Exchange( Left, Right);
          if Top > Bottom
          then Exchange( Bottom, Top);
        end;

        DrawFocusRect( aDC, aRect);
      finally
        ReleaseDC( 0, aDC);
      end;
    end;

    procedure   TKnobsWirePanel.SelectModulesInRect( aRect: TRect);
    var
      i : Integer;
    begin
      with aRect
      do begin
        if Left > Right
        then Exchange( Left, Right);

        if Top > Bottom
        then Exchange( Bottom, Top);
      end;

      for i := 0 to ControlCount - 1
      do begin
        if Controls[ i] is TKnobsCustomModule
        then begin
          with TKnobsCustomModule( Controls[ i])
          do Selected := RectanglesOverlap( aRect, Rect( Left, Top, Left + Width, Top + Height));
        end;
      end;
    end;

    function    TKnobsWirePanel.IndexOfModule( aModule: TKnobsCustomModule): Integer;
    begin
      Result := FModules.IndexOf( aModule);
    end;

    procedure   TKnobsWirePanel.Notification( aComponent: TComponent; anOperation: TOperation); // override;
    var
      i : Integer;
    begin
      inherited;

      if aComponent is TKnobsCustomModule
      then begin
        if anOperation = opInsert
        then begin
          i := IndexOfModule( TKnobsCustomModule( aComponent));

          if i < 0
          then FModules.Add( aComponent);
        end
        else if anOperation = opRemove
        then begin
          i := IndexOfModule( TKnobsCustomModule( aComponent));

          if i >= 0
          then begin
            with FModules
            do begin
              Delete( i);
              Pack;
            end;
          end;
        end;
      End
    end;

    procedure   TKnobsWirePanel.Connect( aSource, aDestination: TKnobsConnector);
    begin
      if Assigned( FWireOverlay)
      then FWireOverlay.Connect( aSource, aDestination);
    end;

    procedure   TKnobsWirePanel.Restack( DoRestack, RedrawWires: Boolean);
    begin
      if ModuleCount > 0
      then Module[ 0].Restack( DoRestack, RedrawWires); // Which will restack all modules
    end;

    procedure   TKnobsWirePanel.Resnap;
    var
      i : Integer;
      p : TPoint;
    begin
      for i := 0 to ModuleCount  - 1
      do begin
        with Module[ i]
        do begin
          p    := Snap( Point( Left, Top));
          Left := P.X;
          Top  := P.Y;
        end;
      end;
    end;

    procedure   TKnobsWirePanel.FixNames;
    var
      i       : Integer;
      aModule : TKnobsCustomModule;
    begin
      for i := 0 to ModuleCount - 1
      do Module[ i].Name := Format( 'a%d', [ i], AppLocale);

      for i := 0 to ModuleCount - 1
      do begin
        aModule := Module[ i];
        aModule.Name := Format( '%s%d', [ Copy( aModule.ClassName, 2, Length( aModule.ClassName)), i], AppLocale);
      end;
    end;

    procedure   TKnobsWirePanel.SolveNamingConflictsWith( aWirePanel: TKnobsWirePanel);
    var
      i        : Integer;
      NextName : Integer;
      aModule  : TKnobsCustomModule;
    begin
      NextName := aWirePanel.ModuleCount;

      for i := 0 to ModuleCount - 1
      do begin
        aModule := Module[ i];

        if Assigned( aWirePanel.FindModule( aModule.Name))
        then begin
          aModule.Name := Format( '%s%d', [ Copy( aModule.ClassName, 2, Length( aModule.ClassName)), NextName], AppLocale);
          Inc( NextName);
        end;
      end;
    end;

    function    TKnobsWirePanel.Snap( const aPoint: TPoint): TPoint;
    begin
      if SnapActive and ( SnapX <> 0) and ( SnapY <> 0)
      then begin
        Result.x := OffsetLeft + (( aPoint.x - OffsetLeft + SnapX div 2) div SnapX) * SnapX;
        Result.y := OffsetTop  + (( aPoint.y - OffsetTop  + SnapY div 2) div SnapY) * SnapY;
      end
      else Result := aPoint;
    end;

    procedure   TKnobsWirePanel.DoHistoryChanged( const aSender: TObject; anUndoCount, aRedoCount: Integer);
    begin
      if Assigned( FOnHistoryChange)
      then begin
        try
          FOnHistoryChange( aSender, anUndoCount, aRedoCount);
        except
          on E: Exception
          do KilledException( E);
        end;
      end;
    end;

    procedure   TKnobsWirePanel.ValueChanged( aSender: TObject; const aPath, aControlType: string; aValue: TSignal; IsFinal: Boolean); // virtual;
    begin
      if Assigned( FOnValueChanged)
      then begin
        try
          FOnValueChanged( aSender, aPath, aControlType, aValue, IsFinal);
        except
          on E: Exception
          do KilledException( E);
        end;
      end;
    end;

    procedure   TKnobsWirePanel.TextChanged( aSender: TObject; const aPath, aValue: string);
    begin
      if Assigned( FOnTextChanged)
      then begin
        try
          FOnTextChanged( aSender, aPath, aValue);
        except
          on E: Exception
          do KilledException( E);
        end;
      end;
    end;

    procedure   TKnobsWirePanel.PatchChanged;
    begin
      if Assigned( FOnPatchChanged)
      then begin
        try
          FOnPatchChanged( Self);
        except
          on E: Exception
          do KilledException( E);
        end;
      end;
    end;

    procedure   TKnobsWirePanel.WiresChanged;
    begin
      if Assigned( FOnWiresChanged)
      then begin
        try
          FOnWiresChanged( Self);
        except
          on E: Exception
          do KilledException( E);
        end;
      end;
    end;

    procedure   TKnobsWirePanel.ScrollPointInView( aPoint: TPoint); // overload;
    begin
      ScrollPointInView( aPoint.X, aPoint.Y);
    end;

    procedure   TKnobsWirePanel.ScrollPointInView( anX, anY: LongInt); // overload;
    const
      Clearance = 33;
    var
      aRect: TRect;
    begin
      aRect.TopLeft     := Point( anX - Clearance, anY - Clearance);
      aRect.BottomRight := Point( anX + Clearance, anY + Clearance);

      Dec( aRect.Left  , HorzScrollBar.Margin);
      Inc( aRect.Right , HorzScrollBar.Margin);
      Dec( aRect.Top   , VertScrollBar.Margin);
      Inc( aRect.Bottom, VertScrollBar.Margin);

      if aRect.Left < 0
      then begin
        with HorzScrollBar
        do Position := Position + aRect.Left
      end
      else if aRect.Right > ClientWidth
      then begin
        if aRect.Right - aRect.Left > ClientWidth
        then aRect.Right := aRect.Left + ClientWidth;
        with HorzScrollBar
        do Position := Position + aRect.Right - ClientWidth;
      end;
      if aRect.Top < 0
      then begin
        with VertScrollBar
        do Position := Position + aRect.Top;
      end
      else if aRect.Bottom > ClientHeight
      then begin
        if aRect.Bottom - aRect.Top > ClientHeight
        then aRect.Bottom := aRect.Top + ClientHeight;
        with VertScrollBar
        do Position := Position + aRect.Bottom - ClientHeight;
      end;
    end;

//  public

    constructor TKnobsWirePanel.Create( anOwner: TComponent); // override;
    begin
      inherited;

      ControlStyle :=
        [
          csAcceptsControls,
          csCaptureMouse,
          csClickEvents,
          csSetCaption,
          csDoubleClicks,
          csPannable
        ];

      OffsetLeft       := 4;
      OffsetTop        := 4;
      AutoScroll       := True;
      Width            := 185;
      Height           :=  41;
      SnapX            := 250;
      SnapY            :=  15;
      SnapActive       := True;
      FModules         := TKnobsModuleList     .Create;
      FWireOverlay     := TKnobsWireOverlay    .Create( Self);
      FHistory         := TKnobsUndoRedoHistory.Create;
      MaxHistory       := 100;
      FControlMode     := dmHorizontal;
      ControlMode      := dmCircular;
      ModuleColor      := clWhite;
      ModuleOpacity    := 223;
      WheelSensitivity := 50;

      FWireOverlay.Parent          := Self;
      FHistory    .OnHistoryChange := DoHistoryChanged;
    end;

    destructor  TKnobsWirePanel.Destroy; // override;
    begin
      inherited;                    // This will call into "Notifiction", which
      FreeAndNil( FModules);        // accesses FModules, so free FModules last.
      FreeAndNil( FHistory);        // and do this later even ...
    end;

    procedure   TKnobsWirePanel.FixAllSynthParams;
    var
      i : Integer;
    begin
      for i := 0 to ModuleCount - 1
      do Module[ i].FixAllSynthParams;
    end;

    procedure   TKnobsWirePanel.ShowAllPopups;
    var
      i : Integer;
    begin
      for i := ModuleCount - 1 downto 0
      do Module[ i].ShowAllPopups;
    end;

    procedure   TKnobsWirePanel.HideAllPopups;
    var
      i : Integer;
    begin
      for i := 0 to ModuleCount - 1
      do Module[ i].HideAllPopups;
    end;

    procedure   TKnobsWirePanel.ForceHideAllPopups;
    var
      i : Integer;
    begin
      for i := 0 to ModuleCount - 1
      do Module[ i].ForceHideAllPopups;
    end;

    procedure   TKnobsWirePanel.BeginStateChange;
    begin
      if Assigned( FPatchReader) and Assigned( FPatchWriter)
      then FHistory.SaveState( Self);
      FHistory.Lock;
    end;

    procedure   TKnobsWirePanel.EndStateChange( PatchWasChanged: Boolean);
    begin
      FHistory.UnLock;
      if ( FHistory.FLockCount = 0) and PatchWasChanged
      then PatchChanged;
    end;

    procedure   TKnobsWirePanel.Undo;
    begin
      if Assigned( FPatchReader) and Assigned( FPatchWriter)
      then begin
        FHistory.Undo( Self);
        PatchChanged;
        UnSelectAllModules;
      end
    end;

    procedure   TKnobsWirePanel.Redo;
    begin
      if Assigned( FPatchReader) and Assigned( FPatchWriter)
      then begin
        FHistory.Redo( Self);
        PatchChanged;
        UnSelectAllModules;
      end;
    end;

    procedure   TKnobsWirePanel.ClearUndoRedo;
    begin
      FHistory.Clear;
    end;

    procedure   TKnobsWirePanel.LockUndoRedo;
    begin
      FHistory.Lock;
    end;

    procedure   TKnobsWirePanel.UnLockUndoRedo;
    begin
      FHistory.UnLock;
    end;

    procedure   TKnobsWirePanel.Dump( const aFileName: string; DoAppend: Boolean); // overload;
    var
      aFile: TextFile;
    begin
      AssignFile( aFile, aFileName);
      if FileExists( aFileName)
      then begin
        if DoAppend
        then Append ( aFile)
        else Rewrite( aFile);
      end
      else Rewrite( aFile);
      try
        Dump( aFile);
      finally
        CloseFile( aFile);
      end;
    end;

    procedure   TKnobsWirePanel.Dump( var aFile: TextFile); // overload;
    begin
      WriteLn     ( aFile,         '---- wirepanel.dump -----------------------------------');
      WriteLn     ( aFile, Format( '---- %s', [ FormatDateTime( 'yyyy-mm-dd hh:nn:ss', Now)], AppLocale));
      DumpModules ( aFile, 1);
      WriteLn     ( aFile);
      DumpWires   ( aFile, 1);;
      WriteLn     ( aFile,         '---- end wirepanel.dump -------------------------------');
      WriteLn     ( aFile);
    end;

    procedure   TKnobsWirePanel.DumpModules( var aFile: TextFile; anInd: Integer);
    begin
      WriteLnInd( aFile, anInd, 'modules');
      if Assigned( FModules)
      then FModules.Dump( aFile, anInd + 1);
      WriteLnInd( aFile, anInd, 'end modules');
    end;

    procedure   TKnobsWirePanel.DumpWires( var aFile: TextFile; anInd: Integer);
    begin
      WriteLnInd( aFile, anInd, 'wires');
      if Assigned( FWireOverlay)
      then FWireOverlay.Dump( aFile, anInd + 1);
      WriteLnInd( aFile, anInd, 'end wires');
    end;

    procedure   TKnobsWirePanel.Log( const aMsg: string);
    begin
      if Assigned( FOnLog)
      then FOnLog( Self, aMsg);
    end;

    function    TKnobsWirePanel.FindModule( const aName: string): TKnobsCustomModule;
    var
      i : Integer;
    begin
      Result := nil;

      for i := 0 to ModuleCount - 1
      do begin
        if SameText( Module[ i].Name, aName)
        then begin
          Result := Module[ i];
          Break;
        end;
      end;
    end;

    function    TKnobsWirePanel.FindFirstSelectedModule: TKnobsCustomModule;
    var
      i : Integer;
    begin
      Result := nil;

      for i := 0 to ModuleCount - 1
      do begin
        if Module[ i].Selected
        then begin
          Result := Module[ i];
          Break;
        end;
      end;
    end;

//  public

    procedure   TKnobsWirePanel.InvalidateWires;
    begin
      FWireOverlay.InvalidateWires;
    end;

    procedure   TKnobsWirePanel.InvalidateDisplays;
    begin
      FModules.InvalidateDisplays;
    end;

    procedure   TKnobsWirePanel.Disconnect( aConnector: TKnobsConnector);
    var
      Changed : Boolean;
    begin
      if Assigned( FWireOverlay)
      then begin
        Changed := False;
        BeginStateChange;
        try
          Changed := FWireOverlay.Disconnect( aConnector);
        finally
          EndStateChange( Changed)
        end;
      end;
    end;

    procedure   TKnobsWirePanel.FreeModules;
    var
      Changed : Boolean;
    begin
      BeginStateChange;
      BlockScreenUpdates;
      Changed := False;
      try
        Changed := FModules.FreeModules;
      finally
        UnBlockScreenUpdates;
        EndStateChange( Changed);
      end;
    end;

    function    TKnobsWirePanel.HasEditPopup: Boolean;
    begin
      Result :=
        ( Assigned( GDisplayEditor) and GDisplayEditor.Visible) or
        ( Assigned( GLabelEditor  ) and GLabelEditor  .Visible)
    end;

    procedure   TKnobsWirePanel.SetSelectedModuleColor( aValue: TColor);
    begin
      FModules.SetSelectedModuleColor( aValue);
    end;

    procedure   TKnobsWirePanel.FreeSelectedModules;
    var
      Changed : Boolean;
    begin
      Assert( not HasEditPopup, 'can not free modules when a popup editor is active');
      BeginStateChange;
      BlockScreenUpdates;
      Changed := False;
      try
        Changed := FModules.FreeSelectedModules;
      finally
        UnBlockScreenUpdates;
        EndStateChange( Changed);
      end;
    end;

    procedure   TKnobsWirePanel.SelectAllModules;
    begin
      FModules.SelectAll;
    end;

    procedure   TKnobsWirePanel.UnSelectAllModules;
    begin
      FModules.UnSelectAll;
    end;

    procedure   TKnobsWirePanel.SelectUnique( aModule: TKnobsCustomModule);
    begin
      FModules.SelectUnique( aModule);
    end;

    procedure   TKnobsWirePanel.FixSelection( aModule: TKnobsCustomModule);
    // When no modules are selected select aModule uniquely
    begin
      FModules.FixSelection( aModule);
    end;

    procedure   TKnobsWirePanel.InvertModuleSelection;
    begin
      FModules.InvertSelection;
    end;

    function    TKnobsWirePanel.CopyToString( aWriteMode: TKnobsWriteMode): string;
    begin
      Result := '';
      if Assigned( FPatchWriter)
      then Result := FPatchWriter.WriteString( Self, '', aWriteMode);
    end;

    procedure   TKnobsWirePanel.CopyFromString( const S: string; aReadMode: TKnobsReadMode);
    begin
      if Assigned( FPatchReader)
      then FPatchReader.ReadString( S, Self, Point( 0, 0), rmReplace, '');
    end;

    procedure   TKnobsWirePanel.CopyTo( aPanel: TKnobsWirePanel; aWriteMode: TKnobsWriteMode; aReadMode: TKnobsReadMode);
    begin
      if Assigned( aPanel)
      then aPanel.CopyFromString( CopyToString( aWriteMode), aReadMode);
    end;

    function    TKnobsWirePanel.SameStructure( aPanel: TKnobsWirePanel): Boolean;
    // For a structural match to be present there needs be a one to one module type mapping from one panel to the other.
    // Module order is important here (imagine the case where two identically typed modules are involved).
    // We only look at the selected modules in self, but look at all the modules in aPanel, regardless of selection.
    // The easy way would be to make a copy of the selected modules of Self and compare that to aPanel. The wires are
    // irrelevamt for a structural match.
    var
      i : Integer;
      p : Integer;
    begin
      Result := False;
      if Assigned( aPanel)
      then begin
        Result := True;
        p      := 0;

        for i := 0 to ModuleCount - 1
        do begin
          if Module[ i].Selected
          then begin
            if
              ( p >= aPanel.ModuleCount) or
              ( aPanel.Module[ p].ModuleType <> Module[ i].ModuleType)
            then begin
              Result := False;
              Break;
            end;
            Inc( p);
          end;
        end;
      end;
    end;

    procedure   TKnobsWirePanel.CopyParamsFrom( aSrc: TKnobsWirePanel);
    //* When the structure of aSrc matches the structure of the selection in Self
    //* then copy the parameter values from aSrc over to the selection
    //* of Self
    var
      i : Integer;
      p : Integer;
    begin
      // done: implement parameter copying from aSrc to the selection in Self.
      // First check structures, the structure of the selection in Self needs to match the structure of aSrc (regardless
      // of selection there). If so copy over all parameter values from aSrc to the selection in Self. For a structural
      // match to be present there needs be a one to one module type mapping from one panel to the other. Only when that
      // helds a parameter copy can be performed. Module order is important here (imagine the case where two identically
      // typed modules are involved), so a sort should be performed here, likely, on the screen positons ... hmm. The
      // origin of aSrc is assumed to be a copy (action) of selected modules from a patch .. this may not be a current
      // invariant .. need to check that ... ok, modules are written in module order always when using a PatchWriter.
      // The easy way would be to make a copy of the selected modules of Self and copy that to aPanel. The wires are
      // irrelevamt for a parameter copy.
      if Assigned( aSrc)
      then begin
        if SameStructure( aSrc)
        then begin
          p := 0;
          for i := 0 to ModuleCount - 1
          do begin
            if Module[ i].Selected and ( p < aSrc.ModuleCount)
            then begin
              // The SameStructure call checked Module[ i] and aSrc.Module[] p] to be of the same type so we should not
              // check that here in order to let p and i not get out of sync. The CopyParamsFrom method for a module
              // however should still perform a type check, and as such it may not actually perform a param copy.
              Module[ i].CopyParamsFrom( aSrc.Module[ p]);
              Inc( p);
            end;
          end;
        end;
      end;
    end;

    procedure   TKnobsWirePanel.AddModules( aSrc: TKnobsWirePanel; aDistance: TPoint; DoRestack, RedrawWires, MustDrag: Boolean); // override;
    //* This one is called only from the KnobParser either when a patch files
    //* (or a set of modules) is loaded or when a wirepanel was pasted from the
    //* clipboard; when pasted from the clipboard, or when it is a set of modules
    //* to be added in, the newly pasted modules should be set into drag mode,
    //* for other cases see __2 (also in KnobParser although that one will end
    //* up here and nothing special is done there except for setting the
    //* Dragging parameter to true when the Addition was the result of a paste
    //* from the clip board, or when a set of modules was loaded from a file).
    //* Correction : also after a Ctrl+Drag (copy) operation we end up here ...
    //* in that case dragging will be false, which is the correct way to handle
    //* this ... luck ... :shock:
    //* See Clone method - that one fixes it for single module insertions.
    var
      i : Integer;
      M : TKnobsCustomModule;
    begin
      Assert( not HasEditPopup, 'can not add modules when a popup editor is active');
      BeginStateChange;          // Blocks undo/redo updates
      BlockScreenUpdates;        // Blocks screen paints
      DisableAlign;
      try
        UnSelectAllModules;
        FixNames;

        with aSrc
        do begin
          FixNames;
          SolveNamingConflictsWith( Self);

          // When dragging the following code should not be performed as this
          // represents a final insertion into the wirepanel which will move
          // any existing modules around ... this will be tricky .... __2
          // we could maybe treat this by inserting the modules a a negative
          // offset (which is not normally possible) and setting an approriate
          // distance instead of the one passed as an argument to us.
          for i := 0 to ModuleCount - 1
          do begin
            M := Module[ i];
            if M.Color = TColor( $ffffffff)
            then M.Color := Self.ModuleColor;
            M.Opacity := Self.ModuleOpacity;
            M.Clone(
              Self,
              Self,
              Self.Snap( Point( M.Left, M.Top ) + aDistance),
              MustDrag
            ).Restack( DoRestack, RedrawWires);
          end;
        end;

        with aSrc
        do begin
          for i := 0 to WireCount - 1
          do begin
            Self.StrToConnection(
              Format
              (
                '%s %s',
                [
                  ConnectorName( Source     [ i]),
                  ConnectorName( Destination[ i])
                ],
                AppLocale
              )
            );
          end;
        end;
      finally
        FixControlMode;
        UnBlockScreenUpdates;
        EndStateChange( True);
        EnableAlign;
      end;
    end;

    procedure   TKnobsWirePanel.AddWire( const aSrc, aDst: string);
    begin
      BeginStateChange;
      try
        StrToConnection( Format( '%s %s', [ aSrc, aDst], AppLocale));
      finally
        EndStateChange( True);
      end;
    end;

    procedure   TKnobsWirePanel.WiggleWires;
    begin
      FWireOverlay.WiggleWires;
    end;

    procedure   TKnobsWirePanel.FixAllWires;
    begin
      if Assigned( FWireOverlay)
      then FWireOverlay.FixAllWires;
    end;

    procedure   TKnobsWirePanel.FixControlMode;
    var
      i : Integer;
    begin
      for i := 0 to ModuleCount - 1
      do begin
        Module[ i].ControlMode         := ControlMode;
        Module[ i].WheelSupportOnKnobs := WheelSupportOnKnobs;
        Module[ i].WheelSensitivity    := WheelSensitivity;
      end;
    end;

    procedure   TKnobsWirePanel.BlockScreenUpdates;
    begin
      FWireOverlay.BlockScreenUpdates;
    end;

    procedure   TKnobsWirePanel.UnblockScreenUpdates;
    begin
      FWireOverlay.UnBlockScreenUpdates;
    end;

    procedure   TKnobsWirePanel.Search( const aValue: string);
    var
      i : Integer;
      M : TKnobsCustomModule;
      U : string;
      N : string;
    begin
      U := UpperCase( aValue);
      for i := 0 to ModuleCount - 1
      do begin
        M := Module[ i];
        N := UpperCase( M.Title);
        M.Selected := ( Pos( U, N) > 0) and ( Length( U) > 0);
      end;
    end;

    procedure   TKnobsWirePanel.FindUnconnectedModules;
    var
      i       : Integer;
      aModule : TKnobsCustomModule;
    begin
      for i := 0 to ModuleCount - 1
      do begin
        aModule          := Module[ i];
        aModule.Selected := not aModule.IsConnected;
      end;
    end;

    procedure   TKnobsWirePanel.UnFocus( const aSender: TObject);
    begin
      if Assigned( FOnUnFocus)
      then FOnUnFocus( aSender);
    end;

    procedure   TKnobsWirePanel.FindGhosts( const aGhosts: TStringList);
    var
      i : Integer;
    begin
      if Assigned( aGhosts)
      then begin
        for i := 0 to ModuleCount - 1
        do begin
          if ( Module[ i].Top < 0) or ( Module[ i].Left < 0)
          then aGhosts.Add( Module[ i].Name);
        end;
      end;
    end;

    function    TKnobsWirePanel.CountGhosts: Integer;
    var
      i : Integer;
    begin
      Result := 0;
      for i := 0 to ModuleCount - 1
      do begin
        if ( Module[ i].Top < 0) or ( Module[ i].Left < 0)
        then Inc( Result);
      end;
    end;

//  public

    procedure   TKnobsWirePanel.ConnectorDelete( const aConnector: TKnobsConnector);
    var
      anOutput: TKnobsConnector;
    begin
      // Disconnects every connector on this tree
      if Assigned( aConnector)
      then begin
        anOutput := aConnector.FindOutput;
        if Assigned( anOutput)
        then anOutput  .DisConnectDownStreamRecursive
        else aConnector.DisConnectDownStreamRecursive;
      end;
    end;

    procedure   TKnobsWirePanel.ConnectorBreak( const aConnector: TKnobsConnector);
    begin
      // Should disconnect the downstream links, currently breaks the upstream one too
      if Assigned( aConnector)
      then aConnector.DisConnectDownStream;
    end;

    procedure   TKnobsWirePanel.ConnectorDisconnect( const aConnector: TKnobsConnector);
    begin
      // Takes this connector out of the connection tree
      if Assigned( aConnector)
      then aConnector.DisConnectAll;
    end;


{ ========
  TKnobsCustomModule = class( TCustomPanel)
  // Base type for TModule, it has a bitmap and a title mainly
  private
    FAnchor               : TPoint;                 // Drag origin
    FDistance             : TPoint;                 // Drag distance
    FTresholdPast         : Boolean;                // Set when dragged over a minimum distance
    FFlags                : TKnobsModuleFlags;      // Selected / Capturing
    FTitleLabel           : TKnobsEditLabel;        // All editor modules can have a title label, it may be nil
    FTitle                : string;                 // User editible module title
    FComment              : string;                 // Comment as displayed on mouseover in the module selector
    FModuleType           : TKnobsModuleType;       // The type, used to couple editor modules to synth modules
    FPageName             : string;                 // The page in the module selector this module will be on,
                                                    // when the parent is a TTabSheet and that TTabSheet has a non empty
                                                    // Caption that value will be used instead.
    FBitmap               : TBitmap;                // Background bitmap for module
    FOpacity              : Byte;                   // Opacity for background bitmap
    FControlMode          : TDistanceMode;          // Knobs control mode, circular, horizontal or vertical
    FWheelSupportOnKnobs  : Boolean;
    FWheelSensitivity     : Integer;
    FOnValueChanged       : TKnobsOnValueChange;    // Called with the new value
    FOnTextChanged        : TKnobsOnTextChanged;    // Called with the new text
    FOnLog                : TKnobsOnLog;            // To atach an external logger
    FOnUnFocus            : TOnUnFocus;             // Called when a control could be unfocussed
  public
    property    Selected            : Boolean                read GetSelected          write SetSelected;
    property    Capturing           : Boolean                read GetCapturing         write SetCapturing;
    property    Title               : string                 read FTitle               write SetTitle;
    property    Comment             : string                 read FComment             write FComment;
    property    PageName            : string                 read GetPageName          write SetPageName;
    property    ModuleType          : TKnobsModuleType       read FModuleType          write SetModuleType;
    property    OnLog               : TKnobsOnLog            read FOnLog               write FOnLog;
  published
    property    Color;
    property    TitleLabel          : TKnobsEditLabel        read FTitleLabel          write SetTitleLabel;
    property    Opacity             : Byte                   read FOpacity             write SetOpacity             default 223;
    property    ControlMode         : TDistanceMode          read FControlMode         write SetControlMode;
    property    WheelSupportOnKnobs : Boolean                read FWheelSupportOnKnobs write SetWheelSupportOnKnobs;
    property    WheelSensitivity    : Integer                read FWheelSensitivity    write SetWheelSensitivity;
    property    OnValueChanged      : TKnobsOnValueChange    read FOnValueChanged      write FOnValueChanged;
    property    OnTextChanged       : TKnobsOnTextChanged    read FOnTextChanged       write FOnTextChanged;
    property    OnUnFocus           : TOnUnFocus             read FOnUnFocus           write FOnUnFocus;
  private
}

    procedure   TKnobsCustomModule.IncludeFlags( aFlags: TKnobsModuleFlags);
    begin
      FFlags := FFlags + aFlags;
    end;

    procedure   TKnobsCustomModule.ExcludeFlags( aFlags: TKnobsModuleFlags);
    begin
      FFlags := FFlags - aFlags;
    end;

    function    TKnobsCustomModule.HasFlags( aFlags: TKnobsModuleFlags): Boolean;
    begin
      Result := aFlags <= FFlags;
    end;

//  private

    function    TKnobsCustomModule.GetSelected: Boolean;
    begin
      Result := HasFlags( [ mfSelected]);
    end;

    procedure   TKnobsCustomModule.SetSelected( aValue: Boolean);
    const
      Colors : array[ Boolean] of TColor = ( clSilver, clWhite);
    begin
      if aValue <> Selected
      then begin
        if aValue
        then IncludeFlags( [ mfSelected])
        else ExcludeFlags( [ mfSelected]);

        if Assigned( FTitleLabel)
        then begin
          with FTitleLabel
          do begin
            Color       := Colors[ Selected];
            Transparent := not Selected;
          end;
        end;
      end;
    end;

    function    TKnobsCustomModule.GetCapturing: Boolean;
    begin
      Result := HasFlags( [ mfCapturing]);
    end;

    procedure   TKnobsCustomModule.SetCapturing( aValue: Boolean);
    begin
      if aValue <> Capturing
      then begin
        if aValue
        then IncludeFlags( [ mfCapturing])
        else ExcludeFlags( [ mfCapturing]);
      end;
    end;

//  private

    function    TKnobsCustomModule.GetPageName: string;
    begin
      if HasParent and ( Parent is TTabSheet) and ( TTabSheet( Parent).Caption <> '')
      then Result := TTabSheet( Parent).Caption
      else Result := FPageName;
    end;

    procedure   TKnobsCustomModule.SetPageName( const aValue: string);
    begin
      if HasParent and ( Parent is TTabSheet) and ( TTabSheet( Parent).Caption <> '')
      then FPageName := TTabSheet( Parent).Caption
      else FPageName := aValue;
    end;

    procedure   TKnobsCustomModule.SetModuleType( aValue : TKnobsModuleType);
    begin
      if aValue <> FModuleType
      then begin
        FModuleType := aValue;
      end;
    end;

    procedure   TKnobsCustomModule.SetTitle( const aValue: string);
    begin
      if aValue <> FTitle
      then begin
        FTitle := aValue;
        if Assigned( FTitleLabel)
        then FTitleLabel.Text := FTitle;
      end;
    end;

    procedure   TKnobsCustomModule.SetTitleLabel( const aValue: TKnobsEditLabel);
    begin
      FTitleLabel := aValue;

      if Assigned( FTitleLabel) and not ( ComponentState * [ csDestroying, csLoading] = [])
      then FTitleLabel.Caption := Title;
    end;

    procedure   TKnobsCustomModule.SetOpacity( aValue: Byte);
    begin
      if FOpacity <> aValue
      then begin
        FOpacity := aValue;
        Invalidate;
      end;
    end;

    procedure   TKnobsCustomModule.SetControlMode( aValue: TDistanceMode);
    var
      i : Integer;
    begin
      if aValue <> FControlMode
      then begin
        FControlMode := aValue;
        for i := 0 to ControlCount - 1
        do begin
          if Controls[ i] is TKnobsKnob
          then TKnobsKnob( Controls[ i]).ControlMode := FControlMode;
        end;
      end;
    end;

    procedure   TKnobsCustomModule.SetWheelSupportOnKnobs( aValue: Boolean);
    var
      i : Integer;
    begin
      for i := 0 to ComponentCount - 1
      do begin
        if Components[ i] is TKnobsValuedControl
        then TKnobsValuedControl( Components[ i]).WheelSupport := aValue;
      end;
    end;

    procedure   TKnobsCustomModule.SetWheelSensitivity( aValue: Integer);
    var
      i : Integer;
    begin
      for i := 0 to ComponentCount - 1
      do begin
        if Components[ i] is TKnobsValuedControl
        then TKnobsValuedControl( Components[ i]).WheelSensitivity := aValue;
      end;
    end;

//  protected

    procedure   TKnobsCustomModule.WMEraseBackground( var aMsg: TWMEraseBkgnd); // message WM_ERASEBKGND;
    begin
      InvalidateKnobs;
      aMsg.Result := -1;
    end;

    procedure   TKnobsCustomModule.MouseDown( Button: TMouseButton; Shift: TShiftState; X, Y: Integer); // override;
    var
      aWP: TKnobsWirePanel;
    begin
      inherited;

      if Capturing
      then Exit;

      if Button = mbLeft
      then begin
        HandleSelection( ssCtrl in Shift);
        SetCapture( Handle);
        Capturing     := True;
        FTresholdPast := False;
        BeginMove( Point( X, Y));
      end
      else if Button = mbRight
      then begin
        aWP := FindWirePanel;
        if Assigned( aWP)
        then aWP.HandleRightClick( Self);
      end;

      SendToBack; // See __1 remarks  - needed to keep wires on top of modules
    end;

    procedure   TKnobsCustomModule.MouseMove( Shift: TShiftState; X, Y: Integer); // override;
    begin
      inherited;

      if  Capturing
      then begin
        if FTresholdPast
        then MoveSelected( FDistance);

        FDistance := Point( X, Y) - FAnchor;

        if not FTresholdPast
        then FTresholdPast := Distance( FDistance) > 8;

        if FTresholdPast
        then begin
          if ssCtrl in Shift
          then Selected := True;

          MoveSelected( FDistance);
        end;
      end;
    end;

    procedure   TKnobsCustomModule.MouseUp( Button: TMouseButton; Shift: TShiftState; X, Y: Integer); // override;
    begin
      inherited;

      if Capturing
      then begin
        ReleaseCapture;
        ReleasePeerCapture;

        if FTresholdPast
        then begin
          if HasParent
          then Parent.DisableAlign;

          try
            FinalMoveSelected( FDistance, ssCtrl in Shift)
          finally
            if HasParent
            then Parent.EnableAlign;
          end;
        end;

        SendToBack;  // See __1 remarks  - needed to keep wires on top of modules
      end;
    end;

//  protected

    procedure   TKnobsCustomModule.InvalidateKnobs;
    var
      i : Integer;
    begin
      for i := 0 to ComponentCount - 1
      do begin
        if Components[ i] is TKnobsKnob
        then TKnobsKnob( Components[ i]).Invalidate;
      end;
    end;

    procedure   TKnobsCustomModule.InvalidateDisplays;
    var
      i : Integer;
    begin
      for i := 0 to ComponentCount - 1
      do begin
        if Components[ i] is TKnobsKnob
        then TKnobsKnob( Components[ i]).InvalidateDisplay;
      end;
    end;

    procedure   TKnobsCustomModule.ValueChanged( aSender: TObject; const aPath, aControlType: string; aValue: TSignal; IsFinal: Boolean); // virtual;
    var
      aWp: TKnobsWirePanel;
    begin
      if Assigned( FOnValueChanged)
      then FOnValueChanged( aSender, aPath, aControlType, aValue, IsFinal)
      else begin
        aWp := FindWirePanel;
        if Assigned( aWp)
        then aWp.ValueChanged( aSender, MakePath( [ aWp.Name, aPath]), aControlType, aValue, IsFinal);
      end;
    end;

    procedure   TKnobsCustomModule.TextChanged( aSender: TObject; const aPath, aValue: string);
    var
      aWp: TKnobsWirePanel;
    begin
      if Assigned( FOnTextChanged)
      then FOnTextChanged( aSender, aPath, aValue)
      else begin
        aWp := FindWirePanel;
        if Assigned( aWp)
        then aWp.TextChanged( aSender, MakePath( [ aWp.Name, aPath]), aValue);
      end;
    end;

    function    TKnobsCustomModule.FindWirePanel: TKnobsWirePanel;
    var
      aParent : TWinControl;
    begin
      Result  := nil;
      aParent := Parent;

      while Assigned( aParent) and not ( aParent is TKnobsWirePanel)
      do aParent :=  aParent.Parent;

      if aParent is TKnobsWirePanel
      then Result := TKnobsWirePanel( aParent);
    end;

    function    TKnobsCustomModule.Snap( const aPoint: TPoint): TPoint;
    var
      aWp : TKnobsWirePanel;
    begin
      aWp := FindWirePanel;

      if Assigned( aWp)
      then Result := aWp.Snap( aPoint)
      else Result := aPoint;
    end;


    procedure   TKnobsCustomModule.Paint; // override;
    var
      aBmp: TBitmap;
    begin
      aBmp := TBitmap.Create;
      try
        aBmp.Width  := Self.Width  - 2;
        aBmp.Height := Self.Height - 2;
        aBmp.Canvas.StretchDraw( Rect( 0, 0, aBmp.Width, aBmp.Height), FBitmap);

        with Canvas
        do begin
          Pen.Color := Color;
          Pen.Width := 1;
          Pen.Style := psSolid;

          Brush.Color := Color;
          Brush.Style := bsSolid;

          Rectangle( 1, 1, Width - 1, Height - 1);
          Draw( 1, 1, aBmp, Opacity);

          Pen.Color := InterpolateColors( clWhite, Color, 50);
          MoveTo( 0, Height - 1);
          LineTo( 0, 0);
          LineTo( Width, 0);

          Pen.Color := InterpolateColors( clBlack, Color, 50);
          MoveTo( 0, Height - 1);
          LineTo( Width - 1, Height - 1);
          LineTo( Width - 1, 1);
        end;
      finally
        aBmp.Free;
      end;
    end;

    procedure   TKnobsCustomModule.Loaded; // override;
    var
      p : TPoint;
    begin
      inherited;

      p := Snap( Point( Left, Top));
      Left := P.X;
      Top  := P.Y;
    end;

    procedure   TKnobsCustomModule.AssignBitmap; // virtual;
    begin
      FBitmap := bmSkin;
    end;

    procedure   TKnobsCustomModule.ScrollSelfInView;
    var
      aWp : TKnobsWirePanel;
    begin
      aWp := FindWirePanel;

      if Assigned( aWp) and ( aWp.BlockCount = 0)
      then aWp.ScrollInView( Self);
    end;

    procedure   TKnobsCustomModule.HandleSelection( CtrlKeyActive: Boolean);
    begin
      if CtrlKeyActive
      then Selected := not Selected
      else begin
        if not Selected
        then SelectUnique;
      end;

      BringToFront; // See __1 remarks around
    end;

    procedure   TKnobsCustomModule.BeginMove( anAnchor: TPoint);
    begin
      FAnchor   := anAnchor;
      FDistance := Point( 0, 0);
      MoveSelected( FDistance);
    end;

    procedure   TKnobsCustomModule.Move( aDistance: TPoint);
    var
      aWp   : TKnobsWirePanel;
      aDC   : HDC;
      aRect : TRect;
      aRgn  : HRGN;
    begin
      aWp := FindWirePanel;
      if Assigned( aWp)
      then begin
        aRect.Left   := Left + aDistance.X + 1;
        aRect.Top    := Top  + aDistance.Y + 1;
        aRect.Right  := aRect.Left + Width  - 2;
        aRect.Bottom := aRect.Top  + Height - 2;

        aDC := GetDCEx( aWp.Handle, 0, DCX_PARENTCLIP);
        try
          with aWp.ClientRect
          do aRgn := CreateRectRgn( Left, Top, Right, Bottom);
          SelectClipRgn( aDC, aRgn);
          DeleteObject( aRgn);
          DrawFocusRect( aDC, aRect);
        finally
          ReleaseDC( Parent.Handle, aDC);
        end;
      end;
    end;

    procedure   TKnobsCustomModule.MoveSelected( aDistance: TPoint);
    var
      i     : Integer;
      aWp   : TKnobsWirePanel;
      aPt   : TPoint;
      aRect : TRect;
    begin
      aWp   := FindWirePanel;
      aPt   := aDistance;
      aRect := Rect( Left, Top, Left + Width, Top + Height);
      aPt.Offset( aRect.CenterPoint);

      if Assigned( aWp)
      then begin
        with aWp
        do begin
          aWp.ScrollPointInView( aWp.ScreenToClient( ClientToScreen( aPt)));
          for i := 0 to ModuleCount - 1
          do begin
            with Module[ i]
            do begin
              if Selected
              then Move( aDistance);
            end;
          end;
        end;
      end;
    end;

    procedure   TKnobsCustomModule.FinalMoveSelected( aDistance: TPoint; CtrlKeyActive: Boolean);
    //* When CtrlKeyActive is true this is a clone operation and the set of
    //* newly created modules should be set into drag mode. See __2 for other
    //* cases (also in KnobParser). This will end up in TWirePanel.AddModules
    //* when CtrlKeyActive is true, but as we set the filename to '' and not
    //* to '<clip>' it will not go into drag mode (which was just finished
    //* for the ctrl+drag mode ...).
    var
      P   : TPoint;
      i   : Integer;
      aWp : TKnobsWirePanel;
    begin
      aWp := FindWirePanel;

      if Assigned( aWp)
      then begin
        MoveSelected( FDistance);

        with aWp
        do begin
          if CtrlKeyActive // When CtrlKeyActive this was a /copy/ action, a /move/ otherwise
          then PatchReader.ReadString( PatchWriter.WriteString( aWp, '', wmSelected), aWp, aDistance, rmAppend)
          else begin
            aWp.BeginStateChange;
            aWp.BlockScreenUpdates;
            DisableAlign;
            try
              for i := 0 to ModuleCount - 1
              do begin
                with Module[ i]
                do begin
                  if Selected
                  then begin
                    P := Snap( Point( Left, Top) + ScrollOffset + aDistance);
                    if P.X < OffsetLeft then P.X := OffsetLeft;
                    if P.Y < OffsetTop  then P.Y := OffsetTop;
                    P := P - ScrollOffset;
                    SetBounds( P.X, P.Y, Width, Height);
                  end;
                end;
              end;
              ReStack( DO_RESTACK, NO_REDRAW_WIRES);
            finally
              aWp.UnBlockScreenUpdates; // This will invalidate wires, when the lock count gets zero
              aWp.EndStateChange( False);
              EnableAlign;
            end;
          end;
        end;
        ScrollSelfInView;
      end;
    end;

    procedure   TKnobsCustomModule.ReleasePeerCapture;
    var
      i   : Integer;
      aWp : TKnobsWirePanel;
    begin
      aWp := FindWirePanel;
      if Assigned( aWp)
      then begin
        with aWp
        do begin
          for i := 0 to ModuleCount - 1
          do Module[ i].Capturing := False;
        end;
      end;
    end;

    procedure   TKnobsCustomModule.Notification( aComponent: TComponent; anOperation: TOperation); // override;
    var
      aWp : TKnobsWirePanel;
    begin
      inherited;

      if ( anOperation = opRemove) and ( aComponent is TKnobsConnector)
      then begin
        aWp := FindWirePanel;

        if Assigned( aWp)
        then aWp.Disconnect( TKnobsConnector( aComponent));
      end;
    end;

    procedure   TKnobsCustomModule.FixOwnership( aRoot, aControl: TWinControl);
    //
    // Recursively make all controls parented by 'aControl' to be owned by
    // 'aRoot', this in order to allow 'aRoot' to write such controls on
    // a stream easily.
    //
    var
      i : Integer;
    begin
      if Assigned( aControl) and Assigned( aRoot)
      then begin
        with aControl
        do begin
          for i := 0 to ControlCount - 1
          do begin
            with Controls[ i]
            do begin
              if ( Owner <> aRoot) and ( Parent = aRoot)
              then begin
                Owner.RemoveComponent( Controls[ i]);
                aRoot.InsertComponent( Controls[ i]);

                if Controls[ i] is TWinControl
                then FixOwnerShip( aRoot, TWinControl( Controls[ i]));
              end;
            end;
          end;
        end;
      end;
    end;

    procedure   TKnobsCustomModule.FixNames;
    var
      aWp : TKnobsWirePanel;
    begin
      aWp := FindWirePanel;
      if Assigned( aWp)
      then aWp.FixNames;
    end;

    function    TKnobsCustomModule.FindValuedControl( const aName: string): TKnobsValuedControl;
    var
      i : Integer;
    begin
      Result := nil;

      for i := 0 to ComponentCount - 1
      do begin
        if ( Components[ i] is TKnobsValuedControl) and SameText( aName, Components[ i].Name)
        then begin
          Result := TKnobsValuedControl( Components[ i]);
          break;
        end;
      end;
    end;

    function    TKnobsCustomModule.FindXYControl( const aName: string): TKnobsXYControl;
    var
      i : Integer;
    begin
      Result := nil;

      for i := 0 to ComponentCount - 1
      do begin
        if ( Components[ i] is TKnobsXYControl) and SameText( aName, Components[ i].Name)
        then begin
          Result := TKnobsXYControl( Components[ i]);
          break;
        end;
      end;
    end;

    function    TKnobsCustomModule.FindDisplay( const aName: string): TKnobsDisplay;
    var
      i : Integer;
    begin
      Result := nil;

      for i := 0 to ComponentCount - 1
      do begin
        if ( Components[ i] is TKnobsDisplay) and SameText( aName, Components[ i].Name)
        then begin
          Result := TKnobsDisplay( Components[ i]);
          break;
        end;
      end;
    end;

    function    TKnobsCustomModule.FindFileSelector( const aName: string): TKnobsFileSelector;
    var
      i : Integer;
    begin
      Result := nil;

      for i := 0 to ComponentCount - 1
      do begin
        if ( Components[ i] is TKnobsFileSelector) and SameText( aName, Components[ i].Name)
        then begin
          Result := TKnobsFileSelector( Components[ i]);
          break;
        end;
      end;
    end;

//  public

    constructor TKnobsCustomModule.Create( anOwner: TComponent); // override;
    begin
      inherited;

      ControlStyle :=
        [
          csAcceptsControls,
          csCaptureMouse,
          csClickEvents,
          csDoubleClicks,
          csOpaque
        ];

      AssignBitmap;
      FBitmap.Transparent := False;

      Width          := 250;
      Height         :=  60;
      Name           := '';
      FOpacity       := 223;
      FTitle         := '';
      DoubleBuffered := True;
    end;

    destructor  TKnobsCustomModule.Destroy; // override;
    begin
      Disconnect;
      inherited;
    end;

    procedure   TKnobsCustomModule.ShowAllPopups;
    var
      i : Integer;
    begin
      for i := ControlCount - 1 downto 0
      do begin
        if Controls[ i] is TKnobsKnob
        then TKnobsKnob( Controls[ i]).ShowPopup;
      end;
    end;

    procedure   TKnobsCustomModule.HideAllPopups;
    var
      i : Integer;
    begin
      for i := 0 to ControlCount - 1
      do begin
        if Controls[ i] is TKnobsKnob
        then TKnobsKnob( Controls[ i]).HidePopup;
      end;
    end;

    procedure   TKnobsCustomModule.ForceHideAllPopups;
    var
      i : Integer;
    begin
      for i := 0 to ControlCount - 1
      do begin
        if Controls[ i] is TKnobsKnob
        then TKnobsKnob( Controls[ i]).ForceHidePopup;
      end;
    end;

    procedure   TKnobsCustomModule.FixAllSynthParams;
    var
      i : Integer;
    begin
      for i := 0 to ControlCount - 1
      do begin
        if Controls[ i] is TKnobsValuedControl
        then begin
          if TKnobsValuedControl( Controls[ i]).StepCount > 1
          then TKnobsValuedControl( Controls[ i]).ValueChanged( True);
        end
        else if
          (( Controls[ i] is TKnobsDisplay     ) and TKnobsDisplay( Controls[ i]).HasEditor) or
           ( Controls[ i] is TKnobsFileSelector)
        then begin
          with TKnobsTextControl( Controls[ i])
          do TextChanged( TextValue, NO_NOTIFY);
        end;
      end;
    end;

    function    TKnobsCustomModule.Clone( anOwner: TComponent; aParent: TWinControl; anOffset: TPoint; MustDrag: Boolean): TKnobsCustomModule;
    //
    // Make all children parented by self to be owned by self, so they will be
    // streamed out. Then write self on a stream and read it back into a new
    // TCustomModule. Finally make the clone owned and parented as indicated in
    // the parameters passed, then position it relative to it's parent's (top, left).
    //
    var
      S : TStream;
      i : Integer;
    begin
      FixOwnerShip( Self, Self);
      S := TMemoryStream.Create;

      try
        S.WriteComponent( Self);

        S.Seek( 0, 0);
        Result := TKnobsCustomModule( S.ReadComponent( nil));

        // ReadComponent fux up zero values in TValuedControls ... fix 'm manually then ...

        for i := 0 to ControlCount - 1
        do begin
          if Controls[ i] is TKnobsValuedControl
          then ( Result.Controls[ i] as TKnobsValuedControl).KnobPosition := TKnobsValuedControl( Controls[ i]).KnobPosition;
        end;

        if Assigned( anOwner)
        then anOwner.InsertComponent( Result);

        with Result
        do begin
          if MustDrag
          then begin
            Left := anOffset.X;
            Top  := anOffset.Y - 1000;
          end
          else begin
            Left := anOffset.X;
            Top  := anOffset.Y;
            if Top < 0
            then Top := 0;
          end;
          Parent   := aParent;
          Selected := True;
        end;

        with Result
        do begin
          if MustDrag and HasParent and Parent.HasParent
          then begin
            SetCapture( Handle);
            Capturing := True;
          end;
          SelectUnique;
        end;
      finally
        S.Free;
      end;
    end;

    procedure   TKnobsCustomModule.SetKnobValue( const aType, aName: string; aValue: Integer);
    var
      aValuedControl : TKnobsValuedControl;
      anXYControl    : TKnobsXYControl;
      aSuffix        : string;
    begin
      if
        SameText( aType, 'TKnobsSelector'     ) or
        SameText( aType, 'TKnobsKnob'         ) or
        SameText( aType, 'TKnobsSmallKnob'    ) or
        SameText( aType, 'TKnobsNoKnob'       ) or
        SameText( aType, 'TKnobsSlider'       ) or
        SameText( aType, 'TKnobsValuedButton' ) or
        SameText( aType, 'TKnobsValuedButtons')
      then begin
        aValuedControl := FindValuedControl( aName);

        if Assigned( aValuedControl)
        then aValuedControl.KnobPosition := aValue;
      end
      else if
        SameText( aType, 'TKnobsPad')
      then begin
        aSuffix     := Copy( aName, Length( aName), 1);
        anXYControl := FindXYControl( Copy( aName, 1, Length( aName) - 1));

        if Assigned( anXYControl)
        then begin
          if SameText( aSuffix, 'x')
          then anXYControl.KnobPositionX := aValue
          else if SameText( aSuffix, 'y')
          then anXYControl.KnobPositionY := aValue;
        end;
      end;
    end;

    procedure   TKnobsCustomModule.SetDetailValue( const aType, aName: string; const aValue: string);
    var
      aDisplay      : TKnobsDisplay;
      aFileSelector : TKnobsFileSelector;
    begin
      if SameText( aType, 'TKnobsDisplay')
      then begin
        if SameText( aName, 'waveplayer_display_filename')   // Some trickery to be able to read old patches
        then begin                                           // which did not have the file selector component
          aFileSelector := FindFileSelector( 'waveplayer_fileselector_filename');

          if Assigned( aFileSelector)
          then aFileSelector.SetText( aValue);
        end
        else begin
          aDisplay := FindDisplay( aName);

          if Assigned( aDisplay)
          then aDisplay.SetText( aValue);
        end;
      end
      else if SameText( aType, 'TKnobsFileSelector')
      then begin
        aFileSelector := FindFileSelector( aName);

        if Assigned( aFileSelector)
        then aFileSelector.SetText( aValue);
      end;
    end;

    procedure   TKnobsCustomModule.SetLockOn( const aType, aname: string);
    var
      aValuedControl : TKnobsValuedControl;
    begin
      if
        SameText( aType, 'TKnobsKnob'     ) or
        SameText( aType, 'TKnobsSmallKnob') or
        SameText( aType, 'TKnobsNoKnob'   ) or
        SameText( aType, 'TKnobsSlider'   )
      then begin
        aValuedControl := FindValuedControl( aName);

        if Assigned( aValuedControl)
        then aValuedControl.Locked := True;
      end;
    end;

    procedure   TKnobsCustomModule.SetLight( const aName: string; aValue: TSignal);
    var
      i          : Integer;
      anIndicator: TKnobsIndicator;
      anIndBar   : TKnobsIndicatorBar;
      anIndText  : TKnobsIndicatorText;
    begin
      for i := 0 to ControlCount - 1
      do begin
        if Controls[ i] is TKnobsIndicator
        then begin
          anIndicator := TKnobsIndicator( Controls[ i]);
          if SameText( aName, CleanControlName( anIndicator.Name))
          then anIndicator.Active := aValue > 0;
        end
        else if Controls[ i] is TKnobsIndicatorBar
        then begin
          anIndBar := TKnobsIndicatorBar( Controls[ i]);
          if SameText( aName, CleanControlName( anIndBar.Name))
          then anIndBar.Value := aValue * TKnobsIndicatorBar( Controls[ i]).IndicatorCount;
        end
        else if Controls[ i] is TKnobsIndicatorText
        then begin
          anIndText := TKnobsIndicatorText( Controls[ i]);
          if SameText( aName, CleanControlName( anIndText.Name))
          then anIndText.Value := aValue;
        end;
      end;
    end;

    procedure   TKnobsCustomModule.SetData( const aName: string; const aData: TSignalArray);
    var
      i           : Integer;
      aDataViewer : TKnobsDataViewer;
    begin
      for i := 0 to ControlCount - 1
      do begin
        if Controls[ i] is TKnobsDataViewer
        then begin
          aDataViewer := TKnobsDataViewer( Controls[ i]);
          if SameText( aName, CleanControlName( aDataViewer.Name))
          then aDataViewer.YData := aData;
        end;
      end;
    end;

    procedure   TKnobsCustomModule.SetXYData( const aName: string; const aData: TSignalPairFifo);
    var
      i           : Integer;
      aDataViewer : TKnobsDataViewer;
    begin
      for i := 0 to ControlCount - 1
      do begin
        if Controls[ i] is TKnobsDataViewer
        then begin
          aDataViewer := TKnobsDataViewer( Controls[ i]);
          if SameText( aName, CleanControlName( aDataViewer.Name))
          then aDataViewer.XYData := aData;
        end;
      end;
    end;


    procedure   TKnobsCustomModule.SetInfo( const aName: string; const aValue: string);
    var
      i        : Integer;
      aDisplay : TKnobsDisplay;
    begin
      for i := 0 to ControlCount - 1
      do begin
        if Controls[ i] is TKnobsDisplay
        then begin
          aDisplay := TKnobsDisplay( Controls[ i]);
          if SameText( aName, CleanControlName( aDisplay.Name))
          then aDisplay.Caption := aValue;
        end
      end;
    end;

    procedure   TKnobsCustomModule.Dump( var aFile: TextFile; anInd: Integer);
    var
      i : Integer;
    begin
      WriteLnInd( aFile, anInd, Format( 'module : %s', [ Name], AppLocale));
      for i := 0 to ControlCount - 1
      do begin
        if Controls[ i] is TKnobsConnector
        then TKnobsConnector( Controls[ i]).Dump( aFile, anInd + 1)
        else if Controls[ i] is TKnobsValuedControl
        then TKnobsValuedControl( Controls[ i]).Dump( aFile, anInd + 1);
      end;
      WriteLnInd( aFile, anInd, Format( 'end module : %s', [ Name], AppLocale));
    end;

    procedure   TKnobsCustomModule.Log( const aMsg: string);
    begin
      if Parent is TKnobsWirePanel
      then TKnobsWirePanel( Parent).Log( Format( 'custom module %s( %s)', [ Name, aMsg], AppLocale))
      else if Assigned( FOnLog)
      then FOnLog( Self, Format( 'custom module %s( %s)', [ Name, aMsg], AppLocale));
    end;

    procedure   TKnobsCustomModule.LogFmt( const aFmt: string; const anArgs: array of const);
    begin
      Log( Format( aFmt, anArgs, AppLocale));
    end;

    function    TKnobsCustomModule.AsBitmap: TBitmap;
    begin
      Result := TBitmap.Create;
      Result.Width  := Width;
      Result.Height := Height;
      PaintTo( Result.Canvas, 0, 0);
    end;

//  public

    procedure   TKnobsCustomModule.Restack( DoRestack, RedrawWires: Boolean);

      function DoSort( Item1, Item2: Pointer): Integer;

        function CompareModulePositions( F1, F2: TKnobsCustomModule): Integer;
        begin
          if F1.Left < F2.Left
          then Result := -1
          else if F1.Left > F2.Left
          then Result := 1
          else begin
            if F1.Top < F2.Top
            then Result := -1
            else if F1.Top > F2.Top
            then Result := 1
            else Result := 0;
          end;
        end;

      begin
        Result := CompareModulePositions( TKnobsCustomModule( Item1), TKnobsCustomModule( Item2));
      end;

    var
      i           : Integer;
      CurrentTop  : Integer;
      CurrentLeft : Integer;
      MinTop      : Integer;
      aWirePanel  : TKnobsWirePanel;
    begin
      if not DoRestack
      then Exit;

      aWirePanel := FindWirePanel;
      if Assigned( aWirePanel) and aWirePanel.SnapActive
      then begin
        with aWirePanel
        do begin

          CurrentLeft := OffsetLeft - ScrollOffset.X;
          CurrentTop  := OffsetTop  - ScrollOffset.Y;

          for i := 0 to ModuleCount - 1
          do begin
            with Module[ i]
            do begin
              if Top < CurrentTop
              then CurrentTop := Top;
              if Left < CurrentLeft
              then CurrentLeft := Left;
            end;
          end;

          MinTop := CurrentTop;

          FModules.Sort( @ DoSort);
          for i := 0 to ModuleCount - 1
          do begin
            with Module[ i]
            do begin
              TabOrder := i;

              if Left > CurrentLeft
              then begin
                CurrentLeft := Left;
                CurrentTop  := MinTop;
              end;

              if Top < CurrentTop
              then Top := CurrentTop;

              CurrentTop := Top + Height;

            end;
          end;

          if RedrawWires
          then InvalidateWires;
        end;
      end;
    end;

    procedure   TKnobsCustomModule.FixupInsertion;
    begin
      // note: first Restack then FixNames, otherwise naming conflicts will result
      // between the compileed patch and the editor patch.
      Restack( DO_RESTACK, DO_REDRAW_WIRES);
      FixNames;
    end;

    procedure   TKnobsCustomModule.FixDisplays;
    var
      i : Integer;
    begin
      for i := 0 to ControlCount - 1
      do begin
        if Controls[ i] is TKnobsValuedControl
        then TKnobsValuedControl( Controls[ i]).FixDisplay;
      end;
    end;

    procedure   TKnobsCustomModule.Disconnect;
    var
      i : Integer;
    begin
      for i := 0 to ControlCount - 1
      do begin
        if Controls[ i] is TKnobsConnector
        then TKnobsConnector( Controls[ i]).DisConnectAll;
      end;
    end;

    procedure   TKnobsCustomModule.SelectUnique;
    // Select this module uniquely in the patch
    var
      aWp : TKnobsWirePanel;
    begin
      aWp := FindWirePanel;

      if Assigned( aWp) and ( aWp.BlockCount = 0)
      then aWp.SelectUnique( Self);
    end;

    procedure   TKnobsCustomModule.FixSelection;
    // When no modules are selected select this Module uniquely
    var
      aWp : TKnobsWirePanel;
    begin
      aWp := FindWirePanel;

      if Assigned( aWp) and ( aWp.BlockCount = 0)
      then aWp.FixSelection( Self);
    end;

    procedure   TKnobsCustomModule.CollectInputsWithBaseName( var anInputs: TKnobsConnectors; const aBaseName: string);
    var
      i        : Integer;
      aParts   : TStringList;
      aResults : TStringList;
    begin
      SetLength( anInputs, 0);
      aResults := TStringList.Create;
      try
        aResults.Sorted := True;
        for i := 0 to ControlCount - 1
        do begin
          if Controls[ i] is TKnobsInput
          then begin
            aParts := Explode( Controls[ i].Name, '_');
            try
              if ( aParts.Count = 2) and ( Pos( aBaseName, aParts[ 1]) = 1)
              then aResults.AddObject( aParts[ 1], Controls[ i]);
            finally
              aParts.Free;
            end;
          end;
        end;
        for i := 0 to aResults.Count - 1
        do begin
          SetLength( anInputs, Length( anInputs) + 1);
          anInputs[ Length( anInputs) - 1] := TKnobsOutput( aResults.Objects[ i]);
        end;
      finally
        aResults.Free;
      end;
    end;

    procedure   TKnobsCustomModule.CollectOutputsWithBaseName( var anOutputs: TKnobsConnectors; const aBaseName: string);
    var
      i        : Integer;
      aParts   : TStringList;
      aResults : TStringList;
    begin
      SetLength( anOutputs, 0);
      aResults := TStringList.Create;
      try
        aResults.Sorted := True;
        for i := 0 to ControlCount - 1
        do begin
          if Controls[ i] is TKnobsOutput
          then begin
            aParts := Explode( Controls[ i].Name, '_');
            try
              if ( aParts.Count = 2) and ( Pos( aBaseName, aParts[ 1]) = 1)
              then aResults.AddObject( aParts[ 1], Controls[ i]);
            finally
              aParts.Free;
            end;
          end;
        end;
        for i := 0 to aResults.Count - 1
        do begin
          SetLength( anOutputs, Length( anOutputs) + 1);
          anOutputs[ Length( anOutputs) - 1] := TKnobsOutput( aResults.Objects[ i]);
        end;
      finally
        aResults.Free;
      end;
    end;

    procedure   TKnobsCustomModule.UnFocus( const aSender: TObject); // virtual;
    var
      aWp : TKnobsWirePanel;
    begin
      if Assigned( FOnUnFocus)
      then FOnUnFocus( aSender)
      else begin
        aWp := FindWirePanel;

        if Assigned( aWp)
        then aWp.UnFocus( aSender);
      end;
    end;

    function    TKnobsCustomModule.IsConnected: Boolean;
    // todo : this needs some more work ...
    //
    // It depends on the side effects a module has ... like a ModOutput
    // has the side effect of exporting signals to the outer world, so it is connected
    // when one of its inputs is connected.
    // For modules that have no side effects we'd want it to be unconnected when no outputs are connected
    var
      i                  : Integer;
      aConnector         : TKnobsConnector;
      HasInputs          : Boolean;
      HasOutputs         : Boolean;
      AnyInputConnected  : Boolean;
      AnyOutputConnected : Boolean;
    begin
      HasInputs          := False;
      HasOutputs         := False;
      AnyInputConnected  := False;
      AnyOutputConnected := False;

      for i := 0 to ControlCount - 1
      do begin
        if Controls[ i] is TKnobsConnector
        then begin
          aConnector := TKnobsConnector( Controls[ i]);

          if aConnector is TKnobsOutput
          then begin
            HasOutputs := True;
            if aConnector.Connected
            then AnyOutputConnected := True;
          end;

          if aConnector is TKnobsInput
          then begin
            HasInputs := True;
            if aConnector.Connected
            then AnyInputConnected := True;
          end;

        end;
      end;

      if not HasInputs and not HasOutputs then Result := True                  // no ins and outs, module is connected
      else if not HasOutputs              then Result := AnyInputConnected     // no outputs, connected when any input is connected
      else                                     Result := AnyOutputConnected;   // otherwise only connected when an output is connected
    end;

    procedure   TKnobsCustomModule.CopyParamsFrom( aModule: TKnobsCustomModule);
    var
      i : Integer;
    begin
      if Assigned( aModule) and ( aModule.ModuleType = ModuleType) and ( aModule.ControlCount = ControlCount)
      then begin
        for i := 0 to ControlCount - 1
        do begin
          if Controls[ i].ClassType = aModule.Controls[ i].ClassType
          then begin
            if Controls[ i] is TKnobsValuedControl
            then TKnobsValuedControl( Controls[ i]).KnobPosition := TKnobsValuedControl( aModule.Controls[ i]).KnobPosition
            else if Controls[ i] is TKnobsXYControl
            then begin
              TKnobsXYControl( Controls[ i]).KnobPositionX := TKnobsXYControl( aModule.Controls[ i]).KnobPositionX;
              TKnobsXYControl( Controls[ i]).KnobPositionY := TKnobsXYControl( aModule.Controls[ i]).KnobPositionY;
            end;
            // todo: maybe also copy over stuff like filename for waveplayers etc.
          end;
        end;
      end;
    end;


{ ========
  TKnobsModuleButton = class( TSpeedButton)
  // Button as used in module selector, on click it creates a module
  // of the type that got registered with it.
  private
    FPageName   : string;
    FModuleType : TKnobsModuleType;
    FModuleName : string;
  public
    property    PageName   : string           read FPageName;
    property    ModuleType : TKnobsModuleType read FModuleType;
    property    ModuleName : string           read FModuleName;
  public
}

    procedure   TKnobsModuleButton.LoadGlyph( const aGlyph: TBitmap);
    var
      aFileName : string;
    begin
      if Assigned( aGlyph)
      then Glyph.Assign( aGlyph)
      else begin
        aFileName := ApplicationPath + 'bitmaps\' + ModuleName + '.bmp';

        if FileExists( aFileName)
        then
          with Glyph
          do begin
            LoadFromFile( aFileName);
            NumGlyphs := Width div Height;
          end
        else Caption := IntToStr( ModuleType);
      end;
    end;

    procedure   TKnobsModuleButton.Initialize( const aPageName: string; aModuleType: TKnobsModuleType; const aModuleName: string; const aGlyph: TBitmap);
    begin
      FPageName   := aPageName;
      FModuleType := aModuleType;
      FModuleName := aModuleName;
      AllowAllUp  := True;
      GroupIndex  := 1;
      Flat        := True;
      LoadGlyph( aGlyph);
    end;


{ ========
  TKnobsModuleSelector = class( TPageControl)
  // A tabbed thingy having TModuleButtons on it's pages.
  private
    FRegisteredModules   : TKnobsModuleList;
    FOnModuleButtonClick : TKnobsOnModuleButtonClick;
    FOnGetGlyph          : TKnobsOnGetGlyph;
    FButtonsEnabled      : Boolean;
    FAllowDuplicates     : Boolean;
    FTimer               : TTimer;
  published
    property    OnModuleButtonClick : TKnobsOnModuleButtonClick read FonModuleButtonClick write FOnModuleButtonClick;
    property    OnGetGlyph          : TKnobsOnGetGlyph          read FOnGetGlyph          write FonGetGlyph;
    property    ButtonsEnabled      : Boolean                   read FButtonsEnabled      write SetButtonsEnabled;
    property    AllowDuplicates     : Boolean                   read FAllowDuplicates     write FAllowDuplicates     default True;
  private
}

    procedure   TKnobsModuleSelector.StartTimeout;
    begin
      if not Assigned( FTimer)
      then FTimer := TTimer.Create( nil);
      if Assigned( FTimer)
      then begin
        FTimer.Enabled  := False;
        FTimer.OnTimer  := TimeoutFired;
        FTimer.Interval := 1000;
        FTimer.Enabled  := True;
      end;
    end;

    procedure   TKnobsModuleSelector.StopTimeout;
    begin
      if Assigned( FTimer)
      then begin
        FTimer.Enabled := False;
        FTimer.OnTimer := nil;
        FreeAndNil( FTimer);
      end;
    end;

    procedure   TKnobsModuleSelector.TimeOutFired( aSender: TObject);
    begin
      AllButtonsUp;
      StopTimeout;
    end;

    procedure   TKnobsModuleSelector.AllButtonsUp;
    var
      i : Integer;
    begin
      with FRegisteredModules
      do begin
        for i := 0 to Count - 1
        do begin
          TKnobsModuleButton( Items[ i]).Down := True;
          TKnobsModuleButton( Items[ i]).Down := False;
        end;
      end;
    end;

//  private

    procedure   TKnobsModuleSelector.SetButtonsEnabled( aValue: Boolean);
    var
      i : Integer;
    begin
      if aValue <> FButtonsEnabled
      then begin
        FButtonsEnabled := aValue;

        with FRegisteredModules
        do begin
          for i := 0 to Count - 1
          do TKnobsModuleButton( Items[ i]).Enabled := FButtonsEnabled;
        end;
      end;
    end;

    function    TKnobsModuleSelector.AddPage( const aName: string): TTabSheet;
    begin
      Result := TTabSheet.Create( Self);

      with Result
      do begin
        Visible     := True;
        Caption     := aName;
        PageControl := Self;
      end;
    end;

    function    TKnobsModuleSelector.PageButtonCount( const aName: string): Integer;
    var
      i : Integer;
    begin
      Result := 0;
      if PageExists( aName)
      then begin
        with PageByName( aName)
        do begin
          for i := 0 to ControlCount - 1
          do begin
            if Controls[ i] is TKnobsModuleButton
            then Inc( Result);
          end;
        end;
      end;
    end;

    function    TKnobsModuleSelector.PageExists( const aName: string): Boolean;
    begin
      Result := PageByName( aName) <> nil;
    end;

    function    TKnobsModuleSelector.PageByName( const aName: string): TTabSheet;
    var
      i : Integer;
    begin
      Result := nil;

      for i := 0 to PageCount - 1
      do begin
        if Pages[ i].Caption = aName
        then begin
          Result := Pages[ i];
          Break;
        end;
      end;
    end;

    function    TKnobsModuleSelector.FindModulePage( aModuleType: TKnobsModuleType): TTabSheet;
    var
      i : Integer;
    begin
      Result := nil;

      with FRegisteredModules
      do begin
        for i := 0 to Count - 1
        do begin
          with TKnobsModuleButton( Items[ i])
          do begin
            if ModuleType = aModuleType
            then begin
              Result := PageByName( PageName);
              Break;
            end;
          end;
        end;
      end;
    end;

    function    TKnobsModuleSelector.FindModuleButton( aModuleType: TKnobsModuleType): TKnobsModuleButton;
    var
      aPage : TTabSheet;
      i     : Integer;
    begin
      aPage := FindModulePage( aModuleType);
      Result := nil;
      if Assigned( aPage)
      then begin
        with aPage
        do begin
          for i := 0 to ControlCount - 1
          do begin
            if Controls[ i] is TKnobsModuleButton
            then begin
              with TKnobsModuleButton( Controls[ i])
              do begin
                if ModuleType = aModuleType
                then begin
                  Result := TKnobsModuleButton( Controls[ i]);
                  Break;
                end;
              end;
            end;
          end;
        end;
      end;
    end;

    procedure   TKnobsModuleSelector.DoModuleMouseDown( aSender: TObject; aButton: TMouseButton; aShift: TShiftState; X, Y: Integer);
    begin
      if Assigned( FOnModuleButtonClick) and ( aSender is TKnobsModuleButton) and ( aShift = [ ssLeft])
      then begin
        FOnModuleButtonClick(
          TKnobsModuleButton( aSender),
          TKnobsModuleButton( aSender).ModuleType
        );
        StartTimeout;  // should release the button after a while ... button stays down though ... grrr .. this sort of works ...
      end;
    end;

    procedure   TKnobsModuleSelector.FormatPage( aPage: TTabSheet);
    var
      LastLeft : Integer;
      i        : Integer;
    begin
      Assert(
        Assigned( aPage),
        'TKnobsModuleSelecteor.FormatPage : can''t format an unassigned <nil> page.'
      );

      if PageButtonCount( aPage.Caption) = 0
      then begin
        aPage.Free;
        if PageCount > 0
        then ActivePage := Pages[ 0];
      end
      else begin
        with aPage
        do begin
          LastLeft := 0;

          for i := 0 to ControlCount - 1
          do begin
            if Controls[ i] is TKnobsModuleButton
            then begin
              with TKnobsModuleButton( Controls[ i])
              do begin
                Left := LastLeft;
                if Controls[ i] is TKnobsModuleSeparator
                then Inc( LastLeft, 12)
                else Inc( LastLeft, Width + 2);
              end;
            end;
          end;
        end;
      end;
    end;

{  protected }

    procedure   TKnobsModuleSelector.Notification( aComponent: TComponent; anOperation: TOperation); // override;
    begin
      if aComponent is TKnobsModuleButton
      then begin
        if anOperation = opInsert
        then FRegisteredModules.Add   ( aComponent)
        else FRegisteredModules.Remove( aComponent);
      end;
    end;

{   public }

    constructor TKnobsModuleSelector.Create( anOwner: TComponent); // override;
    begin
      inherited;

      FRegisteredModules := TKnobsModuleList.Create;
      HotTrack           := True;
      FAllowDuplicates   := True;
    end;

    destructor  TKnobsModuleSelector.Destroy; // override;
    begin
      StopTimeout;
      inherited;
      FRegisteredModules.Free;
     end;

    procedure   TKnobsModuleSelector.RegisterModuleType( const aPageName: string; aModuleType: TKnobsModuleType; const aModuleName: string);
    var
      aPage         : TTabSheet;
      aModuleButton : TKnobsModuleButton;
      aGlyph        : TBitmap;
    begin
      if not FAllowDuplicates
      then begin
        if Assigned( FindModuleButton( aModuleType)) and ( aModuleType >= 0)
        then begin
          raise
            EKnobsNoDuplicates.
              CreateFmt(
                'A module type %d with name ''%s''is already present on page ''%s''',
                [
                  aModuleType,
                  aModuleName,
                  FindModulePage( aModuleType).Caption
                ]
              );
        end;
      end;

      aPage := PageByName( aPageName);

      if not Assigned( aPage)
      then aPage := AddPage( aPageName);

      ActivePage := aPage;
      ActivePage := Pages[ 0];
      if aModuleType >= 0
      then aModuleButton := TKnobsModuleButton   .Create( Self)
      else aModuleButton := TKnobsModuleSeparator.Create( Self);

      if Assigned( FOnGetGlyph)
      then aGlyph := FOnGetGlyph( Self, aModuleName)
      else aGlyph := nil;

      with aModuleButton
      do begin
        Parent   := aPage;
        Left     := 0;
        Top      := 0;
        Flat     := True;
        Hint     := Format( '%s~%d~', [ aModuleName, aModuleType], AppLocale);
        ShowHint := True;
        Enabled  := FButtonsEnabled;
        Visible  := aModuleType >= 0;
        Initialize( aPageName, aModuleType, aModuleName, aGlyph);
        OnMouseDown := DoModuleMouseDown;
      end;

      FormatPage( aPage);
    end;

    procedure   TKnobsModuleSelector.UnregisterModuleType( aModuleType: TKnobsModuleType);
    var
      aModuleButton : TKnobsModuleButton;
      aPage         : TTabSheet;
    begin
      aModuleButton := FindModuleButton( aModuleType);
      aPage         := FindModulePage  ( aModuleType);

      if Assigned( aModuleButton) and Assigned( aPage)
      then begin
        aModuleButton.Free;
        FormatPage( aPage); // This will remove the page when it beccomes empty.
      end;
    end;


{ ========
  TKnobsHintWindow = class( THintWindow)
  private
    FModuleType : string;
    FBitmap     : TBitmap;
    FComment    : string;
  published
    property    Caption;
  protected
}

    function    TKnobsHintWindow.MakeBitmap( const aModuleType: string): TBitmap;
    var
      aModuleTypeInt : TKnobsModuleType;
    begin
      aModuleTypeInt := StrToIntDef( aModuleType, -1);
      if aModuleTypeInt >= 0
      then begin
        Result   := CreateModuleBitmap( aModuleTypeInt, True);
        FComment := ReadModuleComment ( aModuleTypeInt);
      end
      else begin
        Result   := nil;
        FComment := '';
      end;
    end;

    procedure   TKnobsHintWindow.ParseHint( var aHintStr: string);
    type
      TParseState = ( psNothing, psTilde);
    var
      aHint      : string;
      i          : Integer;
      ParseState : TParseState;
    begin
      FModuleType   := '';
      aHint         := '';
      ParseState    := psNothing;

      for i := 1 to Length( aHintStr)
      do begin

        case ParseState of

          psNothing :
            begin
              if aHintStr[ i] = '~'
              then ParseState := psTilde
              else aHint := aHint + aHintStr[ i];
            end;

          psTilde :
            begin
              if aHintStr[ i] = '~'
              then ParseState := psNothing
              else FModuleType := FModuleType + aHintStr[ i];
            end;
        end;

      end;

      if Assigned( FBitmap)
      then FreeAndNil( FBitmap);

      FBitmap := MakeBitmap( FModuleType);

      aHintStr := FComment;
    end;

    procedure   TKnobsHintWindow.Paint; // override;
    var
      R          : TRect;
      aTextRect  : TRect;
      aTopMargin : Integer;
    const
      TEXT_MARGIN_LEFT =  2;
      IMG_MARGIN_TOP   =  4;
      TEXT_MARGIN_TOP  =  2;
    begin
      if Assigned( FBitmap)
      then aTopMargin := IMG_MARGIN_TOP + TEXT_MARGIN_TOP + FBitmap.Height
      else aTopMargin := TEXT_MARGIN_TOP;

      R := ClientRect;
      Inc( R.Left, 2);
      Inc( R.Top , 2);

      with Canvas
      do begin
        Rectangle( 0, 0, R.Width + 2, R.Bottom + 2);
        if Assigned( FBitmap)
        then Draw( 2, IMG_MARGIN_TOP, FBitmap);
        aTextRect := R;
        Inc( aTextRect.Top , aTopMargin);
        Inc( aTextRect.Left, TEXT_MARGIN_LEFT);
        DrawText( Handle, PChar( Caption), -1, aTextRect, dt_NoClip or dt_WordBreak);
      end;
    end;

//  public

    constructor TKnobsHintWindow.Create( anOwner: TComponent); // override;
    begin
      inherited;

      with Canvas.Font do
      begin
        Name  := 'Ariel';
        Color := clBlack;
      end;
    end;

    destructor  TKnobsHintWindow.Destroy; // override;
    begin
      if Assigned( FBitmap)
      then FreeAndNil( FBitmap);
      inherited;
    end;

    procedure   TKnobsHintWindow.ActivateHint( aRect: TRect; const aHint: string); // override;
    const
      EXTRA_WIDTH  =  8;
      EXTRA_HEIGHT =  8;
    var
      anExtraHeight : Integer;
      aWidth        : Integer;
      aHintStr      : string;
    begin
      aHintStr := aHint;
      ParseHint( aHintStr);
      Caption := aHintStr;

      if Assigned( FBitmap)
      then begin
        anExtraHeight := EXTRA_HEIGHT + FBitmap.Height;
        aWidth        := EXTRA_WIDTH  + FBitmap.Width;
      end
      else begin
        anExtraHeight := EXTRA_HEIGHT;
        aWidth        := EXTRA_WIDTH + aRect.Width;
      end;

      Inc( aRect.Bottom, anExtraHeight);
      aRect.Right := aRect.Left + aWidth;
      UpdateBoundsRect( aRect);

      if aRect.Top + Height > Screen.DesktopHeight
      then aRect.Top := Screen.DesktopHeight - Height;

      if aRect.Left + Width > Screen.DesktopWidth
      then aRect.Left := Screen.DesktopWidth - Width;

      if aRect.Left < Screen.DesktopLeft
      then aRect.Left := Screen.DesktopLeft;

      if aRect.Bottom < Screen.DesktopTop
      then aRect.Bottom := Screen.DesktopTop;

      SetWindowPos( Handle, HWND_TOPMOST, aRect.Left, aRect.Top, Width, Height, SWP_SHOWWINDOW or SWP_NOACTIVATE);
      Invalidate;
    end;




initialization

  CreateBitmaps;

finalization

  FreeBitmaps;

end.

